/****************************************************************************
!C

!File: comp_sds_stat.c

!Description:
  Compute mean of sds from input files 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original June 2003. Version 1.0

  Modified in February and April, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	3) speedup comp_sds1_stat() by computing lines not single line

  Modified February, 2020 Version 3.6 by Maosheng Zhao, SSAI
  Correct the issue of output incorrect or missing ESDT, SDSs for VNP43M by
  increasing the length of the strings in the code

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/         

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "sds_rw.h"
#include "main_util.h"
#include "meta.h"

#define HELP \
"NAME \n" \
"     comp_sds_stat - Compute pixel level mean of SDS in input HDF files\n" \
" \n" \
"SYNOPSIS \n" \
"     comp_sds_stat -help \n" \
"     comp_sds_stat -sds=<sds1,sds2. .  .> -of=<fname> f1, f2, . .  \n" \
" \n" \
"DESCRIPTION \n" \
"     Compute pixel level mean of SDS in input HDF files \n" \
" \n" \
"OPTIONS \n" \
"     -help                          Display this help message \n" \
"     -sds=<sds1,sd2,. . >           sdsnames separated by comma\n" \
"     -of=<fname>                    output filename \n" \
" \n" \
"Example: \n" \
"     comp_sds_stat -sds=sur_refl_b01,sur_refl_b04,sur_refl_b03 \n" \
"                   -of=MOD09A1.mean_2016209_2016217.h09v05.006.hdf \n" \
"                   MOD09A1.A2016209.h09v05.006.2016222123023.hdf \n"  \
"                   MOD09A1.A2016217.h09v05.006.2016226075317.hdf \n"  \
"     comp_sds_stat -sds=SurfReflect_M5,SurfReflect_M4,SurfReflect_M3  \n" \
"                   -of=VNP09A1.mean_2016289_2016297.h11v08.001.hdf \n" \
"                   VNP09A1.A2016289.h11v08.001.2017189045223.h5 \n" \
"		    VNP09A1.A2016297.h11v08.001.2017189113645.h5 \n" \
"\n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:\tcomp_sds_stat -sds=<sds1,sds2. .  .> -of=<fname> f1, f2, . . \n" \
" \n" \
"OPTIONS \n" \
"     -help                          Display this help message \n" \
"     -sds=<sds1,sd2,. . >           sdsnames separated by comma\n" \
"     -of=<fname>                    output filename \n" \
" \n"

int parse_cmd_comp_vi_stat(int argc, char **argv, int *nsds, char **sds_names, int *nf, char **in_fnames, char *out_fname);
void comp_sds1_stat(sds_t *in_sds_info, int nf, sds_t *out_sds_info);

int main(int argc, char *argv[])
{
  int st;
  int isds, nsds;
  int ifile, nfile;
  char **sds_names, **in_fnames;
  char out_fname[MAX_PATH_LENGTH];
  sds_t *in_sds_info, out_sds_info;

  if ((argc == 2) && ((strcmp(argv[1], "-h") == 0) || (strcmp(argv[1], "-help") == 0))) {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }         

  /* MSZ 4/5/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 4/5/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names in comp_sds_stat(main)\n");
  if ((in_fnames = (char **)Calloc2D(MAX_NUM_FILE, MAX_PATH_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for in_fnames in comp_sds_stat(main)\n");
  if ((st = parse_cmd_comp_vi_stat(argc, argv, &nsds, sds_names, &nfile, in_fnames, out_fname)) == -1)
    fprintf(stderr, "%s\n", USAGE);
  else {
    if ((in_sds_info = (sds_t *)calloc(nfile, sizeof(sds_t))) == NULL)
      fprintf(stderr, "Cannot allocate memory for in_sds_info in comp_sds_stat(main)\n");
    else {
      out_sds_info.H5OUT = 0;  /* for far, just output to HDF4 */
      out_sds_info.H5IN = out_sds_info.H5OUT;
      out_sds_info.sd_id = out_sds_info.sds_id = -1;
      if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
        fprintf(stderr, "Cannot create output file %s\n", out_fname);
      else {
        for (isds=0; isds<nsds; isds++) {
          for (ifile=0, st=0; ifile<nfile; ifile++) {
            strcpy(in_sds_info[ifile].name, sds_names[isds]);
            in_sds_info[ifile].sd_id = in_sds_info[ifile].sds_id = -1;
            st += get_sds_info(in_fnames[ifile], &in_sds_info[ifile]);
	    /* MSZ 4/6/2018, add this message */
	    if(in_sds_info[ifile].rank != 2)
	    {
		fprintf(stderr, "%s is not 2D data and this tool doesn't support non-2D data, exit...\n", in_sds_info[ifile].name);
		exit(0);
	    }
            if (st == 1) {
              out_sds_info.sds_id = -1;
              out_sds_info.rank = in_sds_info[ifile].rank;
	      out_sds_info.dim_size[0] = in_sds_info[ifile].dim_size[0];
	      out_sds_info.dim_size[1] = in_sds_info[ifile].dim_size[1];
	      out_sds_info.data_type = in_sds_info[ifile].data_type;
	      out_sds_info.data_size = in_sds_info[ifile].data_size;
              strcpy(out_sds_info.name, in_sds_info[ifile].name);
	      if (open_sds((char *)NULL, &out_sds_info, 'W') == -1)
		break;
	    }
          } /* for ifile */
          if ((st == nfile) && (out_sds_info.sds_id != FAIL)) { 
	    comp_sds1_stat(in_sds_info, nfile, &out_sds_info);
	   
	    close_sds(&out_sds_info, out_sds_info.H5OUT); 
          }
          for (ifile=0, st=0; ifile<nfile; ifile++) {
	    close_sds(&in_sds_info[ifile], in_sds_info[ifile].H5IN);
	    close_sd(&in_sds_info[ifile], in_sds_info[ifile].H5IN);
          }
        } /* for isds */
	close_sd(&out_sds_info, out_sds_info.H5OUT);
      }
      free(in_sds_info);
    }
  }
  Free2D((void **)sds_names);
  Free2D((void **)in_fnames);
}

int parse_cmd_comp_vi_stat(int argc, char **argv, int *nsds, char **sds_names, int *nf, char **in_fnames, char *out_fname)
{
  int i, n;
  int st = 1;

  *nf = *nsds = n  = 0;
  out_fname[0] = '\0';
  
  for (i=1; i<argc; i++)
  {
    if (is_arg_id(argv[i], "-sds=") == 0) {
      get_arg_val_arr(argv[i], sds_names, nsds);
    }
    else if (is_arg_id(argv[i], "-of=") == 0)
      get_arg_val(argv[i], out_fname);
    else if (argv[i][0] == '-') 
      fprintf(stderr, "Ignoring unknown option %s\n", argv[i]);
    else {
      strcpy(in_fnames[n], argv[i]);
      n++;
    }
  }
  *nf = n;

  if (out_fname[0] == '\0') {
    st = -1;
    fprintf(stderr, "Missing output filename\n");
  }
  if (*nf == 0) {
    st = -1;
    fprintf(stderr, "No input files to process\n");
  }
  if (*nsds == 0) {
    st = -1;
    fprintf(stderr, "No input SDS to process\n");
  }
  return st;
}

void comp_sds1_stat(sds_t *in_sds_info, int nfile, sds_t *out_sds_info)
{
  /* Maosheng Zhao, SSAI, 4/6/2018 speedup by comping lines not single line */
  uint8 *np;
  double *sum_data, avg_val;
  void *data_in, *data_out;
  int val, ifile;
  int i, irow, icol;
  int nrows, ncols;
  int32 start[2], edge[2];
  int GROWS, last_rows;
  int nlines_a_copy = 20;

  if(in_sds_info->rank != 2)
  {
    fprintf(stderr, "so far, this function just works for 2D data, not others, exit...\n");
    exit(0);
  }

  for (i=0; i<in_sds_info->rank; i++)
  {
    start[i] = 0;
    edge[i] = in_sds_info->dim_size[i];
  }
  if (in_sds_info->rank != 1)
  {
    if(in_sds_info->dim_size[in_sds_info->row_id] < nlines_a_copy) nlines_a_copy = in_sds_info->dim_size[in_sds_info->row_id];
    edge[in_sds_info->row_id] = 1 * nlines_a_copy;
  }
  nrows = (in_sds_info->rank == 1) ? 1 : in_sds_info->dim_size[in_sds_info->row_id];

  /*nrows = in_sds_info->dim_size[in_sds_info->row_id];*/
  ncols = in_sds_info->dim_size[in_sds_info->row_id + 1];

  /*
  nrows = in_sds_info[0].dim_size[0]; 
  ncols = in_sds_info[0].dim_size[1];
  */

  /* MSZ 4/06/2018 */
  GROWS = (int)ceil((double)nrows / (double)nlines_a_copy);
  last_rows = nrows % nlines_a_copy;

  if ((data_in = (void *)calloc(ncols*nlines_a_copy, in_sds_info[0].data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_in in comp_sds1_stat()\n");
  if ((data_out = (void *)calloc(ncols*nlines_a_copy, out_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_out in comp_sds1_stat()\n");
  if ((sum_data = (void *)calloc(ncols*nlines_a_copy, sizeof(double))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sum_data in comp_sds1_stat()\n");
  if ((np = (void *)calloc(ncols*nlines_a_copy, sizeof(uint8))) == NULL)
    fprintf(stderr, "Cannot allocate memory for np in comp_sds1_stat()\n");
  if ((data_in != NULL) && (data_out != NULL) && (np != NULL) && (sum_data != NULL)) {
    edge[0] = 1 * nlines_a_copy;
    start[1] = 0;
    edge[1] = ncols;
    for (irow=0; irow<GROWS; irow++) {
      start[in_sds_info->row_id] = irow * nlines_a_copy;

      /* MSZ, 1/25/2018, be cautious about the last group rows because it may be different from nlines_a_copy */
      if(last_rows > 0 && irow == GROWS - 1)
	edge[in_sds_info->row_id] = last_rows;

      for (icol=0; icol<ncols*nlines_a_copy; icol++) { 
        sum_data[icol] = 0;
        np[icol] = 0;
      }
      for (ifile=0; ifile<nfile; ifile++) {
	if (read_sel_sds_data(&in_sds_info[ifile], start, edge, data_in) == FAIL) {
          fprintf(stderr, "Error reading data line %d from SDS %s\n", irow, in_sds_info[ifile].name);
          continue;
        }
        for (icol=0; icol<ncols*nlines_a_copy; icol++) {
          switch(in_sds_info[ifile].data_type) {
            case 20: val = ((int8 *)data_in)[icol];
		     break;
            case 21: val = ((uint8 *)data_in)[icol];
		     break;
            case 22: val = ((int16 *)data_in)[icol];
		     break;
            case 23: val = ((uint16 *)data_in)[icol];
		     break;
            case 24: val = ((int32 *)data_in)[icol];
		     break;
            case 25: val = ((uint32 *)data_in)[icol];
		     break;
          }
	  if (val > 0) {
	    sum_data[icol] += val;
	    np[icol] += 1;
          }
        } /* for icol */
      } /* for ifile */
      for (icol=0; icol<ncols*nlines_a_copy; icol++) {
        avg_val = (np[icol] == 0) ? 0 : sum_data[icol]/np[icol] + 0.5;
        switch(out_sds_info->data_type) {
          case 20: ((int8 *)data_out)[icol] = (int8)avg_val;
                   break;
          case 21: ((uint8 *)data_out)[icol]= (uint8)avg_val;
                   break;
          case 22: ((int16 *)data_out)[icol]= (int16)avg_val;
                   break;
          case 23: ((uint16 *)data_out)[icol]= (uint16)avg_val;
                   break;
          case 24: ((int32 *)data_out)[icol]= (int16)avg_val;
                   break;
          case 25: ((uint32 *)data_out)[icol]= (uint16)avg_val;
                   break;
        }
      }
      if(set_sel_sds_data(out_sds_info, start, edge, data_out) == FAIL)
        fprintf(stderr, "Error writing data line %d to SDS %s\n", irow, out_sds_info->name);
    } /* for irow */
  }
  free(data_in);
  free(data_out);
  free(sum_data);
  free(np);
  return;
}


