/****************************************************************************
!C

!File: cp_hdfeos.c

!Description:
  Copy geolocation related metadata to a L3 HDF file. 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original March 2001 Version 1.4
  Modified May 2001 (accepts mutiple tile ids to accomodate input mosaic of many tiles)
  Modified March 2013 by Xiaoping Zhang, Sigma Space Corp. 
 	(fixed the bugs to set out_sds_info.sd_id and to allocate memory of out_sds_info) 

  Modified in February and April, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	3)update all HDFEOS API to a higher-level APIs to handle 
	  both HDFEOS2 and HDFEOS5 in eos_sds_rw.c
	4)modify the previous copy_sds_data_new() which can copy 2-D data. Now
	 copy_an_sds_data_attr() can do from 1 to 4
	5)to speedup, copy every 20 lines by 20 lines not line by line in
	 copy_an_sds_data_attr()  

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)
    Need to copy HDFEOS structural metadata and bounding box.
    Somehow copy GRING and Bounding box metadata.

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/dir.h>         

#include "eos_sds_rw.h"
#include "qa_tool.h"
#include "sds_rw.h"
#include "str_op.h"
#include "meta.h"
#include "alloc_mem.h"
#include "main_util.h"         

#define MAX_GRID_NAME_LEN 40
#define MAX_NUM_TILES 25
#define TILE_ID_STR_LEN 10
#define PARAM_FNAME "LAND_TILE_ISIN_PROJ_PARAM.TXT"

#define HELP \
"NAME \n" \
"     cp_hdfeos - Copy the hdfeos metadata to the input L3 HDF files. \n" \
" \n" \
"SYNOPSIS \n" \
"     cp_hdfeos -help \n" \
"     cp_hdfeos -of=<output_file> [-ref=<reference L3 file>] [-tile=<tile_id(s)>] filename \n" \
" \n" \
"     echo -of=<output_file> [-ref=<reference L3 file>] [-tile=<tile_id(s)>] filename |  cp_hdfeos \n" \
" \n" \
"DESCRIPTION \n" \
"     Copy the hdfeos structural metadata to the input L3 HDF file. \n" \
"     The information can either be copied from a L3 reference file or by \n" \
"     providing the L3 tile id. When tile_id is provided integerized \n" \
"     sinusoidal projection parameter for the specified tile is copied from \n" \
"     an ancillary table. All the SDS in the input HDF file are written into \n" \
"     a new grid named SAMPLE_L3_GRID.\n" \
"     If the input file contains many tiles created by mosaicing two or more \n" \
"     L3 tiles then multiple tile ids are input. \n" \
" \n" \
"OPTIONS \n" \
"     -help              Display this help message \n" \
"     -of=filename       Output file \n" \
"     -ref=<L3 file/dir> reference L3 product file \n" \
"                        (overrides the -tile option) \n" \
"     -tile=<tile id(s)> L3 tile id(s) \n" \
"                        Multiple tile ids are separated by \n" \
"                        comma. (input if the input file is \n" \
"                        a mosaic of many L3 or L2G tiles. \n" \
" \n" \
"Examples: \n" \
"\tcp_hdfeos -of=Grid.Masked.MOD09A1.A2001145.h20v10.003.2001214125825.hdf -tile=h20v10 Maksed.MOD09A1.A2001145.h20v10.003.2001214125825.hdf \n" \
"\tcp_hdfeos -of=Grid.Masked.MOD09A1.A2001145.h20v10.003.2001214125825.hdf -ref=MOD09A1.A2001145.h20v10.003.2001214125825.hdf  Masked.MOD09A1.A2001145.h20v10.003.2001214125825.hdf\n" \
"\tcp_hdfeos -of=Grid.NPP_D16VI1KM_L3D.A2003193.h09v05.P1_00602.2009257123847.hdf -tile=h09v05 NPP_D16VI1KM_L3D.A2003193.h09v05.P1_00602.2009257123847.hdf\n" \
"\tcp_hdfeos -of=Grid.NPP_D16VI1KM_L3D_masked.A2003193.h09v05.P1_00602.2009257123847.hdf -ref=NPP_D16VI1KM_L3D.A2003193.h09v05.P1_00602.2009257123847.hdf NPP_D16VI1KM_L3D_masked.A2003193.h09v05.P1_00602.2009257123847.hdf\n" \
"\n" \
"Note: \n" \
"     Since the current version creates only a single grid in the output HDF-EOS file \n" \
"all SDS in the input HDF file are expected to have the same resolution. \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"


#define USAGE \
"usage: cp_hdfeos -of=<output_file> -ref=<reference L3 file> [-tile=<tile_id>] filename \n" \
" \n" \
"OPTIONS \n" \
"     -help              Display this help message \n" \
"     -of=filename       Output file \n" \
"     -ref=<L3 file/dir> reference L3 product file \n" \
"                        (overrides the -tile option) \n" \
"     -tile=<tile id>    L3 tile id \n" \
"                        Multiple tile ids are separated by \n" \
"                        comma. (input if the input file is \n" \
"                        a mosaic of many L3 or L2G tiles. \n" \
" \n"                                 


int parse_cmd_cp_hdfeos(int argc, char **argv, char *in_fname, char *out_fname, char *ref_fname, char **tile_id, int *n_tile);
int get_hdfeos(char *file_name, grid_mapinfo *h, char *grid_name);
void add_hdfeos_meta(char *file_name, char *grid_name, grid_mapinfo *h, sds_t *sds_info, int nsds);
int read_param_table(char **tile_id, int n_tile, grid_mapinfo *hdfeos);
void copy_an_sds_data_attr(sds_t *in_sds_info, sds_t *out_sds_info);

int main(int argc, char **argv)
{
  int n_tile;
  int i_rank;
  int st, isds, nsds;
  grid_mapinfo hdfeos;
  sds_t *sds_info, *out_sds_info;  
  char **sds_names_exclusion = NULL, **sds_names = NULL, **tile_ids = NULL;
  char ref_fname[MAX_PATH_LENGTH], in_fname[MAX_PATH_LENGTH];
  char out_fname[MAX_PATH_LENGTH], grid_name[MAX_PATH_LENGTH];
  int nattrs = 22;  

  if ((argc == 2) && ((strcmp(argv[1], "-h") == 0) || ( strcmp(argv[1], "-help") == 0)))
    fprintf(stderr, "%s\n", HELP);
  else 
  {
    /* MSZ 2/9/2018 silent version mismatch warning standard print out */
    setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

    /* MSZ 2/9/2018 Disable error reporting */
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

    /* for HDF5 input, excluded StructMetadata.0 for sds_names */
    if ((sds_names_exclusion = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for sds_names_exclusion in main(), exit... \n");
      exit(0);
    }
    
    strcpy(sds_names_exclusion[0], "StructMetadata.0");
    strcpy(sds_names_exclusion[1], "CoreMetadata.0");
    strcpy(sds_names_exclusion[2], "ArchiveMetadata.0");

    if ((tile_ids = (char **)Calloc2D(MAX_NUM_TILES, TILE_ID_STR_LEN, sizeof(char))) == NULL)
      fprintf(stderr, "Cannot allocate memory for tile_ids in main()\n");
    else
    {
      if (parse_cmd_cp_hdfeos(argc, argv, in_fname, out_fname, ref_fname, tile_ids, &n_tile) == -1)
        fprintf(stderr, "%s\n", USAGE);
      else 
      {
	if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for sds_names in main() \n");
	else
	{        	
	  fprintf(stdout, "Number of tiles: %d\n", n_tile);
          if (n_tile > 0)
            st = read_param_table(tile_ids, n_tile, &hdfeos);
          else
          {
	    nsds = get_sds_names_exclusions(ref_fname, sds_names, sds_names_exclusion, 3);  
	    if ((sds_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)		
	      fprintf(stderr, "Cannot allocate memory for sds_info and out_sds_info in main()\n");
	    else
	    {  		      
		
	      for (isds=0; isds<nsds; isds++)
	      {
		sds_info[isds].sd_id = (isds == 0) ? -1 : sds_info[0].sd_id;
		sds_info[isds].H5IN = sds_info[0].H5IN;
		sds_info[isds].sds_id = -1;
		strcpy(sds_info[isds].name, sds_names[isds]);  		
		get_sds_info(ref_fname, &sds_info[isds]);	
	      }	  
	    }           
	    if ((st = get_grid_name(ref_fname, grid_name)) == 1)
	      st = get_hdfeos(ref_fname, &hdfeos, grid_name);
            fprintf(stdout, "grid_name=%s\n",grid_name);    
          }
          if (st == -1)
            fprintf(stderr, "Cannot read projection parameter\n");
          else
          {
	    nsds = get_sds_names_exclusions(in_fname, sds_names, sds_names_exclusion, 3);
	    if ((sds_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL ||
	    	(out_sds_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)	
	      fprintf(stderr, "Cannot allocate memory for sds_info and out_sds_info in main()\n");
	    else
	    {  		      
	      for (isds=0; isds<nsds; isds++)
	      {
		sds_info[isds].sd_id = (isds == 0) ? -1 : sds_info[0].sd_id;
		sds_info[isds].H5IN = sds_info[0].H5IN;
		sds_info[isds].sds_id = -1;
		sds_info[isds].file_name = in_fname;
		strcpy(sds_info[isds].name, sds_names[isds]);  		
		get_sds_info(in_fname, &sds_info[isds]);						
	      }
              add_hdfeos_meta(out_fname, "SAMPLE_L3_GRID", &hdfeos, sds_info, nsds);

	      for (isds=0; isds<nsds; isds++) {	  
		if (isds == 0)
		{
		  out_sds_info[0].H5OUT = 0;   /* output to HDF4 */
		  out_sds_info[0].H5IN = out_sds_info[0].H5OUT;  /* in case the output file is opened for inquiry */
		  out_sds_info[0].sd_id = -1;
		  out_sds_info[0].file_name = out_fname;

		  if(open_sd(out_fname, &out_sds_info[0], 'W') == FAIL)
		  {
		    fprintf(stderr, "fail to open %s \n", out_fname);
		    return -1;
		  }
		}
		else 
		{
		  out_sds_info[isds].H5OUT = out_sds_info[0].H5OUT;
		  out_sds_info[isds].H5IN = out_sds_info[isds].H5OUT; /* in case the output file is opened for inquiry */
		  out_sds_info[isds].sd_id = out_sds_info[0].sd_id;	  
		  out_sds_info[isds].file_name=out_sds_info[0].file_name;
		}

		out_sds_info[isds].sds_id = -1;

		/* sds name must be known before using copy_an_sds_data_attr() */
		strcpy(out_sds_info[isds].name, sds_info[isds].name);

		/* copy sds by sds and also its attributes */
		copy_an_sds_data_attr(&sds_info[isds], &out_sds_info[isds]);

		close_sds(&sds_info[isds], sds_info[isds].H5IN);
		close_sds(&out_sds_info[isds], out_sds_info[isds].H5OUT);
	      }
		
	      close_sd(&sds_info[0], sds_info[0].H5IN);
	      close_sd(&out_sds_info[0], out_sds_info[0].H5OUT);

	      free(sds_info);	
	      free(out_sds_info);	      
	    }
	    Free2D((void **)sds_names);

          }
	}
      }  
      Free2D((void **)tile_ids);
    }
  }

  if(sds_names_exclusion) Free2D((void**)sds_names_exclusion);
  return 0;
}

int parse_cmd_cp_hdfeos(int argc, char **argv, char *in_fname, char *out_fname, char *ref_fname, char **tile_ids, int *n_tile)
{
  int st, iarg;

  *n_tile = 0;
  ref_fname[0] = '\0';
  in_fname[0] = out_fname[0] = '\0';
  for (st=1, iarg=1; iarg<argc; iarg++)
  {
    if (is_arg_id(argv[iarg], "-of") == 0)
      get_arg_val(argv[iarg], out_fname);
    else if (is_arg_id(argv[iarg], "-ref") == 0)
      get_arg_val(argv[iarg], ref_fname);
    else if (is_arg_id(argv[iarg], "-tile") == 0)
      get_arg_val_arr(argv[iarg], tile_ids, n_tile);
    else if (argv[iarg][0] == '-')
      fprintf(stderr, "Ignoring unknown option %s\n", argv[iarg]);
    else strcpy(in_fname, argv[iarg]);
  }
  if (in_fname[0] == '\0')
  {
    fprintf(stderr, "Missing input filename . . \n");
    st = -1;
  }
  if (out_fname[0] == '\0')
  {
    fprintf(stderr, "Missing output filename . . \n");
    st = -1;
  }
  if ((*n_tile == 0) && (strlen(ref_fname) <= 0))
  {
    fprintf(stderr, "Missing reference L3 filename or tile id \n");
    st = -1;
  }
  if (st == 1)
  {
    if ((*n_tile > 0) && (strlen(ref_fname) > 0))
    {
      fprintf(stderr, "Using reference L3 file %s only\n", ref_fname);
      *n_tile = 0;
    }
  }
  return st;
}

int get_hdfeos(char *file_name, grid_mapinfo *h, char *grid_name)
{
  int st = 1, i;
  hid_t file_id, grid_id;
  int HDF_ver = 0, he5_fmt;

  h->ns = (long)(-1);
  h->nl = (long)(-1);
  h->ulCornerGrid[0] = (float64)(-999.0);
  h->ulCornerGrid[1] = (float64)(-999.0);
  h->lrCornerGrid[0] = (float64)(-999.0);
  h->lrCornerGrid[1] = (float64)(-999.0);
  h->projCode = (int)(-1);
  h->sphereCode = (int)(-1);
  for (i = 0; i < 13; i++)
    h->projParameters[i] = (float64)(-999.0);
  h->zoneCode = (int)(-1);
  h->originCode = (int)(-1);

  HDF_ver = HDF_version_check(file_name);

  if(HDF_ver == 4) he5_fmt = 0;
  else if(HDF_ver == 5) he5_fmt = 1;
  else
  {
    fprintf(stderr, "%s is neither HDF4 or HDF5.\n", file_name);
    st = -1;
  }

  if ((file_id = EOS_GDopen(file_name, 'R', he5_fmt)) == FAIL) 
  {
    fprintf(stderr, "Input file %s open failure (EOS_GDopen) in get_hdfeos\n", file_name);
    st = -1;
  }
  else
  {
    if ((grid_id = EOS_GDattach(file_id, grid_name, he5_fmt)) == FAIL)
    {
      fprintf(stderr, "Can't attach to %s (EOS_GDattach) in get_hdfeos\n", grid_name);
      st = -1;
    }
    else
    {
      if (EOS_GDgridinfo(grid_id, &h->ns, &h->nl, h->ulCornerGrid, h->lrCornerGrid, he5_fmt) == FAIL)
      {
        fprintf(stderr, "Can't get grid info (EOS_GDgridinfo) in get_hdfeos\n");
        st = -1;
      }
      
      if (EOS_GDprojinfo(grid_id, &h->projCode, &h->zoneCode, &h->sphereCode, h->projParameters, he5_fmt) == FAIL)
      {
        fprintf(stderr, "Can't get projection info (EOS_GDprojinfo) in get_hdfeos\n");
        st = -1;
      }
  
      if (EOS_GDorigininfo(grid_id, &h->originCode, he5_fmt) == FAIL)
      {
        fprintf(stderr,  "Can't get origin defintion (EOS_GDorigininfo) in get_hdfeos\n");
        st = -1;
      }
  
      if (EOS_GDdetach(grid_id, he5_fmt) == FAIL)
        fprintf(stderr, "Can't detach from grid (EOS_GDdetach) in get_hdfeos\n");
    }
  
    if (EOS_GDclose(file_id, he5_fmt) == FAIL)
      fprintf(stderr, "Can't close file (EOS_GDclose) in get_hdfeos\n");
  }
  return st;
}
      
void add_hdfeos_meta(char *file_name, char *grid_name, grid_mapinfo *h, sds_t *sds_info, int nsds)
{
  int isds;
  hid_t gfid, gid, EOS_data_type;
  char dim_str[20];
  int he5_fmt = 0; /*so far, output to HDFEOS2 */

  if ((gfid = EOS_GDopen(file_name, 'C', he5_fmt)) == FAIL)
    fprintf(stderr, "Can't open HDF file %s (EOS_GDopen) in add_hdfeos_meta\n", file_name);
  else
  {
    h->ns = sds_info[0].dim_size[0];
    h->nl = sds_info[0].dim_size[1];
    if ((gid = EOS_GDcreate(gfid, grid_name, h->ns, h->nl, h->ulCornerGrid, h->lrCornerGrid, he5_fmt)) == FAIL)
      fprintf(stderr, "Can't open output file (EOS_GDcreate) in add_hdfeos_meta\n");
    else
    {
      if (EOS_GDdefproj(gid, h->projCode, h->zoneCode, h->sphereCode, h->projParameters, he5_fmt) == FAIL)
        fprintf(stderr, "Cannot define projection parameter (EOS_GDdefproj) in add_hdfeos_meta\n");
    }

    for (isds=0; isds<nsds; isds++)  
    {
      if (sds_info[isds].rank == 2) 
        strcpy(dim_str, "YDim,XDim");
      else if (sds_info[isds].rank == 3) 
        strcpy(dim_str, "ZDim,YDim,XDim");
      hid_t EOS_data_type;
      if(he5_fmt == 0)
        EOS_data_type = (hid_t)sds_info[isds].data_type;
      else if(he5_fmt == 1)
      {
	if(convert_datatype_HDF4toHDFEOS5(sds_info[isds].data_type, &EOS_data_type) == FAIL)
	  fprintf(stderr, "Error in calling convert_datatype_HDF4toHDFEOS5()!!!\n");
      }

      if (EOS_GDdeffield(gid, sds_info[isds].name, dim_str, EOS_data_type, he5_fmt) == FAIL)
        fprintf(stderr, "Error defining the SDS field %s\n", sds_info[isds].name);
    }
  
    if (EOS_GDdetach(gid, he5_fmt) == FAIL)
      fprintf(stderr, "Error detaching from grid %s (EOS_GDdetach) in add_hdfeos_meta\n", grid_name);

    if (EOS_GDclose(gfid, he5_fmt) == FAIL)
      fprintf(stderr, "Cannot close file %s (EOS_GDclose) in add_hdfeos_meta() \n", file_name);
  }
}

void copy_an_sds_data_attr(sds_t *in_sds_info, sds_t *out_sds_info) 
{
    /* Maosheng Zhao, SSAI, 5/5/2018
     * modified from copy_an_sds from copy_sds.c.
     * This is a more general and quick function to copy an sds.
     * Just make sure that in and out files are open. e.g., sd_d
     * != -1 and sds_info name is known. */
    void *sds_data;
    int i;
    int32 start[4], edge[4];
    int irow, icol;
    int nrows, ncols;
    int GROWS, last_rows;
    int nlines_a_copy = 20;
    int nattrs=22;
    
    out_sds_info->sds_id = in_sds_info->sds_id = FAIL;
    if (get_sds_info((char *)NULL, in_sds_info) != FAIL)
    {
        out_sds_info->rank = in_sds_info->rank;
        for (i=0; i<in_sds_info->rank; i++)
          out_sds_info->dim_size[i] = in_sds_info->dim_size[i]; 

        out_sds_info->data_type = in_sds_info->data_type;
        out_sds_info->data_size = in_sds_info->data_size;

        if (open_sds((char *)NULL, out_sds_info, 'W') != FAIL)
        {
	    fprintf(stderr, "copying SDS %s\n", in_sds_info->name);
            for (i=0; i<in_sds_info->rank; i++)
            {
                start[i] = 0;
                edge[i] = in_sds_info->dim_size[i];
            }
            if (in_sds_info->rank != 1) 
	    {
	      if(in_sds_info->dim_size[in_sds_info->row_id] < nlines_a_copy) nlines_a_copy = in_sds_info->dim_size[in_sds_info->row_id];
              edge[in_sds_info->row_id] = 1 * nlines_a_copy;
	    }
            nrows = (in_sds_info->rank == 1) ? 1 : in_sds_info->dim_size[in_sds_info->row_id];
            /*nrows = in_sds_info->dim_size[in_sds_info->row_id];*/
            ncols = in_sds_info->dim_size[in_sds_info->row_id + 1];
            
            write_all_sds_attr(in_sds_info->H5IN, in_sds_info->sds_id, out_sds_info->H5OUT, out_sds_info->sds_id, nattrs);

	    /* MSZ 1/25/2018 */
	    GROWS = (int)ceil((double)nrows / (double)nlines_a_copy);
	    last_rows = nrows % nlines_a_copy;

	    /* MSZ, 1/25/2018, to speedup, copy lines by lines not line by line */
            for (irow=0; irow<GROWS; irow++)
            {
		start[in_sds_info->row_id] = irow * nlines_a_copy;

		/* MSZ, 1/25/2018, be cautious about the last group rows because it may be different from nlines_a_copy */
		if(last_rows > 0 && irow == GROWS - 1)
		   edge[in_sds_info->row_id] = last_rows;

                if ((sds_data = get_sel_sds_data(in_sds_info, start, edge)) == NULL)
                  fprintf(stderr, "Error reading lines from %d to %d from SDS %s\n", irow * nlines_a_copy, irow * nlines_a_copy + nlines_a_copy, in_sds_info->name);
                else
                {
                    if (set_sel_sds_data(out_sds_info, start, edge, sds_data) == FAIL)
                      fprintf(stderr, "Error writing lines %d to %d to SDS %s\n", start[in_sds_info->row_id], start[in_sds_info->row_id] + edge[in_sds_info->row_id] - 1, out_sds_info->name);

		    if(sds_data) free(sds_data);
                }
            }
        }    
	close_sds(in_sds_info, in_sds_info->H5IN);
	close_sds(out_sds_info, out_sds_info->H5OUT);
    }
}

int read_param_table(char **tile_id, int n_tile, grid_mapinfo *h)
{
  FILE *fp;
  int d1, d2, d3, d4;
  int nt_fnd, c, i, id;
  double c1, c2, c3, c4, f[13];
  char *anc_path, in_tile_id[10];
  char param_fname[MAX_PATH_LENGTH];

  anc_path = getenv("ANCPATH");
  sprintf(param_fname, "%s/%s", anc_path, PARAM_FNAME);
  if ((fp = fopen(param_fname, "r")) == NULL)
    fprintf(stderr, "Cannot open projection parameter file %s\n", param_fname);
  else
  {
    nt_fnd = 0;
    c = fgetc(fp);
    while ((nt_fnd < n_tile) && (c != EOF))
    {
      ungetc(c, fp);
      fscanf(fp, "%s", in_tile_id);
      fscanf(fp, "%lf %lf %lf %lf", &c1, &c2, &c3, &c4);
      fscanf(fp, "%d %d %d %d", &d1, &d2, &d3, &d4);
      for (i=0; i<13; i++)
        fscanf(fp, "%lf", &f[i]);  
      for (id=0; id<n_tile; id++)
      {
        if (strcmp(in_tile_id, tile_id[id]) == 0)
        {
	  fprintf(stderr, "Found tile %d: %s\n", nt_fnd, tile_id[id]);
          if (nt_fnd == 0)
	  {
            h->ulCornerGrid[0] = c1; h->ulCornerGrid[1] = c2;
            h->lrCornerGrid[0] = c3; h->lrCornerGrid[1] = c4;
	    h->projCode = d1; h->sphereCode = d2;
	    h->zoneCode = d3; h->originCode = d4;
            for (i=0; i<13; i++)
              h->projParameters[i] = f[i];
	  }
	  else
	  {
	    if (c1 < h->ulCornerGrid[0]) h->ulCornerGrid[0] = c1;
            if (c2 > h->ulCornerGrid[1]) h->ulCornerGrid[1] = c2;
            if (c3 > h->lrCornerGrid[0]) h->lrCornerGrid[0] = c3;
	    if (c4 < h->lrCornerGrid[1]) h->lrCornerGrid[1] = c4;
	  }
	  ++nt_fnd;
        }
      }
      c = fgetc(fp);
    }
    fclose(fp);
  }
  if (nt_fnd != n_tile) return -1;
  else return 1;
}


