/****************************************************************************
!C

!File: create_mask.c

!Description:
  This file contains the routines for creating a 2D mask SDS 

!Input Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original March 1999. QA Tool Version 1.2
  Updated July 1999. Handle 3rd or 4th dimension in the input SDS 

  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "mask_sds_lib.h"

#define OFF_VAL 0
#define ON_VAL 255

#define HELP \
"NAME \n" \
"	create_mask - create an SDS that can be used as a mask \n" \
" \n" \
"SYNOPSIS \n" \
"       create_mask [-help] \n" \
"       create_mask -mask=<mask string> [-on=<output ON value> [-off=<output OFF value>] \n" \
"	-of=<filename> \n" \
" \n" \
"       echo -mask=<mask string> [-on=<output ON value>] [-off=<output OFF value>] \n" \
"	-of=<filename> | create_mask \n" \
" \n" \
" or    create_mask < argument file \n" \
" \n" \
"DESCRIPTION \n" \
"	Create a masking SDS using the user specified masking logic. Output is an \n" \
"	HDF file containing the mask SDS. Output SDS pixel values are set to user \n" \
"	specified ON value where mask logic is true and is set to OFF value where \n" \
"	mask logic is false. Handles 2D/3D/4D input SDS(s) and L2G files. Output is \n" \
"	always a 2D SDS. \n" \
" \n" \
"OPTIONS \n" \
"	-help                  Display this help message \n" \
"	-mask=string           Define the mask to apply as combination of individual \n" \
"	                       masks separated by AND or OR. Each individual mask \n" \
"	                       consists of a file, a SDS, a bit descriptor, equality \n" \
"	                       operator (<, >, <=, >=, !=, ==) and a binary bit value \n" \
"	                       string to match: \n" \
"	                         filename,SDS_name,bits=string. \n" \
"	                       Rather than duplicate the file and SDS names, the * \n" \
"	                       character can be used after initial specification. \n" \
"	                       The bit descriptor is a list of bits or ranges of bits, \n" \
"	                       similar to the Unix filename substitution. \n" \
"	                         Example: -mask=file1,SDS1,0-2,4==0101,AND,*,SDS2,2,3==10 \n" \
"	                       Warning: Although a bit range in the bit descriptor is \n" \
"	                       specified by the lower bit followed by the higher bit, \n" \
"	                       bit value string are in reverse order, i.e. the lower \n" \
"	                       bit on the right (in the example 0-2,4==0101 is for \n" \
"	                       4,2,1,0==0101). If the bit descriptor is omitted, then \n" \
"	                       the whole range of bits is assumed, and the value string \n" \
"	                       is parsed as a decimal value. This is a convenient way \n" \
"	                       to refer to a specific value, instead of a list of bits. \n" \
"	                         Example: -mask=file,SDS,==-200 \n" \
"	                       In case of 3D and 4D input SDS user must specify the layer \n" \
"	                       numbers in the 3rd and 4th dimension by entering the layer \n" \
"	                       number as dot extension in the SDS name within the -mask option \n" \
"	                         Example -mask=file,SDS.n1.m2,0-2==011,AND,file.SDS.n1,0-1==11 \n" \
"	                       In case of L2G specify the observation number as dot extension \n" \
"	                       in the SDS name (without the _1 or _c) in the -mask option.\n" \
"	                         Example -mask=file,QC_500m.2,10-13=1100\n" \
"	-on=<ON value>         Output SDS value where mask is true (default 255) \n" \
"	-off=<OFF value>       Output SDS value where mask is false (default 0) \n" \
"	-of=filename           Output filename \n" \
" \n" \
"Examples: \n" \
"	create_mask -mask=MOD09A1.A1996214.h12v04.002.hdf,sur_refl_state_500m,3-5==001 -of=land.hdf \n" \
"	create_mask -mask=\"MOD09A1.A1996214.h12v04.002.hdf,sur_refl_state_500m,3-5==001,AND,*,*,01==00\" -of=land_clr.hdf \n" \
"	create_mask -mask=MOD35_L2.A1996213.1024.002.hdf,Cloud_Mask.1,1-2==00 -of=mod35_cld.hdf \n" \
"	create_mask -mask=\"MOD35_L2.A1996213.1024.002.hdf,Cloud_Mask.1,6-7==11,AND,*,*,1-2==00\" -of=mod35_land_cld.hdf \n" \
"	create_mask -mask=\"MOD35_L2.A1996213.1024.002.hdf,Cloud_Mask.1,6-7==11,OR,*,*,1-2==00\" -of=mod35_land_cld.hdf \n" \
"	create_mask -mask=MODAGAGG.A1996214.h12v04.001.hdf,Band_QC.1.1,2-5==1100 -of=agg_b01_obs1_qc.hdf \n" \
"	create_mask -mask=\"MODAGAGG.A1996214.h12v04.001.hdf,Band_QC.1.1,2-5==1100,AND,*,Band_QC.2.1,*\" -of=agg_b01_obs1-2_qc.hdf \n" \
"	create_mask -mask=\"MODAGAGG.A1996214.h12v04.001.hdf,Band_QC.1.1,2-5==1100,AND,*,Aggregate_QC.1.1,3-5==001\" -of=agg_b01_obs1_land_qc.hdf \n" \
"	create_mask -mask=\"MOD09GHK.A1996214.h12v04.001.hdf,QC_500m.2,10-13==1100\" -of=l2g_test.hdf \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n\n" \
"When bit numbers are used in the mask option bit values must be in binary. All relational \n" \
"operators can be used with the interpreation of decimal value for the binary string. \n" \
"Please report problems to Sadashiva Devadiga (devadiga@ltpmail.gsfc.nasa.gov) \n"

#define USAGE \
"usage:         create_mask [-mask=<mask string>] [-on=<output ON value>] \n" \
"		[-off=<output OFF value>] -of=<filename> \n" \
" \n" \
"OPTIONS \n" \
"	-help                  Display this help message \n" \
"	-mask=string           Define the mask to apply as combination of individual \n" \
"	                       masks separated by AND or OR. Each individual mask \n" \
"	                       consists of a file, a SDS, a bit descriptor, and a \n" \
"	                       binary bit value string to match: \n" \
"	                         filename,SDS_name,bits=string. \n" \
"	                       Rather than duplicate the file and SDS names, the * \n" \
"	                       character can be used after initial specification. \n" \
"	                       The bit descriptor is a list of bits or ranges of bits, \n" \
"	                       similar to the Unix filename substitution. \n" \
"	                       Example: -mask=file1,SDS1,0-2,4==0101,AND,*,SDS2,2,3==10 \n" \
"	                       Warning: Although a bit range in the bit descriptor is \n" \
"	                       specified by the lower bit followed by the higher bit, \n" \
"	                       bit value string are in reverse order, i.e. the lower \n" \
"	                       bit on the right (in the example 0-2,4==0101 is for \n" \
"	                       4,2,1,0==0101). If the bit descriptor is omitted, then \n" \
"	                       the whole range of bits is assumed, and the value string \n" \
"	                       is parsed as a decimal value. This is a convenient way \n" \
"	                       to refer to a specific value, instead of a list of bits. \n" \
"	                         Example: -mask=file,SDS,==-200 \n" \
"	                       In case of 3D and 4D input SDS user must specify the layer \n" \
"	                       numbers in the 3rd and 4th dimension by entering the layer \n" \
"	                       number as dot extension in the SDS name within the -mask option \n" \
"	                         Example -mask=file,SDS.n1.m2,0-2==011,AND,file.SDS.n1,0-1==11 \n" \
"	-on=<ON value>         Output SDS value where mask is true (default 255) \n" \
"	-off=<OFF value>       Output SDS value where mask is false (default 0) \n" \
"	-of=filename           Output filename \n" \
" \n"

int parse_cmd_create_mask(int argc, char **argv, char *mask_str, char *out_fname, 
		int *on_val, int *off_val, int *H5OUT);
void generate_mask(char *m_str, char *out_fname, int on_val, int off_val, int H5OUT);

int main(int argc, char **argv)
{
  int st, stdin_argc;
  int on_val, off_val;
  int H5OUT;
  char mask_str[MAX_STR_LEN];
  char out_fname[MAX_PATH_LENGTH];
  char **stdin_argv;

  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }

  /* MSZ 2/6/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/6/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  /*
  if (argc == 1)
  {
    if ((stdin_argv = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for stdin_argv in create_mask: main()\n");
      exit(1);
    }
    if (parse_stdin(&stdin_argc, stdin_argv) != -1)
      st = parse_cmd_create_mask(stdin_argc, stdin_argv, mask_str, out_fname, &on_val, &off_val, &H5OUT);
    else st = 0;
    Free2D((void **)stdin_argv);
  }
  else
  */
  /* MSZ 4/3/2018, comment out the above and change the else to if(argc > 2) below */
  if(argc > 2)
    st = parse_cmd_create_mask(argc, argv, mask_str, out_fname, &on_val, &off_val, &H5OUT);

  if (st == -1)
    fprintf(stderr, "%s\n", USAGE);
  else if (st != 0)
    generate_mask(mask_str, out_fname, on_val, off_val, H5OUT);
  return 0;
}

int parse_cmd_create_mask(int argc, char **argv, char *mask_str, char *out_fname, int *on_val, 
	int *off_val, int *H5OUT)
{
  int i, st;
  char on_val_str[20];
  char off_val_str[20];

  st = 1;
  *on_val = ON_VAL;
  *off_val = OFF_VAL;
  mask_str[0] = out_fname[0] = '\0';
  on_val_str[0] = off_val_str[0] = '\0';
  *H5OUT = 0; /* Default is to assume that both I/O will be in hdf4 */
  
  for (i=1; i<argc; i++) {
    if ((strstr(argv[i], "-mask=") != NULL) || (strstr(argv[i], "-m=") != NULL)) 
	get_arg_val(argv[i], mask_str);
    else if (strstr(argv[i], "-on=") != NULL) get_arg_val(argv[i], on_val_str);
    else if (strstr(argv[i], "-off=") != NULL) get_arg_val(argv[i], off_val_str);
    else if (strstr(argv[i], "-of=") != NULL)
    {
        get_arg_val(argv[i], out_fname);
        if ((((strcasestr(out_fname,".h5") != '\0')) ||
            ((strcasestr(out_fname,".he5") != '\0')) ||
            ((strcasestr(out_fname,".hdf5") != '\0')) ||
            ((strcasestr(out_fname,".nc") != '\0'))))
             *H5OUT = 1;
        else
             *H5OUT = 0;
    }
    else fprintf(stderr, "Igonoring invalid argument %s\n", argv[i]);
  }
  if ((mask_str[0] == '\0') || (out_fname[0] == '\0')) st = -1;
  if (mask_str[0] == '\0') fprintf(stderr, "Missing input mask_str . . \n");
  if (out_fname[0] == '\0') fprintf(stderr, "Missing output filename . . \n");
  if (st == 1) 
  {
    *off_val = (int)atoi(off_val_str);
    *on_val = (int)atoi(on_val_str);
    if (*off_val == *on_val)
    {
      *off_val = OFF_VAL; *on_val = ON_VAL;
      if ((strlen(on_val_str) > 0) || (strlen(off_val_str) > 0))
        fprintf(stderr, "Output mask SDS values (ON and OFF) invalid: %s %s\n", on_val_str, off_val_str);
      fprintf(stderr, "Ouput SDS values (ON and OFF) set to default: %d %d\n", *on_val, *off_val);
    }
    else if ((*off_val < OFF_VAL) || (*off_val > ON_VAL) || (*on_val < OFF_VAL) || (*on_val > ON_VAL))
    {
      *off_val = OFF_VAL; *on_val = ON_VAL;
      fprintf(stderr, "Output mask SDS values (ON and OFF) invalid: %s %s\n", on_val_str, off_val_str);
      fprintf(stderr, "Ouput SDS values (ON and OFF) set to default: %d %d\n", *on_val, *off_val);
    }
  }
  return st;
}

void generate_mask(char *m_str, char *out_fname, int on_val, int off_val, int H5OUT)
{
    int st = 1;
    uint8 *mask_row;
    int i, j, p1;
    int rank, len, irow;
    int n_op, i_op, j_op;
    int obs_num[MAX_NUM_OP];
    void *data_qa[MAX_NUM_OP];
    int32 edge[4] = {0, 0};
    int32 start[4] = {0, 0};
    int32 *data_qa_nadd[MAX_NUM_OP];
    char sdsi_name[MAX_SDS_NAME_LEN];
    char sdsj_name[MAX_SDS_NAME_LEN];
    char num_str[5], **mask_str, **qa_fnames;
    sds_t out_sds_info, qa_sds_nobs_info[MAX_NUM_OP];
    sds_t qa_sds_info[MAX_NUM_OP], qa_sdsc_info[MAX_NUM_OP];
    int val_opt[MAX_NUM_OP], sel_qa_op[MAX_NUM_OP], fqa_l2g[MAX_NUM_OP];
    int res_s[MAX_NUM_OP], res_l[MAX_NUM_OP], rel_op[MAX_NUM_OP];
    unsigned long bit_mask_arr[MAX_NUM_OP], mask_val_arr[MAX_NUM_OP];
  
    if ((mask_str = (char **)Calloc2D(4*MAX_NUM_OP, 2*MAX_PATH_LENGTH, sizeof(char))) == NULL)
      fprintf(stderr, "Cannot allocate memory for mask_str generate_mask\n");
    else n_op = get_mask_string(m_str, mask_str, val_opt, fqa_l2g);
  
    if ((qa_fnames = (char **)Calloc2D(MAX_NUM_OP, MAX_PATH_LENGTH, sizeof(char))) == NULL)
      fprintf(stderr, "Cannot allocate memory for qa_fnames generate_mask\n");
    else 
      st = get_parameters(mask_str, n_op, sel_qa_op, qa_fnames, qa_sds_info, bit_mask_arr, 
              mask_val_arr, val_opt, rel_op);
    if ((n_op != -1) && (st != -1))
      st = get_qa_sds_info(qa_fnames, qa_sds_info, qa_sdsc_info, fqa_l2g, n_op);
    if (st != -1) 
    {
        if ((st = get_res_factors(&qa_sds_info[0], qa_sds_info, n_op, res_l, res_s)) != -1)
        {
            strcpy(out_sds_info.name, "Mask_sds");
            out_sds_info.data_type = DFNT_UINT8;
	    out_sds_info.H5OUT = H5OUT;                /* so far, output to HDF4 */
            out_sds_info.sd_id = out_sds_info.sds_id = -1;
            rank = qa_sds_info[0].rank;
            out_sds_info.rank = 2; 
            if ((rank == 2) || (qa_sds_info[0].dim_size[0] > qa_sds_info[0].dim_size[rank-2]))
            {
                out_sds_info.dim_size[0] = qa_sds_info[0].dim_size[0];
                out_sds_info.dim_size[1] = qa_sds_info[0].dim_size[1];
            }
            else
            {
                out_sds_info.dim_size[0] = qa_sds_info[0].dim_size[rank-2];
                out_sds_info.dim_size[1] = qa_sds_info[0].dim_size[rank-1];
            }
            if (open_sds(out_fname, &out_sds_info, 'W') != -1)
            {
                if ((mask_row = (uint8 *)calloc(out_sds_info.dim_size[1], sizeof(uint8))) == NULL)
                  fprintf(stderr, "Cannot allocate memory for mask_row in generate_mask\n");
                st = open_qa_sds_nsds((char *)NULL, (sds_t *)NULL, (sds_t *)NULL, (sds_t *)NULL, 1, 
                    qa_fnames, qa_sds_info, qa_sdsc_info, qa_sds_nobs_info, fqa_l2g, n_op);
                if (st != -1)
                  st = malloc_qa_sds(qa_sds_info, n_op, fqa_l2g, data_qa, data_qa_nadd);
        
                if ((st != -1) && (mask_row != NULL))
                {
                  for (i_op=0; i_op<=n_op; i_op++)
                    if (fqa_l2g[i_op] == 1)
                    {
                      len = (int)strlen(qa_sdsc_info[i_op].name);
                      p1 = charpos(qa_sdsc_info[i_op].name, '.', 0);
                      strmid(qa_sdsc_info[i_op].name, p1+1, len-p1-1, num_str);
                      obs_num[i_op] = (int)atoi(num_str);
                    }
                    else obs_num[i_op] = 1;
        
              edge[0] = 1; edge[1] = out_sds_info.dim_size[1];
                  for (irow=0; irow<out_sds_info.dim_size[0]; irow++)
                  {
                start[0] = irow;
                    read_qa_sds(qa_sds_info, qa_sdsc_info, qa_sds_nobs_info, n_op, data_qa, data_qa_nadd,
                                    irow, res_l, fqa_l2g, obs_num);
                    process_mask_data(data_qa, out_sds_info.dim_size[1], qa_sds_info, n_op, sel_qa_op,
                bit_mask_arr, mask_val_arr, rel_op, res_s, mask_row, on_val, off_val);
	
		if(set_sel_sds_data(&out_sds_info, start, edge, mask_row) == FAIL)	
                  fprintf(stderr, "Error writing a line of data to output SDS in generate_mask\n");
                  } /* for (irow=0; . . . */
                }
        
                /* close all SDS and HDF files */
                for (i_op=0; i_op<=n_op; i_op++)
                  if (fqa_l2g[i_op] == 1)
                  {
                    for (j_op=0; j_op<i_op; j_op++)
                      if ((qa_sds_nobs_info[i_op].sd_id == qa_sds_nobs_info[j_op].sd_id) &&
                                      (qa_sds_nobs_info[i_op].sds_id == qa_sds_nobs_info[j_op].sds_id))
			;
                      if (j_op >= i_op) close_sds(&qa_sds_nobs_info[i_op], qa_sds_nobs_info[i_op].H5IN);
                  }
        
                for (i_op=0; i_op<=n_op; i_op++)
                  if (fqa_l2g[i_op] == 1)
                  {
                    for (j_op=0; j_op<i_op; j_op++)
                      if (qa_sds_info[i_op].sd_id == qa_sds_info[j_op].sd_id) break;
                    if (j_op >= i_op) free(data_qa_nadd[i_op]);
                  }
        
                close_qa_hdf((char *)NULL, (sds_t *)NULL, qa_fnames, qa_sds_info, n_op);
                if (out_sds_info.sds_id != -1) close_sds(&out_sds_info, out_sds_info.H5OUT);
                if (out_sds_info.sd_id != -1) close_sd(&out_sds_info, out_sds_info.H5OUT);
                if (mask_row != NULL) free(mask_row);
                if (data_qa[0] != NULL) free(data_qa[0]);
                for (i=1; i<=n_op; i++)
                {
                  for (j=0; j<i; j++)
              {
                    if ((p1 = charpos(qa_sds_info[i].name, '.', 0)) == -1)
                      strcpy(sdsi_name, qa_sds_info[i].name);
                    else
                      strmid(qa_sds_info[i].name, 0, p1, sdsi_name);
                    if ((p1 = charpos(qa_sds_info[j].name, '.', 0)) == -1)
                      strcpy(sdsj_name, qa_sds_info[j].name);
                    else
                      strmid(qa_sds_info[j].name, 0, p1, sdsj_name);
                    if ((strcmp(qa_fnames[i], qa_fnames[j]) == 0) && (strcmp(sdsi_name, sdsj_name) == 0)) 
                  break;
              }
                  if (j >= i)
                  {
                    if (data_qa[i] != NULL) free(data_qa[i]);
                  }
                } /* for (i=1; . . . */
            }
        }
    }
    Free2D((void **)mask_str);
    Free2D((void **)qa_fnames);
}


