#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/dir.h>                        

#include "hdf5.h"
#include "mfhdf.h"
#include "qa_tool.h"
#include "sds_rw.h"
#include "str_op.h"
#include "meta.h"
#include "alloc_mem.h"
#include "main_util.h"                     

#define MAX_NUM_REG 255 
#define FILL_VALUE 255 
#define DATA_TYPE 21 

typedef struct {
  int r_min;
  int r_max;
  int c_min;
  int c_max;
  int g_val;
} IMG_REG;

#define HELP \
"NAME \n" \
"       create_reg_img - Cretae HDF image with blocks of gray values \n" \
" \n" \
"SYNOPSIS \n" \
"       create_reg_img [-help] \n" \
"       create_reg_img [-sds=<sds_name>] [-sz=<row,col>] -of=<output filename> \n" \
"                      -reg=r_min,r_max,c_min,c_max,val [-dt=<data type>] [-fill=<fill value>] \n" \
" \n" \
"       echo [-sds=<sds_name>] [-rows=<nrows>] [-cols=<ncols>] -of=<output filename> \n" \
"                     -reg=r_min,r_max,c_min,c_max,val [-dt=<data type>] [-fill=<fill value>] | create_reg_img \n" \
" \n" \
"DESCRIPTION \n" \
"       Create an image containing many regions of user specified size and gray values. User can \n" \
"       specify upto a maximum of 255 regions bounded by minimum and maximum row column numbers. \n" \
" \n" \
"OPTIONS \n" \
"       -help                       Display this help message \n" \
"       -sds=<SDS name>             Ouput SDS name (default: Sample_SDS) \n" \
"       -sz=<rows,cols>             Image size in number of rows and columns. \n" \
"                                   (default: 1200,1200) \n" \
"       -of=<filename>              Output filename \n" \
"       -reg=<region size and val>  Regions with bounding box and gray values. \n" \
"                                      Example: -reg=100,149,100,149,25 \n" \
"                                   The option can be repeated for as many times as \n" \
"                                   required to specify mutliple regions within the \n" \
"                                   same SDS \n" \
"                                   Regions can be entered in a file and the filename can \n" \
"                                   be input as argument value. Each region is entered in\n" \
"                                   a separate line in the same format as above.\n" \
"                                   (e.g. a line in the file: 100,149,100,149,25)\n" \
"                                      Example: -reg=img_reg.txt \n" \
"       -dt=<data type>             Data type (CHAR8|UCHAR8|INT8|UINT8|INT16|UINT16|INT32|UINT32) \n" \
"                                   (default: UINT8) \n" \
"       -fill=<fill value>          SDS fill value (default: 255) \n" \
" \n" \
"Examples:\n" \
"\tcreate_reg_img -sds=location_test -sz=2400,2400 -of=reg_img.hdf -reg=100,149,100,149,25 -reg=200,299,500,749,100\n" \
"\tcreate_reg_img -sds=location_test -sz=2400,2400 -of=reg_img.hdf -reg=reg_img.data\n" \
"\n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:\t[-sds=<sds_name>] [-sz=<row,col>] -of=<output filename> -reg=r_min,r_max,c_min,c_max,val [-dt=<data type>] [-fill=<fill value>]\n" \
" \n" \
"OPTIONS \n" \
"       -help                       Display this help message \n" \
"       -sds=<SDS name>             Ouput SDS name (default: Sample_SDS) \n" \
"       -sz=<rows,cols>             Image size in number of rows and columns. \n" \
"                                   (default: 1200,1200) \n" \
"       -of=<filename>              Output filename \n" \
"       -reg=<region size and val>  Regions with bounding box and gray values. \n" \
"                                      Example: -reg=100,149,100,149,25 \n" \
"                                   The option can be repeated for as many times as \n" \
"                                   required to specify mutliple regions within the \n" \
"                                   same SDS \n" \
"                                   Regions can be entered in a file and the filename can \n" \
"                                   be input as argument value. Each region is entered in\n" \
"                                   a separate line in the same format as above.\n" \
"                                      Example: -reg=img_reg.txt \n" \
"       -dt=<data type>             Data type (CHAR8|UCHAR8|INT8|NINT8|INT16|UINT16|INT32|UINT32) \n" \
"       -fill=<fill value>          SDS fill value \n" \
" \n" \
" \n"                                                                                      

int parse_cmd_create_reg_img(int argc, char **argv, char *out_fname, char *sds_name, int *nrows, int *ncols, 
	IMG_REG *img_reg, int *n_reg, int *dt, int *fill_value);
void create_image(sds_t *sds_info, IMG_REG *img_reg, int n_reg);

int main(int argc, char **argv)
{
  sds_t sds_info;
  IMG_REG *img_reg;
  int dt, fill_val, nrows, ncols, n_reg;
  char sds_name[MAX_SDS_NAME_LEN];
  char out_fname[MAX_PATH_LENGTH];

  /* MSZ 2/6/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/6/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
    fprintf(stderr, "%s\n", HELP);
  else
  {
    if ((img_reg = (IMG_REG *)calloc(MAX_NUM_REG, sizeof(IMG_REG))) ==NULL)
      fprintf(stderr, "Cannot allocate memory for img_reg in main() \n");
    else 
    { 
      if (parse_cmd_create_reg_img(argc, argv, out_fname, sds_name, &nrows, &ncols, img_reg, &n_reg, &dt, &fill_val) == -1)
        fprintf(stderr, "%s\n", USAGE);
      else
      {
	sds_info.H5OUT = 0;  /* so far, output to HDF4 */
	sds_info.sd_id = sds_info.sds_id = -1;
	if(open_sd(out_fname, &sds_info, 'W') == FAIL)
	  fprintf(stderr, "Error create the ouptut HDF file %s\n", out_fname);
	else 
	{
	  sds_info.rank = 2;
	  sds_info.sds_id = -1;
	  sds_info.dim_size[0] = nrows;
	  sds_info.dim_size[1] = ncols;
	  switch(dt) {  
            case 20: sds_info.data_type = DFNT_INT8; break;
            case 21: sds_info.data_type = DFNT_UINT8; break;
            case 22: sds_info.data_type = DFNT_INT16; break;
            case 23: sds_info.data_type = DFNT_UINT16; break;
            case 24: sds_info.data_type = DFNT_INT32; break;
            case 25: sds_info.data_type = DFNT_UINT32; break;
          }
	  sds_info.fill_val = fill_val; 
	  strcpy(sds_info.name, sds_name);
          sds_info.data_size = DFKNTsize(sds_info.data_type);
	  create_image(&sds_info, img_reg, n_reg);
	  close_sd(&sds_info, sds_info.H5OUT);
        }
      }
      free(img_reg);
    }
  }
  return 0;
}

int parse_cmd_create_reg_img(int argc, char **argv, char *out_fname, char *sds_name, int *nrows, int *ncols, 
	IMG_REG *img_reg, int *n_reg, int *dt, int *f_val)
{
  FILE *fp;
  int ln, st, ireg;
  int i, tmp_num;
  char **tmp_str;
  char tmp1_str[20];

  ireg = 0;
  *dt = DATA_TYPE;
  *f_val = FILL_VALUE;
  *nrows = *ncols = 1200;
  out_fname[0] = sds_name[0] = '\0';
  if ((tmp_str = (char **)Calloc2D(MAX_NUM_REG, 50, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for tmp_str in parse_cmd_create_reg_img \n" );
  else
  {
    for (i=1, st=1; i<argc; i++)
    {
      if ((is_arg_id(argv[i], "-of=") == 0) || (is_arg_id(argv[i], "-o=") == 0))
        get_arg_val(argv[i], out_fname);
      else if (is_arg_id(argv[i], "-sds=") == 0)
        get_arg_val(argv[i], sds_name);
      else if (is_arg_id(argv[i], "-dt=") == 0)
      {
        get_arg_val(argv[i], tmp1_str);
	if (strcmp(tmp1_str, "INT8") == 0) { *dt = 20; } 
	else if (strcmp(tmp1_str, "UINT8") == 0) { *dt = 21; } 
	else if (strcmp(tmp1_str, "INT16") == 0) { *dt = 22; } 
	else if (strcmp(tmp1_str, "UINT16") == 0) { *dt = 23; } 
	else if (strcmp(tmp1_str, "INT32") == 0) { *dt = 24; } 
	else if (strcmp(tmp1_str, "UINT32") == 0) { *dt = 25; } 
      }
      else if (is_arg_id(argv[i], "-fill=") == 0)
      {
        get_arg_val(argv[i], tmp1_str);
        *f_val = atoi(tmp1_str);
      }
      else if (is_arg_id(argv[i], "-sz=") == 0)
      {
        tmp_num = 0;
        get_arg_val_arr(argv[i], tmp_str, &tmp_num);
        if (tmp_num != 2) {
          fprintf(stderr, "Error in image dimension option -sz: %s\n", argv[i]);
          st = -1;
        }
        else {
          *nrows = (int)atoi(tmp_str[0]);
          *ncols = (int)atoi(tmp_str[1]);
        }
      }
      else if (is_arg_id(argv[i], "-reg=") == 0)
      {
        tmp_num = 0;
        get_arg_val_arr(argv[i], tmp_str, &tmp_num);
	if (tmp_num == 1) {
	  if ((fp = fopen(tmp_str[0], "r")) != NULL) {
	    ln = 0;
	    while (fscanf(fp, "%s", tmp1_str) == 1) {
	      tmp_num = 0;
	      split_string(tmp1_str, tmp_str, &tmp_num);
              if (tmp_num != 5)
                fprintf(stderr, "Error in line %d in file %s\n", ln, tmp1_str);
	      else {
                img_reg[ireg].r_min = (int)atoi(tmp_str[0]);
                img_reg[ireg].r_max = (int)atoi(tmp_str[1]);
                img_reg[ireg].c_min = (int)atoi(tmp_str[2]);
                img_reg[ireg].c_max = (int)atoi(tmp_str[3]);
                img_reg[ireg].g_val = (int)atoi(tmp_str[4]);
	        if ((img_reg[ireg].r_min >= img_reg[ireg].r_max) || (img_reg[ireg].c_min >= img_reg[ireg].c_max))
	          fprintf(stderr, "Ignoring invalid region input at line %d in file %s\n", ln, tmp1_str);
	        else ireg++;
	      }
	      ln++;
	    }
	    fclose(fp);
	  }
	  else
	    fprintf(stderr, "Cannot open file %s\nIgnoring option %s\n", tmp_str[0], argv[i]);
	}
        else if (tmp_num != 5)
          fprintf(stderr, "Error in region option -reg: %s\n", argv[i]);
        else {
          img_reg[ireg].r_min = (int)atoi(tmp_str[0]);
          img_reg[ireg].r_max = (int)atoi(tmp_str[1]);
          img_reg[ireg].c_min = (int)atoi(tmp_str[2]);
          img_reg[ireg].c_max = (int)atoi(tmp_str[3]);
          img_reg[ireg].g_val = (int)atoi(tmp_str[4]);
	  if ((img_reg[ireg].r_min >= img_reg[ireg].r_max) || (img_reg[ireg].c_min >= img_reg[ireg].c_max))
	    fprintf(stderr, "Ignoring invalid region input: %s\n", argv[i]);
	  else ireg++;
        }
      }
      else if (argv[i][0] == '-') 
        fprintf(stderr, "Unknown option %s\n", argv[i]);
    }
    if (out_fname[0] == '\0') {
      st = -1;
      fprintf(stderr, "Missing output filename . .\n");
    }
    if (ireg == 0) {
      st = -1;
      fprintf(stderr, "No valid regions input . .\n");
    }
    if (st == 1)
    {
      *n_reg = ireg; 
      if (sds_name[0] == '\0')
        strcpy(sds_name, "Sample_sds");
    }
    Free2D((void **)tmp_str);
  }                   
  return st;
}

void create_image(sds_t *sds_info, IMG_REG *img_reg, int n_reg)
{
  void *sds_data;
  int ir, ic, ireg;
  int nrows, ncols;
  int ir_min, ic_min;
  int ir_max, ic_max, ig_val;
  int32 edge[2], start[2];

  nrows = sds_info->dim_size[0];
  ncols = sds_info->dim_size[1];
  if (open_sds((char *)NULL, sds_info, 'W') != FAIL)
  {
    write_attr_fval(sds_info->H5OUT, sds_info->sds_id, sds_info->data_type, 1, sds_info->fill_val);
    if ((sds_data = (void *)calloc(nrows*ncols, sds_info->data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memroy for sds_data in create_image()\n");
    else
    {
      for (ir=0; ir<nrows; ir++)
        for (ic=0; ic<ncols; ic++)
	  switch(sds_info->data_type) {	  
	    case 20: ((int8 *)sds_data)[ir*ncols + ic] = (int8)(sds_info->fill_val); break;
	    case 21: ((uint8 *)sds_data)[ir*ncols + ic] = (uint8)(sds_info->fill_val); break;
	    case 22: ((int16 *)sds_data)[ir*ncols + ic] = (int16)(sds_info->fill_val); break;
	    case 23: ((uint16 *)sds_data)[ir*ncols + ic] = (uint16)(sds_info->fill_val); break;
	    case 24: ((int32 *)sds_data)[ir*ncols + ic] = (int32)(sds_info->fill_val); break;
	    case 25: ((uint32 *)sds_data)[ir*ncols + ic] = (uint32)(sds_info->fill_val); break;
	  }
      for (ireg=0; ireg<n_reg; ireg++)
      {
        ir_min = img_reg[ireg].r_min;
        ir_max = img_reg[ireg].r_max;
        ic_min = img_reg[ireg].c_min;
        ic_max = img_reg[ireg].c_max;
	ig_val = img_reg[ireg].g_val;
        if ((ir_min < 0) || (ir_min >= nrows) || (ir_max < 0) || (ir_max >= nrows) || (ic_min < 0) || (ic_min >= ncols) || (ic_max < 0) || (ic_max >= ncols))
	  fprintf(stderr, "Ignoring invalid region: (%d,%d,%d,%d)\n", ir_min, ir_max, ic_min, ic_max);
	else
	{
          for (ir=ir_min; ir<=ir_max; ir++)
            for (ic=ic_min; ic<=ic_max; ic++)
	      switch(sds_info->data_type) {        
		case 20: ((int8 *)sds_data)[ir*ncols + ic] = ig_val; break;
	        case 21: ((uint8 *)sds_data)[ir*ncols + ic] = ig_val; break;
	        case 22: ((int16 *)sds_data)[ir*ncols + ic] = ig_val; break;
	        case 23: ((uint16 *)sds_data)[ir*ncols + ic] = ig_val; break;
	        case 24: ((int32 *)sds_data)[ir*ncols + ic] = ig_val; break;
	        case 25: ((uint32 *)sds_data)[ir*ncols + ic] = ig_val; break;
	      }
	}
      }
      start[0] = start[1] = 0;
      edge[0] = nrows; edge[1] = ncols;
      if (set_sel_sds_data(sds_info, start, edge, sds_data) == FAIL)
	fprintf(stderr, "Error writing SDS %s to output file\n", sds_info->name);
      free(sds_data);
    }
  }
  close_sds(sds_info, sds_info->H5OUT);
}
