/****************************************************************************
!C

!File: enlarge_sds.c

!Description:
  Contains routines for rescaling (dimension) an image/sds

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original June 1999 Version 1.2 
  Original July 1999 Version 1.2 
  Modified September 1999 (to include -meta option) 
  Modified July 2011 by Xiaoping Zhang, Sigma space corp.
  	(to add fill value option on command line)

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "qa_tool.h"
#include "sds_rw.h"
#include "str_op.h"
#include "main_util.h"
#include "alloc_mem.h"
#include "meta.h"

#define HELP \
"NAME \n" \
"     enlarge_sds - enlarge an sds \n" \
" \n" \
"SYNOPSIS \n" \
"	enlarge_sds -help \n" \
"	enlarge_sds -sf=<scale factor> -of=output_file [-meta] \n" \
"                [-sds=SDSname1[,SDSname2...]] -fill_value=<fillVal> filename \n" \
" \n" \
"	echo -sf=<scale factor> -of=output_file [-meta] \n" \
"		[-sds=SDSname1[,SDSname2...]] filename | enlarge_sds \n" \
" or  enlarge_sds < argument_file \n" \
" \n" \
"DESCRIPTION \n" \
"	Enlarge specified SDS(s) of an input HDF file. Pixel values at (i, j) in the input \n" \
"	SDS is duplicated in the region (sf*i, sf*j) (sf*i + sf - 1, sf*j + sf - 1) of the \n" \
"	output SDS. The output is an HDF file. Tool handles 2D/3D/4D SDS(s). \n" \
" \n" \
"OPTIONS \n" \
"	-help            Display this help message \n" \
"	-sf=factor       Scale factor (must be a nonzero integer) \n" \
"	-of=filename     Output filename \n" \
"	-meta            Copy metadata from input to output file \n" \
"	-fill_value=fillVal       User inputs fill value \n" \
"	-sds=<SDS list>  List of SDS to enlarge (separated by commas) \n" \
"	                 To enlarge a specific layer in a 3D or 4D SDS, specify the \n" \
"	                 layer number (1-based) as dot extension in the SDS name: \n" \
"	                 (sds_name.n.m)   Ex: sur_refl_b02.1 Surf_Refl.1.2 \n" \
"	                 sds_name.*.m will enlarge all 3D layers in the mth layer of \n" \
"	                 4th dimesion. A range of layers can be specified as in \n" \
"	                 sds_name.n1-n2.m. Multiple SDS layer numbers in the 3rd or \n" \
"	                 4th dimension cannot be separated by comma. Default is all \n" \
"	                 SDS, all layer, and retain original interleaving. \n" \
" \n" \
"Examples: \n" \
"       enlarge_sds -sds=\"500m Surface Reflectance Band 1\" -sf=2 -of=test.hdf ./DATA/MOD09/MOD09.A1996214.0811.002.hdf \n" \
"	enlarge_sds -sds=EV_500_Aggr1km_RefSB.2 -sf=2 -of=test.hdf ./DATA/SAMPLES/MOD021KM.A1996213.1024.002.hdf \n" \
"	enlarge_sds -sds=EV_500_Aggr1km_RefSB.2-4 -sf=2 -of=test.hdf ./DATA/SAMPLES/MOD021KM.A1996213.1024.002.hdf \n" \
"	enlarge_sds -sds=Surface_Refl.1.1 -sf=2 -of=test.hdf ./DATA/SAMPLES/MODAGAGG.A1996214.h12v04.001.hdf \n" \
"	enlarge_sds -sds=\"Surface_Refl.*.2\" -meta -sf=2 -of=test.hdf ./DATA/SAMPLES/MODAGAGG.A1996214.h12v04.001.hdf \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:      enlarge_sds -sf=<scale factor> -of=output_file [-meta] \n" \
"            [-sds=SDSname1[,SDSname2...]] filename \n" \
" \n" \
"OPTIONS \n" \
"	-help            Display this help message \n" \
"	-sf=factor       Scale factor (must be a nonzero integer) \n" \
"	-of=filename     Output filename \n" \
"	-meta            Copy metadata from input to output file \n" \
"	-sds=<SDS list>  List of SDS to enlarge (separated by commas) \n" \
"	                 To enlarge a specific layer in a 3D or 4D SDS, specify the \n" \
"	                 layer number (1-based) as dot extension in the SDS name: \n" \
"	                 (sds_name.n.m)   Ex: sur_refl_b02.1 Surf_Refl.1.2 \n" \
"	                 sds_name.*.m will enlarge all 3D layers in the mth layer of \n" \
"	                 4th dimesion. A range of layers can be specified as in \n" \
"	                 sds_name.n1-n2.m. Multiple SDS layer numbers in the 3rd or \n" \
"	                 4th dimension cannot be separated by comma. \n" \
"	                 Default is all SDS, all layer. \n" \
"      -fill_value=fillVal       User inputed fill value \n"

int parse_cmd_enlarge_sds(int argc, char **argv, int *if_cnt, char **sds_names, int *sds_cnt,
	char *of_name, int *sf, int *m_opt, char *fillVal); 
void enlarge_an_sds(sds_t *in_sds_info, int32 out_sd_id, int sf, char *tmp_str, int if_cnt, char *fillVal, int H5OUT);
void enlarge_mult_sds(char *in_fname, char **sds_names, int sds_cnt, int32 out_sd_id, 
	int if_cnt, int sf, int m_opt, char *fillVal, int H5OUT);

int main(int argc, char *argv[])
{
  int m_opt;
  int stdin_argc;
  int status, sf;
  int fid, i, k, if_cnt;
  int isds, sds_cnt;
  char **stdin_argv;
  char **sds_names;
  char **tmp_sds_names;
  char out_fname[MAX_PATH_LENGTH];
  char in_fname[MAX_PATH_LENGTH];
  char fillVal[10];
  char dim_str[MAX_STR_LEN];
  int32 sd_id, sds_id, dt;
  char name[MAX_SDS_NAME_LEN], sds_name[MAX_SDS_NAME_LEN];
  sds_t out_sds_info;

  /* MSZ 2/7/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/7/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  else
  {
    /* Print all SDS names when -help option is input with the input filename */
    for (i=0, status=0; i<argc; i++)
      if ((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "-help") == 0)) status = i;
    if (status != 0)
    {
      for (i=1; i<argc; i++)
	if (argv[i][0] != '-')
	  /* Print all the SDS names in the input file */
	  /* MSZ 2/7/2018 */
	  display_sds_info_of_file(argv[i]);
      exit(0);
    }
  }
  
  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names in enlarge_sds(main)\n");
  else
  {
    if (argc == 1) 
    {
      if ((stdin_argv = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
      {
        fprintf(stderr, "Cannot allocate memory for stdin_argv in data_fedex(main)\n");
        exit(1);
      }
      if (parse_stdin(&stdin_argc, stdin_argv) != -1)
	status = parse_cmd_enlarge_sds(stdin_argc, stdin_argv, &if_cnt, sds_names, &sds_cnt, 
							out_fname, &sf, &m_opt, fillVal);
      else status = 0;
      Free2D((void **)stdin_argv);
    }
    else
      status = parse_cmd_enlarge_sds(argc, argv, &if_cnt, sds_names, &sds_cnt, out_fname, &sf, &m_opt, fillVal);
    if (status == -1) fprintf(stderr, "%s\n", USAGE);
    else if (status != 0)
    {
      if ((tmp_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
        fprintf(stderr, "Cannot allocate memory for tmp_sds_names in enlarge_sds(main)\n");
      else
      {
	out_sds_info.sd_id = out_sds_info.sds_id = -1;
	out_sds_info.H5OUT = 0;
	if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
          fprintf(stderr, "Cannot create the output file %s\n", out_fname);
        else
        {
	  if (argc != 1) stdin_argc = argc;
          for (fid=1; fid<stdin_argc; fid++)
          {
            if (argc == 1)
              strcpy(in_fname, stdin_argv[fid]);
            else
              strcpy(in_fname, argv[fid]);
            if (in_fname[0] == '-') continue;
            fprintf(stderr, "Processing input HDF file %s\n", in_fname);
	    fprintf(stderr, "---------------------------------------------------------------\n");
            for (isds=0; isds<sds_cnt; isds++)
              strcpy(tmp_sds_names[isds], sds_names[isds]);
	    update_nd_sdsnames(tmp_sds_names, &sds_cnt, in_fname);
            enlarge_mult_sds(in_fname, tmp_sds_names, sds_cnt, out_sds_info.sd_id, if_cnt, sf, m_opt, fillVal, out_sds_info.H5OUT);
          }
	  close_sd(&out_sds_info, out_sds_info.H5OUT);
        }
        Free2D((void **)tmp_sds_names);
      }
    }
    Free2D((void **)sds_names);
  }
  return 0;
}

int parse_cmd_enlarge_sds(int argc, char **argv, int *if_cnt, char **sds_names, int *sds_cnt,
	char *of_name, int *sf, int *m_opt, char *fillVal) 
{
  int i;
  int status = 1;
  char sf_str[10];

  *m_opt = 0;
  *if_cnt = 0;
  *sds_cnt = 0;
  of_name[0] = sf_str[0] = fillVal[0] = '\0';
  for (i=1; i<argc; i++)
  {
    if (is_arg_id(argv[i], "-sds=") == 0) get_arg_val_arr(argv[i], sds_names, sds_cnt);
    else if (is_arg_id(argv[i], "-sds") == 0) ;
    else if ((is_arg_id(argv[i], "-of=") == 0) || (is_arg_id(argv[i], "-o=") == 0))
      get_arg_val(argv[i], of_name);
    else if (is_arg_id(argv[i], "-sf=") == 0) get_arg_val(argv[i], sf_str);
    else if (strcmp(argv[i], "-meta") == 0) *m_opt = 1;
    else if (is_arg_id(argv[i], "-fill_value=") == 0) get_arg_val(argv[i], fillVal);
    else if (argv[i][0] == '-') fprintf(stderr, "Unknonwn option %s\n", argv[i]);
    else ++*if_cnt;
  }
  *sf = (int)atoi(sf_str);

  if ((*if_cnt == 0) || (strlen(of_name) <= 0) || (*sf <= 0)) status = -1;
  if (*if_cnt == 0) fprintf(stderr, "No input filename specified\n");
  if (strlen(of_name) <= 0) fprintf(stderr, "No output filename specified\n");
  if (sf_str[0] == '\0') fprintf(stderr, "Dimension scale factor not input\n");
  else if (*sf <= 0) fprintf(stderr, "Invalid value for scale factor %s\n", sf_str);
  if ((status > 0) && (*sds_cnt == 0))
  {
    fprintf(stderr, "No SDS name input. Reading all SDS . . \n");
    *sds_cnt = 1;
    strcpy(sds_names[0], "all");
  }
  return status;
}

void enlarge_mult_sds(char *in_fname, char **sds_names, int sds_cnt, int32 out_sd_id, 
	int if_cnt, int sf, int m_opt, char *fillVal, int H5OUT)
{
  int isds;
  sds_t in_sds_info;
  char tmp_str[MAX_PATH_LENGTH];

  if ((sds_cnt == 1) && (strcmp(sds_names[0], "all") == 0))
  {
    sds_cnt = get_sds_names(in_fname, sds_names);
    if (sds_cnt == 0)
    {
      fprintf(stderr, "No SDS found in %s\n", in_fname);
      return;
    }
  }
  in_sds_info.sd_id = in_sds_info.sds_id = -1;
  in_sds_info.H5OUT = 0;

  if(open_sd(in_fname, &in_sds_info, 'R') == FAIL)
  {
    fprintf(stderr, "Cannot open input HDF file in reduce_mult_sds\n");
    return;
  }
  if (if_cnt > 1)
  {
    strcpy(tmp_str, in_fname);
    rm_path(tmp_str);
    tmp_str[24] = '\0';
  }
  for (isds=0; isds<sds_cnt; isds++)
  {
    strcpy(in_sds_info.name, sds_names[isds]);
    fprintf(stderr, "\tProcessing SDS %s\n", in_sds_info.name);
    in_sds_info.sds_id = -1;
    if (get_sds_info((char *)NULL, &in_sds_info) == -1)
    {
      fprintf(stderr, "Ignoring invalid input sds %s\n", in_sds_info.name);
      close_sds(&in_sds_info, in_sds_info.H5IN);
      continue;
    }
    if (fillVal[0] != '\0')       
      switch(in_sds_info.data_type)
      {
	case  5: in_sds_info.fill_val = (float32)atof(fillVal); break;
	case  6: in_sds_info.fill_val = (float64)atof(fillVal); break;
	case 20: in_sds_info.fill_val = (int8)atoi(fillVal); break;
	case 21: in_sds_info.fill_val = (uint8)atoi(fillVal); break;
	case 22: in_sds_info.fill_val = (int16)atoi(fillVal); break;
	case 23: in_sds_info.fill_val = (uint16)atoi(fillVal); break;
	case 24: in_sds_info.fill_val = (int32)atoi(fillVal); break;
	case 25: in_sds_info.fill_val = (uint32)atoi(fillVal); break;
      }       
    
    strcpy(in_sds_info.name, sds_names[isds]);	/* added on 05/02/2011	*/
    enlarge_an_sds(&in_sds_info, out_sd_id, sf, tmp_str, if_cnt, fillVal, H5OUT);
    close_sds(&in_sds_info, in_sds_info.H5IN);
  }
  if (m_opt == 1)
    copy_metadata(in_sds_info.H5IN, in_sds_info.sd_id, H5OUT, out_sd_id);
  close_sd(&in_sds_info, in_sds_info.H5IN);
}

void enlarge_an_sds(sds_t *in_sds_info, int32 out_sd_id, int sf, char *tmp_str, int if_cnt, char *fillVal, int H5OUT)
{
  float32 fval1;
  int ir, ic, rank;
  int n, m, i, j, k;
  int val1, val[100];
  int irow, icol, nrow;
  int st_c, offset;
  int out_ncol, ndata_in_sm;
  int ndata, ndata_in, ndata_out;
  int32 attr_type, attr_cnt;
  int32 in_edge[4] = {0, 0, 0, 0};
  int32 out_edge[4] = {0, 0, 0, 0};
  int32 in_start[4] = {0, 0, 0, 0};
  int32 out_start[4] = {0, 0, 0, 0};
  sds_t out_sds_info;
  char sds_name[MAX_SDS_NAME_LEN];
  void  *attr_val, *data_in, *data_out;
  float fill_fval;
  int fill_val;
  
  if ((attr_val = (void *)get_sds_attr(in_sds_info, "_FillValue", &attr_type,
              &attr_cnt)) == NULL)
    fprintf(stderr, "Attribute "_FillValue" not defined for output\n");
  if (fillVal[0] != '\0')       
  {
    switch(in_sds_info->data_type)
      {
	case  5: ((float32 *)attr_val)[0] = (float32)atof(fillVal); fill_fval = (float32)atof(fillVal); break;
	case  6: ((float64 *)attr_val)[0] = (float64)atof(fillVal); fill_fval = (float)(float64)atof(fillVal); break;
	case 20: ((int8 *)attr_val)[0] = (int8)atoi(fillVal); fill_val = (int)(int8)atoi(fillVal); break;
	case 21: ((uint8 *)attr_val)[0] = (uint8)atoi(fillVal); fill_val = (int)(uint8)atoi(fillVal); break;
	case 22: ((int16 *)attr_val)[0] = (int16)atoi(fillVal); fill_val = (int)(int16)atoi(fillVal); break;
	case 23: ((uint16 *)attr_val)[0] = (uint16)atoi(fillVal); fill_val = (int)(uint16)atoi(fillVal); break;
	case 24: ((int32 *)attr_val)[0] = (int32)atoi(fillVal); fill_val = (int)(int32)atoi(fillVal); break;
	case 25: ((uint32 *)attr_val)[0] = (uint32)atoi(fillVal); fill_val = (int)(uint32)atoi(fillVal); break;
      } 
  }
  rank = in_sds_info->rank;
  out_sds_info.sds_id = -1;
  out_sds_info.sd_id = out_sd_id;
  out_sds_info.H5OUT = H5OUT;
  if (charpos(in_sds_info->name, '.', 0) != -1)
  {
    out_sds_info.rank = 2;
    if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
    {
      out_sds_info.dim_size[0] = sf*in_sds_info->dim_size[0];
      out_sds_info.dim_size[1] = sf*in_sds_info->dim_size[1];
    }
    else
    {
      out_sds_info.dim_size[0] = sf*in_sds_info->dim_size[rank-2];
      out_sds_info.dim_size[1] = sf*in_sds_info->dim_size[rank-1];
    }
  }
  else
  {
    out_sds_info.rank = in_sds_info->rank;
    for (i=0; i<rank; i++)
      out_sds_info.dim_size[i] = in_sds_info->dim_size[i];
    if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
    {
      out_sds_info.dim_size[0] = sf*in_sds_info->dim_size[0];
      out_sds_info.dim_size[1] = sf*in_sds_info->dim_size[1];
    }
    else
    {
      out_sds_info.dim_size[rank-2] = sf*in_sds_info->dim_size[rank-2];
      out_sds_info.dim_size[rank-1] = sf*in_sds_info->dim_size[rank-1];
    }
  }
  if (in_sds_info->interleave == 0)
  {
    out_ncol = in_sds_info->dim_size[1];
    if (out_sds_info.rank == 2) ndata_in_sm = 1;
    else
    {
      ndata_in_sm = out_sds_info.dim_size[2];
      if (out_sds_info.rank == 4) ndata_in_sm *= out_sds_info.dim_size[3];
    }
  }
  out_sds_info.data_size = in_sds_info->data_size;
  out_sds_info.data_type = in_sds_info->data_type;
  if (if_cnt == 1) sprintf(out_sds_info.name, "%s_%dx", in_sds_info->name, sf);
  else sprintf(out_sds_info.name, "%s_%dx_%s", in_sds_info->name, sf, tmp_str);
  if (open_sds((char *)NULL, &out_sds_info, 'W') == -1)
  {
    if (attr_val != NULL) free(attr_val);
    return;
  }
  else if (attr_val != NULL)
  { 
    if (out_sds_info.data_type == 5 || out_sds_info.data_type == 6)
       write_attr_fvalf(out_sds_info.H5OUT, out_sds_info.sds_id, out_sds_info.data_type, 1, (double)fill_fval);
    else
       write_attr_fval(out_sds_info.H5OUT, out_sds_info.sds_id, out_sds_info.data_type, 1, fill_val);
  }

  get_sdsname_dim(in_sds_info->name, sds_name, &n, &m);
  if (rank == 2)
  {
    nrow = in_sds_info->dim_size[0];
    ndata = ndata_in = in_sds_info->dim_size[1];
    ndata_out = sf*in_sds_info->dim_size[1];
  }
  else if (rank > 2)
  {
    if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
    {
      nrow = in_sds_info->dim_size[0];
      ndata_in = in_sds_info->dim_size[1];
      for (i=2; i<rank; i++)
        ndata_in *= in_sds_info->dim_size[i];
      ndata = ((n == -1) && (m == -1)) ? ndata_in : in_sds_info->dim_size[1];
      ndata_out = sf*ndata;
    }
    else
    {
      nrow = in_sds_info->dim_size[rank-2];
      ndata_in = in_sds_info->dim_size[rank-1];
      for (i=0; i<rank-2; i++)
        ndata_in *= in_sds_info->dim_size[i];
      ndata = ((n == -1) && (m == -1)) ? ndata_in : in_sds_info->dim_size[rank-1];
      ndata_out = sf*ndata;
    }
  }

  if ((data_in = (void *)calloc(ndata_in, in_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_in in reduce_an_sds_by_sub\n");
  if ((data_out = (void *)calloc(ndata_out, out_sds_info.data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_out in reduce_an_sds_by_sub\n");
  if ((data_in != NULL) && (data_out != NULL))
  {
    if (rank == 2)
    {
      in_edge[0] = out_edge[0] = 1;
      in_edge[1] = in_sds_info->dim_size[1];
      out_edge[1] = out_sds_info.dim_size[1];
    }
    else
    {
      for (i=0; i<rank; i++)
        in_edge[i] = in_sds_info->dim_size[i];
      for (i=0; i<out_sds_info.rank; i++)
        out_edge[i] = out_sds_info.dim_size[i];
      if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
	in_edge[0] = out_edge[0] = 1;
      else
      {
	in_edge[rank-2] = 1;
        if ((n == -1) && (m == -1))
          out_edge[rank-2] = 1;
        else
	  out_edge[0] = 1; 
      }
    }

    compute_sds_start_offset(in_sds_info, n, m, &st_c, &offset);

    for (irow=0; irow<nrow; irow++)
    {
      if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
        in_start[0] = irow;
      else
        in_start[rank-2] = irow;

      if (read_sel_sds_data(in_sds_info, in_start, in_edge, data_in) == FAIL) 
      {
        fprintf(stderr, "Error reading SDS data line in enlarge_sds: %s", in_sds_info->name);
        break;
      }	    
      
/*      fprintf(stdout, "offset: %d\n",offset);  
      fprintf(stdout, "st_c: %d\n",st_c);*/

      
      if (in_sds_info->interleave == 1)
      {
        switch(in_sds_info->data_type)
        {
          case 5 : for (i=0, ic=0, icol=st_c; i<ndata; icol += offset, i++)
		   {
		     fval1 = ((float32 *)data_in)[icol];
		     for (k=0; k<sf; k++, ic++)
		       ((float32 *)data_out)[ic] = fval1;
		   }
		   break;
          case 20: for (i=0, ic=0, icol=st_c; i<ndata; icol += offset, i++)
		   {
		     val1 = ((int8 *)data_in)[icol];
		     for (k=0; k<sf; k++, ic++)
		       ((int8 *)data_out)[ic] = val1;
		   }
		   break;
          case 21: for (i=0, ic=0, icol=st_c; i<ndata; icol += offset, i++)
		   {
		     val1 = ((uint8 *)data_in)[icol];
		     for (k=0; k<sf; k++, ic++)
		       ((uint8 *)data_out)[ic] = val1;
		   }
		   break;
          case 22: for (i=0, ic=0, icol=st_c; i<ndata; icol += offset, i++)
		   {
		     val1 = ((int16 *)data_in)[icol];
		     for (k=0; k<sf; k++, ic++)
		       ((int16 *)data_out)[ic] = val1;
		   }
		   break;
          case 23: for (i=0, ic=0, icol=st_c; i<ndata; icol += offset, i++)
		   {
		     val1 = ((uint16 *)data_in)[icol];
		     for (k=0; k<sf; k++, ic++)
		       ((uint16 *)data_out)[ic] = val1;
		   }
		   break;
          case 24: for (i=0, ic=0, icol=st_c; i<ndata; icol += offset, i++)
		   {
		     val1 = ((int32 *)data_in)[icol];
		     for (k=0; k<sf; k++, ic++)
		       ((int32 *)data_out)[ic] = val1;
		   }
		   break;
          case 25: for (i=0, ic=0, icol=st_c; i<ndata; icol += offset, i++)
		   {
		     val1 = ((uint32 *)data_in)[icol];
		     for (k=0; k<sf; k++, ic++)
		       ((uint32 *)data_out)[ic] = val1;
		   }
		   break;
	} /* switch(...) */
      }
      else
      {
        for (i=0, ic=0, icol=st_c; i<out_ncol; i++)
	{
	  for (j=0; j<ndata_in_sm; j++, icol += offset)
            switch(in_sds_info->data_type)
            {
              case 20: val[j] = ((int8 *)data_in)[icol]; break;
              case 21: val[j] = ((uint8 *)data_in)[icol]; break;
              case 22: val[j] = ((int16 *)data_in)[icol]; break;
              case 23: val[j] = ((uint16 *)data_in)[icol]; break;
              case 24: val[j] = ((int32 *)data_in)[icol]; break;
              case 25: val[j] = ((uint32 *)data_in)[icol]; break;
            }
	  for (k=0; k<sf; k++)
	    for (j=0; j<ndata_in_sm; j++, ic++)
              switch(in_sds_info->data_type)
              {
                case 20: ((int8 *)data_out)[ic] = (int8)val[j]; break;
                case 21: ((uint8 *)data_out)[ic] = (uint8)val[j]; break;
                case 22: ((int16 *)data_out)[ic] = (int16)val[j]; break;
                case 23: ((uint16 *)data_out)[ic] = (uint16)val[j]; break;
                case 24: ((int32 *)data_out)[ic] = (int32)val[j]; break;
                case 25: ((uint32 *)data_out)[ic] = (uint32)val[j]; break;
              }
	} /* for (i=0; . . . ) */
      }
      for (ir=0; ir<sf; ir++)
      {
	if(set_sel_sds_data(&out_sds_info, out_start, out_edge, data_out) == FAIL)
          fprintf(stderr, "Error writing SDS data line for %s\n", out_sds_info.name);
        if ((rank > 2) && (n == -1) && (m == -1) && (in_sds_info->dim_size[0] < in_sds_info->dim_size[rank-1]))
          ++out_start[rank-2];
        else
          ++out_start[0];
      }
    } /* for (irow=0; . . )   */
  }
  close_sds(&out_sds_info, out_sds_info.H5OUT);
  if (data_in != NULL) free(data_in);
  if (data_out != NULL) free(data_out);
  if (attr_val != NULL) free(attr_val);
}


