/****************************************************************************
!C

!File: geolocation.c

!Description:
  Compute the geolocation for a pixel in L3 product 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Version 1.0    December 2003
  Version 2.0    March 2013 by Xiaoping Zhang Sigma Space Corp.
  	updated to work for both MODIS and NPP products.

  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
        3)update all HDFEOS API to a higher-level APIs to handle 
          both HDFEOS2 and HDFEOS5 in eos_sds_rw.c

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Yi (Sophyi) Zhang
    LDOPE                             Science Systems and Applications Inc.
    sophyi@ltpmail.gsfc.nasa.gov      NASA/GSFC Code 922
    phone: 301-614-5497               Greenbelt, MD 20771

    David Roy
    LDOPE                             University of Maryland, Department of Geography
    droy@kratmos.gsfc.nasa.gov        NASA/GSFC Code 922 (B32)
    phone: 301-614-5571               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "eos_sds_rw.h"
#include "qa_tool.h"
#include "str_op.h"
#include "meta.h"
#include "main_util.h"
#include "alloc_mem.h"
#include "sds_rw.h"

#define PI 	3.141592653589793238

#define TILESZ_DEG 10
#define NROWS  1200
#define NCOLS  1200

/*#include "tile_bound_info.h"*/

#define HELP \
"NAME \n" \
"       Compute the geolocation of a pixel in a L3 MODIS Land HDF-EOS product. \n" \
"\n" \
"SYNOPSIS \n" \
"       geolocation -help \n" \
"       geolocation -proj=<SIN|ISIN> -xy=<column,row> [-res=qkm|hkm|1km] \n" \
"		[-hv=<horizontal,vertical>] [filename|local granule id] \n" \
" \n" \
"DESCRIPTION \n" \
"	Compute the geolocation of a pixel in L3 MODISLand product file. Pixel \n" \
"	location is specified as row column number at a specified resolution and \n" \
"	the tile id is either explicitly input or is implied by the local granule \n" \
"	id of the L2G/L3/L4 MODISLand product file.\n" \
" \n" \
"	The tool command arguments can be specified in any order. \n" \
" \n" \
"OPTIONS \n" \
"	-help				Display this help message \n" \
"	-res=qkm|hkm|1km		Reference resolution (default = 1km) \n" \
"	-xy=<column,row>		Row and column location of pixel. \n" \
"					(valid: 0-1199 for 1km, 0-2399 for hkm and \n" \
"					0-4799 for qkm) \n" \
"	-hv=<horizontal,vertical>	Horizontal and vertical tile number. \n" \
"					Valids for horizontal: 0-35, vertical: 0-17 \n" \
"	filename			MODISLand product filename/local granule \n" \
"					id of a L2G|L3|L4 product. \n" \
" \n" \
"Examples: \n" \
"	geolocation -xy=750,500 -hv=12,4 -proj=SIN\n" \
"	geolocation -xy=750,500 MOD09GHK.A1997223.h17v03.001.1999153204552.hdf \n" \
"	geolocation -xy=750,500 NPP_DSRF1KD_L2GD.A2003222.h20v11.C1_00602.2009253063846.hdf\n" \
" \n" \
"AUTHOR \n" \
"	Code: S. Devadiga and Xiaoping Zhang \n" \
"	Documentation: S. Devadiga and D. Roy \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage: geolocation -proj=<SIN|ISIN> -xy=<column,row> [-res=qkm|hkm|1km] \n" \
"		[-hv=<horizontal,vertical>] [filename|local granule id] \n" \
" \n" \
"OPTIONS \n" \
"	-help				Display this help message \n" \
"	-res=qkm|hkm|1km		Reference resolution (default = 1km) \n" \
"	-xy=<column,row>		Row and column location of pixel. \n" \
"					(valid: 0-1199 for 1km, 0-2399 for hkm and \n" \
"					0-4799 for qkm) \n" \
"	-hv=<horizontal,vertical>	Horizontal and vertical tile number. \n" \
"					Valids for horizontal: 0-35, vertical: 0-17 \n" \
"	filename			MODISLand product filename/local granule \n" \
"					id of a L2G|L3|L4 product. \n" \
" \n"

/*****************************************************************************
                            Prototypes.
*****************************************************************************/
int parse_cmd_geoloc(int argc, char **argv, char *xy_str, char *hv_str, int *res, 
		     char *proj_name);
int read_bound_table(int hn, int vn, char* map_fname, double *lon_min, double *lon_max,
                     double *lat_min, double *lat_max);
int get_xy_coord(char *xy_str, int *x, int *y);
int get_tileid(char *hv_str, int *hn, int *vn, char *proj_name);
int check_tileid(char *hv_str, int *p);
int compute_geoloc(int row, int col, int hn, int vn, double *lon, double *lat, 
		   int res, char *proj_name);
long isinusinvinit(double sphere, double lon_cen_mer, double false_east, 
		   double false_north, double dzone, double djustify);
int isinusinv(double x, double y, double *lon, double *lat);
int sininvint(double r, double center_long, double false_east, double false_north);
int sininv(double x, double y, double *lon, double *lat);
		     
/*************************************************************************************/

int main(int argc, char *argv[])
/*
!C************************************************************************************

!Description:
  The geolocation routine computes the geolocation of the user specified location of 
  a MODLand L2G, L3, L4 pixel coordinate. Tool works on both Integerized Sinusoidal 
  projection (MODIS collection 3 product) and Sinusoidal projection (MODIS collection 
  4 product).
 
!Input Parameters: (none)

!Output Parameters:
  (returns)  Completion status:
             0 - successful completion
             1 - error exit

!Revision History: (none)

!Team-unique Header: (see file prolog)

!References and Credits: (see file prolog)

!Design Notes: (none)
  The tables containing the bounding coordinates for each tile are provided by 
  Robert Wolfe (Raytheon).
!END
**************************************************************************************/
{
  int res;
  int st = 0;
  int hn, vn;
  int row, col;
  int row_in, col_in;
  double lat, lon;
  char xy_str[255], hv_str[MAX_PATH_LENGTH];
  char proj_name[255];

   
  /* missing input tileid and projection parameter */
  if (argc == 1)
    {
      st = -1;
      fprintf(stderr, "Missing input tileid and projection parameter \n");
      fprintf(stderr, "%s\n", USAGE);
      exit(EXIT_FAILURE);
    }
  
  if ((argc == 2) && (strcmp(argv[1], "-help")==0))
    {
      /* Print command line help when -help option is input */
      fprintf(stderr, "%s\n", HELP);
      exit(EXIT_SUCCESS);
    }
   
  if (parse_cmd_geoloc(argc, argv, xy_str, hv_str, &res, proj_name) == -1)
    {
      fprintf(stderr, "%s\n", USAGE);
      exit(EXIT_FAILURE);
    }
  else
    {     
      /* MSZ 2/10/2018 silent version mismatch warning standard print out */
      setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

      /* MSZ 2/10/2018 Disable error reporting */
      H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

      if (get_xy_coord(xy_str, &col_in, &row_in) == -1) return 1;
          
      if (get_tileid(hv_str, &hn, &vn, proj_name) == -1) return 1;
      
      row = row_in/res; col = col_in/res;
      
      if ((row < 0) || (row > NROWS-1)) { 
	fprintf(stderr, "Invalid row number: %d\n", row_in);
	st = -1;
      }
      if ((col < 0) || (col > NCOLS-1)) {
	fprintf(stderr, "Invalid col number: %d\n", col_in);
	st = -1;
      }
      if ((hn < 0) || (hn > 35)) {
	fprintf(stderr, "Invalid Horizontal tile number: %d\n", hn);
	st = -1;
      }
      if ((vn < 0) || (vn > 17)) {
	fprintf(stderr, "Invalid Vertical tile number: %d\n", vn);
	st = -1;
      }
   
      if (st != -1)
	{
	  if (compute_geoloc(row_in, col_in, hn, vn, &lon, &lat, res, proj_name) == -1)
	    fprintf(stderr, "Error computing the geolocation at (%d %d) in tile (%d %d)\n",
		    row_in, col_in, hn, vn);
	  else
	    fprintf(stdout, "Geolocation at (%d %d) in tile (%d %d): %lf %lf\n", row_in, col_in, 
		    hn, vn, lon, lat);
	}
    }
  
  fprintf(stdout, "Processing done ! \n");
  return 0;
}

int parse_cmd_geoloc(int argc, char **argv, char *xy_str, char *hv_str, int *res,
		     char *proj_name)
/*
!C*********************************************************************************

!Function: parse_cmd_geoloc
       
!Description:
   Function parse_cmd_geoloc parse the argument string to the corresponding
   fields.

!Input Parameters: 
   argc:     input argument count
   argv:     input argument vector

!Input/output Parameters:
   xy_str:    Colow and row of the pixel location.
   hv_str:    horizontal and vertical tile number.
   res:       Resolution of the tile: 1km, hkm, qkm
   proj_name: projection type: ISIN or SIN   

  in_fname  Input HDF file name.
  out_fname Output file name.
  sds_names User input sds_name. Should only input one sds name since sds2bin is 
            intend to write one SDS(all or some layers) out to output file.

!Output Parameters:

  ret   :    1  - successful completion
	     -1 - error exit   

!Revision History: (see file prolog) 

!Team-unique Header: (see file prolog)
 
!References and Credits: (see file prolog)

!Design Notes: (none)

!END
*****************************************************************************/     
{
  int i, st;
  char fname[MAX_PATH_LENGTH];
  int projflag = 0;
  char res_str[50];
  
  st = 1;
  *res = 1;
  res_str[0] = xy_str[0] = hv_str[0] = fname[0] =proj_name[0] = '\0';
  
  for (i=1; i<argc; i++)
    {
      if (is_arg_id(argv[i], "-xy=") == 0) get_arg_val(argv[i], xy_str);
      else if (is_arg_id(argv[i], "-res=") == 0)
	{
	  get_arg_val(argv[i], res_str);
	  {
	     if (strcmp(res_str, "qkm") == 0) *res = 4; 
	     else if (strcmp(res_str, "hkm") == 0) *res = 2; 
	     else if (strcmp(res_str, "1km") == 0) *res = 1; 
	     else {
	       fprintf(stderr, "Invalid resolution input. Valid resolution input are : 1km, qkm, hkm \n");
	       exit(EXIT_FAILURE);
	     }
	  }
	}
      else if (is_arg_id(argv[i], "-hv=") == 0) get_arg_val(argv[i], hv_str);
      else if (is_arg_id(argv[i], "-proj") == 0)
	{
	  get_arg_val(argv[i], proj_name);
	  
	  if ( proj_name == NULL )
	    {
	      fprintf(stderr, "Invalid projection type entered. Valid projection \n");
	      fprintf(stderr, "type is ISIN or SIN. \n");
	      exit(EXIT_FAILURE);
	    }
	  else 
	    {
	      projflag = 1;
	    }	
	}
      else if (argv[i][0] == '-') fprintf(stderr, "Ignoring invalid option %s\n", argv[i]); 
      else strcpy(fname, argv[i]);
    }
  
  if (strlen(xy_str) <= 0) 
    { 
      fprintf(stderr, "Missing input coordinate \n");
      st = -1;
    }

  if (strlen(fname) <= 0)
    {
      if ((strlen(hv_str) <=0) ||  projflag == 0)
	{
	  fprintf(stderr, "Tile ID with projection type or L2G/L3 filename is a required input\n");
	  st = -1;
	}
      else
	{
	  st = 0;
	}
    } 

  if (strlen(hv_str) > 0) 
    {
      if (projflag == 0)
	{
	  fprintf(stderr, "Tile ID with projection type is a required input\n");
	  st = -1;
	}
      else
	{
	  st = 0;
	}
    } 
  
  if ((st != -1) && (strlen(hv_str) <= 0))
    strcpy(hv_str, fname);
  
  return st;
}

int get_xy_coord(char *xy_str, int *x, int *y)
/* get the coordinates frm the user input string. */
{
  int st = 1;
  int p1, len;
  char tmp_str[10];

  if ((p1 = charpos(xy_str, ',', 0)) == -1) st = -1;
  else
  {
    len = (int)strlen(xy_str);
    strmid(xy_str, 0, p1, tmp_str);
    *x = (int)atoi(tmp_str);
    strmid(xy_str, p1+1, len-p1-1, tmp_str);
    *y = (int)atoi(tmp_str);
  }
  return st;
}

int get_tileid(char *hv_str, int *hn, int *vn, char *proj_name)
/* get the tile id number and projection parameter from user input
 * string. If user didn't input the projection name, it retrieve 
 * the projection information from meta data of the input HDF file.
 */
{
  int i, j;
  int st = 1;
  int p, meta_cnt;
  int version_num;
  char tmp_str[10];
  char version_str[MAX_META_NAME_LEN];
  char *cmeta_str, **meta_val;
  char fname[MAX_PATH_LENGTH];
  char meta_name[MAX_META_NAME_LEN];
  FILE *fp;
  char in_fname[MAX_PATH_LENGTH];  
  char HDFEOSgridname[MAX_PATH_LENGTH]; 
  grid_mapinfo *hdfeos; 
  int32 projCode;   
  int status;
  long bufsiz;  
  
  hdfeos = (grid_mapinfo *)calloc(1, sizeof(grid_mapinfo));  
  if (get_xy_coord(hv_str, hn, vn) == -1)
  {
    if ((fp = fopen(hv_str, "r")) == NULL)
      {
	fprintf(stderr, "Cannot open input file %s\n", hv_str);
	exit(EXIT_FAILURE);
      }
    else fclose(fp);
  
    status = (strpos(hv_str, "NPP_", 0) != -1) ? 1 : 0;   
    if (status == 1)
    {
      /* check the input HDF version */
      int he5_fmt;
      int HDF_ver = 0;

      HDF_ver = HDF_version_check(in_fname);

      if( !(HDF_ver ==4 || HDF_ver == 5) )
      {
        printf("%s is neither HDF4 nor HDF5, exit...\n", in_fname);
        exit(0);
      }

      if(HDF_ver == 4) he5_fmt = 0;
      else if(HDF_ver == 5) he5_fmt = 1;

      if(EOS_GDinqgrid(hv_str, HDFEOSgridname, &bufsiz, he5_fmt)==FAIL || GetHDFEOS(hv_str, HDFEOSgridname, hdfeos)==-1)
      {
	printf("Could not find grid name\n");  
        return 1;
      } 
      else 
      {
	printf("Grid name:%s\n", HDFEOSgridname);  	      	  
	projCode = hdfeos->projCode;
	printf("projCode====%d\n",projCode); 
	if (projCode == 16)  
	{    
	  strcpy(proj_name, "SIN");
	  fprintf(stdout, "Use Sinusoidal projection \n"); 
	}
	else if (projCode == 31)
	{    
	  strcpy(proj_name, "ISIN");
	  fprintf(stdout, "Use Integerized Sinusoidal projection \n"); 
	}	
	else
	{
	  fprintf(stderr, "Projection type not know for file %s \n", in_fname);
	  return -1;	 
	}
	st = check_tileid(hv_str, &p);	     
      }      
    }  
    else
    {
      meta_cnt = 0;
      if ((meta_val = (char **)Calloc2D(5, MAX_META_VAL_LEN, sizeof(char))) == NULL)
	fprintf(stderr, "Cannot allocate memory for meta_val in get_tileid\n");
      else 
	{
	  if ((cmeta_str = get_attr_metadata(hv_str, "CoreMetadata.0")) != NULL)
	    {
	      strcpy(meta_name, "VERSIONID");
	      get_sel_metadata(cmeta_str, meta_name, meta_val, &meta_cnt, 0);
	      if (meta_cnt == 0)
		fprintf(stderr, "No metadata LOCALGRANULEID found in file %s", hv_str);
	      else
		{
		  strcpy(version_str, meta_val[0]);
		  version_num = atoi(version_str);
		  if (version_num == 3) 
		    {
		      if (strcmp(proj_name, "ISIN") != 0)
			{
			  strcpy(proj_name, "ISIN");
			  fprintf(stdout, "Use Integerized Sinusoidal projection \n"); 
			}
		    }
		  else if (version_num == 4)
		    {    if (strcmp(proj_name, "SIN") != 0)
			{
			  strcpy(proj_name, "SIN");
			  fprintf(stdout, "Use Sinusoidal projection \n"); 
			}		 
		    }
		  else {
		    fprintf(stderr, "Projection type not know for file %s \n", hv_str);
		    return -1;
		  }
		}

	      st = check_tileid(hv_str, &p);

	      if (st == -1)
		{ 
		  st = 1;
		  strcpy(meta_name, "LOCALGRANULEID");
		  get_sel_metadata(cmeta_str, meta_name, meta_val, &meta_cnt, 0);

		  if (meta_cnt == 0)
		    fprintf(stderr, "No metadata LOCALGRANULEID found in file %s", hv_str);
		  else
		    {
		      strcpy(hv_str, meta_val[0]);
		      st = check_tileid(hv_str, &p);
		    }
		} 
	      free(cmeta_str);
	    }
          Free2D((void **)meta_val);
	}
    }
    
    if (st == -1)
      fprintf(stderr, "Cannot find tile id in local granule id string: %s\n", hv_str);
    else 
      {  
	for (i=0, j = p+2; i<2; i++, j++) 
	  tmp_str[i] = hv_str[j];

	tmp_str[i] = '\0';
	*hn = (int)atoi(tmp_str);
	for (i=0, j = p+5; i<2; i++, j++) 
	  tmp_str[i] = hv_str[j];
	tmp_str[i] = '\0';
	*vn = (int)atoi(tmp_str);
      }
  }
  free(hdfeos);
  return st;
}

int check_tileid(char *hv_str, int *p)
     /*
      * check if the tile id is provided in the user input.  
      */
{
  int st = 1;
  int p1, p2, p3;

  rm_path(hv_str);


  p1 = strpos(hv_str, ".h", 0);
  p2 = charpos(hv_str, 'v', p1);
  p3 = charpos(hv_str, '.', p1+1);
  if ((p2 == p1+4) && (p3 == p1+7)) st = 1;

  *p = p1;
  return st;
}

int compute_geoloc(int row, int col, int hn, int vn, double *lon, double *lat, 
		   int res, char* proj_name)
     /* computer the geolocation based on the map projection */
{
  int st = 0;
  double xproj, yproj;
  double lon_min, lon_max, lat_min, lat_max;

  double   xproj_ul_corner = -20015109.354;
  double   yproj_ul_corner = 10007554.677;
  double   x_pixel_size = 926.62543305;
  double   y_pixel_size = 926.62543305;
  double   xtile_size = 1200;
  double   ytile_size = 1200;
  double   xn, yn;

  if (strcmp(proj_name, "ISIN") == 0)
    {  
      /* initialize the Integerized Sinusoidal projection */
      st = (int)isinusinvinit(6371007.181, 0.0, 0.0, 0.0, 86400.0, 1.0);       
    }

  if (strcmp(proj_name, "SIN") == 0)
    {
      /* initialize the Sinusoidal projection */
      st = sininvint(6371007.181, 0.0, 0.0, 0.0);   
    }

  /* calculate project parameter */
  xn = (double) hn * xtile_size;
  yn = (double) vn * ytile_size;
  
  xproj = xn * x_pixel_size + (col+0.5) * x_pixel_size / (double)res + xproj_ul_corner;
  yproj = - yn * y_pixel_size - (row+0.5) * y_pixel_size / (double)res + yproj_ul_corner;

  /* calculate longitude and latitude */
  if (strcmp(proj_name, "ISIN") == 0)
	{      
	  st = isinusinv(xproj, yproj, lon, lat);	  
	}
  
  if (strcmp(proj_name, "SIN") == 0)
    {    
      st = sininv(xproj, yproj, lon, lat);
    }

  *lon = (*lon * 180.0)/PI;
  *lat = (*lat * 180.0)/PI;
  printf("lon==%f lat==%f\n", *lon, *lat); 
     
  /* check to see if it's projected on the map projection of MODIS LAND tiles */
  st = read_bound_table(hn, vn, proj_name, &lon_min, &lon_max, &lat_min, &lat_max);

  if ((*lon < lon_min) || (*lon > lon_max) || (*lat < lat_min) || (*lat > lat_max))
  {
    fprintf(stderr, "Invalid input parameter. Not projected onto map projection! ");
    st = -1;
  }
  return st;
}



