/****************************************************************************
!C

!File: geolocation_image.c

!Description:
  Compute the geolocation for a pixel in L3 product 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Version 1.0    November 2006 

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mfhdf.h"
#include "hdf.h"

typedef struct
{
  char name[80];
  int32 data_type;
  int32 data_size;
  int32 rank;
  int32 dim_size[2];
  int32 sd_id;
  int32 sds_id;
  int32 sds_index;
} sds_t;

#define PI 3.141592653589793238

#define TILE_SIZE_1KM 1200
#define GEOLOC_FILL_VAL -9999.  /* Maosheng Zhao, Feb, 28, 2019, changed 999.999 to -9999. */

#define HELP \
"NAME \n" \
"       Compute the geolocation image for a MODIS Land tile \n" \
"\n" \
"SYNOPSIS \n" \
"       geolocation_image -help \n" \
"       geolocation_image [-res=qkm|hkm|1km] -hv=<horizontal,vertical> [-lf=<filename>] -of=<filename> \n" \
" \n" \
"DESCRIPTION \n" \
"	Compute the geolocation image for a MODIS tile at specified resolution \n" \
" \n" \
"	The tool command arguments can be specified in any order. \n" \
" \n" \
"OPTIONS \n" \
"	-help				Display this help message \n" \
"	-res=qkm|hkm|1km		Reference resolution (default = 1km) \n" \
"	-hv=<horizontal,vertical>	Horizontal and vertical tile number. \n" \
"					Valids for horizontal: 0-35, vertical: 0-17 \n" \
"	-lf=<filename>			File containing lat/long limit for projection \n" \
"	                                If not input the points outside of projection \n" \
"                                       will not be identified. The file sn_bound_10deg.txt \n" \
"                                       is provided with this source code\n" \
"	-of=<filename>			Output filename \n" \
" \n" \
"Examples: \n" \
"	geolocation -hv=12,04 \n" \
"	geolocation -hv=12,04 -res=hkm \n" \
"	geolocation -hv=24,02 -res=qkm -lf=sn_bound_10deg.txt\n" \
" \n" \
"AUTHOR \n" \
"	Code: S. Devadiga \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage: geolocation [-res=qkm|hkm|1km] -hv=<horizontal,vertical> [-lf=<filename>] -of=<filename> \n" \
" \n" \
"OPTIONS \n" \
"	-help				Display this help message \n" \
"	-res=qkm|hkm|1km		Reference resolution (default = 1km) \n" \
"	-hv=<horizontal,vertical>	Horizontal and vertical tile number. \n" \
"					Valids for horizontal: 0-35, vertical: 0-17 \n" \
"	-lf=<filename>			File containing lat/long limit for projection \n" \
"	                                If not input the points outside of projection \n" \
"                                       will not be identified. The file sn_bound_10deg.txt \n" \
"                                       is provided with this source code\n" \
"	-of=<filename>			output filename \n" \
" \n"

/*****************************************************************************
                            Prototypes.
*****************************************************************************/
int parse_cmd_geoloc(int argc, char **argv, char *hv_str, int *res, char *out_fname, char *map_fname);
int charpos(char *s, char c, int p);
void strmid(char *s1, int p1, int cnt, char *s2);
int is_arg_id(char *arg_str, char *arg_id);
void get_arg_val(char *arg_str, char *arg_val);
int get_tileid(char *hv_str, int *hn, int *vn);
int compute_geoloc(int row, int col, int hn, int vn, double *lon, double *lat, int res);
long sininvint(double r, double center_long, double false_east, double false_north);
long sininv(double x, double y, double *lon, double *lat);
int read_bound_table(int hn, int vn, char* map_fname, double *lon_min, double *lon_max,
                     double *lat_min, double *lat_max);

/*************************************************************************************/

int main(int argc, char *argv[])
/*
!C************************************************************************************

!Description:
  This routine computes the geolocation of every pixel in a given tile and then outputs
  latitude and longitude as two data sets in an output HDF file.
 
!Input Parameters: (none)

!Output Parameters:
  (returns)  Completion status:
             0 - successful completion
             1 - error exit

!Revision History: (none)

!Team-unique Header: (see file prolog)

!References and Credits: (see file prolog)

!Design Notes: (none)
  The tables containing the bounding coordinates for each tile are provided by 
  Robert Wolfe (Raytheon).
!END
**************************************************************************************/
{
  int res;
  int lim_st;
  int st = 0;
  int hn, vn;
  int i, j, k;
  int nrows, ncols;
  int32 start[2], edge[2];
  double lat, lon;
  double lon_min, lon_max;
  double lat_min, lat_max;
  float32 *lat_val, *lon_val;
  sds_t sds_info[2];
  char hv_str[10], out_fname[255], map_fname[255];
  char *sds_names[] = {"Longitude", "Latitude"};

  /* missing input tileid and projection parameter */
  if (argc == 1) {
    st = -1;
    fprintf(stderr, "Missing required input parameters \n");
    fprintf(stderr, "%s\n", USAGE);
    exit(1);
  }
  
  if ((argc == 2) && (strcmp(argv[1], "-help")==0)) {
    /* Print command line help when -help option is input */
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
   
  if (parse_cmd_geoloc(argc, argv, hv_str, &res, out_fname, map_fname) == -1) {
    fprintf(stderr, "%s\n", USAGE);
    exit(0);
  }
  else {
    if (get_tileid(hv_str, &hn, &vn) == -1) return 1;
    nrows = ncols = TILE_SIZE_1KM*res;
    for (k=0; k<2; k++) {
      sds_info[k].sd_id = (k == 0) ? -1 : sds_info[0].sd_id;
      sds_info[k].sds_id = -1;
      sds_info[k].data_type = DFNT_FLOAT32;
      strcpy(sds_info[k].name, sds_names[k]);
      sds_info[k].rank = 2;
      sds_info[k].dim_size[0] = nrows;
      sds_info[k].dim_size[1] = ncols;
      if (open_geoloc_sds(out_fname, &sds_info[k]) == -1) 
        st = -1;
    }
    if (st != -1) {
      if ((lat_val = (float32 *)calloc(ncols, sizeof(float32))) == NULL) {
        fprintf(stderr, "Error allocating memory for lat_val\n");
        st = -1;
      }
      if ((lon_val = (float32 *)calloc(ncols, sizeof(float32))) == NULL) {
        fprintf(stderr, "Error allocating memory for lon_val\n");
        st = -1;
      }
    }
    if (strlen(map_fname) > 0)
      lim_st = read_bound_table(hn, vn, map_fname, &lon_min, &lon_max, &lat_min, &lat_max);
    if (st != -1) {
      start[1] = 0;
      edge[0] = 1;
      edge[1] = ncols;
      for (i=0; i<nrows; i++) {
	start[0] = i;
        for (j=0; j<ncols; j++) {
	  if (compute_geoloc(i, j, hn, vn, &lon, &lat, res) == -1) {
	    fprintf(stderr, "Error computing geolocation at %d %d\n", i, j);
            lon_val[j] = (float32)GEOLOC_FILL_VAL;
            lat_val[j] = (float32)GEOLOC_FILL_VAL;
	  }
	  else {
	    lon_val[j] = (float32)lon;
	    lat_val[j] = (float32)lat;
	    if (lim_st == 1) {
	      if ((lon < lon_min) || (lon > lon_max) || (lat < lat_min) || (lat > lat_max)) {
                lon_val[j] = (float32)GEOLOC_FILL_VAL;
                lat_val[j] = (float32)GEOLOC_FILL_VAL;
	      }
	    }
	  }
        } /* for j . . */
        if (SDwritedata(sds_info[0].sds_id, start, NULL, edge, (VOIDP)lon_val) == FAIL) {
          fprintf(stderr, "Error writing line %d of SDS  %s\n", i, sds_info[0].name);  
          exit(0);
        }
        if (SDwritedata(sds_info[1].sds_id, start, NULL, edge, (VOIDP)lat_val) == FAIL) {
          fprintf(stderr, "Error writing line %d of SDS  %s\n", i, sds_info[1].name);  
          exit(0);
        }
      } /* for i . . */
      free(lon_val);
      free(lat_val);
      SDendaccess(sds_info[0].sds_id);
      SDendaccess(sds_info[1].sds_id);
      SDend(sds_info[0].sd_id);
    }
  }
  
  fprintf(stdout, "Processing done ! \n");
  return 0;
}

int parse_cmd_geoloc(int argc, char **argv, char *hv_str, int *res, char *out_fname, char *ln_fname)
/*
!C*********************************************************************************

!Function: parse_cmd_geoloc
       
!Description:
   Function parse_cmd_geoloc parse the argument string to the corresponding
   fields.

!Input Parameters: 
   argc:     input argument count
   argv:     input argument vector

!Input/output Parameters:
   hv_str:    horizontal and vertical tile number.
   res:       Resolution of the tile: 1km, hkm, qkm
   out_fname: projection type: ISIN or SIN   

!Output Parameters:

  ret   :    1  - successful completion
	     -1 - error exit   

!Revision History: (see file prolog) 

!Team-unique Header: (see file prolog)
 
!References and Credits: (see file prolog)

!Design Notes: (none)

!END
*****************************************************************************/     
{
  int i, st;
  char res_str[50];
  
  st = 1;
  res_str[0] = hv_str[0] = '\0';
  ln_fname[0] = out_fname[0] = '\0';
  
  for (i=1; i<argc; i++) {
    if (is_arg_id(argv[i], "-hv=") == 0) get_arg_val(argv[i], hv_str);
    else if (is_arg_id(argv[i], "-res=") == 0) get_arg_val(argv[i], res_str);
    else if (is_arg_id(argv[i], "-of") == 0) get_arg_val(argv[i], out_fname);
    else if (is_arg_id(argv[i], "-lf") == 0) get_arg_val(argv[i], ln_fname);
    else if (argv[i][0] == '-') fprintf(stderr, "Ignoring invalid option %s\n", argv[i]); 
  }
  
  if (strlen(ln_fname) <= 0) {
    fprintf(stderr, "The file containing lat/long limits for map projection is not input\n");
    fprintf(stderr, "Pixels outside of map projections in tiles on edges of global projection \n");
    fprintf(stderr, "cannot be identified and hence may contain incorrect geolocation values\n");
  }
  if (strlen(out_fname) <= 0) {
    fprintf(stderr, "Missing output filename\n");
    st = -1;
  }
  if (strlen(hv_str) <=0) {
    fprintf(stderr, "Missing tile id\n");
    st = -1;
  }
  if (strlen(res_str) <= 0) { 
    fprintf(stderr, "Resolution not input, set to default 1km\n"); 
    *res = 1;
  }
  else if (strcmp(res_str, "1km") == 0) *res = 1;
  else if (strcmp(res_str, "hkm") == 0) *res = 2;
  else if (strcmp(res_str, "qkm") == 0) *res = 4;
  else {
    fprintf(stderr, "Invalid input resolution \n"); 
    st = -1;
  }

  return st;
}

int get_tileid(char *hv_str, int *hn, int *vn)
/* get the coordinates frm the user input string. */
{
  int st = 1;
  int p1, len;
  char tmp_str[10];

  if ((p1 = charpos(hv_str, ',', 0)) == -1) st = -1;
  else
  {
    len = (int)strlen(hv_str);
    strmid(hv_str, 0, p1, tmp_str);
    *hn = (int)atoi(tmp_str);
    if ((*hn < 0) || (*hn > 35)) {
      fprintf(stderr, "Invalid Horizontal tile number: %s\n", tmp_str);
      st = -1;
    }
    strmid(hv_str, p1+1, len-p1-1, tmp_str);
    *vn = (int)atoi(tmp_str);
    if ((*vn < 0) || (*vn > 17)) {
      fprintf(stderr, "Invalid Vertical tile number: %s\n", tmp_str);
      st = -1;
    }
  }

  return st;
}

int open_geoloc_sds(char *out_fname, sds_t *sds_info) 
{
  int len;
  int status = 1;
  float32 *fill_val;
  char units[] = "Degrees";

  if (sds_info->sd_id == -1) {
    if ((sds_info->sd_id = SDstart(out_fname, DFACC_CREATE)) == FAIL) {
      fprintf(stderr, "Cannot create the hdf file %s\n", out_fname);
      status = -1;
    }
  }
  if (sds_info->sd_id != -1) {
    if ((sds_info->sds_id = SDcreate(sds_info->sd_id, sds_info->name,
                sds_info->data_type, sds_info->rank, sds_info->dim_size)) == FAIL) {
      fprintf(stderr, "Cannot create the SDS %s\n", sds_info->name);
      status = -1;
    }
    else {
      len = (int)strlen(sds_info->name);
      if (SDsetattr(sds_info->sds_id, "long_name", DFNT_CHAR8, len,
                        (VOIDP)sds_info->name) == FAIL)
        fprintf(stderr, "Could not write the attribute 'long_name' in %s\n", sds_info->name);

      /* Maosheng Zhao, 3/1/2018, add units  */
      len = (int)strlen(units);
       if (SDsetattr(sds_info->sds_id, "units", DFNT_CHAR8, len,
			(VOIDP)units) == FAIL)
	fprintf(stderr, "Could not write the attribute 'units' in %s\n", units);
      len = 1;
      fill_val = (float32 *)calloc(1, sizeof(float32));
      fill_val[0] = GEOLOC_FILL_VAL;
      if (SDsetattr(sds_info->sds_id, "_FillValue", DFNT_FLOAT32, len,
                        (VOIDP)fill_val) == FAIL)
        fprintf(stderr, "Could not write the attribute '_FillValue' in %s\n", sds_info->name);
      free(fill_val);
    }
  }
  return status;
}

int compute_geoloc(int row, int col, int hn, int vn, double *lon, double *lat, int res)
     /* compute the geolocation based on the map projection */
/* Maosheng Zhao, Mar, 06, 2019, as GCTP SIN forward and invers functions have issue, I just use mathmatic equations */
{
  long st = 1;
  double xproj, yproj;
  double lon_min, lon_max, lat_min, lat_max;

  double   xproj_ul_corner = -20015109.354;
  double   yproj_ul_corner = 10007554.677;
  double   x_pixel_size = 926.62543305;
  double   y_pixel_size = 926.62543305;
  double   xtile_size = 1200;
  double   ytile_size = 1200;
  double   xn, yn;
  char proj_name[] = "File_name";
  /* Maosheng Zhao, Mar 6, 2019, add the below */
  double R = 6371007.181;
  double lon_negtv_180 = -1. * (double)PI;
  double lon_postv_180 = 1. * (double)PI;

  /* calculate project parameter */
  xn = (double) hn * xtile_size;
  yn = (double) vn * ytile_size;

  xproj = xn * x_pixel_size + (col+0.5) * x_pixel_size / (double)res + xproj_ul_corner;
  yproj = - yn * y_pixel_size - (row+0.5) * y_pixel_size / (double)res + yproj_ul_corner;

  /* Maosheng Zhao, Mar 6, 2019, forget about using GCTP functions of sininvint() and sininv()
   * because they give normal longitude values for area out of SIN map projection, ridiculous */
  *lat = yproj / R;
  *lon = xproj / (R * cos(*lat));

  if(*lon < lon_negtv_180 || *lon > lon_postv_180)
  {
     *lon = (double)GEOLOC_FILL_VAL;
     *lat = (double)GEOLOC_FILL_VAL;
  }
  else
  {
     *lon = (*lon * 180.0)/PI;
     *lat = (*lat * 180.0)/PI;
  }

  return st;
}

int read_bound_table(int hn, int vn, char* table_fname, double *lon_min, double *lon_max,
                     double *lat_min, double *lat_max)
{
  FILE *fp;
  int in_tile_hn, in_tile_vn;
  int tile_find, c;
  double c1, c2, c3, c4;

  tile_find = -1;
  if ((fp = fopen(table_fname, "r")) == NULL)
    fprintf(stderr, "Cannot open projection boundary file %s\n", table_fname);
  else
  {
    c = fgetc(fp);
    while (c != EOF)
    {
      ungetc(c, fp);
      fscanf(fp, "%d %d", &in_tile_vn, &in_tile_hn);
      fscanf(fp, "%lf %lf %lf %lf", &c1, &c2, &c3, &c4);
      if ((in_tile_vn == vn) && (in_tile_hn == hn))
        {
          *lon_min = c1;
          *lon_max = c2;
          *lat_min = c3;
          *lat_max = c4;
          tile_find = 1;
          break;
        }
      c = fgetc(fp);
    }
    fclose(fp);
  }
  return tile_find;
}
