/****************************************************************************
!C

!File: main_util.c

!Description:
  Contains routines for follwoing tasks
    -get the product short name
    -get the pointer filename
    -locate a given modis file
    -read and parse the input pointer to get input filenames
    -routines for parsing the argument name and values
    -check the validity of bit string for a given sds
    -parse the bit string to get bit number values
    -rm pathname in a filename
    -routine used in command line parsing
    -get qa tool environment variable
    -read the time for a granule or tile number for a tile
    -read a line from the file

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:

  Version 1.0 October, 2002

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Yi (Sophyi) Zhang
    LDOPE                             Science Systems and Applications Inc.
    sophyi@ltpmail.gsfc.nasa.gov      NASA/GSFC Code 922
    phone: 301-614-5497               Greenbelt, MD 20771

    David Roy
    LDOPE                             University of Maryland, Department of Geography
    droy@kratmos.gsfc.nasa.gov        NASA/GSFC Code 922 (B32)
    phone: 301-614-5571               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "mfhdf.h"

#include "str_op.h"
#include "alloc_mem.h"
#include "qa_tool.h"
#include "meta.h"
#include "main_util.h"


int parse_stdin(int *argc, char **argv)
/*****************************************************************************
        Reads the argument from file input as stdin. Modified to follow exact
        unix syntax (space, line breaks, quotes and combination of quotes etc).
        Assumes that the stdin conatins only arguments.
******************************************************************************/
{
  int cnt = 0;
  int len = 0;
  int i = 0, j = 0;
  int st_quote = 0;
  char quote = '\0';
  char tmp[MAX_STR_LEN];

  cnt = 1;
  j = 0;
  st_quote = 0;
  while (fscanf(stdin, "%s", tmp) == 1)
  {
    len = (int)strlen(tmp);
    if (st_quote == 1) argv[cnt][j++] = ' ';
    for (i=0; i<len; i++)
    {
      if ((tmp[i] != '"') && (tmp[i] != '\'')) argv[cnt][j++] = tmp[i];
      else
      {
        if (st_quote == 0)
        {
          st_quote = 1;
          quote = tmp[i];
          if (j > 0)
          {
            argv[cnt][j] = '\0';
            j = 0;
            cnt++;
          }
        }
        else if (tmp[i] == quote)
        {
          st_quote = 0;
          if (j > 0)
          {
            argv[cnt][j] = '\0';
            j = 0;
            cnt++;
          }
        }
        else argv[cnt][j++] = tmp[i];
      }
    }
    if ((st_quote != 1) && (j > 0))
    {
      argv[cnt][j] = '\0';
      j = 0;
      cnt++;
    }
    if (cnt == MAX_NUM_PARAM) break;
  }
  if (cnt == MAX_NUM_PARAM)
    fprintf(stderr, "Too many input arguments. Reading only first %d arguments . . .\n", cnt);
  *argc = cnt;
  if (st_quote == 1) return -1;
  else return 1;
}


int is_arg_id(char *arg_str, char *arg_id)
/**************************************************************************************
	return 0 if arg_id is initial part of the arg_str else return -1
**************************************************************************************/
{
  int p1 = 0;

  p1 = charpos(arg_str, '=', 0);
  if (p1 == -1) return -1;
  if (strncmp(arg_str, arg_id, p1) == 0)
   return 0;
  else return -1;
}

void get_arg_val_arr(char *arg_str, char **arg_val, int *arg_cnt)
/************************************************************************************
	read the argument values from arg_str and return the number of argument 
	values as arg_cnt and the argument values in the array arg_val.
************************************************************************************/
{
  int p1 = 0, p2 = 0, len = 0;

  p1 =  0;
  p2 = charpos(arg_str, '=', p1);
  if (p2 == -1) return;
  p1 = p2 + 1;
  p2 = charpos(arg_str, ',', p1);
  while ((p2 != -1) && (*arg_cnt < MAX_NUM_PARAM))
  {
    strmid(arg_str, p1, p2-p1, arg_val[*arg_cnt]);
    ++*arg_cnt;
    p1 = p2 + 1;
    p2 = charpos(arg_str, ',', p1);
  }
  if (*arg_cnt >= MAX_NUM_PARAM)
  {
    fprintf(stderr, "Too many parameters in option %s\n", arg_str);
    fprintf(stderr, "Considering only %d number of parameter values\n", MAX_NUM_PARAM);
  }
  else
  {
    len = (int)strlen(arg_str);
    if (p1 < len)
    {
      strmid(arg_str, p1, len-p1, arg_val[*arg_cnt]);
      ++*arg_cnt;
    }
  }
} 


void get_arg_val(char *arg_str, char *arg_val)
/************************************************************************************
	Read single argument val arg_val from arg_str
************************************************************************************/
{
  int p1 = 0, len = 0;

  p1 = charpos(arg_str, '=', 0);
  if (p1 == -1) return;
  p1 = p1 + 1;
  len = (int)strlen(arg_str);
  if (len == p1) return;
  else
    strmid(arg_str, p1, len-p1, arg_val);
}

void rm_path(char *fname)
/*****************************************************************************
	Remove the path from the file name fname and return the filename in
	fname itself
****************************************************************************/
{
  int i = 0, j = 0, k = 0;
  int len = 0;

  len = (int)strlen(fname);
  for (i=len; i>0; i--)
    if (fname[i] == '/') break;
  if (i != 0)
  {
    for (j = 0, k = i+1; k < len; j++, k++)
      fname[j] = fname[k];
    fname[j] = '\0';
  }
}


void get_qa_tool_env(char *env_var, char **env_val)
/********************************************************************************
	Get the value of the environment variable env_var and return it in
	env_val. The environment variable values follow unix syntax.
********************************************************************************/ 
{
  int cnt = 0;
  int len = 0;
  int p1 = 0, p2 = 0;
  char *tmp_val = NULL;
  char tmp_path[MAX_PATH_LENGTH];

  tmp_val = getenv(env_var);
  if (tmp_val != NULL)
  {
    cnt = 0;
    p1 = 0;
    while ((p2 = charpos(tmp_val, ':', p1)) != -1)
    {
      strmid(tmp_val, p1, p2-p1, tmp_path);
      strtrim(tmp_path);
      sprintf(env_val[cnt++], "%s/",  tmp_path);
      p1 = p2 + 1;
    }
    len = (int)strlen(tmp_val);
    strmid(tmp_val, p1, len-p1, tmp_path);
    strtrim(tmp_path);
    sprintf(env_val[cnt++], "%s/",  tmp_path);
  }
}

int get_pnt_fname(char *hdf_fname, char **modis_fpath, char *ptr_fname, char *pnt_fname)
/***********************************************************************************
	Gets pointer filename used in the production of product file hdf_fname
	modis_fpath contains a list of pathnames to search for a modis product
	On success returns 1 and pnt_fname contains the pointer filename
	else returns -1 and pnt_fname is not meaningful
***********************************************************************************/
{
  FILE *fp = NULL;
  int st = 0, status = 0;
  int pos1 = 0, pos2 = 0;
  int len = 0, meta_cnt = 0;
  char *cmeta_str = NULL, **meta_val = NULL;
  char fname[MAX_PATH_LENGTH] = {'\0'};
  char tmp_fname[MAX_PATH_LENGTH] = {'\0'};
  char meta_name[MAX_META_NAME_LEN] = {'\0'};
  char pnt_sname[10] = {'\0'}, prod_sname[10] = {'\0'};
  char hdf_fname_no_path[MAX_PATH_LENGTH] = {'\0'};
  char pnt_esdt_name[6] = {'\0'};

  status = -1;
  if ((meta_val = (char **)Calloc2D(5, MAX_META_VAL_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for meta_val in get_pnt_fname\n");
  else if ((cmeta_str = get_attr_metadata(hdf_fname, "CoreMetadata.0")) != NULL)
  {
    meta_cnt = 0; 
    strcpy(meta_name, "INPUTPOINTER");
    get_sel_metadata(cmeta_str, meta_name, meta_val, &meta_cnt, 0);
    if (meta_cnt == 0)
      fprintf(stderr, "Metadata INPUTPOINTER not found in file %s", hdf_fname);
    else
    {

      pos1 = charpos(meta_val[0], '"', 0);
      while ((pos1 != -1) && (status == -1))
      {
	st = 0; pos1++;
        pos2 = charpos(meta_val[0], '"', pos1);
        strmid(meta_val[0], pos1, pos2-pos1, fname);
        strtrim(fname);
        if (strstr(ptr_fname, fname) != NULL)
        {
	  st = 1;
	  strcpy(fname, ptr_fname);
        }
	else
	{
	  sprintf(tmp_fname, "%s/%s", ptr_fname, fname);
	  if ((fp = fopen(tmp_fname, "r")) != NULL)
	  {
	    st = 1;
	    strcpy(fname, tmp_fname);
	    fclose(fp);
	  }
	}
	if (st == 0)
          st = find_file(modis_fpath, fname);

	if (st > 0)
        {
          if (get_prod_sname(fname, prod_sname) != -1)
          {
            len = (int)strlen(prod_sname);
	    if (len > 5)
            {
              strmid(prod_sname, 0, 5, pnt_sname); 
              if ((strcmp(pnt_sname, "MODPT") == 0) || 
	      	   strcmp(pnt_sname, "MYDPT") == 0)
		   {
		  	 status = 1;
		   }	
	    }
          }
        }
        pos1 = charpos(meta_val[0], '\"', pos2+1);
      }
      if (status == -1)
      {
	/* MSZ 3/7/2015, to make error information more informative to users */
	strcpy(hdf_fname_no_path, hdf_fname);
	rm_path(hdf_fname_no_path);
	strncpy(pnt_esdt_name, hdf_fname_no_path, 3);
	/*
	strcat(pnt_esdt_name, "PT");
        fprintf(stderr, "No pointer file %s found. Please copy the associatated granule pointer file to current path or use -ptr= to specify the location of the pointer file. Exit...\n", pnt_esdt_name);
	*/
      }
      else
        strcpy(pnt_fname, fname);
    }
    free(cmeta_str);
    Free2D((void **)meta_val);
  }
  return status;
}


int find_file(char **fpath, char *fname)
/*************************************************************************************
	Returns 1 if file fname is found in one of the paths listed in fpath
	else return -1
*************************************************************************************/
{
  FILE *fp = NULL;
  int i = 0;
  /* MSZ 7/19/2018, correct the bug by increasing the length */
  /*char tmp_fname[80];*/
  char tmp_fname[1024];

  for (i=0; i<MAX_NUM_PATH; i++)
    if (strlen(fpath[i]) > 1)
    {
      sprintf(tmp_fname, "%s%s", fpath[i], fname);
      if ((fp = fopen(tmp_fname, "r")) != NULL) break;
    }
  if (i < MAX_NUM_PATH) 
  {
    fclose(fp);
    strcpy(fname, tmp_fname);
    return 1;
  }
  else
  {
    if ((fp = fopen(fname, "r")) != NULL) 
    {
      fclose(fp);
      return 1;
    } else return -1;
  }
}


int get_prod_sname(char *hdf_fname, char *prod_sname)
/*********************************************************************************
	Read the product SHORTNAME of the hdf files
	On success the function returns 1 and prod_sname contains the SHORTNAME
	else function returns -1 and prod_sname is not meaningfull
**********************************************************************************/
{
  int len = 0;
  int status = 0;
  int meta_cnt = 0;
  char *cmeta_str = NULL, **meta_val = NULL;
  char meta_name[MAX_META_NAME_LEN];

  status = -1;
  meta_cnt = 0;
  if ((meta_val = (char **)Calloc2D(5, MAX_META_VAL_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for meta_val in get_prod_sname\n");
  else if ((cmeta_str = get_attr_metadata(hdf_fname, "CoreMetadata.0")) != NULL)
  {
    strcpy(meta_name, "SHORTNAME");
    get_sel_metadata(cmeta_str, meta_name, meta_val, &meta_cnt, 0);
    if (meta_cnt == 0)
      fprintf(stderr, "No metadata SHORTNAME found in file %s", hdf_fname);
    else
    {
      status = 1;
      len = (int)strlen(meta_val[0]);
      strmid(meta_val[0], 1, len-2, prod_sname);
    }
    free(cmeta_str);
    Free2D((void **)meta_val);
  }
  return status;
}

void get_bit_num_arr(char *bn_str, int *cnt, int *bn_arr)
/*************************************************************************************
	Read the bit numbers from the bit number string bn_str into an ineger 
	array bn_arr and number of bits into cnt
*************************************************************************************/
{
  int i = 0, j = 0, tmp = 0;
  int len = 0, len2 = 0;
  int val = 0, val1 = 0, val2 = 0;
  int pos1 = 0, pos2 = 0, pos3 = 0, bcnt = 0;
  char num[10], t1[5], t2[5];

  bcnt = 0;
  pos1 = 0;
  strtrim(bn_str);
  len = (int)strlen(bn_str);
  while (pos1 < len)
  {
    pos2 = charpos(bn_str, ',', pos1);
    if (pos2 == -1) pos2 = len;
    strmid(bn_str, pos1, pos2-pos1, num);
    pos3 =  charpos(num, '-', 0);
    if (pos3 == -1)
    {
      strtrim(num);
      val = atoi(num);
      bn_arr[bcnt++] = val;
    }
    else
    {
      len2 = (int)strlen(num);
      strmid(num, 0, pos3, t1);
      strmid(num, pos3+1, len2-pos3, t2);
      strtrim(t1);
      strtrim(t2);
      val1 = atoi(t1);
      val2 = atoi(t2);
      for (i=val1; i<=val2; i++)
        bn_arr[bcnt++] = i;
    }
    pos1 = pos2+1;
  }
  *cnt = bcnt;

  for (i=0; i<bcnt-1; i++)
  for (j=bcnt-1; j>i; j--)
    if (bn_arr[j] < bn_arr[j-1])
    {
      tmp = bn_arr[j];
      bn_arr[j] = bn_arr[j-1];
      bn_arr[j-1] = tmp;
    }
}

int validate_bit_str(char *bit_str)
/* *************************************************************************************
      validate user input bit string,  return 1 if the input bit string is valid. return
      0 if bit string is invalid. 
***************************************************************************************/
{
  int i = 0;
  int len = 0, len2 = 0;
  int val = 0, val1 = 0, val2 = 0;
  int pos1 = 0, pos2 = 0, pos3 = 0, blen = 0;
  char num[10], t1[5], t2[5];
  int ret = 0;

  ret = 1;
  pos1 = 0;
  strtrim(bit_str);
  len = (int)strlen(bit_str);

  /*  return ret; */
  
  while (pos1 < len)
    {
      pos2 = charpos(bit_str, ',', pos1);
      if (pos2 == -1) pos2 = len;
      strmid(bit_str, pos1, pos2-pos1, num);
      pos3 =  charpos(num, '-', 0);
      if (pos3 == -1)
	{
	  strtrim(num);
	  val = atoi(num);
	  if (val >= MAX_NUM_BITS)
		{
		  fprintf(stderr, "Input bit parameter out of range. \n");
		  fprintf(stderr, "The bits you entered is %s, ignore.\n", num);
		  pos1 = pos2 + 1;
		  ret = 0;
		  continue;
		}
	}
      else
	{
	  if (pos3 == 0)
	    {
	      fprintf(stderr, "Bit number can't be negative. \n");
	      fprintf(stderr, "The bits you entered is %s, ignore.\n", num);
	      pos1 = pos2 +1;
	      ret = 0;
	      continue;
	    }
          else
	    {
	      len2 = (int)strlen(num);
	      strmid(num, 0, pos3, t1);
              
              if ((len2-pos3) > 5)
		{
		  fprintf(stderr, "Input for bit number is too large. \n");
		  ret = 0;
		  return ret;
		}
	      else
		{
		  strmid(num, pos3+1, len2-pos3, t2);
		  strtrim(t1);
		  strtrim(t2);
		  
		  blen = strlen(t1);
		  for (i=0; i<blen; i++)
		    {
		      if (isdigit(t1[i]) == 0)
			{
			  fprintf(stderr, "Input bit number in wrong format. Should be a positive integer.\n");
			  ret = 0;
			  return ret;
			}
		    }

		  blen = strlen(t2);
		  for (i=0; i<blen; i++)
		    {
		      if (isdigit(t2[i]) == 0)
			{
			  fprintf(stderr, "Input bit number in wrong format. Should be a positive integer.\n");
			  ret = 0;
			  return ret;
			}
		    }

		  val1 = atoi(t1);
		  val2 = atoi(t2);
	
		  if (val2 >= MAX_NUM_BITS || val1 >= MAX_NUM_BITS)
		    {
		      fprintf(stderr, "Bit number out of range. \n");
		      fprintf(stderr, "The bits you entered is %s, ignore.\n", num);
		      ret = 0;
		      return ret;
		    }
		  if (val2 <= val1)
		    {
		      fprintf(stderr, "Ending Bit number should be greater than starting Bit. \n");
		      fprintf(stderr, "The bits you entered is %s, ignore.\n", num);
		      ret =0;
		      return ret;
		    }
		}
	    }
	}
      pos1 = pos2+1;
    }
  return ret;
}



void get_esdt_tileid(char *fname, char *esdt, char *tile_id, char *jday, char *ver_id)
{
  int i = 0, j = 0, k = 0, len = 0;

  esdt[0] = tile_id[0] = '\0';
  jday[0] = ver_id[0] = '\0';
  len = (int)strlen(fname);
  for (k=len-1; k>=0; k--)
    if (fname[k] == '/') break;
  k++;
  for (j=k, i=0; j<len; j++, i++)
  {
    if (fname[j] == '.') { esdt[i] = '\0'; break; }
    else esdt[i] = fname[j];
  }
  k = j+2;
  for (j=k, i=0; j<len; j++, i++)
  {
    if (fname[j] == '.') { jday[i] = '\0'; break; }
    else jday[i] = fname[j];
  }
  k = j+1;
  for (j=k, i=0; j<len; j++, i++)
  {
    if (fname[j] == '.') { tile_id[i] = '\0'; break; }
    else tile_id[i] = fname[j];
  }
  k = j+1;
  for (j=k, i=0; j<len; j++, i++)
  {
    if (fname[j] == '.') { ver_id[i] = '\0'; break; }
    else ver_id[i] = fname[j];
  }
  if ((k = charpos(esdt, '_', 0)) != -1)
    esdt[k] = '\0';
}


void sort_values(int *values, int n)
/* sort the input values(integer) to ascending order. Total input  
 * values is n */ 
{
  int i = 0, j = 0, tmp = 0;
 
  for (i=n-1; i>0; i--)
    for (j=i-1; j>=0; j--)
      if (values[i] < values[j])
      {
	tmp = values[i];
        values[i] = values[j];
        values[j] = tmp;
      }
}

void sort_fvalues(float *values, int n)
/* sort the input values(float) to ascending order. Total input 
 * values is n */
{
  int i = 0, j = 0;
  float tmp = 0.;
  
  for (i=n-1; i>0; i--)
    for (j=i-1; j>=0; j--)
      if (values[i] < values[j])
      {
	tmp = values[i];
        values[i] = values[j];
        values[j] = tmp;
      }
}

int read_clr_table(char *in_fname, int *clr_table)
/* read in the color table */
{
  FILE *fp = NULL;
  int st = -1;
  int i = 0, id = 0, r = 0, g = 0, b = 0;

  if ((fp = fopen(in_fname, "r")) == NULL)
    fprintf(stderr, "Cannot open input file: %s\n", in_fname);
  else
  {
    i = id = 0;
    while ((fscanf(fp, "%d %d %d", &r, &g, &b) == 3) && (i < 256))
    {
      clr_table[id++] = r;
      clr_table[id++] = g;
      clr_table[id++] = b;
      ++i;
    }
    fclose(fp);
    fprintf(stderr, "Number of colors read from color table: %d\n", i);
    st = i;
  }
  return st;
}                             

int get_npp_input_files(char *fname, char *iptr_id, char **iptr_fnames)
/***********************************************************************************
	Returns number of input files used in the production of fname that matches
	the string iptr_id. iptr_fname contains the resulting filenames. if iptr_id 
	== 'all', all filenames are read from metadata INPUTPOINTER. else matching 
	filenames (match to iptr_id) are returned. else filenames from INPUTPOINTER
	with matching SHORTNAME (match to iptr_id) are returned.
***********************************************************************************/ 
{
  int i, is_l2g, len;
  int pos1, pos2, pos3;
  int m_cnt, tot_fcnt;
  int iptr_fcnt, iptr_id_len;
  int num_op_gran, num_ip_gran;
  char **tmp_fnames, **modis_fpath;
  char *meta_str = NULL;
  char tmp[MAX_STR_LEN], meta_name[MAX_META_NAME_LEN];
  int32 attr_type, attr_cnt;

  if ((tmp_fnames = (char **)Calloc2D(MAX_NUM_FILE, MAX_STR_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for tmp_fnames in get_input_files \n");
  if ((modis_fpath = (char **)Calloc2D(MAX_NUM_PATH, MAX_PATH_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for modis_fpath in get_input_files \n");

  if ((tmp_fnames == NULL) || (modis_fpath == NULL))
  {
    Free2D((void **)tmp_fnames);
    Free2D((void **)modis_fpath);
    return -1;
  }

  tot_fcnt = 0;
  iptr_fcnt = 0;

  m_cnt = 0;
  strcpy(meta_name, "InputPointer");
  meta_str = (char *)get_attr_sel_metadata(fname, meta_name, &attr_type, &attr_cnt);  

  if (meta_str == NULL)
  {
    fprintf(stderr, "Metadata INPUTPOINTER not found in %s\n", fname);
    exit(1);
  }
  else
  {
    get_qa_tool_env(MODIS_ENV, modis_fpath);
    len = (int)strlen(meta_str);
    if (strcmp(iptr_id, "all") == 0)
    {
      /*fprintf(stderr, "\nReading all the filenames in the InputPointer\n");*/
      pos1 = 0;
      pos2 = 0;
      pos2 = charpos(meta_str, ',', 0);
      pos3 = charpos(meta_str, ':', 0);      
      while ((pos2 != -1) || (pos3 != -1))
      {
        if (pos2 == -1) pos2 = pos3;
        else if ((pos3 != -1) && (pos3 < pos2)) pos2 = pos3;
	strmid(meta_str, pos1, pos2 - pos1, tmp);
	strtrim(tmp); 
        rm_path(tmp);

	/* MSZ 3/9/2018, comment this out to get granule file names */	
        /*if (find_file(modis_fpath, tmp) == 1)*/
        strcpy(iptr_fnames[iptr_fcnt++], tmp); 	
	pos1 = pos2 + 1;
	pos2 = charpos(meta_str, ',', pos1);
        pos3 = charpos(meta_str, ':', pos1); 			
      }
      strmid(meta_str, pos1, len - pos1, tmp);
      strtrim(tmp);        
      rm_path(tmp); 
      /* MSZ 3/9/2018, comment this out to get granule file names */
      /*if (find_file(modis_fpath, tmp) == 1)*/
      strcpy(iptr_fnames[iptr_fcnt++], tmp);  
/*printf("iptr_fnames[0]===%s\n", iptr_fnames[0]);*/
/*printf("iptr_fcnt===%d\n", iptr_fcnt);*/               
    }
    else
    {
      fprintf(stderr, "\nComparing %s to filenames in the INPUTPOINTER\n", iptr_id);
      iptr_id_len = (int)strlen(iptr_id);
      pos1 = 0;
      pos2 = 0;
      pos2 = charpos(meta_str, ',', 0);
      pos3 = charpos(meta_str, ':', 0);   
      while ((pos2 != -1) || (pos3 != -1))
      {
        if (pos2 == -1) pos2 = pos3;
        else if ((pos3 != -1) && (pos3 < pos2)) pos2 = pos3;
	strmid(meta_str, pos1, pos2 - pos1, tmp);
	strtrim(tmp); 
        rm_path(tmp);
        /* MSZ 3/9/2018, comment this out to get granule file names */
        /*if (find_file(modis_fpath, tmp) == 1)*/
          strcpy(tmp_fnames[tot_fcnt++], tmp); 	
	pos1 = pos2 + 1;
	pos2 = charpos(meta_str, ',', pos1);
        pos3 = charpos(meta_str, ':', pos1); 			
      }
      strmid(meta_str, pos1, len - pos1, tmp);
      strtrim(tmp); 
      rm_path(tmp);
      /* MSZ 3/9/2018, comment this out to get granule file names */
      /*if (find_file(modis_fpath, tmp) == 1)*/
        strcpy(tmp_fnames[tot_fcnt++], tmp);   	      
      for (i=0; i<tot_fcnt; i++)
      {   
	if (strncmp(iptr_id, tmp_fnames[i], iptr_id_len) == 0)	    
	  /* MSZ 3/9/2018, comment this out to get granule file names */
          /*if (find_file(modis_fpath, tmp_fnames[i]) == 1)*/
	  {
            strcpy(iptr_fnames[0], tmp_fnames[i]);
	    iptr_fcnt++;
	  }  	               
      }              
    }
    if (iptr_fcnt == 0)
      fprintf(stderr, "No matching input files found in the InputPointer \n");        		

    if(meta_str) free(meta_str);

  }    
  Free2D((void **)tmp_fnames);
  Free2D((void **)modis_fpath);

  return iptr_fcnt;
}


