/****************************************************************************
!C

!File: mask_sds.c

!Description:
  Contains routines for masking sds.

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original January 1998. Version 1.0
  Modifid March/April 1998. (Changes to incorporate command line format)
			    (Changes to incorporate new mask specification style)
  Modified June 1998 Version 1.1 (Changes to process multiple input files)
  Modified Jan 1999 Version 1.1 (Changes to apply relational operators on sds values )
  Modified July 1999 Version 1.2 (Changes to accomodate 3D/4D sds)
			         (Changes to process multiple input SDS)
  Modified August 1999 Version 1.2 (Handle L2G transparent to L2/L3/L4)
  				   (Create output ascii log file and(or) HDF file)
  Modified September 1999. (included -meta option)
  Modified September 2001. (Fixed bug while improting to Linux)
  Modified July 2011 by Xiaoping Zhang, Sigma space corp.
  	(to add fill value option on command line)
  Modified March 2013 by Xiaoping Zhang, Sigma space corp. 	
  	(to work for MODIS/NPP l2g lite)
  Modified February 2016 by Xiaoping Zhang, SSAI. 	
  	modified to get modis/viirs data status.	
	fixed a bug for output 2D SDS from 3D/4D if a SDS layer is specified.
  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	3) update the L2G check function and mask the tool exit if the input is L2G

  Modified in February, 2020. Version 3.6 by Maosheng Zhao, SSAI.
  1. update process_mask_data() in mask_sds_lib.c to deal with the negative range input
  2. output HDF4 with chunked internal compression
	
!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "qa_tool.h"
#include "sds_rw.h"
#include "alloc_mem.h"
#include "main_util.h"
#include "meta.h"
#include "str_op.h"
#include "mask_sds_lib.h"
#include "l2g.h"

#define MAX_NSDS 20

#define HELP \
"NAME \n" \
"     mask_sds - use a combination of SDS to mask another SDS \n" \
" \n" \
"SYNOPSIS \n" \
"     mask_sds -help \n" \
"     mask_sds -mask=mask1[,AND|OR,mask2...] -sds=SDS_name filename \n" \
"              [-of=output_file] [-log=logfile] [-meta] -fill_value=<fillVal> filename \n" \
" \n" \
"     echo -mask=mask1[,AND|OR,mask2...] -sds=SDS_name filename \n" \
"          [-of=output_file] [-log=logfile] [-meta] -fill_value=<fillVal> filename | mask_sds \n" \
" or  mask_sds < argument_file \n" \
" \n" \
"DESCRIPTION \n" \
"     Produces data of pixels from the specified SDS that verify a given mask. \n" \
"     The mask is a combination of individual masks combined with the logical \n" \
"     AND or OR operator. Each individual mask is based on a separate SDS from \n" \
"     same or separate files. The mask is activated according to a list of bits \n" \
"     and associated bit values of each individual SDS (or the SDS value itself). \n" \
"     The SDS used to compose the mask can belong to any file, including the \n" \
"     input file itself. \n" \
"     The values of pixels satisfying the mask are output into an ASCII file and(or)\n" \
"     stored as SDS in a user specified output HDF file. When more than one input \n" \
"     SDS is masked multiple SDS output are created and(or) values of output pixels \n" \
"     from different SDS are listed in different columns of the same ASCII file. \n" \
"     The tool handles user specified layers of 3D and 4D in the mask and also in \n" \
"     the input HDF file to be masked. Effective mask is always a 2D SDS. SDS to be \n" \
"     masked can be 2D/3D/4D. The same 2D mask is used to mask all user specified \n" \
"     layers in 3D and 4D. \n" \
"     To specify a L2G SDS enter the layer number as dot extension in the SDS name. However \n" \
"     do not use the _1 or _c as given in the HDF file. They stand for first layer and \n" \
"     additional layers in compact form. In MOD09GQK SDS sur_refl_b03_1 and sur_refl_b03_c \n" \
"     contain 1st layer and additional layers in compact form. To specify observation layer \n" \
"     number 4 of sds sur_refl_b03 file, in the command enter -sds=file,sur_refl_b03.4 \n" \
" \n" \
"OPTIONS \n" \
"     -help         Display this help message \n" \
"     -mask=string  Define the mask to apply as combination of individual masks \n" \
"                   separated by AND or OR. Each individual mask consists of \n" \
"                   a file, a SDS, a bit descriptor, and a binary bit value \n" \
"                   string to match: filename,SDS_name,bits<operator>string. \n" \
"                   Valid operators are >, <, <=, >=, ==, !=. Rather than duplicate \n" \
"                   the file and SDS names, the * character can be used after \n" \
"                   initial specification. The bit descriptor is a list of bits or \n" \
"                   ranges of bits, similar to the Unix filename substitution. \n" \
"                     Ex: -mask=file1,SDS1,0-2,4==0101,AND,*,SDS2,2,3==10 \n" \
"                   Warning: Although a bit range in the bit descriptor is \n" \
"                   specified by the lower bit followed by the higher bit, the \n" \
"                   bit value string are in reverse order, i.e. the lower bit on \n" \
"                   the right (in the example 0-2,4=0101 is for 4,2,1,0=0101). \n" \
"                   If the bit descriptor is omitted, then the whole range of \n" \
"                   bits is assumed, and the value string is parsed as a decimal \n" \
"                   value. This is a convenient way to refer to a specific \n" \
"                   value, instead of a list of bits. \n" \
"                     Ex: -mask=file,SDS,>=200 \n" \
"                   If SDS is a 3D or 4D specify the dimension value as dot \n" \
"                   extension in the sds name: \n" \
"                     Ex: -mask=file,SDS.n1,0-2==001,AND,file,SDS.n2.m2,0-1==01 \n" \
"                   If SDS is a L2G in compact form specify the observation layer \n" \
"                   number as dot extension in the sds name: \n" \
"                     Ex: -mask=file,SDS.n1,0-2==001,AND,file,SDS.n2.m2,0-1==01 \n" \
"     -sds=SDS_name SDS to mask. If SDS is a 3D or 4D specify the dimension value \n" \
"                   as dot extension in the sds name. If SDS is L2G in compact form \n" \
"                   specify the observation layer number as dot extension in the \n" \
"                   name.  \n" \
"                   Warning: All input SDS must be of same resolution. \n" \
"     -of=filename  Output file \n" \
"     -log=filename ASCII logfile that lists all the pixels matching the \n" \
"                   specified mask. Values from each SDS are stored in different \n" \
"                   columns within the same file. \n" \
"     -meta         Copy metadata from input file to output \n" \
"     -fill_value=fillVal       User inputs fill value \n" \
" \n" \
"Example: \n" \
"	mask_sds -sds=sur_refl_b04,sur_refl_b03 -mask=\"./MOD09A1.A2017129.h11v03.006.2017138034901.hdf,sur_refl_qc_500m,14-17==0000,AND,*,*,10-13==0000,AND,*,*,30==1,AND,*,sur_refl_state_500m,3-5==001,AND,*,*,6-10>=00001,AND,*,*,6-10<=00010,AND,*,*,15==0\" -of=MOD09A1.masked.hdf MOD09A1.A2017129.h11v03.006.2017138034901.hdf \n" \
"	../src/mask_sds -sds=SurfReflect_M5,SurfReflect_M4,SurfReflect_M3 -mask=\"./VNP09A1.A2016297.h11v08.001.2017189113645.h5,SurfReflect_M5,<=10000,AND,*,SurfReflect_QC,0-1==00,AND,*,*,30==1,AND,*,SurfReflect_State,0-5==001000,AND,*,*,6-9>=0001,AND,*,*,6-9<=0010,AND,*,*,13==0\" -of=VNP09A1.masked.hdf VNP09A1.A2016297.h11v08.001.2017189113645.h5 \n" \
"	mask_sds -sds=EV_500_Aggr1km_RefSB.1 -mask=\"./DATA//MOD35/1997223/MOD35_L2.A1997223.0219.002.1999141012404.hdf,Cloud_Mask.1,1-2==11,AND,*,*,6-7==11\" -of=mod02_dmask.hdf ./DATA/MOD021KM/1997223/MOD021KM.A1997223.0219.211.1999131214304.hdf \n" \
"	mask_sds -sds=\"EV_500_Aggr1km_RefSB.1\" -mask=\"./DATA//MOD35/1997223/MOD35_L2.A1997223.0219.002.1999141012404.hdf,Cloud_Mask.1,1-2==11,AND,*,*,6-7==11\" -of=mod02_2dmask.hdf ./DATA/MOD021KM/1997223/MOD021KM.A1997223.0219.211.1999131214304.hdf \n" \
"	mask_sds -sds=\"Surface_Refl.1.1\" -mask=\"./DATA/SAMPLES/MODAGAGG.A1996214.h12v04.001.hdf,Aggregate_QC.1.1,3-5==001\" -of=modagg_2dmask.hdf ./DATA/SAMPLES/MODAGAGG.A1996214.h12v04.001.hdf \n" \
"	mask_sds -sds=sur_refl_b01.5 -mask=\"./DATA/MOD09GQK/1997223/MOD09GQK.A1997223.h33v08.001.1999191055309.hdf,QC_250m.5,0-1==10\" -of=mod09l2g_obs5_2dmask.hdf ./DATA/MOD09GQK/1997223/MOD09GQK.A1997223.h33v08.001.1999191055309.hdf \n" \
"	mask_sds -sds='500m 16 days NDVI' -of=NPP_D16VI1KM_L3D_masked.A2003193.h09v05.P1_00602.2009257123847.hdf -mask='NPP_D16VI1KM_L3D.A2003193.h09v05.P1_00602.2009257123847.hdf,500m 16 days VI Quality,6-8>=001,AND,*,*,6-8<=111,AND,*,*,1==0,AND,*,*,2-5<=1100,AND,*,*,10-15==000010,AND,*,500m 16 days NDVI,<=10000' NPP_D16VI1KM_L3D.A2003193.h09v05.P1_00602.2009257123847.hdf\n" \
"\n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n\n" \
"	Please note that this code is still under revision. Tool will run very slow if \n" \
"data is output to an ascii file (too many I/O operations). Tool has not been tested \n" \
"for every possible combination of inputs and outputs. In some ocassions on error tool \n" \
"might core dump while exiting. However errors are reported to stdout in all cases. \n" \
"When bit numbers are used in the mask option bit values must be in binary. All relational \n" \
"operators can be used with the interpreation of decimal value for the binary string. \n" \
"Please report problems to Sadashiva Devadiga (devadiga@ltpmail.gsfc.nasa.gov) \n"

#define USAGE \
"usage:      mask_sds -mask=mask1[,AND|OR,mask2...] -sds=SDS_name filename \n" \
"              [-of=output_file] [-log=logfile] [-meta] \n" \
" \n" \
"OPTIONS \n" \
"     -help         Display this help message \n" \
"     -mask=string  Define the mask to apply as combination of individual masks \n" \
"                   separated by AND or OR. Each individual mask consists of \n" \
"                   a file, a SDS, a bit descriptor, and a binary bit value \n" \
"                   string to match: filename,SDS_name,bits<operator>string. \n" \
"                   Valid operators are >, <, <=, >=, ==, !=. Rather than duplicate \n" \
"                   the file and SDS names, the * character can be used after \n" \
"                   initial specification. The bit descriptor is a list of bits or \n" \
"                   ranges of bits, similar to the Unix filename substitution. \n" \
"                     Ex: -mask=file1,SDS1,0-2,4==0101,AND,*,SDS2,2,3==10 \n" \
"                   Warning: Although a bit range in the bit descriptor is \n" \
"                   specified by the lower bit followed by the higher bit, the \n" \
"                   bit value string are in reverse order, i.e. the lower bit on \n" \
"                   the right (in the example 0-2,4=0101 is for 4,2,1,0=0101). \n" \
"                   If the bit descriptor is omitted, then the whole range of \n" \
"                   bits is assumed, and the value string is parsed as a decimal \n" \
"                   value. This is a convenient way to refer to a specific \n" \
"                   value, instead of a list of bits. \n" \
"                     Ex: -mask=file,SDS,>=200 \n" \
"                   If SDS is a 3D or 4D specify the dimension value as dot \n" \
"                   extension in the sds name: \n" \
"                     Ex: -mask=file,SDS.n1,0-2==001,AND,file,SDS.n2.m2,0-1==01 \n" \
"                   If SDS is a L2G in compact form specify the observation layer \n" \
"                   number as dot extension in the sds name: \n" \
"                     Ex: -mask=file,SDS.n1,0-2==001,AND,file,SDS.n2.m2,0-1==01 \n" \
"     -sds=SDS_name SDS to mask. If SDS is a 3D or 4D specify the dimension value \n" \
"                   as dot extension in the sds name. If SDS is L2G in compact form \n" \
"                   specify the observation layer number as dot extension in the \n" \
"                   name. \n" \
"                   Warning: All input SDS must be of same resolution. \n" \
"     -of=filename  Output file \n" \
"     -log=filename ASCII logfile that lists all the pixels matching the specified mask \n" \
"     -meta         Copy metadata from input file to output \n" \
"     -fill_value=fillVal       User inputed fill value \n" 

int parse_cmd_mask_sds(int argc, char **argv, int *arg_cnt, char **arg_list, int *opt, int *fqa_l2g,
        char *m_str, char **sds_names, int *sds_cnt, int *if_cnt, int *m_opt, char *fillVal);
int mask_nsds(int fin_l2g, int *fqa_l2g, char *m_str, char **arg_list, int arg_cnt, int *opt_arr, 
	char **sds_names, int nsds, int32 out_sd_id, FILE *fp, long *min, long *max, long *mu, 
	long *std, int *npix, int m_opt, char *fillVal);
void mask_nsds_data_row(void **data_in, void **data_out, uint8 *data_mask, int *ndata_out, 
	int ndata_mask, int nsds, int bsq, int *st_c, int *offset, sds_t *sds_info, 
	double *sum_val, double *sum_val2, int *npix, long *min, long *max, FILE *fp);

int main(int argc, char *argv[])
{
  FILE *fp;
  int m_opt;
  int32 out_sd_id;
  int status = 1;
  int L2G_boolean;
  int sds_cnt, meta_cnt, arg_cnt;
  int nobs, nsds, stdin_argc, if_cnt;
  int p1, len, fid, ii, isds, fin_l2g, fcnt;
  int opt[MAX_NUM_OP], fqa_l2g[MAX_NUM_OP], npix[MAX_NSDS];
  long mu[MAX_NSDS], std[MAX_NSDS], min[MAX_NSDS], max[MAX_NSDS];
  char org_name[80] ={'\0'}, org_num[5] ={'\0'};
  char meta_name[MAX_META_NAME_LEN] = {'\0'};
  char **stdin_argv = NULL, **meta_val = NULL, **arg_list = NULL;
  char *tile = NULL, *metadata = NULL, **sds_names = NULL, **tmp_sds_names = NULL;
  char hdf_fname[MAX_PATH_LENGTH] = {'\0'}, m_str[MAX_STR_LEN] = {'\0'};
  char fillVal[10]={'\0'};
  char *meta_val1 = NULL, *meta_val2 = NULL;  
  char tmp1_sds_names[MAX_PATH_LENGTH][MAX_STR_LEN] = {'\0'}; 
  sds_t out_sds_info;
   
  if ((argc == 2) && ((strcmp(argv[1], "-h") == 0) || (strcmp(argv[1], "-help") == 0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }

  /* MSZ 2/8/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/8/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if (argc != 1) stdin_argc = argc;
  fprintf(stdout, "Number of arguments: %d\n", stdin_argc);
  if ((arg_list = (char **)Calloc2D(3*MAX_NUM_OP+5, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for arg_list in mask_sds: main()\n");
  if ((sds_names = (char **)Calloc2D(MAX_NSDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names mask_sds: main()\n");
  if ((arg_list != NULL) && (sds_names != NULL))
  {
    status = parse_cmd_mask_sds(argc, argv, &arg_cnt, arg_list, opt, fqa_l2g, m_str, sds_names, 
			&sds_cnt, &if_cnt, &m_opt, fillVal);
    if (status == -1) 
      fprintf(stderr, "%s\n", USAGE);
    else if (status != 0)
    { 
      if ((meta_val = (char **)Calloc2D(5, MAX_GID_LEN, sizeof(char))) == NULL)
        fprintf(stderr, "Cannot allocate memory for meta_val mask_sds: main()\n");
      if ((tmp_sds_names = (char **)Calloc2D(MAX_NSDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
        fprintf(stderr, "Cannot allocate memory for tmp_sds_names mask_sds: main()\n");
      if ((meta_val != NULL) && (tmp_sds_names != NULL)) 
      {
        if (arg_list[arg_cnt-2][0] != '\0')
        {
	  out_sds_info.H5OUT = 0;
	  out_sds_info.sd_id = out_sds_info.sds_id = -1;
	  if(open_sd(arg_list[arg_cnt-2], &out_sds_info, 'W') == FAIL)
            fprintf(stderr, "Cannot create ouput file %s in mask_sds: main() \n", arg_list[arg_cnt-2]);
	  else fprintf(stderr, "\tAll output SDS data goes to HDF file: %s\n", arg_list[arg_cnt-2]);
        }
        else out_sds_info.sd_id = FAIL;
        if (arg_list[arg_cnt-1][0] != '\0')
        {
          if ((fp = fopen(arg_list[arg_cnt-1], "w")) == NULL)
            fprintf(stderr, "Cannot create ouput log file %s in mask_sds: main() \n", arg_list[arg_cnt-1]);
	  else fprintf(stderr, "\tAll output SDS data goes to ASCII file: %s\n", arg_list[arg_cnt-1]);
        }
        else fp = NULL;

        if ((out_sds_info.sd_id != FAIL) || (fp != NULL))
        {
          for (fid=1, fcnt=0; fid<stdin_argc; fid++)
          {
	    if (argv[fid][0] == '-') continue;
            strcpy(hdf_fname, argv[fid]);
	    strcpy(arg_list[0], hdf_fname);
            fprintf(stderr, "\nProcessing HDF file: %s\n", hdf_fname);
            fprintf(stderr, "---------------------------------------------------------------------\n");
	    fin_l2g = 0;

	    if ((metadata = get_attr_metadata(hdf_fname, "CoreMetadata.0")) != NULL) status = 0;
	    else status = 1;

	    printf("status: %d\n", status);  

	    if (status == 0)
	    {	    
              /* Search for the tile/granule indentifier */
              meta_cnt = 0;
	      strcpy(meta_name, "LOCALGRANULEID");
              get_sel_metadata(metadata, meta_name, meta_val, &meta_cnt, 0);
	      if(metadata)
	      {
		 free(metadata);
		 metadata = NULL;
	      }
              if(meta_cnt != 0)
              {
                ii=0;
                tile = strstr(meta_val[0], ".A") + 1;
                while(tile[++ii] != '.' ) ;
                while(tile[++ii] != '.' ) ;
                tile[ii] = '\0';
              }

	      /* is input an l2g file? */
	      if(L2G_format_check(hdf_fname) != -1)
		fin_l2g = 1;
	      else fin_l2g = 0;

	      if (fin_l2g == 1)
	      {
		printf("%s is a L2G data file, this version doesn't support L2G, exit...\n", hdf_fname);
		exit(0);
        	meta_cnt = 0;
		strcpy(meta_name, "MAXIMUMOBSERVATIONS");
        	get_sel_metadata(metadata, meta_name, meta_val, &meta_cnt, 0);
		if (meta_cnt != 0) nobs = (int)atoi(meta_val[0]);
	      }
	      if (metadata)
	      {
		free(metadata);
		metadata = NULL;
	      }
	    }
	    else if (status == 1)
	    {
	      strcpy(meta_name, "LocalGranuleID");
	      int32 attr_type, attr_cnt;
	      meta_val1 = (char *)get_attr_sel_metadata(hdf_fname, meta_name, &attr_type, &attr_cnt); 
	      printf("LocalGranuleID=%s\n", meta_val1);   		  		     
	      if (meta_val1 != NULL)  
	      {
                ii=0;
                tile = strstr(meta_val1, ".A") + 1;
                while(tile[++ii] != '.' ) ;
                while(tile[++ii] != '.' ) ;
                tile[ii] = '\0';   		      
	      }	 
	     
	      if(L2G_format_check(hdf_fname) != -1) 
		fin_l2g = 1;

	      if (fin_l2g == 1)
	      {
	        printf("%s is a L2G data file, this version doesn't support L2G, exit...\n", hdf_fname);
	        exit(0);
	      }
	      free(meta_val1);  
	    }
	    
	    /* get SDS names and update the names if it is l2g/3D/4D SDS */
            for (isds=0; isds<sds_cnt; isds++)
              strcpy(tmp_sds_names[isds], sds_names[isds]);
            if ((sds_cnt == 1) && (strcmp(tmp_sds_names[0], "all") == 0))
              nsds = get_sds_names(hdf_fname, tmp_sds_names);
            else nsds = sds_cnt;
	    printf("fin_l2g = %d\n",   fin_l2g);
	    if (nsds > 0)
	    {
	      if (fin_l2g == 0) update_nd_sdsnames(tmp_sds_names, &nsds, hdf_fname);
	      else update_l2g_sdsnames(tmp_sds_names, &nsds, hdf_fname, nobs);
              if ((fcnt == 0) && (fp != NULL))
              {
                for (isds=0; isds<nsds; isds++)
                  fprintf(fp, "%s ", tmp_sds_names[isds]);
                fprintf(fp, "\n");
              }
              fcnt++;  
  	      if (if_cnt > 1)
  	      {
    	        rm_path(hdf_fname);
    	        strncpy(arg_list[arg_cnt], hdf_fname, 28);
    	        arg_list[arg_cnt][28] = '\0';
  	      }
	      else arg_list[arg_cnt][0] = '\0';
	      
	      for (isds=0; isds<nsds; isds++) {
   	      	strcpy(tmp1_sds_names[isds], tmp_sds_names[isds]);
		printf("tmp_sds_names[%d]=%s\n", isds, tmp_sds_names[isds]);  
	      }
	      if (fin_l2g == 1)
	      {
	        for (isds=0; isds<nsds; isds++)
	        {   	          
		  len = (int)strlen(tmp_sds_names[isds]);
    	          p1 = charpos(tmp_sds_names[isds], '.', 0);
    	          strmid(tmp_sds_names[isds], 0, p1, org_name);
    	          strmid(tmp_sds_names[isds], p1, len-p1, org_num);
    	          sprintf(tmp_sds_names[isds], "%s_1%s", org_name, org_num);	
	          printf("tmp_sds_names[%d]=%s (l2g==1)\n", isds, tmp_sds_names[isds]);  
	        }
	      }

	      if (mask_nsds(fin_l2g, fqa_l2g, m_str, arg_list, arg_cnt, opt, tmp_sds_names, nsds, 
				out_sds_info.sd_id, fp, min, max, mu, std, npix, m_opt, fillVal) == 1)
	      {
/*                if (status != 2 ) 
		{
		  fprintf(stdout, "\t\tOutput Statistics\n");
                  fprintf(stdout, " tile/granule   \t   SDS   \t   npix \t mean \t stdev \t min \t max\n");
	          for (isds=0; isds<nsds; isds++)
	            fprintf(stdout, "%s \t %s \t %d \t %ld \t %ld \t %ld \t %ld\n", tile, tmp1_sds_names[isds], 
				    npix[isds], mu[isds], std[isds], min[isds], max[isds]);
	        }	*/
		printf("mask_sds processing done!\n");
	      }
	    }
	  } /* for (fid=0; . . . ) */
	} /* if ((out_sds_info.sd_id != -1) . . ) */
        if (fp != NULL) fclose(fp);
	close_sd(&out_sds_info, out_sds_info.H5OUT);
	if(meta_val) Free2D((void **)meta_val);
	if(tmp_sds_names) Free2D((void **)tmp_sds_names);
      } /* if ((meta_val != NULL) . . .) */
    } /* if (status != -1) */
    if(arg_list) Free2D((void **)arg_list);
    if(sds_names) Free2D((void **)sds_names);
  } /* if ((arg_list != NULL) . . .  ) */
  return 0;
}

int parse_cmd_mask_sds(int argc, char **argv, int *arg_cnt, char **arg_list, int *opt, int *fqa_l2g,
        char *m_str, char **sds_names, int *sds_cnt, int *if_cnt, int *m_opt, char *fillVal)
{
  int i, k;
  int status =1;
  char out_fname[MAX_PATH_LENGTH];
  char log_fname[MAX_PATH_LENGTH];

  *if_cnt = *sds_cnt = *m_opt = 0;
  m_str[0] = out_fname[0] = log_fname[0] = fillVal[0] = '\0';

  for (i=1; i<argc; i++)
  {
    if ((is_arg_id(argv[i], "-of=") == 0) || (is_arg_id(argv[i], "-o=") == 0))
      get_arg_val(argv[i], out_fname);
    else if (is_arg_id(argv[i], "-sds=") == 0)
      get_arg_val_arr(argv[i], sds_names, sds_cnt);
    else if ((is_arg_id(argv[i], "-m=") == 0) || (is_arg_id(argv[i], "-mask=") == 0))
      get_arg_val(argv[i], m_str);
    else if (is_arg_id(argv[i], "-log=") == 0) get_arg_val(argv[i], log_fname);
    else if (strcmp(argv[i], "-meta") == 0) *m_opt = 1;
    else if (is_arg_id(argv[i], "-fill_value=") == 0) get_arg_val(argv[i], fillVal);
    else
    {
      if (argv[i][0] == '-') fprintf(stderr, "Unknown option %s\n", argv[i]);
      else ++*if_cnt;
    }
  }
  if ((*if_cnt == 0) || (m_str[0] == '\0') || ((out_fname[0] == '\0') && (log_fname[0] == '\0')))
  {
    status = -1;
    if (*if_cnt == 0) fprintf(stderr,"Missing input file \n");
    if (m_str[0] == '\0') fprintf(stderr,"Missing mask option \n");
    if ((out_fname[0] == '\0') && (log_fname[0] == '\0')) fprintf(stderr, "No output option specified \n");
  }
  else
  {
    if (*sds_cnt == 0)
    {
      *sds_cnt = 1;
      strcpy(sds_names[0], "all");
      fprintf(stderr, "No SDS name input. Masking all SDS in the input file. . \n");
    }
    if ((k = get_mask_string(m_str, arg_list, opt, fqa_l2g)) < 0) status = 0;
    if (status == 1)
    {
      *arg_cnt = 3 + k*3;
      strcpy(arg_list[*arg_cnt], out_fname);
      *arg_cnt = *arg_cnt + 1;
      strcpy(arg_list[*arg_cnt], log_fname);
      *arg_cnt = *arg_cnt + 1;
    }
  }
  return status;
}
	    
int mask_nsds(int fin_l2g, int *fqa_l2g, char *m_str, char **arg_list, int arg_cnt, int *opt_arr, 
	char **sds_names, int nsds, int32 out_sd_id, FILE *fp, long *min, long *max, long *mu, 
	long *std, int *npix, int m_opt, char *fillVal)
{
  long mu2;
  uint8 *data_mask;
  int rank, data_size;
  int len, p1, obs_num_in[MAX_NSDS];
  int st, i_op, j_op, n_op;
  int i, j, isds, jsds, irow, nrow;
  int bsq, *ndata_in, *ndata_out, ndata_mask, tmp, Isds;
  int sel_qa_op[MAX_NUM_OP], rel_op[MAX_NUM_OP];
  int res_s[MAX_NUM_OP], res_l[MAX_NUM_OP], obs_num[MAX_NUM_OP];
  int st_c[MAX_NSDS], offset[MAX_NSDS], n[MAX_NSDS], m[MAX_NSDS];
  char **qa_fnames, num_str[10];
  double sum_val[MAX_NSDS], sum_val2[MAX_NSDS];
  int32 *data_qa_nadd[MAX_NUM_OP], *data_in_nadd, *data_in_nadd_1km, *data_in_nadd_500m; 
  int32 **in_edge, **in_start, **out_edge, **out_start;
  sds_t sds_info, in_sds_nobs_info, *in_sds_info, *in_sdsc_info, *out_sds_info;
  sds_t qa_sdsc_info[MAX_NUM_OP], qa_sds_info[MAX_NUM_OP], qa_sds_nobs_info[MAX_NUM_OP];
  unsigned long bit_mask_arr[MAX_NUM_OP], mask_val_arr[MAX_NUM_OP];
  void **data_in, **data_out, *data_qa[MAX_NUM_OP];
  char qa_fname[255], qa_name[20];
  int l2g_lite = 0;
    
  if ((qa_fnames = (char **)Calloc2D(MAX_NUM_OP, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for qa_fnames mask_nsds\n"); return -1;
  }
  
  n_op = (arg_cnt - 3)/3;  
  printf("nsds=%d\n",nsds);
  st = get_parameters(arg_list, n_op, sel_qa_op, qa_fnames, qa_sds_info, bit_mask_arr, 
			mask_val_arr, opt_arr, rel_op);
  if (st != -1)
    st = get_qa_sds_info(qa_fnames, qa_sds_info, qa_sdsc_info, fqa_l2g, n_op);
  if (st != -1)
  {
    if ((out_sds_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
      fprintf(stderr, "Cannot allocate memory for out_sds_info in mask_nsds()\n");
    if ((in_sds_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
      fprintf(stderr, "Cannot allocate memory for in_sds_info in mask_nsds()\n");
    
    if ((ndata_in = (int *)calloc(nsds, sizeof(int))) == NULL)       
      fprintf(stderr, "Cannot allocate memory for ndata_in in mask_nsds()\n"); 
    if ((ndata_out = (int *)calloc(nsds, sizeof(int))) == NULL)       
      fprintf(stderr, "Cannot allocate memory for ndata_out in mask_nsds()\n");       
    
    if (fin_l2g == 1)
    {
      if ((in_sdsc_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
        fprintf(stderr, "Cannot allocate memory for in_sdsc_info in mask_nsds()\n");
    }
    if ((out_sds_info != NULL) && (in_sds_info != NULL))
    {
      for (isds=0; isds<nsds; isds++)
        strcpy(in_sds_info[isds].name, sds_names[isds]);
      st = get_in_sds_info(arg_list[0], in_sds_info, in_sdsc_info, &in_sds_nobs_info, 
		fin_l2g, nsds);			
      if (st != -1)
        st = get_res_factors(&in_sds_info[0], qa_sds_info, n_op, res_l, res_s);
      if (st != -1)
      {
        if (fillVal[0] != '\0')
	{  
	  for (isds=0; isds<nsds; isds++)
          {
	    switch(in_sds_info[isds].data_type)
	    {
	      case  5: in_sds_info[isds].fill_val = (float32)atof(fillVal); break;
	      case  6: in_sds_info[isds].fill_val = (float64)atof(fillVal); break;
	      case 20: in_sds_info[isds].fill_val = (int8)atoi(fillVal); break;
	      case 21: in_sds_info[isds].fill_val = (uint8)atoi(fillVal); break;
	      case 22: in_sds_info[isds].fill_val = (int16)atoi(fillVal); break;
	      case 23: in_sds_info[isds].fill_val = (uint16)atoi(fillVal); break;
	      case 24: in_sds_info[isds].fill_val = (int32)atoi(fillVal); break;
	      case 25: in_sds_info[isds].fill_val = (uint32)atoi(fillVal); break;
	    } 	 
	  }  
	  if (fin_l2g == 1)
          {
	    switch(in_sds_info[isds].data_type)
	    {
	      case  5: in_sdsc_info[isds].fill_val = (float32)atof(fillVal); break;
	      case  6: in_sdsc_info[isds].fill_val = (float64)atof(fillVal); break;
	      case 20: in_sdsc_info[isds].fill_val = (int8)atoi(fillVal); break;
	      case 21: in_sdsc_info[isds].fill_val = (uint8)atoi(fillVal); break;
	      case 22: in_sdsc_info[isds].fill_val = (int16)atoi(fillVal); break;
	      case 23: in_sdsc_info[isds].fill_val = (uint16)atoi(fillVal); break;
	      case 24: in_sdsc_info[isds].fill_val = (int32)atoi(fillVal); break;
	      case 25: in_sdsc_info[isds].fill_val = (uint32)atoi(fillVal); break;
	    } 	 
	  }  	  
	}  
	in_edge = (int32 **)Calloc2D(nsds, 4, sizeof(int32));
        in_start = (int32 **)Calloc2D(nsds, 4, sizeof(int32));
        out_edge = (int32 **)Calloc2D(nsds, 4, sizeof(int32));
        out_start = (int32 **)Calloc2D(nsds, 4, sizeof(int32));
        if (fin_l2g == 1)
        {    
	  if ((strstr(arg_list[0], "MOD09GA") == NULL) && (strstr(arg_list[0], "MYD09GA") == NULL) &&
	      (strstr(arg_list[0], "NPP_DSRFLD_L2GD") == NULL))
          { 
            strcpy(sds_info.name, "nadd_obs_row"); sds_info.rank = 1; 
	    sds_info.sd_id = in_sds_info[0].sd_id; sds_info.sds_id = -1;
            sds_info.dim_size[0] = in_sds_info[0].dim_size[0];
            if ((data_in_nadd = (int32 *)calloc(sds_info.dim_size[0], sizeof(int32))) == NULL)
              fprintf(stderr, "Cannot allocate memory for data_in_nadd in mask_nsds\n");
            else
            {
              get_sds_data(&sds_info, data_in_nadd);
	      close_sds(&sds_info, sds_info.H5IN);
            }
          }
	  else
	  {
	    l2g_lite = 1;	
            strcpy(sds_info.name, sds_name_nadd_obs_1km); 
	    sds_info.sd_id = in_sds_info[0].sd_id; sds_info.sds_id = -1;
	    if (get_sds_info(arg_list[0], &sds_info) == -1) 
	      fprintf(stderr, "Result may be in error for %s \n", sds_info.name);            
            if ((data_in_nadd_1km = (int32 *)calloc(sds_info.dim_size[0], sizeof(int32))) == NULL)
              fprintf(stderr, "Cannot allocate memory for data_in_nadd_1km in mask_nsds\n");
            else
            {
	      get_sds_data(&sds_info, data_in_nadd_1km);
	      close_sds(&sds_info, sds_info.H5IN);
            }	    
            strcpy(sds_info.name, sds_name_nadd_obs_500m); 
	    sds_info.sd_id = in_sds_info[0].sd_id; sds_info.sds_id = -1;
	    if (get_sds_info(arg_list[0], &sds_info) == -1) 
	      fprintf(stderr, "Result may be in error for %s \n", sds_info.name);              
            if ((data_in_nadd_500m = (int32 *)calloc(sds_info.dim_size[0], sizeof(int32))) == NULL)
              fprintf(stderr, "Cannot allocate memory for data_in_nadd_500m in mask_nsds\n");
            else
            {
              get_sds_data(&sds_info, data_in_nadd_500m);
	      close_sds(&sds_info, sds_info.H5IN);
            }	    
	  }
	}
	
        for (isds=0; isds<nsds; isds++)
	{  
	  out_sds_info[isds].sd_id = out_sd_id; out_sds_info[isds].sds_id = -1; 
	}
	
        for (isds=0; isds<nsds; isds++)
          strcpy(in_sds_info[isds].name, sds_names[isds]);        
	create_out_sds(in_sds_info, out_sds_info, nsds, arg_list[arg_cnt], m_str, n, m, out_sd_id);
	  
        for (isds=0; isds<nsds; isds++)
        {
          rank = in_sds_info[isds].rank;
          if (rank == 2)
          {
            in_edge[isds][0] = out_edge[isds][0] = 1;
            in_edge[isds][1] = out_edge[isds][1] = in_sds_info[isds].dim_size[1];
          }
          else if (rank > 2)
          {
            for (i=0; i<rank; i++)
              in_edge[isds][i] = out_edge[isds][i] = in_sds_info[isds].dim_size[i];
            if ((n[isds] == -1) && (m[isds] == -1))
            {
              if (in_sds_info[isds].dim_size[0] > in_sds_info[isds].dim_size[rank-1])
                in_edge[isds][0] = out_edge[isds][0] = 1;
              else in_edge[isds][rank-2] = out_edge[isds][rank-2] = 1;
            }
            else
            {
              out_edge[isds][0] = 1; out_edge[isds][1] = out_sds_info[isds].dim_size[1];
              if (in_sds_info[isds].dim_size[0] > in_sds_info[isds].dim_size[rank-1]) in_edge[isds][0] = 1;
              else in_edge[isds][rank-2] = 1;
            }
          }
          compute_sds_start_offset(&in_sds_info[isds], n[isds], m[isds], &st_c[isds], &offset[isds]); 
	  get_ndata_vals(&in_sds_info[isds], &bsq, &nrow, &ndata_in[isds], &ndata_mask, &ndata_out[isds], n[isds], m[isds]);	  
        } /* for (isds=0; . . . ) */

/*        get_ndata_vals(&in_sds_info[0], &bsq, &nrow, &ndata_in, &ndata_mask, &ndata_out, n[0],
m[0]);*/

/*        for (isds=0, data_size=in_sds_info[0].data_size; isds<nsds; isds++)
          if (in_sds_info[isds].data_size > data_size) data_size = in_sds_info[isds].data_size;
        if ((data_in = (void **)Calloc2D(nsds, ndata_in, data_size)) == NULL)
          fprintf(stderr, "Cannot allocate memory for data_in in mask_nsds()\n"); 
        if ((data_out = (void **)Calloc2D(nsds, ndata_out, data_size)) == NULL)
          fprintf(stderr, "Cannot allocate memory for data_out in mask_nsds()\n"); 
        if ((data_mask = (uint8 *)calloc(ndata_mask, sizeof(uint8))) == NULL)
          fprintf(stderr, "Cannot allocate memory for data_mask in mask_nsds()\n"); */

	tmp = ndata_in[0];
	Isds = 0;
        for (isds=0, data_size=in_sds_info[isds].data_size; isds<nsds; isds++)
	{	
	  if (ndata_in[isds] > tmp) {
	    tmp = ndata_in[isds];
	    Isds = isds;
	  }
          if (in_sds_info[isds].data_size > data_size) data_size = in_sds_info[isds].data_size;
          if ((data_in = (void **)Calloc2D(nsds, tmp, data_size)) == NULL)
            fprintf(stderr, "Cannot allocate memory for data_in in mask_nsds()\n"); 
          if ((data_out = (void **)Calloc2D(nsds, ndata_out[Isds], data_size)) == NULL)
            fprintf(stderr, "Cannot allocate memory for data_out in mask_nsds()\n"); 
          if ((data_mask = (uint8 *)calloc(ndata_mask, sizeof(uint8))) == NULL)
            fprintf(stderr, "Cannot allocate memory for data_mask in mask_nsds()\n");
	}	
	  
	if ((data_in != NULL) && (data_out != NULL) && (data_mask != NULL))
	{	
          st = open_qa_sds_nsds(arg_list[0], in_sds_info, in_sdsc_info, &in_sds_nobs_info, nsds, qa_fnames,
				qa_sds_info, qa_sdsc_info, qa_sds_nobs_info, fqa_l2g, n_op);
	  if (st != -1)
	  {
            if (l2g_lite == 0)
	      st = malloc_qa_sds(qa_sds_info, n_op, fqa_l2g, data_qa, data_qa_nadd);
	    else
	      st = malloc_qa_sds_lite(qa_sds_info, n_op, fqa_l2g, data_qa, data_qa_nadd);	    
  	  }
          if (st != -1)
          {
	    for (isds=0; isds<nsds; isds++)
	    {
	      sum_val[isds] = sum_val2[isds] = 0.0;
	      mu[isds] = std[isds] = npix[isds] = 0;
	      min[isds] = 99999; max[isds] = -99999;
	    }
	    rank = in_sds_info[0].rank;
            for (i_op=0; i_op<=n_op; i_op++)
	    {
              if (fqa_l2g[i_op] == 1)
              {	      	      
		if (get_file_sds_names(arg_list[1+i_op*3], qa_fname, qa_name) == -1) 
		  st = -1;
		else  
		{              		
		  len = (int)strlen(qa_name);
  		  printf("qa_name == %s\n", qa_name);		
                  p1 = charpos(qa_name, '.', 0);
                  strmid(qa_name, p1+1, len-p1-1, num_str);
                  obs_num[i_op] = (int)atoi(num_str);
  		  printf("obs_num[%d] = %d\n", i_op, obs_num[i_op]);		
        	}
    	      }
	      else 
	        obs_num[i_op] = 1;
	    }  
	    for (isds=0; isds<nsds; isds++)
	    {
              if (fin_l2g == 1)
	      {
		len = (int)strlen(sds_names[isds]);		  	  
        	p1 = charpos(sds_names[isds], '.', 0);
		strmid(sds_names[isds], p1+1, len-p1-1, num_str);		  		  		  		  
        	obs_num_in[isds] = (int)atoi(num_str);
		printf("obs_num_in[%d] = %d\n", isds, obs_num_in[isds]);    
	      }
	      else
	        obs_num_in[isds] = 1;
	    }	
    
	    for (irow=0; irow<nrow; irow++)
	    {
	      read_qa_sds(qa_sds_info, qa_sdsc_info, qa_sds_nobs_info, n_op, data_qa, data_qa_nadd,
			    irow, res_l, fqa_l2g, obs_num);
  
              process_mask_data(data_qa, ndata_mask, qa_sds_info, n_op, sel_qa_op, bit_mask_arr, 
			  mask_val_arr, rel_op, res_s, data_mask, YES, NO);
		
  	      for (isds=0; isds<nsds; isds++)
  	      {
                if ((rank == 2) || (in_sds_info[isds].dim_size[0] > in_sds_info[isds].dim_size[rank-1]))
                  in_start[isds][0] = irow;
                else in_start[isds][rank-2] = irow;
                if (fin_l2g == 1)
                {
                  if (obs_num_in[isds] > 1)
		  {		    		    
		    if (l2g_lite == 0)
		    {
		      read_sdsc_data(&in_sdsc_info[isds], &in_sds_nobs_info, data_in[isds], data_in_nadd,
                                            irow, obs_num_in[isds]);
	            }
		    else
		    {
		      if (get_sds_info(arg_list[0], &in_sds_info[isds]) == -1) 
			fprintf(stderr, "Result may be in error for %s \n", in_sds_info[isds].name); 	
		      else
		      {		      
			if (in_sds_info[isds].dim_size[0] == 1200)
		          read_sdsc_data(&in_sdsc_info[isds], &in_sds_nobs_info, data_in[isds], data_in_nadd_1km,
                                        	irow, obs_num_in[isds]);
			else if (in_sds_info[isds].dim_size[0] == 2400)	
		          read_sdsc_data(&in_sdsc_info[isds], &in_sds_nobs_info, data_in[isds], data_in_nadd_500m,
                                        	irow, obs_num_in[isds]);		      		      		    
		      }
		    }		  
		  }
		  else
		    if (read_sel_sds_data(&in_sds_info[isds], in_start[isds], in_edge[isds], data_in[isds]) == FAIL)
      	              fprintf(stderr, "Cannot read data line from SDS %s in mask_sds()\n", in_sds_info[isds].name);
                }
	        else
		  if (read_sel_sds_data(&in_sds_info[isds], in_start[isds], in_edge[isds], data_in[isds]) == FAIL)
      	            fprintf(stderr, "Cannot read data line from SDS %s in mask_sds()\n", in_sds_info[isds].name);
	      } /* for (isds=0; . . .  */

              mask_nsds_data_row(data_in, data_out, data_mask, ndata_out, ndata_mask, nsds, bsq, st_c, 
			offset, in_sds_info, sum_val, sum_val2, npix, min, max, fp);
    	
	      if (out_sd_id != -1)
	      {
	        for (isds=0; isds<nsds; isds++)
	        {
                  if ((rank>2) && (n[isds] == -1) && (m[isds] == -1) && 
				      (in_sds_info[isds].dim_size[rank-1] > in_sds_info[isds].dim_size[0]))
	            out_start[isds][rank-2] = irow;
                  else out_start[isds][0] = irow;

		  /* MSZ 1/25/2020. I don't use set_sel_sds_data_compression because create_out_sds() above has set compression for each out_sds */
		  if(set_sel_sds_data(&out_sds_info[isds], out_start[isds], out_edge[isds], data_out[isds]) == FAIL) 
	            fprintf(stderr, "Cannot write data line to SDS %s in mask_nsds()\n", out_sds_info[isds].name);
	        }
	      }
            } /* for (irow=0;  . .  ) */

	    if (m_opt == 1)
	      copy_metadata(in_sds_info[0].H5IN, in_sds_info[0].sd_id, out_sds_info[0].H5OUT, out_sds_info[0].sd_id);

            for (isds=0; isds<nsds; isds++)
	      if (npix[isds] != 0)
	      {	
	        mu2 = (long)(sum_val2[isds]/(double)npix[isds]);
	        mu[isds] = (long)(sum_val[isds]/(double)npix[isds]);
	        std[isds] = (long)sqrt((double)mu2 - (double)(mu[isds]*mu[isds]));	
	      }

            /* close all SDS and HDF files */
	    if (fin_l2g == 1) close_sds(&in_sds_nobs_info, in_sds_nobs_info.H5IN);
	    for (i_op=0; i_op<=n_op; i_op++)
	      if (fqa_l2g[i_op] == 1)
	      {
	        if ((qa_sds_nobs_info[i_op].sd_id == in_sds_nobs_info.sd_id) && 
			  (qa_sds_nobs_info[i_op].sds_id == in_sds_nobs_info.sds_id))
			    ;
	        else
	        {
	          for (j_op=0; j_op<i_op; j_op++)
		    if ((qa_sds_nobs_info[i_op].sd_id == qa_sds_nobs_info[j_op].sd_id) &&
		       (qa_sds_nobs_info[i_op].sds_id == qa_sds_nobs_info[j_op].sds_id))
		        ;
	          if (j_op >= i_op) close_sds(&qa_sds_nobs_info[i_op], qa_sds_nobs_info[i_op].H5OUT);
	        }
	      }

            for (i_op=0; i_op<=n_op; i_op++)
              if (fqa_l2g[i_op] == 1)
              {
                for (j_op=0; j_op<i_op; j_op++)
                  if (qa_sds_info[i_op].sd_id == qa_sds_info[j_op].sd_id) break;
                if (j_op >= i_op) free(data_qa_nadd[i_op]);
              }
  
	    for (isds=0; isds<nsds; isds++)
              if (out_sds_info[isds].sds_id != -1) close_sds(&out_sds_info[isds], out_sds_info[isds].H5OUT);
            close_qa_hdf_nsds(arg_list[0], in_sds_info, nsds, qa_fnames, qa_sds_info, n_op);
            for (isds=0; isds<nsds; isds++)
              if (in_sds_info[isds].sds_id != -1) 
	      {
		close_sds(&in_sds_info[isds], in_sds_info[isds].H5IN);
	        if (fin_l2g == 1)  close_sds(&in_sdsc_info[isds], in_sdsc_info[isds].H5IN);
	        for (jsds=isds+1; jsds<nsds; jsds++)
	          if (in_sds_info[isds].sds_id == in_sds_info[jsds].sds_id)
	          {
	            in_sds_info[jsds].sds_id = -1; 
	            if (fin_l2g == 1)  close_sds(&in_sdsc_info[jsds], in_sdsc_info[jsds].H5IN);
	          }
	        in_sds_info[isds].sds_id = -1;
		if (fin_l2g == 1)
	          in_sdsc_info[isds].sds_id = -1;
	      }
	    close_sd(&in_sds_info[0], in_sds_info[0].H5IN);

            if (data_qa[0] != NULL) free(data_qa[0]);
            for (i=1; i<=n_op; i++)
            {
              for (j=0; j<i; j++)
                if ((strcmp(qa_fnames[i], qa_fnames[j]) == 0) &&
                      (strcmp(qa_sds_info[i].name, qa_sds_info[j].name) == 0)) break;
              if (j >= i) { if (data_qa[i] != NULL) free(data_qa[i]); }
            } /* for (i=1; . . .  */
	  } /* if (st != -1) . . .  */
	  free(data_mask);
	  Free2D((void **)data_in);
	  Free2D((void **)data_out);
        } /* if ((data_in != NULL) && . . .  ) */
        Free2D((void **)in_edge);
        Free2D((void **)in_start);
        Free2D((void **)out_edge);
        Free2D((void **)out_start);
      } /* if (st != -1)  */
      free(out_sds_info);
      free(in_sds_info);
      if (fin_l2g == 1) { free(in_sdsc_info);  free(data_in_nadd); }
      if (l2g_lite == 1) { free(data_in_nadd_1km);  free(data_in_nadd_500m); }
    } /* if (out_sds_info != . . ) */
  } /* if (st != -1) */ 
  Free2D((void **)qa_fnames);
  return st;
}

void mask_nsds_data_row(void **data_in, void **data_out, uint8 *data_mask, int *ndata_out, 
	int ndata_mask, int nsds, int bsq, int *st_c, int *offset, sds_t *sds_info, 
	double *sum_val, double *sum_val2, int *npix, long *min, long *max, FILE *fp)
{
  int val;
  int i, j, k;
  int ndata = 0;
  int k1, jj, isds;
  int ii[MAX_NSDS];
  int ic[MAX_NSDS];


/*  k = ndata_out/ndata_mask;*/
  if (bsq == 1)
  {
    for (isds=0; isds<nsds; isds++) {  
      ii[isds] = st_c[isds];   
/*    for (k1=0, j=0; k1<k; k1++)*/
      for (k1=0, j=0; k1<ndata_out[isds]/ndata_mask; k1++)    
      {
	for (i=0; i<ndata_mask; i++, j++)
	{
	  jj = ii[isds];  
          if (data_mask[i] == YES)
          {
            switch(sds_info[isds].data_type)
            {
              case 20: val = ((int8 **)data_out)[isds][j] = ((int8 **)data_in)[isds][jj]; break;
              case 21: val = ((uint8 **)data_out)[isds][j] = ((uint8 **)data_in)[isds][jj]; break;
              case 22: val = ((int16 **)data_out)[isds][j] = ((int16 **)data_in)[isds][jj]; break;
              case 23: val = ((uint16 **)data_out)[isds][j] = ((uint16 **)data_in)[isds][jj]; break;
              case 24: val = ((int32 **)data_out)[isds][j] = ((int32 **)data_in)[isds][jj]; break;
              case 25: val = ((uint32 **)data_out)[isds][j] = ((uint32 **)data_in)[isds][jj]; break;
            } 
            if (val != sds_info[isds].fill_val)	 	    
	    {       
	      if (val < min[isds]) min[isds] = val;
              else if (val > max[isds]) max[isds] = val;
              sum_val[isds] += (double)val; 
              sum_val2[isds] += (double)(val*val);
              npix[isds] += 1;
              if (fp != NULL) { fprintf(fp, "%d  ", val); ndata = 1; }
	    }
          }
          else
            switch(sds_info[isds].data_type)
            {
              case 20: ((int8 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 21: ((uint8 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 22: ((int16 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 23: ((uint16 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 24: ((int32 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 25: ((uint32 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
            }
	  ii[isds] += offset[isds];
	  if ((fp != NULL) && (ndata == 1)) { fprintf(fp, "\n"); ndata = 0; }
	} /* for (i=0; . . . ) */
      } /* for (k1=0; . .  .) */
    }
  }
  else
  {
    for (isds=0; isds<nsds; isds++)
    {
      ic[isds] = 0;
      ii[isds] = st_c[isds];
    }
    for (i=0; i<ndata_mask; i++)
    {
      for (isds=0; isds<nsds; isds++)
      {
/*        j = ic[isds];*/
/*        for (k1=0; k1<k; k1++)*/
        for (k1=0; k1<ndata_out[isds]/ndata_mask; ic[isds]++, k1++)	
        {
	  j = ic[isds];
	  jj = ii[isds];
          if (data_mask[i] == YES)
          {
            switch(sds_info[isds].data_type)
            {
              case 20: val = ((int8 **)data_out)[isds][j] = ((int8 **)data_in)[isds][jj]; break;
              case 21: val = ((uint8 **)data_out)[isds][j] = ((uint8 **)data_in)[isds][jj]; break;
              case 22: val = ((int16 **)data_out)[isds][j] = ((int16 **)data_in)[isds][jj]; break;
              case 23: val = ((uint16 **)data_out)[isds][j] = ((uint16 **)data_in)[isds][jj]; break;
              case 24: val = ((int32 **)data_out)[isds][j] = ((int32 **)data_in)[isds][jj]; break;
              case 25: val = ((uint32 **)data_out)[isds][j] = ((uint32 **)data_in)[isds][jj]; break;
            }
            if (val != sds_info[isds].fill_val)
            {
              if (val < min[isds]) min[isds] = val;
              else if (val > max[isds]) max[isds] = val;
              sum_val[isds] += (double)val;
              sum_val2[isds] += (double)(val*val);
              npix[isds] += 1;
              if (fp != NULL) { fprintf(fp, "%d  ", val); ndata = 1; }
            }
          }
          else
            switch(sds_info[isds].data_type)
            {
              case 20: ((int8 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 21: ((uint8 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 22: ((int16 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 23: ((uint16 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 24: ((int32 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
              case 25: ((uint32 **)data_out)[isds][j] = sds_info[isds].fill_val; break;
            }
	  ii[isds] += offset[isds];
        } /* for (k1=0; . . . ) */
      } /* for (isds=0; . . .) */ 
      if ((fp != NULL) && (ndata == 1)) { fprintf(fp, "\n"); ndata = 0; }
    } /* for (i=0; . .  .) */
  }
}


