/******************************************************************************
!C

!File: mosaic_sds.c

!Description:
  This file contains the routines for making mosaic of multiple SDS into
one single SDS. 

!Input Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original September/October 2002. 

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c


!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
********************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "meta.h"
#include "sds_rw.h"
#include "qa_tool.h"
#include "str_op.h"
#include "alloc_mem.h"
#include "main_util.h"


#define HELP \
"NAME \n" \
"	sds_mosaic - prepare a mosaic of SDSs or spatial subset of SDSs from different files \n" \
" \n" \
"SYNOPSIS \n" \
"	sds_mosaic -help [filename]\n" \
"	sds_mosaic [-sds=SDSname1[,SDSname2...]] -of=<filename> [-fill=<fill value>] [-gr=<hb,vb>]\n" \
"		f1 [-row=<row range>] [-col=<col range>] f2 [-row=<row range>] [-col=<col range> [f3 . [f4 . . ]] \n" \
" \n" \
"DESCRIPTION \n" \
"	Make a single mosaic SDS by arranging the same input SDS or user speicified \n" \
"       subset of input SDS from different input files in order of tile id. \n" \
"	The horizontal and vertical tile numbers are either recognized from the\n" \
"       filename or the metadata VERTICALTILENUMBER and HORIZONTALTILENUMBER. Missing \n" \
"       tiles within the recognized range of tile numbers are filled with a fill value\n" \
"       different from the SDS fill value. Multiple mosaic SDSs are created if two or \n" \
"       more SDSs are listed in the input argument. Output is an HDF file containing\n" \
"	one or more mosaic SDSs. \n" \
"       If the tile id is not recognized in one or more of the input files then the tiles \n" \
"       are mosaiced by placing the tiles in an optimum row column block order. \n" \
" \n" \
"OPTIONS \n" \
"\t-help               Display this help message \n" \
"\t-sds=<SDS list>     List of SDS (separated by commas). Accepts only 2D SDS(s). \n" \
"\t-of=<filename>      Output filename \n" \
"\t-fill=<fill value>  Fill value for missing tiles (default: SDS fill value) \n" \
"\t-gr=<hb,vb>         Gravity control; Number of blocks in horzontal and vertical \n" \
"\t                    Default is square root of total number of input files \n" \
"\tf1 f2 f3 .          Input filenames (followed by -row and -col option if subset is required) \n" \
"\t-row=<min,max>      Range of SDS rows to be included in the mosaic. (default: all rows) \n" \
"\t                    Input for each file separately follwoing the input filename. \n" \
"\t-col=<min,max>      Range of SDS columns to be included in the mosaic. (default: all columns) \n" \
"\t                    Input for each file separately follwoing the input filename. \n" \
" \n" \
"Examples: \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:		sds_mosaic [-sds=SDSname1[,SDSname2...]] -of=<filename> [-fill=<fill value>] [-gr=<hb,vb>]\n" \
"		f1 [-row=<row range>] [-col=<col range>] f2 [-row=<row range>] [-col=<col range> [f3 . [f4 . . ]] \n" \
" \n" \
"OPTIONS \n" \
"\t-help               Display this help message \n" \
"\t-sds=<SDS list>     List of SDS (separated by commas). Accepts only 2D SDS(s). \n" \
"\t-of=<filename>      Output filename \n" \
"\t-fill=<fill value>  Fill value for missing tiles (default: SDS fill value) \n" \
"\t-gr=<hb,vb>         Gravity control; Number of blocks in horzontal and vertical \n" \
"\t                    Default is square root of total number of input files \n" \
"\tf1 f2 f3 .          Input filenames (followed by -row and -col option if subset is required) \n" \
"\t-row=<min,max>      Range of SDS rows to be included in the mosaic. (default: all rows) \n" \
"\t                    Input for each file separately follwoing the input filename. \n" \
"\t-col=<min,max>      Range of SDS columns to be included in the mosaic. (default: all columns) \n" \
"\t                    Input for each file separately follwoing the input filename. \n" \

int parse_cmd_mosaic_sds(int argc, char **argv, int *if_cnt, char **sds_names, int *sds_cnt,
	char *out_fname, int **row_range, int **col_range, char *fval_str, int *hb, int *vb);
int get_range(char *rc_str, int *rc_range);
int **create_mosaic_map(int argc, char **argv, int if_cnt, int *n_br, int *n_bc, int hb, int vb);
int get_tile_id(char *fname, int *h_id, int *v_id);
void open_mosaic_sds(sds_t *in_sds_info, sds_t *out_sds_info, int n_br, int n_bc, int n, int m, char *fval_str);
void update_mosaic_sds(sds_t *in_sds_info, sds_t *out_sds_info, int fid, int *rowRange, int *colRange, 
		char *fval_str, int i_br, int i_bc, int n, int m);	
	

int main(int argc, char **argv)
/******************************************************************************
!C

!Description:
  Main function for mosaic_sds

!Input Parameters: (none)
  command line arguments: see help for details.

!Output Parameters: (none)
  return 1 on successful completion of the process

!Revision History:
  Original September/October 2002.

!Team-unique Header:
  See file prologue.

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
********************************************************************************/
{
  int hb, vb;
  int i, j, k, p;
  int i_br, i_bc, n_br, n_bc;
  int **mosaic_map;
  int sds_cnt, len, attr_len;
  int isds, fid, if_cnt;
  int32 sd_id, sds_id, dt;
  int32 nsds, nattr, rank, dim_size[4];
  int n, m, status;
  int **row_range, **col_range, *rowRange, *colRange;
  sds_t in_sds_info, out_sds_info;
  char name[MAX_SDS_NAME_LEN], sds_name[MAX_SDS_NAME_LEN];
  char *iptr_str, **sds_names, dim_str[MAX_STR_LEN];
  char ip_name[MAX_SDS_NAME_LEN], fval_str[10];
  char in_fname[MAX_PATH_LENGTH], out_fname[MAX_PATH_LENGTH];

  /* MSZ 2/7/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/7/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ((argc == 2) && ((strcmp(argv[1], "-h") == 0) || (strcmp(argv[1], "-help") == 0)))
  {
    /* print command line help */
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  else
  {
    /* print SDS names if filename input with help option */
    for (i=0, status=0; i<argc; i++)
      if ((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "-help") == 0)) status = i; 
    if (status != 0)
    {
      /* Print help or all the SDS names from the first file in the list of input files */
      nsds = 0;
      for (i=1; i<argc; i++) 
      {
	if (argv[i][0] != '-') 
	{
	  display_sds_info_of_file(argv[i]);
	  nsds = get_nsds(argv[i]);
          if (nsds == 0)
	    fprintf(stdout, "Input %s do not contain any valid SDS\n", argv[i]);
	  break;
        }
      } /* for (i=2; . .  ) */
      exit(0);
    }
  }

  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names in sds_mosaic(main)\n");
  if ((row_range = (int **)Calloc2D(MAX_NUM_FILE, 2, sizeof(int))) == NULL)
    fprintf(stderr, "Cannot allocate memory for row_range in enlarge_sds(main)\n");
  if ((col_range = (int **)Calloc2D(MAX_NUM_FILE, 2, sizeof(int))) == NULL)
    fprintf(stderr, "Cannot allocate memory for col_range in enlarge_sds(main)\n");
  if ((sds_names != NULL) && (row_range != NULL) && (col_range != NULL))
  {
    /* get all the input parameters */

    status = parse_cmd_mosaic_sds(argc, argv, &if_cnt, sds_names, &sds_cnt, out_fname,
					row_range, col_range, fval_str, &hb, &vb);

    if (status == -1) fprintf(stderr, "%s\n", USAGE);
    else if (status != 0)
    {
      fprintf(stdout, "Started Process sds_mosaic \n");

      /* default: read all SDS */

      for (fid=1; fid<argc; fid++)
	if (argv[fid][0] != '-') break;
      strcpy(in_fname, argv[fid]);
      if ((sds_cnt == 1) && (strcmp(sds_names[0], "all") == 0))
        sds_cnt = get_sds_names(in_fname, sds_names);

      /* Fix SDS names with dot extension when neccessary */

      update_nd_sdsnames(sds_names, &sds_cnt, in_fname);

      out_sds_info.H5OUT = 0;  /* so far, output to HDF4 */
      out_sds_info.sd_id = out_sds_info.sds_id = -1;
      if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
	fprintf(stderr, "Cannot create output file: %s\n", out_fname);
      else
      {
        if ((iptr_str = (char *)calloc(if_cnt*MAX_PATH_LENGTH, sizeof(char))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for iptr_str in sds_mosaic: main()\n");
	else
	{
	  for (i=1, fid=0; i<argc; i++)
	  {
            if (argv[i][0] != '-')
	    {
	      strcpy(ip_name, argv[i]);
	      rm_path(ip_name);
	      j = (int)strlen(ip_name);
	      if (fid == 0) strcpy(iptr_str, ip_name);
	      else {
		strcat(iptr_str, "\n");
		strcat(iptr_str, ip_name);
	      }
	      fid++;
	    }
	  }
	  /* write attribute INPUTPOINTER to output HDF file */
	  /* this attribute contains list of input HDF files */
	  attr_len = (int)strlen(iptr_str) + 1;
	  if (set_sd_attr(out_sds_info.H5OUT, out_sds_info.sd_id, "INPUTPOINTER", DFNT_CHAR8, attr_len, (VOIDP)iptr_str) == FAIL) 
	    fprintf(stderr, "Cannot write SDS attribute INPUTPOINTER in sds_mosaic: main()\n");
	  free(iptr_str);
	}

	/* create mosaic map of pointers to input tiles */

	mosaic_map = (int **)create_mosaic_map(argc, argv, if_cnt, &n_br, &n_bc, hb, vb); 

	/* create mosaic of each input SDS */

	for (isds=0; isds<sds_cnt; isds++)
	{
	  fprintf(stdout, "\tCreating mosaic of SDS: %s\n", sds_names[isds]);
	  out_sds_info.H5OUT = 0;   /* so far, output to HDF4 */
	  out_sds_info.sds_id = -1;
	  get_sdsname_dim(sds_names[isds], sds_name, &n, &m);
	  for (i=1; i<argc; i++) {
	    if (argv[i][0] == '-') continue;
	    else {
	      strcpy(in_fname, argv[i]);
	      strcpy(in_sds_info.name, sds_name);
	      in_sds_info.sds_id = in_sds_info.sd_id = -1;
	      if (get_sds_info(in_fname, &in_sds_info) != -1) { 	      
		open_mosaic_sds(&in_sds_info, &out_sds_info, n_br, n_bc, n, m, fval_str);
		close_sds(&in_sds_info, in_sds_info.H5IN);
		close_sd(&in_sds_info, in_sds_info.H5IN);
		break;
	      }
	    }
          } 
	  for (i=0; i<n_br; i++)
	  {
	    for (j=0; j<n_bc; j++) 
	    {
	      fid = mosaic_map[i][j];
	      if (fid != -1) {
	        for (k=1, p=0; k<argc; k++) {
		  if (argv[k][0] == '-') continue;
		  else {
		    if (p == fid) break;
		    else ++p; 
		  }
	        }
	        strcpy(in_fname, argv[k]);
	        strcpy(in_sds_info.name, sds_name);
	        in_sds_info.sds_id = in_sds_info.sd_id = -1;
	        if (get_sds_info(in_fname, &in_sds_info) == -1) { 
	          fprintf(stderr, "Ignoring input file: %s\n", in_fname);
		  fid = -1;
		}	      	      
		update_mosaic_sds(&in_sds_info, &out_sds_info, fid, row_range[p], col_range[p], fval_str, i, j, n, m);	      
	      }
	      else		
		update_mosaic_sds(&in_sds_info, &out_sds_info, fid, row_range[0], col_range[0], fval_str, i, j, n, m);	    

	      close_sds(&in_sds_info, in_sds_info.H5IN);
	      close_sd(&in_sds_info, in_sds_info.H5IN);
	    }
	  }
	} /* for (isds=0; . .  ) */

	close_sds(&out_sds_info, out_sds_info.H5OUT);
	close_sd(&out_sds_info, out_sds_info.H5OUT);
	Free2D((void **)mosaic_map);
      }
    }
    Free2D((void **)sds_names);
    Free2D((void **)col_range);
    Free2D((void **)row_range);
  }
  fprintf(stdout, "Process sds_mosaic completed\n");
  return 0;
}
     
int parse_cmd_mosaic_sds(int argc, char **argv, int *if_cnt, char **sds_names, int *sds_cnt, char *out_fname,
	int **row_range, int **col_range, char *fval_str, int *hb, int *vb)
/******************************************************************************
!C

!Description:
  Function to parse command line arguments.

!Input Parameters:
  argc: number of input arguments
  argv: string array containing arguments

!Output Parameters:
  if_cnt: number of input files.
  sds_names: input SDS names
  sds_cnt: number of SDS
  out_fname: output filename
  row_range: row range (number of file x 2)
  col_range: column range (number of file x 2)
  fval_str: fill value
  hb: number of tiles in x direction
  vb: number of tiles in y direction.
  return 1 if parsing is succesfull, -1 if not all required parameters input.
 
!Revision History:
  Original September/October 2002.

!Team-unique Header:
    See file prologue

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
********************************************************************************/
{
  int i, j, nb, status;
  int len, rcnt, ccnt, hbvb_arr[2];
  char rc_str[20], gr_str[20], tmp_str[10];

  /* initialize all variables */
  status = 1;
  rcnt = ccnt = 0;
  *hb = *vb = -1;
  *if_cnt = *sds_cnt = 0;
  fval_str[0] = '\0';
  out_fname[0] = '\0';
  for (i=0; i<MAX_NUM_FILE; i++) {
    row_range[i][0] = col_range[i][0] = -1;
    row_range[i][1] = col_range[i][1] = -1;
  }

  /* parse input arguments */
  i = 1;
  while (i < argc)
  {
    if (is_arg_id(argv[i], "-sds=") == 0)
      get_arg_val_arr(argv[i], sds_names, sds_cnt);
    else if (is_arg_id(argv[i], "-of=") == 0)
      get_arg_val(argv[i], out_fname);
    else if (is_arg_id(argv[i], "-fill=") == 0) 
      get_arg_val(argv[i], fval_str);
    else if (is_arg_id(argv[i], "-gr=") == 0)
    {
      get_arg_val(argv[i], gr_str);
      if (get_range(gr_str, hbvb_arr) != 2)
	fprintf(stderr, "Ignoring invalid option: %s\n", argv[i]);
      else {
	*hb = hbvb_arr[0];
	*vb = hbvb_arr[1];
      }
    }
    else if (argv[i][0] == '-') fprintf(stderr, "Unknonwn option %s\n", argv[i]);
    else 
    {
      row_range[*if_cnt][0] = col_range[*if_cnt][0] = -1;
      row_range[*if_cnt][1] = col_range[*if_cnt][1] = -1;
      j = i+1;
      if (j < argc)
      { 
	/* read row and column range of subset if input - 
	   can be input in any order following the input filename */
	if (is_arg_id(argv[j], "-row=") == 0) 
	{
	  i++;
	  get_arg_val(argv[i], rc_str);
	  if (strlen(rc_str) > 0)
	    rcnt = get_range(rc_str, row_range[*if_cnt]);
	  if (rcnt != 2) {
	    fprintf(stderr, "Input subset row range %s not valid. Set to default image size\n", argv[j]);
	    row_range[*if_cnt][0] = row_range[*if_cnt][1] = -1;
	  }
	  else if (row_range[*if_cnt][0] < 0) {
	    fprintf(stderr, "Input subset row range %s is negative. Set to default image size\n", argv[j]);
	    row_range[*if_cnt][0] = -1;
	  }
	  j++;
	  if (j < argc) 
	  {
            if (is_arg_id(argv[j], "-col=") == 0)
            {
              i++;
              get_arg_val(argv[i], rc_str);
              if (strlen(rc_str) > 0)
                ccnt = get_range(rc_str, col_range[*if_cnt]);
              if (ccnt != 2) {
                fprintf(stderr, "Input subset column range %s not valid. Set to default image size\n", argv[j]);
                col_range[*if_cnt][0] = col_range[*if_cnt][1] = -1;
              }
              else if (col_range[*if_cnt][0] < 0) {
                fprintf(stderr, "Input subset row range %s is negative. Set to default image size\n", argv[j]);
                col_range[*if_cnt][0] = -1;
              }
	    }
          }
	}
        else if (is_arg_id(argv[j], "-col=") == 0)
        { 
	  i++;
          get_arg_val(argv[i], rc_str);
          if (strlen(rc_str) > 0)
            ccnt = get_range(rc_str, col_range[*if_cnt]);
	  if (ccnt != 2) {
	    fprintf(stderr, "Input subset column range %s not valid. Set to default image size\n", argv[j]);
	    col_range[*if_cnt][0] = col_range[*if_cnt][1] = -1;
	  }
	  else if (col_range[*if_cnt][0] < 0) {
	    fprintf(stderr, "Input subset row range %s is negative. Set to default image size\n", argv[j]);
	    col_range[*if_cnt][0] = -1;
	  }
	  j++;
	  if (j < argc)
	  {
            if (is_arg_id(argv[j+1], "-row=") == 0)
            {
              i++;
              get_arg_val(argv[i], rc_str);
              if (strlen(rc_str) > 0)
                rcnt = get_range(rc_str, row_range[*if_cnt]);
              if (rcnt != 2) {
                fprintf(stderr, "Input subset row range %s not valid. Set to default image size\n", argv[j]);
                row_range[*if_cnt][0] = row_range[*if_cnt][1] = -1;
              }
              else if (row_range[*if_cnt][0] < 0) {
                fprintf(stderr, "Input subset row range %s is negative. Set to default image size\n", argv[j]);
                row_range[*if_cnt][0] = -1;
              }
	    }
	  }
	}
      }
      ++*if_cnt;
    }
    i++;
  }

  /* check if all required arguments are input */
  if (out_fname[0] == '\0')
  {
    status = -1;
    fprintf(stderr, "No output filename specified\n");
  }
  if (*if_cnt == 0)
  {
    status = -1;
    fprintf(stderr, "No input filename specified\n");
  }
  if ((status != -1) && (*sds_cnt == 0))
  {
    fprintf(stderr, "No SDS name input. Reading all SDS . . \n");
    *sds_cnt = 1;
    strcpy(sds_names[0], "all");
    nb = *hb * *vb;
    if ((nb > 1) && (nb < *if_cnt))
    {
      fprintf(stderr, "Number of input files exceed gravity option limit. Gravity option set to compact\n");
      *hb = *vb = -1;
    }
  }
  return status;
}
	    
int get_range(char *rc_str, int *rc_range)
/******************************************************************************
!C

!Description:
  get range value: min and max from the input string

!Input Parameters:
  rc_str: string containing range value

!Output Parameters: 
  rc_range: integer array to output min and max.
  return nubmer of values extracted from rc_str. 

!Revision History:
  Original September/October 2002.

!Team-unique Header:
    See file prologue

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes:

!END
********************************************************************************/
{
  int p, st, len;
  char tmp_str[10];

  st = 0;
  len = (int)strlen(rc_str);
  p = charpos(rc_str, ',', 0);
  if (p != -1)
  {
    strmid(rc_str, 0, p, tmp_str);
    rc_range[0] = atoi(tmp_str);
    st++;
    p++;
    strmid(rc_str, p, len-p, tmp_str);
    if (strlen(tmp_str) > 0)
      rc_range[1] = atoi(tmp_str);
    if (rc_range[1] > rc_range[0])
      st++;
  }
  return st;
}

int **create_mosaic_map(int argc, char **argv, int if_cnt, int *n_br, int *n_bc, int hb, int vb)
/******************************************************************************
!C

!Description:
  Create a mosaic map - a 2D array containing identification of input files
  that go into that position of the output mosaic.

!Input Parameters:
  argc: number of arguments in the command line
  argv: input command line arguments
  if_cnt: number of input files.
  hb: number of tiles in x direction (when input using -gr option)
  vb: nubmer of tiles in y direction (when input using -gr option)

!Output Parameters: 
  n_br: number of tiles in y direction
  n_bc: number of tiles in x direction
  return the 2D array on sucessful completion or NULL on failure. 

!Revision History:
  Original September/October 2002.

!Team-unique Header:
    See file prologue

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes:

!END
********************************************************************************/
{
  int i, j;
  int nh, nv;
  int fid, st;
  int *h_id, *v_id;
  int h_min, h_max;
  int v_min, v_max;
  int **mosaic_map;
  float nh_f, nv_f;

  st = 0;
  h_id = (int *)calloc(if_cnt, sizeof(int));
  v_id = (int *)calloc(if_cnt, sizeof(int));
  
  /* Identify the tile ids from input MODIS file */
  /* If tile id is not reconginized for one or more input tile */
  /* output mosaic cannot be created based on the tile ids */

  for (i=1, fid=0; i<argc; i++)
  {
    if (argv[i][0] == '-') continue;
    st = get_tile_id(argv[i], &h_id[fid], &v_id[fid]);
    if (st != 1) {
      fprintf(stdout, "No tile id found in %s\n", argv[i]);
    }
    else
    {
      /* search for id and identify the min and max value for 
	 vertical and horizontal tile numbers */

      fprintf(stdout, "\tRecognized file %d %s: TileID(H, V) ->  (%d, %d)\n", fid, argv[i], h_id[fid], v_id[fid]);
      if (fid == 0) 
      {
        h_min = h_max = h_id[fid];
        v_min = v_max = v_id[fid];
      }
      else
      {
        if (h_id[fid] < h_min) h_min = h_id[fid];
        else if (h_id[fid] > h_max) h_max = h_id[fid];
        if (v_id[fid] < v_min) v_min = v_id[fid];
        else if (v_id[fid] > v_max) v_max = v_id[fid];
      }
      fid++;
    }
  }
  if (st == 1) 
  {
    /* tile ids identified for all inputs. nh x nv is the number of tiles in the output mosaic */
    nh = h_max - h_min + 1;
    nv = v_max - v_min + 1;
  }
  else 
  {
    /* tile id not identified for at least one tile */
    /* compute the nh and nv based on least space requirement in the output mosaic */

    if ((hb == -1) && (vb == -1))
    {
      /* no -gr option is input. */
      fprintf(stderr, "Tile ID is not recognized in one or more input files. \n");
      nh_f = (float)sqrt((double)if_cnt);
      nh_f += 0.5;
      nh = (int)nh_f;
      nv_f = (float)if_cnt/(float)nh;
      nv_f = ceilf(nv_f);;
      nv = (int)nv_f;
    }
    else
    {
      /* -gr option is input - nh and nv values are input in the option */
      nh = hb;
      nv = vb;
    }
  }

  /* nh = nv; a requirement by the toolkit for putting projection information 
     at the tile level. Need to check it if it is really an issue */
/*
  if (nh != nv)
  {
    if (nh > nv) nv = nh;
    else nh = nv;
  }
*/

  /* now create the a 2D array mosaic_map that will hold pointer to input files */

  *n_bc = nh; *n_br = nv;
  mosaic_map = (int **)Calloc2D(nv, nh, sizeof(int));
  for (i=0; i<nv; i++)
    for (j=0; j<nh; j++)
      mosaic_map[i][j] = -1;
  if (st == 1)
  {
    for (fid=0; fid<if_cnt; fid++)
    {
      j = h_id[fid] - h_min;
      i = v_id[fid] - v_min;
      mosaic_map[i][j] = fid;
    }                           
  }
  else
  {
    for (fid=0; fid<if_cnt; fid++)
    {
      i = fid/nh; 
      j = fid%nh;
      mosaic_map[i][j] = fid;
    }                           
  }
  free(h_id);
  free(v_id);
  return mosaic_map;
}

int get_tile_id(char *fname, int *h_id, int *v_id)
/******************************************************************************
!C

!Description:
  get the tile id from the input filename

!Input Parameters: (none)
  fname: input filename

!Output Parameters: (none)
  h_id: horizontal tile number
  v_id: vertical tile number
  return 1 on success and -1 on failure

!Revision History:
  Original September/October 2002.

!Team-unique Header:
    See file prologue

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
********************************************************************************/
{
  int p1, p2, i, len;
  int rd_mt, st, m_cnt;
  char gid[80], id_str[10];
  char *cmeta_str, **meta_val;

  /* try finding the tileid in the filename */
  /* else read required metadata */

  rd_mt = st = 0;
  len = (int)strlen(fname);

  /* skip the filepath */

  for (i=len-1; i>=0; i--)
    if (fname[i] == '/') break;

  /* check the filename */

  strmid(fname, i+1, len-i-1, gid);
  p1 = charpos(gid, '.', 0);
  if (p1 != -1)
    p2 = charpos(gid, '.', p1+1);
  else rd_mt = 1;
  if ((p1 != -1) && (p2 != -1))
  {
    /* get tile id from the filename */

    if ((gid[p2+1] == 'h') && (gid[p2+4] == 'v'))
    {
      strmid(gid, p2+2, 2, id_str);                 
      *h_id = (int)atoi(id_str);
      strmid(gid, p2+5, 2, id_str);                 
      *v_id = (int)atoi(id_str);
      st = 1;
    }
    else rd_mt = 1;
  }
  else rd_mt = 1;
  if (rd_mt == 1)
  {
    /* read the metadata and get the tile id */

    if ((cmeta_str = get_attr_metadata(fname, "CoreMetadata.0")) != NULL)
    {
      if ((meta_val = (char **)Calloc2D(5, 20, sizeof(char))) != NULL)
      {
        m_cnt = 0;
        get_sel_metadata(cmeta_str, "HORIZONTALTILENUMBER", meta_val, &m_cnt, 0);
        get_sel_metadata(cmeta_str, "VERTICALTILENUMBER", meta_val, &m_cnt, 0);
        if (m_cnt == 2)
        {
          *h_id = (int)atoi(meta_val[0]);
          *v_id = (int)atoi(meta_val[1]);
	  st = 1;
        }
        Free2D((void **)meta_val);
      }
      free(cmeta_str);
    }                                                                             
  }
  return st;
}

void open_mosaic_sds(sds_t *in_sds_info, sds_t *out_sds_info, int n_br, int n_bc, int n, int m, char *fval_str)
{
  int i, j, k, bsq; 
  int rank, out_rank;
  void *miss_fill;

  /* identify the input SDS layer and data interleaving format and SDS dimension information */
  /* based on the compute output SDS interleaving and dimension information etc. */
                                                                                                                               
  rank = in_sds_info->rank;
  bsq = ((rank == 2) || (in_sds_info->dim_size[0] < in_sds_info->dim_size[rank-1])) ? 1 : 0;
/*  if ((n == -1) && (m == -1)) sprintf(out_sds_info->name, "Mosaic_%s", in_sds_info->name);*/
  if ((n == -1) && (m == -1)) sprintf(out_sds_info->name, "%s", in_sds_info->name);  
  else if ((n != -1) && (m == -1)) sprintf(out_sds_info->name, "Mosaic_%s.%d", in_sds_info->name, n+1);
  else if ((n != -1) && (m != -1)) sprintf(out_sds_info->name, "Mosaic_%s.%d.%d", in_sds_info->name, n+1, m+1);
  if (in_sds_info->data_type == 5 || in_sds_info->data_type == 6)
    out_sds_info->fill_val = in_sds_info->fill_fval;
  else
    out_sds_info->fill_val = in_sds_info->fill_val;
  out_sds_info->data_type = in_sds_info->data_type;
  out_sds_info->data_size = in_sds_info->data_size;
  out_sds_info->rank = ((n == -1) && (m == -1)) ? rank : 2;
  if (rank == out_sds_info->rank)
  {
    for (j=0; j<rank; j++)
      out_sds_info->dim_size[j] = in_sds_info->dim_size[j];
  }
  else
  {
    if (bsq == 1)
    {
      out_sds_info->dim_size[0] = in_sds_info->dim_size[rank-2];
      out_sds_info->dim_size[1] = in_sds_info->dim_size[rank-1];
    }
    else
    {
      out_sds_info->dim_size[0] = in_sds_info->dim_size[0];
      out_sds_info->dim_size[1] = in_sds_info->dim_size[1];
    }
  }
  if (bsq == 0) {
    out_sds_info->dim_size[0] *= n_br;
    out_sds_info->dim_size[1] *= n_bc;
  }
  else {
    if (out_sds_info->rank == 2) {
      out_sds_info->dim_size[0] *= n_br;
      out_sds_info->dim_size[1] *= n_bc;
    }
    else {
      out_sds_info->dim_size[rank-2] *= n_br;
      out_sds_info->dim_size[rank-1] *= n_bc;
    }
  }

  if (open_sds((char *)NULL, out_sds_info, 'W') != -1)
  {
    if (in_sds_info->data_type == 5 || in_sds_info->data_type == 6)
      write_attr_fvalf(out_sds_info->H5OUT, out_sds_info->sds_id, in_sds_info->data_type, 1, in_sds_info->fill_fval);
    else
      write_attr_fval(out_sds_info->H5OUT, out_sds_info->sds_id, in_sds_info->data_type, 1, (int)in_sds_info->fill_val);
    switch(in_sds_info->data_type)
    {
      case 5 : miss_fill = (float32 *)calloc(1, in_sds_info->data_size);
               if (fval_str[0] != '\0') ((float32 *)miss_fill)[0] = (float)atof(fval_str);
               else ((float32 *)miss_fill)[0] = in_sds_info->fill_fval;
               break;
      case 6 : miss_fill = (float64 *)calloc(1, in_sds_info->data_size);
               if (fval_str[0] != '\0') ((float64 *)miss_fill)[0] = (float)atof(fval_str);
               else ((float64 *)miss_fill)[0] = in_sds_info->fill_fval;
               break;
      case 20: miss_fill = (int8 *)calloc(1, in_sds_info->data_size);
               if (fval_str[0] != '\0') ((int8 *)miss_fill)[0] = (int8)atoi(fval_str);
               else ((int8 *)miss_fill)[0] = in_sds_info->fill_val;
               break;
      case 21: miss_fill = (uint8 *)calloc(1, in_sds_info->data_size);
               if (fval_str[0] != '\0') ((uint8 *)miss_fill)[0] = (uint8)atoi(fval_str);
               else ((uint8 *)miss_fill)[0] = in_sds_info->fill_val;
               break;
      case 22: miss_fill = (int16 *)calloc(1, in_sds_info->data_size);
               if (fval_str[0] != '\0') ((int16 *)miss_fill)[0] = (int16)atoi(fval_str);
               else ((int16 *)miss_fill)[0] = in_sds_info->fill_val;
               break;
      case 23: miss_fill = (uint16 *)calloc(1, in_sds_info->data_size);
               if (fval_str[0] != '\0') ((uint16 *)miss_fill)[0] = (uint16)atoi(fval_str);
               else ((uint16 *)miss_fill)[0] = in_sds_info->fill_val;
               break;
      case 24: miss_fill = (int32 *)calloc(1, in_sds_info->data_size);
               if (fval_str[0] != '\0') ((int32 *)miss_fill)[0] = (int32)atoi(fval_str);
               else ((int32 *)miss_fill)[0] = in_sds_info->fill_val;
               break;
      case 25: miss_fill = (uint32 *)calloc(1, in_sds_info->data_size);
               if (fval_str[0] != '\0') ((uint32 *)miss_fill)[0] = (uint32)atoi(fval_str);
               else ((uint32 *)miss_fill)[0] = in_sds_info->fill_val;
               break;
      default: fprintf(stdout, "HDF datatype %d not supported", in_sds_info->data_type);
	       break;
    }
    if(set_sds_attr(out_sds_info->H5OUT, out_sds_info->sds_id, "MissingFile_FillValue", in_sds_info->data_type, 1, (VOIDP)miss_fill) == FAIL)
      fprintf(stderr, "Cannot write SDS attribute MissingFile_FillValue in sds_mosaic: open_mosaic_sds()\n");
  }
}

void update_mosaic_sds(sds_t *in_sds_info, sds_t *out_sds_info, int fid, int *rowRange, int *colRange, 
		char *fval_str, int i_br, int i_bc, int n, int m)
/******************************************************************************
!C

!Description:
  create mosaic of an SDS from all input files. 

!Input Parameters: 
  in_sds_info: array of input SDS information structure (same SDS from all input files)
  if_cnt: number of input files
  row_range: row range of input SDS to mosaic (if_cnt x 2)
  col_range: col range of input SDS to mosaic (if_cnt x 2) 
  fval_str: fill value for missing tiles
  i_br: index of tile in y direction
  i_bc: index of tile in x direction
  n: 3rd dimension value to identify the SDS layer
  m: 4th dimension value to identify the SDS layer

!Input Parameters/output parameters: 
  out_sds_info: pointer to output mosaic SDS structure.

!Output Parameters: (none)

!Revision History:
  Original September/October 2002.

!Team-unique Header:
    See file prologue

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes:

!END
********************************************************************************/
{
  int i, j, k, ii, rank, out_rank;
  int st_c, bsq, offset;
  int nrows, ncols;
  int nobs, n_iter;
  int fpt;
  int fcnt, ir, ic, ic1;
  int st_row, st_col;
  int end_row, end_col;
  int ndata_in, ndata_out;
  int32 in_edge[4], out_edge[4];
  int32 in_start[4] = {0, 0, 0, 0};
  int32 out_start[4] = {0, 0, 0, 0};
  void *data_in, *data_out, *sds_fill, *miss_fill;
  int32 attr_type, attr_cnt;
  
  if ((miss_fill = (void *)get_sds_attr(out_sds_info, "MissingFile_FillValue", &attr_type, &attr_cnt)) == NULL)
    fprintf(stderr, "Attribute "_FillValue" not defined for output\n");  
    
  /* identify the input SDS layer and data interleaving format and SDS dimension information */
  /* based on the compute output SDS interleaving and dimension information etc. */

  rank = in_sds_info->rank;
  bsq = ((rank == 2) || (in_sds_info->dim_size[0] < in_sds_info->dim_size[rank-1])) ? 1 : 0;
  if (bsq == 0) {
    nrows = in_sds_info->dim_size[0];
    ncols = in_sds_info->dim_size[1];
  }
  else {
    nrows = in_sds_info->dim_size[rank-2];
    ncols = in_sds_info->dim_size[rank-1];
  }

  /* compute the number of input and output pixels in one line */

  if (rank == 2)
    ndata_in = ndata_out = in_sds_info->dim_size[1];
  else if (rank > 2)
  {
    if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
    {
      ndata_in = in_sds_info->dim_size[1];
      for (i=2; i<rank; i++)
        ndata_in *= in_sds_info->dim_size[i];
      ndata_out = ((n == -1) && (m == -1)) ? ndata_in : in_sds_info->dim_size[1];
    }
    else
    {
      ndata_in = in_sds_info->dim_size[rank-1];
      for (i=0; i<rank-2; i++)
        ndata_in *= in_sds_info->dim_size[i];
      ndata_out = ((n == -1) && (m == -1)) ? ndata_in : in_sds_info->dim_size[rank-1];
    }
  }
 
  /* check the row and column range and set to default if there is any problem */
  if ((rowRange[0] == -1) && (rowRange[1] == -1))
  {
    rowRange[0] = 0;
    rowRange[1] = nrows;
  }
  if ((colRange[0] == -1) && (colRange[1] == -1))
  {
    colRange[0] = 0;
    colRange[1] = ncols;
  }
  if (bsq == 0) {
    if (rowRange[1] > in_sds_info->dim_size[0]) {
      fprintf(stderr, "Subset row range %d exceeds the image dimension. Set to default image size\n", rowRange[1]);
      rowRange[1] = nrows;
    }
    if (colRange[1] > in_sds_info->dim_size[1]) {
      fprintf(stderr, "Subset column range %d exceeds the image dimension. Set to default image size\n", colRange[1]);
      colRange[1] = ncols;
    }
  }
  else {
    if (rowRange[1] > in_sds_info->dim_size[rank-2]) {
      fprintf(stderr, "Subset row range %d exceeds the image dimension. Set to default image size\n", rowRange[1]);
      rowRange[1] = nrows;
    }
    if (colRange[1] > in_sds_info->dim_size[rank-1]) {
      fprintf(stderr, "Subset column range %d exceeds the image dimension. Set to default image size\n", colRange[1]);
      colRange[1] = ncols;
    }
  }

  /* create an array SDS fill value */

  if ((sds_fill = (void *)calloc(ndata_out, out_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_fill in create_mosaic\n");
  else
  {
    switch(out_sds_info->data_type)
    {
      case 5 : for (ic=0; ic<ndata_out; ic++) ((float32 *)sds_fill)[ic] = ((float32 *)miss_fill)[0]; break;
      case 6 : for (ic=0; ic<ndata_out; ic++) ((float64 *)sds_fill)[ic] = ((float64 *)miss_fill)[0]; break;
      case 20: for (ic=0; ic<ndata_out; ic++) ((int8 *)sds_fill)[ic] = ((int8 *)miss_fill)[0]; break;
      case 21: for (ic=0; ic<ndata_out; ic++) ((uint8 *)sds_fill)[ic] = ((uint8 *)miss_fill)[0]; break;
      case 22: for (ic=0; ic<ndata_out; ic++) ((int16 *)sds_fill)[ic] = ((int16 *)miss_fill)[0]; break;
      case 23: for (ic=0; ic<ndata_out; ic++) ((uint16 *)sds_fill)[ic] = ((uint16 *)miss_fill)[0]; break;
      case 24: for (ic=0; ic<ndata_out; ic++) ((int32 *)sds_fill)[ic] = ((int32 *)miss_fill)[0]; break;
      case 25: for (ic=0; ic<ndata_out; ic++) ((uint32 *)sds_fill)[ic] = ((uint32 *)miss_fill)[0]; break;
      default: fprintf(stdout, "HDF datatype %d not supported", out_sds_info->data_type);
    }
    free(miss_fill);
  }

  /* compute start and offset for the requested SDS layer */

  compute_sds_start_offset(in_sds_info, n, m, &st_c, &offset);

  if ((data_in = (void *)calloc(ndata_in, out_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_in in create_mosaic\n");
  if ((data_out = (void *)calloc(ndata_out, out_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_in in create_mosaic\n");
  nobs = ndata_out/ncols;

  /* create output mosaic. Process one tile at a time */

  if ((data_in != NULL) && (data_out != NULL) && (fid != -1)) 
  {
    fprintf(stdout, "\t\tOutput file %d to output Mosaic as block (H, V): (%d %d)\n", fid, i_bc, i_br); 
    if (rank == 2)
    {
      in_edge[0] = out_edge[0] = 1;
      in_edge[1] = out_edge[1] = ncols;
    }
    else if (rank > 2)
    {
      for (i=0; i<rank; i++)
        in_edge[i] = out_edge[i] = in_sds_info->dim_size[i];
      if ((n == -1) && (m == -1))
      {
        if (bsq == 0) in_edge[0] = out_edge[0] = 1;
        else in_edge[rank-2] = out_edge[rank-2] = 1;
      }
      else
      {
        out_edge[0] = 1;
	if (bsq == 0) {
	  in_edge[0] = 1;
	  out_edge[1] = in_sds_info->dim_size[1];
	}
	else {
	  in_edge[rank-2] = 1;
	  out_edge[1] = in_sds_info->dim_size[rank-1];
	} 
      }
    }

    st_row = rowRange[0];
    end_row = rowRange[1];
    if (end_row < nrows) end_row++;
    st_col = colRange[0];
    end_col = colRange[1];
    if (end_col < ncols) end_col++;
    if (bsq == 0) {
      st_col *= nobs;
      end_col *= nobs;
    }
    if ((bsq == 0) || (out_sds_info->rank == 2)) n_iter = 1;
    else {
      n_iter = nobs;
      ndata_out = ncols;
    }

    for (ir=0; ir<nrows; ir++) 
    {
      if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
        in_start[0] = ir;
      else
        in_start[rank-2] = ir;
      if ((rank>2) && (n == -1) && (m == -1) && (in_sds_info->dim_size[rank-1] > in_sds_info->dim_size[0]))
      {
        out_start[rank-2] = i_br*nrows + ir;
        out_start[rank-1] = i_bc*ncols;
      }
      else
      {
        out_start[0] = i_br*nrows + ir;
        out_start[1] = i_bc*ncols;
      }

      if ((ir < st_row) || (ir > end_row)) {
	if(set_sel_sds_data(out_sds_info, out_start, out_edge, sds_fill) == FAIL)
          fprintf(stderr, "Error writing data line %d to SDS %s\n", ir, out_sds_info->name);
      }
      else 
      {
	if(read_sel_sds_data(in_sds_info, in_start, in_edge, data_in) == FAIL)
	{
          fprintf(stderr, "Error reading data line %d from SDS %s\n", ir, in_sds_info->name);
	  exit(1);
	}
	
	for (k=0, ic=st_c, ii=0; k<n_iter; k++)	
	{
          for (i=0; i<st_col; ic += offset, i++, ii++)
	  {
            switch(in_sds_info->data_type) {
              case  5: ((float32 *)data_out)[ii] = (float32)in_sds_info->fill_fval; break;
              case  6: ((float64 *)data_out)[ii] = (float64)in_sds_info->fill_fval; break;
              case 20: ((int8 *)data_out)[ii] = (int8)in_sds_info->fill_val; break;
              case 21: ((uint8 *)data_out)[ii] = (uint8)in_sds_info->fill_val; break;
              case 22: ((int16 *)data_out)[ii] =  (int16)in_sds_info->fill_val; break;
              case 23: ((uint16 *)data_out)[ii] = (uint16)in_sds_info->fill_val; break;
              case 24: ((int32 *)data_out)[ii] =  (int32)in_sds_info->fill_val; break;
              case 25: ((uint32 *)data_out)[ii] = (uint32)in_sds_info->fill_val; break;
	      default: fprintf(stdout, "HDF datatype %d not supported", in_sds_info->data_type);
	    }
	  }
          for (j=i; j<end_col; ic += offset, j++, ii++)
          {
            switch(in_sds_info->data_type) {
              case 5 : ((float32 *)data_out)[ii] = ((float32 *)data_in)[ic]; break;
              case 6 : ((float64 *)data_out)[ii] = ((float64 *)data_in)[ic]; break;
              case 20: ((int8 *)data_out)[ii] = ((int8 *)data_in)[ic]; break;
              case 21: ((uint8 *)data_out)[ii] = ((uint8 *)data_in)[ic]; break;
              case 22: ((int16 *)data_out)[ii] = ((int16 *)data_in)[ic]; break;
              case 23: ((uint16 *)data_out)[ii] = ((uint16 *)data_in)[ic]; break;
              case 24: ((int32 *)data_out)[ii] = ((int32 *)data_in)[ic]; break;
              case 25: ((uint32 *)data_out)[ii] = ((uint32 *)data_in)[ic]; break;
	      default: fprintf(stdout, "HDF datatype %d not supported", in_sds_info->data_type);
            }         
	  }
/*          for (i=j; i<ndata_out; i++, ic += offset, ii++)*/
          for (i=j; i<ncols; i++, ic += offset, ii++)	  
          { 
            switch(in_sds_info->data_type) { 
              case  5: ((float32 *)data_out)[ii] = (float32)in_sds_info->fill_fval; break;
              case  6: ((float64 *)data_out)[ii] = (float64)in_sds_info->fill_fval; break;
              case 20: ((int8 *)data_out)[ii] = (int8)in_sds_info->fill_val; break;
              case 21: ((uint8 *)data_out)[ii] = (uint8)in_sds_info->fill_val; break;
              case 22: ((int16 *)data_out)[ii] = (int16)in_sds_info->fill_val; break;
              case 23: ((uint16 *)data_out)[ii] = (uint16)in_sds_info->fill_val; break;
              case 24: ((int32 *)data_out)[ii] = (int32)in_sds_info->fill_val; break;
              case 25: ((uint32 *)data_out)[ii] = (uint32)in_sds_info->fill_val; break;
	      default: fprintf(stdout, "HDF datatype %d not supported", in_sds_info->data_type);
            }
          }      
	  if(set_sel_sds_data(out_sds_info, out_start, out_edge, data_out) == FAIL)
	  {
            fprintf(stderr, "Error writing data line %d to SDS %s\n", ir, out_sds_info->name);
	    exit(1);
	  }  
        }/* for (k=0 */ 	 
      }   /* else */ 
    } /* for (ir=0; . . . */
  }  
      
 /* there is not tile to output for this area of mosaic. so put fill value */
  if (sds_fill != NULL && fid == -1)
  {  	   
    fprintf(stdout, "\t\tOutput fill value to output Mosaic as block (H, V): (%d %d)\n", i_bc, i_br);
    out_rank = out_sds_info->rank;
    rank = in_sds_info->rank;
    for (ir=0; ir<rank; ir++)
      out_edge[ir] = in_sds_info->dim_size[ir];
    for (ir=0; ir<nrows; ir++)
    {
      if ((out_rank == 2) || (out_sds_info->dim_size[out_rank-1] < out_sds_info->dim_size[0]))
      {
	out_start[0] = i_br*nrows + ir;
	out_start[1] = i_bc*ncols;
	out_edge[0] = 1;
	out_edge[1] = ncols;
      }
      else 
      {
        out_start[out_rank - 2] = i_br*nrows + ir;
        out_start[out_rank - 1] = i_bc*ncols;
	out_edge[out_rank - 2] = 1;
      }
      if(set_sel_sds_data(out_sds_info, out_start, out_edge, sds_fill) == FAIL)
      {
        fprintf(stderr, "Error writing data line %d to SDS %s\n", ir, out_sds_info->name);
	exit(1);
      }
    }	    
  }/* if(sds_fill */    
  if (sds_fill != NULL) free(sds_fill);
  if (data_in != NULL) free(data_in);
  if (data_out != NULL) free(data_out);  
}


