/****************************************************************************
!C

!File: read_l2g_comp.c

!Description:
  Supporting routines for read l2g compact format files. 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original January/February 1998. Version 1.0
  Modified April 1998 (Couple of bugs were fixed)
  Modified June 1998 Version 1.1 (Changes to improve the speed up. Changes to 
		support batch processing. Routnies do not exit abruptly on 
		error, instead returns to main with error messages.)  
  Modified September 1999 (Update to support -meta option)
  Modified March 2000 (Update to support -ptr option)
  Modified Oceober 2011 by Xiaoping Zhang, Sigma Space Corp. 
  		(updated the sub of get_l2g_datac_by_gran to get Pointer file name 
		from the input file for NPP.)
		
!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include "mfhdf.h"
#include "l2g.h"
#include "meta.h"
#include "qa_tool.h"
#include "sds_rw.h"
#include "alloc_mem.h"
#include "main_util.h"

int read_input_sds_names_ress(char *fname, char **in_sds_names, int sds_cnt, char **out_sds_names, int *res)
{
  int H5IN, i, cnt;
  int32 sd_id;
  int nsds = 0;
  sds_t sds_info;
  char tmp_sds_name[MAX_SDS_NAME_LEN];

  if (sds_cnt == 0)
    nsds = get_l2g_sds_names(fname, out_sds_names);
  else  /* user specifies without _1 or _c in the names */
  {
    /* MSZ 2/15/2018, modify to use sds_name_existence() */
    for (i=0, cnt=0; i<sds_cnt; i++)
    {
      sprintf(tmp_sds_name, "%s_1", in_sds_names[i]);
      if(sds_name_existence(fname, tmp_sds_name) == 0)
        fprintf(stderr, "Ignoring invalid SDS name %s\n", in_sds_names[i]);
      else {
        strcpy(out_sds_names[cnt], in_sds_names[i]);
        cnt++;
      }
    }
    
    nsds = cnt;
  }

  /* MSZ 2/15/2018, update to HDF4/HDF5 wrapper functions */
  sds_info.sd_id = -1;
  if(open_sd(fname, &sds_info, 'R') == FAIL)
    fprintf(stderr, "Cannot open the input hdf file %s in read_input_sds_names_ress()\n", fname);
  else {
    H5IN = sds_info.H5IN;
    sd_id = sds_info.sd_id;
    for (i=0; i<nsds; i++) {
      sds_info.H5IN = H5IN;
      sds_info.sd_id = sd_id;
      sds_info.sds_id = -1;
      sprintf(sds_info.name, "%s_1", out_sds_names[i]);
      if (get_sds_info((char *)NULL, &sds_info) != -1) {
      /* MSZ 7/25/2018, for both MODIS and non-MODIS */
#if 0
	if (sds_info.dim_size[0] == 1200) res[i] = 1;               /* 1km */
	else if (sds_info.dim_size[0] == 2400) res[i] = 2;          /* 500m */
	else if (sds_info.dim_size[0] == 4800) res[i] = 4;          /* 250m */
#endif
	if ((int)(sds_info.dim_size[0] * 0.001) == 1) res[i] = 1;               /* 1km */
	else if ((int)(sds_info.dim_size[0] * 0.001) == 2) res[i] = 2;          /* 500m */
	else if ((int)(sds_info.dim_size[0] * 0.001) == 4) res[i] = 4;          /* 250m */

	close_sds(&sds_info, sds_info.H5IN);
      }
    }
    close_sd(&sds_info, sds_info.H5IN);
  }

  return nsds;
}

int get_l2g_max_obs_layers(char *hdf_fname, char *max_obs_sds_name)
{
   /* Maosheng Zhao, SSAI, 2/16/2018 */
   sds_t in_sds_info;
   void *sds_data = NULL;
   int i;
   int32 start[4], edge[4];
   int irow, icol;
   int nrows, ncols;
   int GROWS, last_rows;
   int nlines_a_copy = 20;  /* to speed up */
   int num_layers = -1;

   in_sds_info.sd_id = in_sds_info.sds_id = -1;
   if(open_sd(hdf_fname, &in_sds_info, 'R') == FAIL)
     fprintf(stderr, "Cannot open the HDF file %s\n", hdf_fname);
   else
   {
     strcpy(in_sds_info.name, max_obs_sds_name);
     if(get_sds_info((char *)NULL, &in_sds_info) != FAIL)
     {
	for (i=0; i<in_sds_info.rank; i++)
	{
          start[i] = 0;
          edge[i] = in_sds_info.dim_size[i];
	}
	if (in_sds_info.rank != 1) 
	{
	  if(in_sds_info.dim_size[in_sds_info.row_id] < nlines_a_copy) nlines_a_copy = in_sds_info.dim_size[in_sds_info.row_id];
          edge[in_sds_info.row_id] = 1 * nlines_a_copy;
	}

	nrows = (in_sds_info.rank == 1) ? 1 : in_sds_info.dim_size[in_sds_info.row_id];
	ncols = in_sds_info.dim_size[in_sds_info.row_id + 1];
            
	/* MSZ 1/25/2018 */
	GROWS = (int)ceil((double)nrows / (double)nlines_a_copy);
	last_rows = nrows % nlines_a_copy;

	/* MSZ, 1/25/2018, to speedup, copy lines by lines not line by line */
        for (irow=0; irow<GROWS; irow++)
        {
	  start[in_sds_info.row_id] = irow * nlines_a_copy;
	  /* MSZ, 1/25/2018, be cautious about the last group rows because it may be different from nlines_a_copy */
	  if(last_rows > 0 && irow == GROWS - 1)
	    edge[in_sds_info.row_id] = last_rows;

	  if ((sds_data = get_sel_sds_data(&in_sds_info, start, edge)) == NULL)
	    fprintf(stderr, "Error reading lines from %d to %d from SDS %s\n", irow * nlines_a_copy, irow * nlines_a_copy + nlines_a_copy, in_sds_info.name);
	  else
	  {
	    switch(in_sds_info.data_type)
	    {
	      case 20:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((int8 *)sds_data)[icol] > (int8)num_layers)
		    num_layers = (int)((int8 *)sds_data)[icol];
		break;
	      case 21:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((uint8 *)sds_data)[icol] > (uint8)num_layers)
		    num_layers = (int)((uint8 *)sds_data)[icol];
		break;
	      case 22:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((int16 *)sds_data)[icol] > (int16)num_layers)
		    num_layers = (int)((int16 *)sds_data)[icol];
		break;
	      case 23:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((uint16 *)sds_data)[icol] > (uint16)num_layers)
		    num_layers = (int)((uint16 *)sds_data)[icol];
		break;
	      case 24:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((int32 *)sds_data)[icol] > (int32)num_layers)
		    num_layers = (int)((int32 *)sds_data)[icol];
		break;
	      case 25:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((uint32 *)sds_data)[icol] > (uint32)num_layers)
		    num_layers = (int)((uint32 *)sds_data)[icol];
		break;
	      default:
		fprintf(stderr, "The data type should be integer\n");
	    }
	    if(sds_data) free(sds_data);
	  }
        }
	close_sds(&in_sds_info, in_sds_info.H5IN);
     }
     close_sd(&in_sds_info, in_sds_info.H5IN);
   }

   return num_layers;
}

void get_l2g_dataf_by_obs(char **arg_list, int32 *out_sd_id, int if_cnt, int *m_opt, int H5OUT) 
{
  int i, c=0;
  int attr_val=0;
  int irun, isds;
  int nrows, ncols;
  int run_cnt, sds_cnt;
  int layer_num[MAX_NUM_LAYERS];
  sds_t out_sds_info;
  int32 edge1[2], edgef[3];
  int32 start1[2] = {0, 0};
  int32 startf[3] = {0, 0, 0};
  char out_sds_name[MAX_SDS_NAME_LEN];
  sds_t sds1_info, sdsf_info;
  void *sds_data = NULL;

  sds1_info.sd_id = -1;
  if(open_sd(arg_list[0], &sds1_info, 'R') == FAIL)
  {
    fprintf(stderr, "Cannot open HDF file %s in get_l2g_dataf_by_obs\n", arg_list[0]);
    return;
  }
  sdsf_info.H5IN = sds1_info.H5IN;
  sdsf_info.sd_id = sds1_info.sd_id;

  sds_cnt = atoi(arg_list[3]);
  run_cnt = atoi(arg_list[4]);
  for (irun=0; irun<run_cnt; irun++)
    layer_num[irun] = atoi(arg_list[6+sds_cnt+irun]);
  for (isds=0; isds<sds_cnt; isds++)
  {
    fprintf(stderr, "Reading SDS %s\n", arg_list[6+isds]);
    sprintf(sds1_info.name, "%s_1", arg_list[6+isds]); 
    sprintf(sdsf_info.name, "%s_f", arg_list[6+isds]); 
    sds1_info.sds_id = sdsf_info.sds_id = -1;
    if (get_sds_info((char *)NULL, &sds1_info) == FAIL)
    {
      close_sds(&sds1_info, sds1_info.H5IN);
      continue;
    }
    if (open_sds((char *)NULL, &sdsf_info, 'R') == FAIL)
    {
      close_sds(&sds1_info, sds1_info.H5IN);
      close_sds(&sdsf_info, sdsf_info.H5IN);
      continue;
    }
    nrows = sds1_info.dim_size[0];
    ncols = sds1_info.dim_size[1];
    if ((sds_data = (void *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for sds1_data in get_l2g_dataf_by_obs\n");
    else
    {
      edge1[0] = nrows;
      edge1[1] = ncols;
      edgef[0] = 1;
      edgef[1] = nrows;
      edgef[2] = ncols;
      for (irun=0; irun<run_cnt; irun++)
      {
	fprintf(stderr, "	Reading Layer %d\n", layer_num[irun]);
	if (if_cnt == 1)
          sprintf(out_sds_name, "%s_layer%d", arg_list[6+isds], layer_num[irun]);
	else
          sprintf(out_sds_name, "%s_layer%d_%s", arg_list[6+isds], layer_num[irun], arg_list[1]);

        out_sds_info.sd_id = *out_sd_id;
        out_sds_info.H5OUT = H5OUT;
        strcpy(out_sds_info.name, out_sds_name);
        out_sds_info.data_type = sds1_info.data_type;
        out_sds_info.rank = sds1_info.rank;
        for(i = 0; i < out_sds_info.rank; i++)
          out_sds_info.dim_size[i] = sds1_info.dim_size[i];

	if (open_sds((char *)NULL, &out_sds_info, 'W') == FAIL)
          fprintf(stderr, "Cannot create the output SDS: %s", out_sds_name);
	else
	{
          write_attr_fval(H5OUT, out_sds_info.sds_id, sds1_info.data_type, c, attr_val);
	  if (layer_num[irun] == 0)
	  {
	    if (read_sel_sds_data(&sds1_info, start1, edge1, sds_data) == FAIL)
	      fprintf(stderr, "Error reading SDS %s in get_l2g_dataf_by_obs\n", sds1_info.name);
	  }
	  else
 	  {
	    startf[0] = layer_num[irun]-1;
	    if (read_sel_sds_data(&sdsf_info, startf, edgef, sds_data) == FAIL)
	      fprintf(stderr, "Error reading SDS %s in get_l2g_dataf_by_obs\n", sdsf_info.name);
	  }
	  /*if(set_sel_sds_data(&out_sds_info, start1, edge1, sds_data) == FAIL)*/
	  if(set_sel_l2g_sds_2D_data_chunked_gzip(&out_sds_info, start1, edge1, sds_data) == FAIL)
	    fprintf(stderr, "Error writing SDS data %s in get_l2g_dataf_by_obs\n", out_sds_name);

	  close_sds(&out_sds_info, out_sds_info.H5OUT);
        }
      } /* for (irun=0; . . . ) */
      if(sds_data) free(sds_data);
    }
    close_sds(&sds1_info, sds1_info.H5IN);
    close_sds(&sdsf_info, sdsf_info.H5IN);
  } /* for (isds=0; . . . . */
  if (*m_opt == 1)
  {
    copy_metadata(sds1_info.H5IN, sds1_info.sd_id, H5OUT, *out_sd_id);
    *m_opt = 0;   /* MSZ 2/21/2018, just write meta data once to the output file */
  }
  close_sd(&sds1_info, sds1_info.H5IN);
}

void get_l2g_datac_by_obs(char **arg_list, int32 *out_sd_id, int if_cnt, int *m_opt, int H5OUT, int max_nobs) 
{
  int c = 0;
  int status;
  int sh_col;
  int i, ir, ic;
  int irun, isds;
  int attr_val = 0;
  int obs_num_in_c;
  int sds_cnt, run_cnt;
  int ic_obs, skip_obs;
  int nrows, ncols, nobs;
  int layer_num[MAX_NUM_LAYERS];
  int32 edge1[2] = {0, 0};
  int32 edgec[2] = {0, 0};
  int32 start1[2] = {0, 0};
  int32 startc[2] = {0, 0};
  int32 max_nadd_obs_row;
  char out_sds_name[MAX_SDS_NAME_LEN];
  sds_t sds_info, sds1_info, sdsc_info, out_sds_info;;
  void *sds_data = NULL, *data_row = NULL, *out_data = NULL;
  int8 *sds_num_obs = NULL;
  int32 *sds_nadd_obs = NULL;

  sds_cnt = atoi(arg_list[3]);
  run_cnt = atoi(arg_list[4]);
  for (i=0; i<run_cnt; i++)
    layer_num[i] = atoi(arg_list[6+sds_cnt+i]);

  sds_info.sd_id = sds_info.sds_id = -1;
  if(open_sd(arg_list[0], &sds_info, 'R') == FAIL)
  {
    fprintf(stderr, "Cannot open HDF file %s in get_l2g_datac_by_obs\n", arg_list[0]);
    return;
  }

  /* read the SDS number of observations */
  status = 1;
  sds_info.sds_id = -1;
  strcpy(sds_info.name, sds_name_nobs);
  if (get_sds_info((char *)NULL, &sds_info) == -1)
    status = -1;
  else
  {
    nrows = edge1[0] = sds_info.dim_size[0];
    ncols = edge1[1] = sds_info.dim_size[1];
    if ((sds_num_obs = (int8 *)calloc(nrows*ncols, sds_info.data_size)) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for sds_num_obs get_l2g_datac_by_obs\n");
      status = -1;
    }
    else if (read_sel_sds_data(&sds_info, start1, edge1, sds_num_obs) == FAIL)
    {
      fprintf(stderr, "Cannot read SDS number of observations in get_l2g_datac_by_obs");
      free(sds_num_obs);
      status = -1;
    }
    close_sds(&sds_info, sds_info.H5IN);
  }

  /* read the SDS additional number of observations */
  max_nadd_obs_row = 0;
  sds_info.use_L2G_max_nobs = 0;
  sds_info.rank = 1;
  sds_info.dim_size[0] = nrows;
  strcpy(sds_info.name, sds_name_nadd_obs);
  if ((sds_nadd_obs = (int32 *)calloc(nrows, sizeof(int32))) == NULL)
  {
      fprintf(stderr, "Cannot allocate memory for sds_nadd_obs in get_l2g_datac_by_obs");
      status = -1;
  }
  else 
  {
    if(max_nobs > 1)  /* MSZ 3/6/2018 */
    {
      if (get_sds_data(&sds_info, sds_nadd_obs) != -1)
      {
        for (ir = 0; ir < nrows; ir++)
          if (sds_nadd_obs[ir] > max_nadd_obs_row)
            max_nadd_obs_row = sds_nadd_obs[ir];
        sds_info.sds_id = -1;
      }
      else
	status = -1;
    }
    else
	status = 1;
  }
  
  if (status == -1)
  {
    if (sds_num_obs != NULL) free(sds_num_obs);
    if (sds_nadd_obs != NULL) free(sds_nadd_obs);
    close_sd(&sds_info, sds_info.H5IN);

    out_sds_info.sd_id = *out_sd_id;
    out_sds_info.H5OUT = H5OUT;
    close_sd(&out_sds_info, out_sds_info.H5OUT);
    *out_sd_id = -1;
    return;
  }

  for (isds=0; isds<sds_cnt; isds++)
  {
    fprintf(stderr, "Uncompacting SDS %s\n", arg_list[6+isds]);

    /* read the information about the first layer */
    sds1_info.sds_id = -1;
    sds1_info.use_L2G_max_nobs = 0;
    sds1_info.sd_id = sds_info.sd_id;
    sds1_info.H5IN = sds_info.H5IN;
    sprintf(sds1_info.name, "%s_1", arg_list[6+isds]);
    if (get_sds_info((char *)NULL, &sds1_info) == -1)
    {
      close_sds(&sds1_info, sds1_info.H5IN);
      continue;
    }

    if ((out_data = (void *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
     fprintf(stderr, "Cannot allocate memory for out_data in read_l2g_by_obs\n");
    if ((data_row = (void *)calloc(max_nadd_obs_row, sds1_info.data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for data_row in get_l2g_datac_by_obs\n");
    if ((out_data == NULL) || (data_row == NULL))
    {
      if (out_data != NULL) free(out_data);
      if (data_row == NULL) free(data_row);
      close_sds(&sds1_info, sds1_info.H5IN);
      break;
    }

    if(max_nobs > 1)  /* MSZ 3/6/2018 */
    {
      sdsc_info.sds_id = -1;
      sdsc_info.use_L2G_max_nobs = 0;
      sdsc_info.sd_id = sds_info.sd_id;
      sdsc_info.H5IN = sds_info.H5IN;
      sprintf(sdsc_info.name, "%s_c", arg_list[6+isds]);
      status = open_sds((char *)NULL, &sdsc_info, 'R');
    }

    for (irun=0; irun<run_cnt; irun++)
    {
      fprintf(stderr, "	Reading layer %d\n", layer_num[irun]);
      if (if_cnt == 1)
        sprintf(out_sds_name, "%s_layer%d", arg_list[6+isds], layer_num[irun]);
      else
        sprintf(out_sds_name, "%s_layer%d_%s", arg_list[6+isds], layer_num[irun], arg_list[1]);

      out_sds_info.sd_id = *out_sd_id;
      out_sds_info.H5OUT = H5OUT;
      strcpy(out_sds_info.name, out_sds_name);
      out_sds_info.data_type = sds1_info.data_type;
      out_sds_info.rank = sds1_info.rank;
      for(i = 0; i < out_sds_info.rank; i++)
        out_sds_info.dim_size[i] = sds1_info.dim_size[i];

      if (open_sds((char *)NULL, &out_sds_info, 'W') == FAIL)
      {
        fprintf(stderr, "Cannot create the output SDS: %s", out_sds_name);
	continue;
      }
      write_attr_fval(H5OUT, out_sds_info.sds_id, sds1_info.data_type, c, attr_val);
      if (layer_num[irun] == 0) 
      {
        if ((sds_data = (void *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
          fprintf(stderr, "Cannot allocate memory for sds_data in get_l2g_datac_by_obs\n");
        else
        {
          if (get_sds_data(&sds1_info, sds_data) != -1)
          {
	    /*if(set_sel_sds_data(&out_sds_info, start1, edge1, sds_data) == FAIL)*/
	    if(set_sel_l2g_sds_2D_data_chunked_gzip(&out_sds_info, start1, edge1, sds_data) == FAIL)
              fprintf(stderr, "Cannot write output SDS %s_layer1\n", arg_list[6+isds]);
          }
          if(sds_data) free(sds_data);
        }
      }
      else if (status != -1 && max_nobs > 1)
      {
        startc[0] = edgec[0] = 0;
        for (ir=0; ir<nrows; ir++)
        {
	  sh_col = ir*ncols;
          startc[0] += edgec[0];
          edgec[0] = sds_nadd_obs[ir]; 
	  if (edgec[0] == 0)
          {
	    for (ic=0; ic<ncols; ic++)
	    {
              switch(sds1_info.data_type)
              {
	        /* MSZ 3/5/2018, update it using fill value obtained from the SDS */
                case 20: ((int8 *)out_data)[sh_col+ic] = FILL_VALUE_INT8;
                         break;
                case 21: ((uint8 *)out_data)[sh_col+ic] = FILL_VALUE_UINT8;
                         break;
                case 22: ((int16 *)out_data)[sh_col+ic] = FILL_VALUE_INT16;
                         break;
                case 23: ((uint16 *)out_data)[sh_col+ic] = FILL_VALUE_UINT16;
                         break;
                case 24: ((int32 *)out_data)[sh_col+ic] = FILL_VALUE_INT32;
                         break;
                case 25: ((uint32 *)out_data)[sh_col+ic] = FILL_VALUE_UINT32;
                         break;
		/*
                case 20: ((int8 *)out_data)[sh_col+ic] = (int8)sds1_info.fill_val;
                         break;
                case 21: ((uint8 *)out_data)[sh_col+ic] = (uint8)sds1_info.fill_val;
                         break;
                case 22: ((int16 *)out_data)[sh_col+ic] = (int16)sds1_info.fill_val;
                         break;
                case 23: ((uint16 *)out_data)[sh_col+ic] = (uint16)sds1_info.fill_val;
                         break;
                case 24: ((int32 *)out_data)[sh_col+ic] = (int32)sds1_info.fill_val;
                         break;
                case 25: ((uint32 *)out_data)[sh_col+ic] = (uint32)sds1_info.fill_val;
                         break;
	        */

              }
	    }
          }
	  else
	  {
	    if (read_sel_sds_data(&sdsc_info, startc, edgec, data_row) == FAIL)
	    {
	      fprintf(stderr, "Cannot read a row of SDS %s\n", sdsc_info.name);
	      break;
	    }
            obs_num_in_c = layer_num[irun] - 1;
            for (ic=0, skip_obs=0; ic<ncols; ic++)
            {
	      nobs = sds_num_obs[ir*ncols + ic];
              if (nobs > layer_num[irun])
              {
                ic_obs = skip_obs + obs_num_in_c;
                switch(sds1_info.data_type)
                {
                  case 20: ((int8 *)out_data)[sh_col+ic] = ((int8 *)data_row)[ic_obs];
                           break;
                  case 21: ((uint8 *)out_data)[sh_col+ic] = ((uint8 *)data_row)[ic_obs];
                           break;
                  case 22: ((int16 *)out_data)[sh_col+ic] = ((int16 *)data_row)[ic_obs];
  	   	           break;
                  case 23: ((uint16 *)out_data)[sh_col+ic] = ((uint16 *)data_row)[ic_obs];
  	   	           break;
                  case 24: ((int32 *)out_data)[sh_col+ic] = ((int32 *)data_row)[ic_obs];
  		           break;
                  case 25: ((uint32 *)out_data)[sh_col+ic] = ((uint32 *)data_row)[ic_obs];
  		           break;
                }
              }
              else 
              {
                switch(sds1_info.data_type)
                {
		  /* MSZ 3/5/2018, update it using fill value obtained from the SDS */ 
                  case 20: ((int8 *)out_data)[sh_col+ic] = FILL_VALUE_INT8; 
  	    	           break;
                  case 21: ((uint8 *)out_data)[sh_col+ic] = FILL_VALUE_UINT8; 
  	    	           break;
                  case 22: ((int16 *)out_data)[sh_col+ic] = FILL_VALUE_INT16;
  		           break;
                  case 23: ((uint16 *)out_data)[sh_col+ic] = FILL_VALUE_UINT16;
  		           break;
                  case 24: ((int32 *)out_data)[sh_col+ic] = FILL_VALUE_INT32;
  		           break;
                  case 25: ((uint32 *)out_data)[sh_col+ic] = FILL_VALUE_UINT32;
  		           break;
		  /*
                  case 20: ((int8 *)out_data)[sh_col+ic] = (int8)sds1_info.fill_val;
                         break;
                  case 21: ((uint8 *)out_data)[sh_col+ic] = (uint8)sds1_info.fill_val;
                         break;
                  case 22: ((int16 *)out_data)[sh_col+ic] = (int16)sds1_info.fill_val;
                         break;
                  case 23: ((uint16 *)out_data)[sh_col+ic] = (uint16)sds1_info.fill_val;
                         break;
                  case 24: ((int32 *)out_data)[sh_col+ic] = (int32)sds1_info.fill_val;
                         break;
                  case 25: ((uint32 *)out_data)[sh_col+ic] = (uint32)sds1_info.fill_val;
                         break;
		  */
                }
              }
	      if (nobs > 1)
                skip_obs += (nobs - 1);
            } /* for ic . . .  */
	  }
        } /* for ir . . . */
	/*if(set_sel_sds_data(&out_sds_info, start1, edge1, out_data) == FAIL)*/
	if(set_sel_l2g_sds_2D_data_chunked_gzip(&out_sds_info, start1, edge1, out_data) == FAIL)
	  fprintf(stderr, "Cannot write sds data to output file in get_l2g_datac_by_obs\n");
      }
      else
        fprintf(stderr, "Cannot uncompact layer number %d\n", irun);
      close_sds(&out_sds_info, out_sds_info.H5OUT);
    } /* for irun . . . */ 
    free(out_data);
    free(data_row);
  } /* for isds */
  free(sds_nadd_obs);
  free(sds_num_obs);
  if (*m_opt == 1)
  {
    copy_metadata(sds_info.H5IN, sds_info.sd_id, H5OUT, *out_sd_id);
    *m_opt = 0;   /* MSZ 2/21/2018, just write meta data once to the output file */
  }
  close_sds(&sds_info, sds_info.H5IN);
}

void get_l2g_dataf_by_gran(char **arg_list, char *ptr_fname, int32 *out_sd_id, int if_cnt, int *m_opt, int H5OUT) 
{
  int c = 0;
  int ir, ic;
  int attr_val = 0;
  int nrows, ncols;
  int status, offset;
  int isds, irun, iobs;
  int cnt, sds_cnt, run_cnt;
  int gran_num, orb_num, gran_pnt;
  int skip_obs, max_nobs, nadd_obs;
  int8 *sds_num_obs, *pnt1_row, *pntf_row;
  int32 edge[2] = {0, 0};
  int32 start[2] = {0, 0};
  int32 edge1[2] = {0, 0};
  int32 start1[2] = {0, 0};
  int32 edgef[3] = {0, 0, 0};
  int32 startf[3] = {0, 0, 0};
  void *sds1_row, *sdsf_row, *out_data;
  sds_t pnt1_info, pntf_info, sds1_info, sdsf_info, out_sds_info;
  char **modis_fpath, out_sds_name[MAX_SDS_NAME_LEN];
  char pnt_fname[MAX_PATH_LENGTH], gran_id[MAX_NAME_LENGTH];
  int st;
  char **iptr_fnames;
  int i, iptr_fcnt;    
  
  if ((modis_fpath = (char **)Calloc2D(MAX_NUM_PATH, MAX_PATH_LENGTH, sizeof(char))) ==
        NULL)
    fprintf(stderr, "Cannot allocate memory for modis_fpath in get_l2g_dataf_by_gran\n");
  else 
    get_qa_tool_env(MODIS_ENV, modis_fpath);
 
  /* MSZ 3/8/2018 use mete data existence to determine MODIS or NPP */ 
  /*st = strpos(arg_list[0], "_L2G", 0);*/
  if(file_attr_name_existence(arg_list[0], "ArchiveMetadata") ||
     file_attr_name_existence(arg_list[0], "CoreMetadata"))
     st = -1;  /* MODIS */
  else
     st = 1;   /* VNP or NPP */

  if (st == -1)  /* MODIS */
  {
    if (get_pnt_fname(arg_list[0], modis_fpath, ptr_fname, pnt_fname) == -1)
      pnt1_info.sd_id = -1;
    else if (open_sd(pnt_fname, &pnt1_info, 'R') == FAIL)
      fprintf(stderr, "Cannot open pointer file %s in get_l2g_dataf_by_gran\n", pnt_fname);  
  }
  else          /* VNP */
  {
    if ((iptr_fnames = (char **)Calloc2D(MAX_NUM_GRAN, MAX_META_VAL_LEN, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for iptr_fnames in parse_cmd_read_l2g");
      return; 
    }  
    if ((iptr_fcnt = get_npp_input_files(arg_list[0], ptr_fname, iptr_fnames)) == 0)
    {
      status = -1;      
    }  
    else 
    {
      strcpy(pnt_fname, iptr_fnames[0]);  
      if (open_sd(pnt_fname, &pnt1_info, 'R') == FAIL)
        fprintf(stderr, "Cannot open pointer file %s in get_l2g_dataf_by_gran\n", pnt_fname);     	
    }
    Free2D((void **)iptr_fnames);	   
  }  
 
  if (open_sd(arg_list[0], &sds1_info, 'R') == FAIL) 
    fprintf(stderr, "Cannot open L2G file %s in get_l2g_dataf_by_gran\n", arg_list[0]);

  status = 1;
  Free2D((void **)modis_fpath);
  if (sds1_info.sd_id != FAIL)
  {
    sds1_info.sds_id = -1;
    strcpy(sds1_info.name, sds_name_nobs);
    if (get_sds_info((char *)NULL, &sds1_info) != -1)
    {
      nrows = sds1_info.dim_size[0];
      ncols = sds1_info.dim_size[1];
      if ((sds_num_obs = (void *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
      {
	fprintf(stderr, "Cannot allocate memory for sds_num_obs in get_l2g_dataf_by_gran\n"); 
	status = -1;
      }
      else
        get_sds_data(&sds1_info, sds_num_obs);
    }
    else 
    {
      status = -1;
      if (sds1_info.sds_id != -1) close_sds(&sds1_info, sds1_info.H5IN);
    }
  }
  if (pnt1_info.sd_id != FAIL)
  {
    pnt1_info.sds_id = pntf_info.sds_id = -1;
    strcpy(pnt1_info.name, sds_name_gran_pntr_1);
    if (get_sds_info((char *)NULL, &pnt1_info) != -1)
    {
      if ((pnt1_row = (int8 *)calloc(ncols, pnt1_info.data_size)) == NULL)
      {
	fprintf(stderr, "Cannot allocate memory for pnt1_row in get_l2g_dataf_by_gran\n");
	status = -1;
      }
    }
    else 
    {
      status = -1;
      close_sds(&pnt1_info, pnt1_info.H5IN);
    }
    pntf_info.sd_id = pnt1_info.sd_id;
    pntf_info.H5IN = pnt1_info.H5IN;
    strcpy(pntf_info.name, sds_name_gran_pntr_f);
    if (get_sds_info((char *)NULL, &pntf_info) != -1)
    {
      max_nobs = pntf_info.dim_size[2];
      if ((pntf_row = (int8 *)calloc(ncols*max_nobs, pntf_info.data_size)) == NULL)
      {
	fprintf(stderr, "Cannot allocate memory for pntf_row in get_l2g_dataf_by_gran\n");
	status = -1;
      }
    }
    else 
    {
      status = -1;
      close_sds(&pntf_info, pntf_info.H5IN);
    }
  }

  if (status != -1)
  {
    edge[0] = ncols;
    edge[1] = nrows;
    sds_cnt = atoi(arg_list[3]);
    run_cnt = atoi(arg_list[4]);
    sdsf_info.sd_id = sds1_info.sd_id;
    sdsf_info.H5IN = sds1_info.H5IN;
    for (isds=0; isds<sds_cnt; isds++)
    {
      fprintf(stderr, "	Reading SDS %s\n", arg_list[6+isds]);
      sdsf_info.sds_id = sds1_info.sds_id = -1;
      sprintf(sds1_info.name, "%s_1", arg_list[6+isds]);
      sprintf(sdsf_info.name, "%s_f", arg_list[6+isds]);
      if ((get_sds_info((char *)NULL, &sds1_info) == -1) || 
		(get_sds_info((char *)NULL, &sdsf_info) == -1))
        status = -1;
      if ((sds1_row = (void *)calloc(ncols, sds1_info.data_size)) == NULL)
        fprintf(stderr, "Cannot allocate memory for sds1_row in get_l2g_dataf_by_gran\n");
      if ((sdsf_row = (void *)calloc(ncols*max_nobs, sdsf_info.data_size)) == NULL)
        fprintf(stderr, "Cannot allocate memory for sdsf_row in get_l2g_dataf_by_gran\n");
      if ((out_data = (void *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
        fprintf(stderr, "Cannot allocate memory for out_data in get_l2g_dataf_by_gran\n");
      if ((sds1_row == NULL) || (sdsf_row == NULL) || (out_data == NULL))
      {
	close_sds(&sds1_info, sds1_info.H5IN);
	close_sds(&sdsf_info, sdsf_info.H5IN);
        if (sds1_row != NULL) free(sds1_row);
        if (sdsf_row != NULL) free(sdsf_row);
        if (out_data != NULL) free(out_data);
        continue;
      }

      for (irun=0; irun<run_cnt; irun++)
      {
        if ((cnt = sscanf(arg_list[6+sds_cnt+irun], "%d %d %d", &gran_pnt, &gran_num,
                &orb_num)) == 3)
	{
	  if (if_cnt == 1)
            sprintf(out_sds_name, "%s_gran%d_orb%d", arg_list[6+isds], gran_num, orb_num);
	  else
            sprintf(out_sds_name, "%s_gran%d_orb%d_%s", arg_list[6+isds], gran_num, orb_num,
			arg_list[1]);
	}
        else
        {
          cnt = sscanf(arg_list[6+sds_cnt+irun], "%d %s", &gran_pnt, gran_id);
          if (cnt == 2)
	  {
	    if (if_cnt == 1)
              sprintf(out_sds_name, "%s_gid%s", arg_list[6+isds], gran_id);
	    else
              sprintf(out_sds_name, "%s_gid%s_%s", arg_list[6+isds], gran_id, arg_list[1]);
	  }
	  else
	  {
	    cnt = sscanf(arg_list[6+sds_cnt+irun], "%d", &gran_pnt);
	    if (if_cnt == 1)
              sprintf(out_sds_name, "%s_gpnt%d", arg_list[6+isds], gran_pnt);
	    else
              sprintf(out_sds_name, "%s_gpnt%d_%s", arg_list[6+isds], gran_pnt, arg_list[1]);
	  }
        }

        out_sds_info.sd_id = *out_sd_id;
        out_sds_info.H5OUT = H5OUT;
        strcpy(out_sds_info.name, out_sds_name);
        out_sds_info.data_type = sds1_info.data_type;
        out_sds_info.rank = sds1_info.rank;
        for(i = 0; i < out_sds_info.rank; i++)
          out_sds_info.dim_size[i] = sds1_info.dim_size[i];

	if (open_sds((char *)NULL, &out_sds_info, 'W') == FAIL)	
        {
          fprintf(stderr, "Cannot create the output SDS: %s", out_sds_name);
	  break;
        }
        write_attr_fval(H5OUT, out_sds_info.sds_id, sds1_info.data_type, c, attr_val);

        if (cnt == 1)
          fprintf(stderr, "		Reading Granule (index=%d)\n", gran_pnt);
        else if (cnt == 2)
          fprintf(stderr, "		Reading Granule (index=%d, gran_id=%s)\n", gran_pnt, gran_id);
        else if (cnt == 3)
          fprintf(stderr, "		Reading Granule (index=%d, gran_num=%d, orb_num=%d)\n", gran_pnt, gran_num, orb_num);

        /* read and process one row at a time for this granules */
        edgef[0] = max_nobs;
        edge1[0] = edgef[1] = 1;
        edge1[1] = edgef[2] = ncols;
        for (ir=0; ir<nrows; ir++)
        {
	  if (read_sel_sds_data(&pnt1_info, start1, edge1, pnt1_row) == FAIL)
	  {
	    fprintf(stderr, "Error reading data from pointer file in get_l2g_dataf_by_gran\n");
	    status = -1;
	  }
	  if (read_sel_sds_data(&pntf_info, startf, edgef, pntf_row) == FAIL)
	  {
	    fprintf(stderr, "Error reading data from pointer file in get_l2g_dataf_by_gran\n");
	    status = -1;
	  }
	  if (read_sel_sds_data(&sds1_info, start1, edge1, sds1_row) == FAIL)
	  {
	    fprintf(stderr, "Error reading data from L2G file in get_l2g_dataf_by_gran\n");
	    status = -1;
	  }
	  if (read_sel_sds_data(&sdsf_info, startf, edgef, sdsf_row) == FAIL)
	  {
	    fprintf(stderr, "Error reading data from L2G file in get_l2g_dataf_by_gran\n");
	    status = -1;
	  }
          if (status == -1) break;
	  for (ic=0, skip_obs=0; ic<ncols; ic++)
	  {
	    offset = ir*ncols+ic;
	    skip_obs = skip_obs + max_nobs;
            nadd_obs = sds_num_obs[offset] - 1;
            if (pnt1_row[ic] == gran_pnt)
            {
              switch(sds1_info.data_type)
              {
                case 20: ((int8 *)out_data)[offset] = ((int8 *)sds1_row)[ic];
      	                 break;
                case 21: ((uint8 *)out_data)[offset] = ((uint8 *)sds1_row)[ic];
      	                 break;
                case 22: ((int16 *)out_data)[offset] = ((int16 *)sds1_row)[ic];
  	  	         break;
                case 23: ((uint16 *)out_data)[offset] = ((uint16 *)sds1_row)[ic];
  	  	         break;
                case 24: ((int32 *)out_data)[offset] = ((int32 *)sds1_row)[ic];
  		         break;
                case 25: ((uint32 *)out_data)[offset] = ((uint32 *)sds1_row)[ic];
  		         break;
              }
            }
            else 
	    {
	      if (nadd_obs > 0)
              {
                for (iobs = 0; iobs < nadd_obs; iobs++)
                {
                  if (pntf_row[skip_obs+iobs] == gran_pnt)
                  {
                    switch(sds1_info.data_type)
                    {
                      case 20: ((int8 *)out_data)[offset] = ((int8 *)sdsf_row)[skip_obs+iobs];
  	    	               break;
                      case 21: ((uint8 *)out_data)[offset] = ((uint8 *)sdsf_row)[skip_obs+iobs];
  	    	               break;
                      case 22: ((int16 *)out_data)[offset] = ((int16 *)sdsf_row)[skip_obs+iobs];
  		               break;
                      case 23: ((uint16 *)out_data)[offset] = ((uint16 *)sdsf_row)[skip_obs+iobs];
  		               break;
                      case 24: ((int32 *)out_data)[offset] = ((int32 *)sdsf_row)[skip_obs+iobs];
  		               break;
                      case 25: ((uint32 *)out_data)[offset] = ((uint32 *)sdsf_row)[skip_obs+iobs];
  		               break;
                    }
                    break;
	          } /* if .. */ 
                } /* for iobs ... */
	        if (iobs == nadd_obs)
                  switch(sds1_info.data_type)
                  {
		    /* MSZ 3/5/2018, update it using fill value obtained from the SDS */
                    case 20: ((int8 *)out_data)[offset] = FILL_VALUE_INT8;
                             break;
                    case 21: ((uint8 *)out_data)[offset] = FILL_VALUE_UINT8;
                             break;
                    case 22: ((int16 *)out_data)[offset] = FILL_VALUE_INT16;
                             break;
                    case 23: ((uint16 *)out_data)[offset] = FILL_VALUE_UINT16;
                             break;
                    case 24: ((int32 *)out_data)[offset] = FILL_VALUE_INT32;
                             break;
                    case 25: ((uint32 *)out_data)[offset] = FILL_VALUE_UINT32;
                             break;
		    /*
                    case 20: ((int8 *)out_data)[offset] = (int8)sds1_info.fill_val;
                         break;
                    case 21: ((uint8 *)out_data)[offset] = (uint8)sds1_info.fill_val;
                         break;
                    case 22: ((int16 *)out_data)[offset] = (int16)sds1_info.fill_val;
                         break;
                    case 23: ((uint16 *)out_data)[offset] = (uint16)sds1_info.fill_val;
                         break;
                    case 24: ((int32 *)out_data)[offset] = (int32)sds1_info.fill_val;
                         break;
                    case 25: ((uint32 *)out_data)[offset] = (uint32)sds1_info.fill_val;
                         break;
		    */
                  }
              } /* if (nadd_obs . . . */ 
	      else
                switch(sds1_info.data_type)
                {
		  /* MSZ 3/5/2018, update it using fill value obtained from the SDS */
                  case 20: ((int8 *)out_data)[offset] = FILL_VALUE_INT8;
                           break;
                  case 21: ((uint8 *)out_data)[offset] = FILL_VALUE_UINT8;
                           break;
                  case 22: ((int16 *)out_data)[offset] = FILL_VALUE_INT16;
                           break;
                  case 23: ((uint16 *)out_data)[offset] = FILL_VALUE_UINT16;
                           break;
                  case 24: ((int32 *)out_data)[offset] = FILL_VALUE_INT32;
                           break;
                  case 25: ((uint32 *)out_data)[offset] = FILL_VALUE_UINT32;
                           break;
		  /*
                  case 20: ((int8 *)out_data)[offset] = (int8)sds1_info.fill_val;
                         break;
                  case 21: ((uint8 *)out_data)[offset] = (uint8)sds1_info.fill_val;
                         break;
                  case 22: ((int16 *)out_data)[offset] = (int16)sds1_info.fill_val;
                         break;
                  case 23: ((uint16 *)out_data)[offset] = (uint16)sds1_info.fill_val;
                         break;
                  case 24: ((int32 *)out_data)[offset] = (int32)sds1_info.fill_val;
                         break;
                  case 25: ((uint32 *)out_data)[offset] = (uint32)sds1_info.fill_val;
                         break;
		  */
                }
	    }
	  } /* for (ic=0; . . . */
        } /* for (ir=0; . . . */
	/*if(set_sel_sds_data(&out_sds_info, start, edge, out_data) == FAIL)*/
	if(set_sel_l2g_sds_2D_data_chunked_gzip(&out_sds_info, start, edge, out_data) == FAIL)
	    fprintf(stderr, "Fail to write data to output file \n");
	close_sds(&out_sds_info, out_sds_info.H5OUT);
      } /* for (irun=0; . . .  */
      close_sds(&sds1_info, sds1_info.H5IN);
      close_sds(&sdsf_info, sdsf_info.H5IN);
      free(sds1_row);
      free(sdsf_row);
      free(out_data);
    } /* for (isds=0; . . .  */
  } /* if (staus != -1) . . . */
  if (*m_opt == 1)
  {
    copy_metadata(sds1_info.H5IN, sds1_info.sd_id, H5OUT, *out_sd_id);
    *m_opt = 0;   /* MSZ 2/21/2018, just write meta data once to the output file */
  }
  close_sd(&sds1_info, sds1_info.H5IN);
  close_sd(&pnt1_info, pnt1_info.H5IN);
}

void get_l2g_datac_by_gran(char **arg_list, char *ptr_fname, int32 *out_sd_id, int if_cnt, int *m_opt, int H5OUT, int max_nobs) 
{
  /* MSZ 3/13/2018, pnt_fname is for VIIRS */
  int c = 1;
  int ir, ic;
  int attr_val = 0;
  int nrows, ncols;
  int status, offset;
  int isds, irun, iobs;
  int skip_obs, nadd_obs;
  int cnt, sds_cnt, run_cnt;
  int gran_pnt, gran_num, orb_num;
  int8 *sds_num_obs = NULL, *pnt1_row = NULL, *pntc_row = NULL;
  int32 *sds_nadd_obs = NULL;
  int32 edge[2] = {0, 0};
  int32 start[2] = {0, 0};
  int32 edgec[2] = {0, 0};
  int32 edge1[2] = {0, 0};
  int32 startc[2] = {0, 0};
  int32 start1[2] = {0, 0};
  int32 max_nadd_obs_row;
  void *sds1_row = NULL, *sdsc_row = NULL, *out_data = NULL;
  char **modis_fpath = NULL, gran_id[MAX_GID_LEN] = {'\0'};
  char out_sds_name[MAX_SDS_NAME_LEN] = {'\0'}, pnt_fname[MAX_PATH_LENGTH] = {'\0'};
  sds_t pnt1_info, pntc_info, sds1_info, sdsc_info, out_sds_info;
  int st, i;
  char **iptr_fnames = NULL;
  int iptr_fcnt; 

  /* get the modis product file path */
  if ((modis_fpath = (char **)Calloc2D(MAX_NUM_PATH, MAX_PATH_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for modis_fpath in get_l2g_profile\n");
  else
    get_qa_tool_env(MODIS_ENV, modis_fpath);

  /* MSZ 3/3/2018 use mete data existence to determine MODIS or NPP */
  /* st = strpos(arg_list[0], "_L2G", 0); */

  if(file_attr_name_existence(arg_list[0], "ArchiveMetadata") ||
     file_attr_name_existence(arg_list[0], "CoreMetadata"))
    st = -1;  /* MODIS */
  else
    st = 1;   /* VNP or NPP */

  if (st == -1) 
  {
    /* MSZ note.  pnt_fname is the derived from meta data named INPUTPOINTER, the first one input file. */
    /* MSZ 3/7/2018 modify the below to make pnt_fname work */
    if (get_pnt_fname(arg_list[0], modis_fpath, ptr_fname, pnt_fname) == -1)
    {
      pnt1_info.sd_id = pnt1_info.sds_id = -1;
    }
    else
    {
      pnt1_info.sd_id = pnt1_info.sds_id = -1;
	/*
      if(find_file("",pnt_fname) == -1)
      {
	fprintf(stderr, "The point file %s doesn't exist, get it to the current fold or specify it using -ptr=, exit...\n", pnt_fname);
	exit(0);
      }
	*/
      if(open_sd(pnt_fname, &pnt1_info, 'R') == FAIL)
        fprintf(stderr, "Cannot open pointer file %s in get_l2g_dataf_by_gran\n", pnt_fname);  
      else
	fprintf(stderr, "pointer file %s is successfully opened!\n", pnt_fname);
    }
  }
  else  
  {
    if ((iptr_fnames = (char **)Calloc2D(MAX_NUM_GRAN, MAX_META_VAL_LEN, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for iptr_fnames in parse_cmd_read_l2g");
      return; 
    }
    if ((iptr_fcnt = get_npp_input_files(arg_list[0], ptr_fname, iptr_fnames)) == 0)
    {  
      status = -1;
    }  
    else 
    {
      /* MSZ 3/8/2018, must initialized sd_id and sds_id to be -1 before open_sd() and open_sds() */
      pnt1_info.sd_id = pnt1_info.sds_id = -1;
      /* MSZ 3/13/2018, in case that pnt_npp_file_name is specified by using -ptr= not from the derived for NPP */
      /* strcpy(pnt_fname, iptr_fnames[0]); */
      strcpy(pnt_fname, arg_list[5]);
	/*
      if(find_file("",pnt_fname) == -1)
	fprintf(stderr, "The point file %s doesn't exist, get it to the current fold or specify it using -ptr= exit...\n", pnt_fname);
	*/

      if(open_sd(pnt_fname, &pnt1_info, 'R') == FAIL)
        fprintf(stderr, "Cannot open pointer file %s in get_l2g_dataf_by_gran\n", pnt_fname);     	
      else
	fprintf(stderr, "pointer file %s is successfully opened!\n", pnt_fname);
    }
    Free2D((void **)iptr_fnames);         	   
  } 

  sds1_info.sd_id = sds1_info.sds_id = -1; 
  if(open_sd(arg_list[0], &sds1_info, 'R') == FAIL) 
    fprintf(stderr, "Cannot open HDF file %s in get_l2g_datac_by_gran\n", arg_list[0]);
  else
    fprintf(stderr, "input file %s is successfully opened!\n", arg_list[0]);

  if(sds1_info.sd_id != FAIL && pnt1_info.sd_id != FAIL)
    /*fprintf(stderr, "both sds1 and pnt1 files are successfully opened!\n")*/;
  else
  {
     fprintf(stderr, "Cannot open either sds1 or pnt1 file. This is mainly caused by missing pointer file. Please copy the corresponding pointer file to this working fold or use -ptr= to specify the pointer file.\n");
     close_sds(&sds1_info, sds1_info.H5IN);
     close_sds(&pnt1_info, pnt1_info.H5IN);
     Free2D((void **)modis_fpath);
     return;
  }

  /* read the SDS number of observations */
  status = 1;
  sds1_info.sds_id = -1;
  strcpy(sds1_info.name, sds_name_nobs);
  if (get_sds_info((char *)NULL, &sds1_info) == -1)
    status = -1;
  else
  {
    nrows = edge1[0] = sds1_info.dim_size[0];
    ncols = edge1[1] = sds1_info.dim_size[1];
    if ((sds_num_obs = (int8 *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for sds_num_obs in get_l2g_datac_by_gran\n");
      status = -1;
    }
    else
    {
      if (read_sel_sds_data(&sds1_info, start1, edge1, sds_num_obs) == FAIL)
      {
        fprintf(stderr, "Cannot read SDS number of observations in get_l2g_datac_by_gran\n");
        free(sds_num_obs);
        status = -1;
      }
      else
	/*fprintf(stderr, "sds1 is successfully read!\n")*/;
    }
    close_sds(&sds1_info, sds1_info.H5IN);
  }

  /* read the SDS additional number of observations */
  if (status != -1)
  {
    sds1_info.rank = 1;
    sds1_info.use_L2G_max_nobs = 0;
    sds1_info.dim_size[0] = nrows;
    strcpy(sds1_info.name, sds_name_nadd_obs);
    if ((sds_nadd_obs = (int32 *)calloc(nrows, sizeof(int32))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for sds_nadd_obs in get_l2g_datac_by_gran\n");      status = -1;
    }
    else
    {
      if (get_sds_data(&sds1_info, sds_nadd_obs) != -1)
      {
        max_nadd_obs_row = sds_nadd_obs[0];
        for (ir = 1; ir < nrows; ir++)
          if (sds_nadd_obs[ir] > max_nadd_obs_row)
            max_nadd_obs_row = sds_nadd_obs[ir];

	/*fprintf(stderr, "sds_nadd_obs is successfully read!\n")*/;
      }
      else
      {
        fprintf(stderr, "Cannot read the SDS sds_nadd_obs in get_l2g_datac_by_gran \n");
        status = -1;
      }
    }
  }

  /* open the sds "granule_pnt_1" and "granule_pnt_c" */
  if (status != -1)
  {
    pntc_info.sd_id = pnt1_info.sd_id; 
    pntc_info.use_L2G_max_nobs = 0;
    pntc_info.H5IN = pnt1_info.H5IN;
    pntc_info.sds_id = pnt1_info.sds_id = -1; 
    strcpy(pnt1_info.name, sds_name_gran_pntr_1);
    strcpy(pntc_info.name, sds_name_gran_pntr_c);
    /*if ((get_sds_info(pnt_fname, &pnt1_info) == -1) ||
		 (get_sds_info(pnt_fname, &pntc_info) == -1))*/
    if ((get_sds_info((char*)NULL, &pnt1_info) == FAIL) ||
		(get_sds_info((char*)NULL, &pntc_info) == FAIL))
    {
      status = -1;
      fprintf(stderr, "Cannot get either pnt1 or pntc sds_info.\n");  /* MSZ 3/8/2018, add the key error information output */
    }
    else
      fprintf(stderr, "pnt1 and pntc sds_info is successfully obtained!\n");

    if ((pnt1_row = (int8 *)calloc(ncols, sizeof(int8))) == NULL)
    {
      status = -1;
      fprintf(stderr, "Cannot allocate memory for pnt1_row in get_l2g_datac_by_gran\n");
    }
    if ((pntc_row = (int8 *)calloc(max_nadd_obs_row, sizeof(int8))) == NULL)
    {
      status = -1;
      fprintf(stderr, "Cannot allocate memory for pntc_row in get_l2g_datac_by_gran.\n");
    }
  }

  if (status == -1)
  {
    if (pnt1_row != NULL) free(pnt1_row);
    if (pntc_row != NULL) free(pntc_row);
    if (pnt1_info.sds_id != -1) close_sds(&pnt1_info, pnt1_info.H5IN);
    if (pntc_info.sds_id != -1) close_sds(&pntc_info, pntc_info.H5IN);
    if (pnt1_info.sd_id != -1) close_sd(&pnt1_info, pnt1_info.H5IN);
    /* MSZ 3/8/2018, this is wrong and sd_id for sds1_info should be closed */
    /*if (sds1_info.sd_id != -1) close_sd(&pntc_info, pntc_info.H5IN);*/
    if (sds1_info.sd_id != -1) close_sd(&sds1_info, sds1_info.H5IN);
    if (sds_num_obs != NULL) free(sds_num_obs);
    if (sds_nadd_obs != NULL) free(sds_nadd_obs);
    Free2D((void **)modis_fpath);
    return;
  }

  edge[0] = ncols;
  edge[1] = nrows;
  sds_cnt = atoi(arg_list[3]);
  run_cnt = atoi(arg_list[4]);
  sdsc_info.sd_id = sds1_info.sd_id;
  sdsc_info.use_L2G_max_nobs = 0;
  sdsc_info.H5IN = sds1_info.H5IN;
  for (isds=0; isds<sds_cnt; isds++)
  {
    /* open the sds1 and sdsc and allocate memory for one row of data */
    fprintf(stderr, "	Reading SDS %s\n", arg_list[6+isds]);
    sdsc_info.sds_id = sds1_info.sds_id = -1;
    sprintf(sds1_info.name, "%s_1", arg_list[6+isds]);
    sprintf(sdsc_info.name, "%s_c", arg_list[6+isds]);
    if ((get_sds_info((char *)NULL, &sds1_info) == -1) || 
		(get_sds_info((char *)NULL, &sdsc_info) == -1))
      status = -1;

    if ((sds1_row = (void *)calloc(ncols, sds1_info.data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for sds1_row in get_l2g_datac_by_gran\n");
    if ((sdsc_row = (void *)calloc(max_nadd_obs_row, sdsc_info.data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for sdsc_row in get_l2g_datac_by_gran\n");
    if ((out_data = (void *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for out_data in get_l2g_datac_by_gran\n");

    if ((sds1_row == NULL) || (sdsc_row == NULL) || (out_data == NULL))
    {
      if (sds1_info.sds_id != -1) close_sds(&sds1_info, sds1_info.H5IN);
      if (sdsc_info.sds_id != -1) close_sds(&sdsc_info, sdsc_info.H5IN);
      if (sds1_row != NULL) free(sds1_row);
      if (sdsc_row != NULL) free(sdsc_row);
      if (out_data != NULL) free(out_data);
      continue;
    }

    /* create all output sds */
    for (irun=0; irun<run_cnt; irun++)
    {
      if ((cnt = sscanf(arg_list[6+sds_cnt+irun], "%d %d %d", &gran_pnt, &gran_num,
              &orb_num)) == 3)
      {
	if (if_cnt == 1)
          sprintf(out_sds_name, "%s_gran%d_orb%d", arg_list[6+isds], gran_num, orb_num);
	else
          sprintf(out_sds_name, "%s_gran%d_orb%d_%s", arg_list[6+isds], gran_num, orb_num,
			arg_list[1]);
      }
      else
      {
        cnt = sscanf(arg_list[6+sds_cnt+irun], "%d %s", &gran_pnt, gran_id);
        if (cnt == 2)
	{
	  if (if_cnt == 1)
            sprintf(out_sds_name, "%s_gid%s", arg_list[6+isds], gran_id);
	  else
            sprintf(out_sds_name, "%s_gid%s_%s", arg_list[6+isds], gran_id, arg_list[1]);
	}
	else
	{
	  cnt = sscanf(arg_list[6+sds_cnt+irun], "%d", &gran_pnt);
	  if (if_cnt == 1)
            sprintf(out_sds_name, "%s_gpnt%d", arg_list[6+isds], gran_pnt);
	  else
            sprintf(out_sds_name, "%s_gpnt%d_%s", arg_list[6+isds], gran_pnt, arg_list[1]);
	}
      }

      out_sds_info.sd_id = *out_sd_id;
      out_sds_info.H5OUT = H5OUT;
      strcpy(out_sds_info.name, out_sds_name);
      out_sds_info.data_type = sds1_info.data_type;
      out_sds_info.rank = sds1_info.rank;
      for(i = 0; i < out_sds_info.rank; i++)
        out_sds_info.dim_size[i] = sds1_info.dim_size[i];

      if (open_sds((char *)NULL, &out_sds_info, 'W') == FAIL)
      {
        fprintf(stderr, "Cannot create the output SDS: %s", out_sds_name);
	break;
      }
      
      if (sds1_info.data_type == 5 || sds1_info.data_type == 6)
        write_attr_fvalf(H5OUT, out_sds_info.sds_id, sds1_info.data_type, c, sds1_info.fill_fval);
      else
        write_attr_fval(H5OUT, out_sds_info.sds_id, sds1_info.data_type, c, sds1_info.fill_val);
      
      if (cnt == 1)
        fprintf(stderr, "			Reading Granule (index=%d)\n", gran_pnt);
      else if (cnt == 2)
        fprintf(stderr, "			Reading Granule (index=%d, gran_id=%s)\n", gran_pnt, gran_id);
      else if (cnt == 3)
        fprintf(stderr, "			Reading Granule (index=%d, gran_num=%d, orb_num=%d)\n", gran_pnt, gran_num, orb_num);
      /* read and process one row at a time for this granules */
      start1[1] = 0;
      edge1[1] = ncols;
      edge1[0] = 1;
      startc[0] = edgec[0] = 0;
      for (ir=0; ir<nrows; ir++)
      {
        start1[0] = ir;
	if (read_sel_sds_data(&pnt1_info, start1, edge1, pnt1_row) == FAIL)
        {
          fprintf(stderr, "Cannot read row %d of SDS %s in pointer file\n", ir, pnt1_info.name);
          break;
        }
	if (read_sel_sds_data(&sds1_info, start1, edge1, sds1_row) == FAIL)
        {
          fprintf(stderr, "Cannot read row %d of SDS %s in input file\n", ir, sds1_info.name);
          break;
        }
        startc[0] += edgec[0];
        edgec[0] = sds_nadd_obs[ir]; 
        if (edgec[0] != 0)
        {
	  if (read_sel_sds_data(&pntc_info, startc, edgec, pntc_row) == FAIL)
	  {
	    fprintf(stderr, "Cannot read row %d of SDS %s in pointer file\n", ir, pntc_info.name);
	    break;
	  }
	  if (read_sel_sds_data(&sdsc_info, startc, edgec, sdsc_row) == FAIL)
	  {
	    fprintf(stderr, "Cannot read row %d of %s SDS in input file\n", ir, sdsc_info.name);
	    break;
	  }
        }
        offset = ir*ncols;
        for (ic=0, skip_obs=0; ic<ncols; ic++)
        {
          nadd_obs = sds_num_obs[offset+ic] - 1;
          if (pnt1_row[ic] == gran_pnt)
          {
            switch(sds1_info.data_type)
            {
              case 20: ((int8 *)out_data)[offset+ic] = ((int8 *)sds1_row)[ic];
      	               break;
              case 21: ((uint8 *)out_data)[offset+ic] = ((uint8 *)sds1_row)[ic];
      	               break;
              case 22: ((int16 *)out_data)[offset+ic] = ((int16 *)sds1_row)[ic];
  		       break;
              case 23: ((uint16 *)out_data)[offset+ic] = ((uint16 *)sds1_row)[ic];
  		       break;
              case 24: ((int32 *)out_data)[offset+ic] = ((int32 *)sds1_row)[ic];
  		       break;
              case 25: ((uint32 *)out_data)[offset+ic] = ((uint32 *)sds1_row)[ic];
  		       break;
            }
          }
          else if (nadd_obs > 0)
          {
            for (iobs = 0; iobs < nadd_obs; iobs++)
            {
              if (pntc_row[skip_obs+iobs] == gran_pnt)
              {
                switch(sds1_info.data_type)
                {
                  case 20: ((int8 *)out_data)[offset+ic] = ((int8 *)sdsc_row)[skip_obs+iobs];
  	    	           break;
                  case 21: ((uint8 *)out_data)[offset+ic] = ((uint8 *)sdsc_row)[skip_obs+iobs];
  	    	           break;
                  case 22: ((int16 *)out_data)[offset+ic] = ((int16 *)sdsc_row)[skip_obs+iobs];
  		           break;
                  case 23: ((uint16 *)out_data)[offset+ic] = ((uint16 *)sdsc_row)[skip_obs+iobs];
  		           break;
                  case 24: ((int32 *)out_data)[offset+ic] = ((int32 *)sdsc_row)[skip_obs+iobs];
  		           break;
                  case 25: ((uint32 *)out_data)[offset+ic] = ((uint32 *)sdsc_row)[skip_obs+iobs];
  		           break;
                }
                break;
	      } /* if .. */ 
            } /* for iobs ... */
	    if (iobs == nadd_obs)
              switch(sds1_info.data_type)
              {
		/* MSZ 3/5/2018, update it using fill value obtained from the SDS */
                case 20: ((int8 *)out_data)[offset+ic] = FILL_VALUE_INT8;
                         break;
                case 21: ((uint8 *)out_data)[offset+ic] = FILL_VALUE_UINT8;
                         break;
                case 22: ((int16 *)out_data)[offset+ic] = FILL_VALUE_INT16;
                         break;
                case 23: ((uint16 *)out_data)[offset+ic] = FILL_VALUE_UINT16;
                         break;
                case 24: ((int32 *)out_data)[offset+ic] = FILL_VALUE_INT32;
                         break;
                case 25: ((uint32 *)out_data)[offset+ic] = FILL_VALUE_UINT32;
                         break;
		/*
                case 20: ((int8 *)out_data)[offset+ic] = (int8)sds1_info.fill_val;
                         break;
                case 21: ((uint8 *)out_data)[offset+ic] = (uint8)sds1_info.fill_val;
                         break;
                case 22: ((int16 *)out_data)[offset+ic] = (int16)sds1_info.fill_val;
                         break;
                case 23: ((uint16 *)out_data)[offset+ic] = (uint16)sds1_info.fill_val;
                         break;
                case 24: ((int32 *)out_data)[offset+ic] = (int32)sds1_info.fill_val;
                         break;
                case 25: ((uint32 *)out_data)[offset+ic] = (uint32)sds1_info.fill_val;
                         break;
		*/
              }
          } /* else */
	  else
            switch(sds1_info.data_type)
            {
	      /* MSZ 3/5/2018, update it using fill value obtained from the SDS */
              case 20: ((int8 *)out_data)[offset+ic] = FILL_VALUE_INT8;
                       break;
              case 21: ((uint8 *)out_data)[offset+ic] = FILL_VALUE_UINT8;
                       break;
              case 22: ((int16 *)out_data)[offset+ic] = FILL_VALUE_INT16;
                       break;
              case 23: ((uint16 *)out_data)[offset+ic] = FILL_VALUE_UINT16;
                       break;
              case 24: ((int32 *)out_data)[offset+ic] = FILL_VALUE_INT32;
                       break;
              case 25: ((uint32 *)out_data)[offset+ic] = FILL_VALUE_UINT32;
                       break;
	      /*
              case 20: ((int8 *)out_data)[offset+ic] = (int8)sds1_info.fill_val;
                         break;
              case 21: ((uint8 *)out_data)[offset+ic] = (uint8)sds1_info.fill_val;
                         break;
              case 22: ((int16 *)out_data)[offset+ic] = (int16)sds1_info.fill_val;
                         break;
              case 23: ((uint16 *)out_data)[offset+ic] = (uint16)sds1_info.fill_val;
                         break;
              case 24: ((int32 *)out_data)[offset+ic] = (int32)sds1_info.fill_val;
                         break;
              case 25: ((uint32 *)out_data)[offset+ic] = (uint32)sds1_info.fill_val;
                         break;
	      */
            }
	  if (nadd_obs > 0)
            skip_obs += nadd_obs;
        }	/* for ic = 0 ..... */
        if (edgec[0] != skip_obs)
	  fprintf(stderr, "Warning: Number of Observations do not match at row: %d\n", ir);
      }  /* for ir . . . */
      /*if (set_sel_sds_data(&out_sds_info, start, edge, out_data) == FAIL)*/
      if (set_sel_l2g_sds_2D_data_chunked_gzip(&out_sds_info, start, edge, out_data) == FAIL)
	  fprintf(stderr, "Fail to write data to output file \n");
      close_sds(&out_sds_info, out_sds_info.H5OUT);
    } /* for irun ....  */
    close_sds(&sds1_info, sds1_info.H5IN);
    close_sds(&sdsc_info, sdsc_info.H5IN);
    free(sds1_row);
    free(sdsc_row);
    free(out_data);
  } /* for isds . . . */

  if (*m_opt == 1)
  {
    copy_metadata(sds1_info.H5IN, sds1_info.sd_id, H5OUT, *out_sd_id);
    *m_opt = 0;   /* MSZ 2/21/2018, just write meta data once to the output file */
  }
  close_sds(&pnt1_info, pnt1_info.H5IN);
  close_sds(&pntc_info, pntc_info.H5IN);
  close_sd(&pnt1_info, pnt1_info.H5IN);
  close_sd(&sds1_info, sds1_info.H5IN);

  Free2D((void **)modis_fpath);
  free(sds_nadd_obs);
  free(sds_num_obs);
  free(pntc_row);
  free(pnt1_row);
}


