/****************************************************************************
!C

!File: read_l2g_lite.c

!Description:
  Contains routines for reading l2g lite.

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original April/May 2005 Version 1.0
  	- developed from read_l2g.c
  tested and updated March 2009

  Modified February, 2013 by Xiaoping Zhang
  fixed bug to read meta data for MOD09GA and MOD09GQ l2g_lite data 
  in subroutine of check_param.
  Made selection of layer number 1-based, instead of 0-based.
  
  Modified March, 2013 by Xiaoping Zhang 
  get the correct maximum observations number and l2g storage format 
  and make it work for NPP	

  Modified February, 2018 Version 3.5.1 by Maosheng Zhao, SSAI
        1)modify it so that HDF5 can be read
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c and read_l2g_lite_lib.c

Maosheng Zhao, SSAI. Note on 3/6/2018
Unlike C6 L2G MxD09, L2G VNP09GHKI and VNP09GHKM in HDF4 don't have ArchiveMetadata.0
and CoreMetadata.0 in the file attribute. Thus the previous method to determine L2G file or not 
won't work for a general purpose in read_l2g and read_l2g_lite tool. The new
method is below
1) use L2G_format_check() to check if it is L2G lite or L2G.
2) develop get_sds_names_num_observations() so that there is no need to read meta data
   for max_obs for different resolution data and also no need to use ESDT to obtain
   associated information.
3) users can output SDSs and the corresponding number layers with -help filenames
4) if there is no file attribute name with "storage_format", the logic of the
   default is "compact" not "full".
5) turn on -layer option and all layers will be decompacted if without =num

# MSZ 3/7/2018, Note on difference and same between read_l2g and read_l2g_lite:
For L2G heavy or L2G lite file with just one SDS with "num_observation" and under -layer option,
both read_l2g and read_l2g_lite tools can handle it.  For just -layer option, read_l2g_lite
can handle both L2G heavy and L2G lite files. Only for granule option -g*, just read_l2g can do it.

  Modified February, 2020 Version 3.6 by Maosheng Zhao, SSAI

1) correct a bug error when running read_l2g and the bug is in main_util.c.
   In function find_file(), the 
     char tmp_fname[80]; is updated to char tmp_fname[1024];
   which is called by get_pnt_fname(), and get_pnt_fname is called by
   get_l2g_datac_by_gran() in read_l2g_comp.c where as get_l2g_datac_by_gran()
   is called in read_l2g.c
2) correct issue for get_sds_names_num_observations() and
   read_input_sds_names_ress() because for some VNP l2g, the dimension of the tile
   data is not 1200, 2400, or/and 4800 but others.
3) increase defined string length to avoid crash as below
   in read_l2g.c, get_npp_gran_id()
   if ((iptr_fnames = (char **)Calloc2D(20, MAX_META_VAL_LEN, sizeof(char))) == NULL) changed to
   if ((iptr_fnames = (char **)Calloc2D(127, MAX_META_VAL_LEN, sizeof(char))) == NULL)

   127 has been change to MAX_NUM_GRAN
   l2g.h:#define MAX_NUM_GRAN 20
     to 127
   change several 20 in read_l2g_comp.c to MAX_NUM_GRAN

4) enable output HDF4 with internal compression

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "hdf5.h"
#include "l2g.h"
#include "meta.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "str_op.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"

#define HELP \
"NAME\n" \
"     read_l2g_lite - read data from L2G files in the L2G_lite format\n" \
"\n" \
"SYNOPSIS\n" \
"     read_l2g_lite -help\n" \
"     read_l2g_lite -help [filenames]\n" \
"     read_l2g_lite -layer[=n1[,n2...]] -of=output_file [-meta]\n" \
"              [-sds=SDSname1[,SDSname2...]] filename\n" \
"\n" \
"DESCRIPTION\n" \
"     Extracts a subset of parameters (full format or compact format) from a L2G\n" \
"     file in lite format, selecting pixels according to the layer number. The\n" \
"     output is an HDF file where all the selected items are stored as separate SDS. \n" \
"\n" \
"OPTIONS\n" \
"     -help                          Display this help message or file information with filename\n" \
"     -layer[=n1[,n2...]]            Select by layer number (1-based)\n" \
"     -sds=<SDS list>                List of SDS to read (separated by commas)\n" \
"                                    Use sds_name without the layer number\n" \
"                                    e.g. use SDS name as sur_refl_b01 to retreive any layer \n" \
"                                    of SDS sur_refl_b01 \n" \
"     -of=filename                   Output file\n" \
"     -meta                          Copy metadata from input file to output \n" \
" \n" \
"Examples: \n" \
"       read_l2g_lite -help MOD09GA.A2001146.h09v05.005.2007024153208.hdf \n" \
"       read_l2g_lite -layer MOD09GA.A2001146.h09v05.005.2007024153208.hdf -of=l2g_layers.hdf -meta \n" \
"	read_l2g_lite -layer -sds=sur_refl_b03,state_1km,SensorZenith MOD09GA.A2001146.h09v05.005.2007024153208.hdf -of=l2g_layers.hdf -meta \n" \
"	read_l2g_lite -layer=1-2 -sds=sur_refl_b03,state_1km,SensorZenith MOD09GA.A2001146.h09v05.005.2007024153208.hdf -of=l2g_layers.hdf -meta \n" \
"	read_l2g_lite -layer=1,3 -sds=sur_refl_b01,sur_refl_b02,state_1km,SensorZenith,SensorAzimuth,SolarZenith,SolarAzimuth,QC_500m MOD09GA.A2001146.h09v05.005.2007024153208.hdf -of=l2g_layers.hdf -meta \n" \
"	read_l2g_lite -layer=1-2 -sds=sur_refl_b01,QC_250m MOD09GQ.A2012339.h09v05.005.2012341105824.hdf -of=l2g_layers.hdf -meta \n" \
"	read_l2g_lite -layer=1-4 -sds=SurfReflect_I1,SurfReflect_M4 VNP09GA.A2017240.h09v05.001.2017241195209.h5 -of=l2g_layer.hdf -meta \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"Usage:      read_l2g_lite -layer=n1[,n2...]] -of=output_file \n" \
"            [-sds=SDSname1[,SDSname2...]]] -meta filename \n" \
" \n" \
"OPTIONS \n" \
"     -help                          Display this help message or all -sds names and their max number of layers if file name is also used\n" \
"     -layer[=n1[,n2...]]            Select by layer number (1-based)\n" \
"     -sds=<SDS list>                List of SDS to read (separated by commas)\n" \
"                                    Use sds_name without the layer number\n" \
"                                    e.g. use SDS name as sur_refl_b01 to retreive any layer \n" \
"                                    of SDS sur_refl_b01 \n" \
"     -of=filename                   Output file \n" \
"     -meta                          Copy metadata from input file to output \n"


int parse_cmd_read_l2g_lite(int argc, char **argv, char **sds_names, int *sds_cnt, 
	int *if_cnt, char **num_str, int *nl, char *out_fname, int *m_opt);
int get_sds_names_num_observations(char *fname, char **SDS_names_num_obs);
int get_l2g_max_obs_layers(char *hdf_fname, char *max_obs_sds_name);
int check_param(char *fname, char *sds_names, int res, char **num_str, int nl, char **arg_list, char **SDS_names_num_obs);
int read_input_sds_names_ress(char *fname, char **in_sds_names, int sds_cnt, char **out_sds_names, int *res);
void get_layer_numbers(char **arg_list, int max_nobs, char **num_str, int nl);
void get_l2g_dataf_by_obs(char **arg_list, int res, int32 out_sd_id, int if_cnt, int *m_opt, int H5OUT);
void get_l2g_lite_datac_by_obs(char **arg_list, char *max_nobs_sds_name, int32 out_sd_id, int if_cnt, int *m_opt, int H5OUT);

int main(int argc, char **argv)
{
  int i;
  sds_t out_sds_info;
  int fid, nl;
  int isds, nsds;
  int m_opt, status;
  int if_cnt, sds_cnt;
  int res[MAX_NUM_SDS];
  char out_fname[MAX_PATH_LENGTH];
  char **num_str, **arg_list, **num_str1;
  char **sds_names = NULL, **out_sds_names = NULL;
  char **file_attr_names = NULL;
  int max_nobs;
  char fname_tmp[MAX_PATH_LENGTH];
  char **SDS_names_num_obs = NULL;         /* MSZ 3/6/2018 */
  int nsds_num_observations;               /* MSZ 3/6/2018 */

  /* MSZ 2/13/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/13/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ( (argc==1) || ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0))))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  else
  {
    /* Print all SDS names when -help option is input with the input filename */
    for (i=0, status=0; i<argc; i++)
      if ((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "-help") == 0)) status = i;
    if (status == 1)
    {
      /* MSZ 2/20/2018, add the below to display the file content of L2G Lite */
      for (fid=1; fid<argc; fid++)
        if (argv[fid][0] != '-')
        {
	  /* just for L2G lite file name patterns */
	  if(L2G_format_check(argv[fid]) >= 1)
	  {
	    if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	      fprintf(stderr, "Cannot allocate memory for sds_names in read_l2g_lite(main)\n");
	    if ((out_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	      fprintf(stderr, "Cannot allocate memory for out_sds_names in read_l2g_lite(main)\n");

	    sds_cnt = 0;
            nsds = read_input_sds_names_ress(argv[fid], sds_names, sds_cnt, out_sds_names, res);

	    fprintf(stderr, "\n\n##############################################################################\n");
	    fprintf(stderr, "Below are the details on the SDSs and their number layers in this L2G file %s:\n", argv[fid]);

	    for (isds=0; isds<nsds; isds++)
	    {
	      max_nobs = 0;
	      if(res[isds] == 1)
	      {
		max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs_1km);
		if(max_nobs == -1)
		  max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs);
		fprintf(stderr, "1km SDS: %s, max_number_layers: %d\n", out_sds_names[isds], max_nobs);
	      }
	      else if (res[isds] == 2)
	      {
		max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs_500m);
		if(max_nobs == -1)
		  max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs);
		fprintf(stderr, "500m SDS: %s, max_number_layers: %d\n", out_sds_names[isds], max_nobs);
	      }
	      else if (res[isds] == 4)
	      {
		max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs);
		if(max_nobs == -1)
		  max_nobs = get_l2g_max_obs_layers(argv[fid], sds_name_nobs);
		fprintf(stderr, "250m SDS: %s, max_number_layers: %d\n", out_sds_names[isds], max_nobs);
	      }
	    } /* for (isds=0; isds<nsds; isds++) */

	    if(sds_names) Free2D((void**)sds_names);
	    if(out_sds_names) Free2D((void**)out_sds_names); 
	    fprintf(stderr, "##############################################################################\n\n\n");
	  } /* if(L2G_format_check(argv[fid]) >= 1) */
	  else
	  {
	    fprintf(stderr, "%s is not L2G data product, exit...\n", argv[fid]);
	    exit(0);
	  }
        } /* if (argv[fid][0] != '-') */
      exit(0);	
    }
  }

  /*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
  /* start the real work */
  if ((SDS_names_num_obs = (char **)Calloc2D(4, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for SDS_names_num_obs...\n");
  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names in read_l2g_lite(main)\n");
  if ((out_sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for out_sds_names in read_l2g_lite(main)\n");
  if ((num_str = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for num_str in read_l2g_lite(main)\n");
  if ((file_attr_names = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for file_attr_names in read_l2g_lite(main)\n");
  if ((arg_list = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for arg_list in read_l2g_lite(main)\n");
  if ((sds_names != NULL) && (arg_list != NULL) && (num_str != NULL) && (out_sds_names != NULL))
  {
    if ((status = parse_cmd_read_l2g_lite(argc, argv, sds_names, &sds_cnt, &if_cnt, num_str, &nl, out_fname, &m_opt)) == -1)
    {  
      fprintf(stderr, "%s\n", USAGE);
      exit(0);
    }
    else if (status != -1)
    {
      out_sds_info.H5OUT = 0;   /* so far, output to HDF4 */
      out_sds_info.sd_id = out_sds_info.sds_id = -1;
      if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
	fprintf(stderr, "Cannot create output HDF file %s in read_l2g_lite:main()\n", out_fname);
      if (out_sds_info.sd_id != FAIL)
      {
        for (fid=1; fid<argc; fid++)
        {
   	  if (argv[fid][0] == '-') continue;
	  if(L2G_format_check(argv[fid]) >= 1)
	  {
	    if((nsds_num_observations = get_sds_names_num_observations(argv[fid], SDS_names_num_obs)) == 0)
	    {
	      fprintf(stderr, "There is no sds named with \"num_observations\" in %s or it is not 2D in square shape, exit...\n", argv[fid]);
	      exit(0);
	    }
	    fprintf(stderr, "\nProcessing input L2G lite file %s\n", argv[fid]);
	    fprintf(stderr, "---------------------------------------------------------------\n");
	    nsds = read_input_sds_names_ress(argv[fid], sds_names, sds_cnt, out_sds_names, res);
	    for (isds=0; isds<nsds; isds++)
	    {        
	      status = check_param(argv[fid], out_sds_names[isds], res[isds], num_str, nl, arg_list, SDS_names_num_obs);
	      strcpy(arg_list[3], out_sds_names[isds]);               /* arg_list[3] = output SDS names in this loop */
	      if (status != 0)
	      {
	        strcpy(arg_list[0], argv[fid]);                       /* arg_list[0] = fname */
	        strcpy(arg_list[1], argv[fid]);
	        rm_path(arg_list[1]);                                 /* arg_list[1] = fname removed path */
	        arg_list[1][24] = '\0'; 
	        /*fprintf(stdout, "arg_list[2] = %s\n", arg_list[2]); */ /* arg_list[2] = storage format from attr or meta data */
	        if (strcmp(arg_list[2], "full") == 0)
                  get_l2g_dataf_by_obs(arg_list, res[isds], out_sds_info.sd_id, if_cnt, &m_opt, out_sds_info.H5OUT);
	        else                                                  /* arg_list[4] = number of layers (max_obs or or -layer=), and 0-based arg_list[5+each_layer] */
		{
		  if(res[isds] >= 1 && res[isds] <= 4)
		  {
                    get_l2g_lite_datac_by_obs(arg_list, SDS_names_num_obs[ res[isds] - 1 ], out_sds_info.sd_id, if_cnt, &m_opt, out_sds_info.H5OUT);
		  }
		  else
		  {
		    fprintf(stderr, "The resolution code must be from 1 to 4, exit...\n");
		    exit(0);
		  }
		}
              } /* if (status = . . .  */
            } /* for (isds = . . . */
	  } /* if(L2G_format_check(argv[fid]) >= 1) */
	  else
	  {
            fprintf(stderr, "%s is not L2G data product, exit...\n", argv[fid]);
            exit(0);
	  }
        } /* for (fid=0 . . . . */
	close_sd(&out_sds_info, out_sds_info.H5OUT);
      }
    }
  }
  if(SDS_names_num_obs) Free2D((void **)SDS_names_num_obs);
  if(num_str) Free2D((void **)num_str);  
  if(arg_list) Free2D((void **)arg_list);
  if(sds_names) Free2D((void **)sds_names);
  if(out_sds_names) Free2D((void **)out_sds_names);
//  if (nsds != 0) fprintf(stderr, "Process read_l2g completed \n");
  return 0;
}

int parse_cmd_read_l2g_lite(int argc, char **argv, char **sds_names, int *sds_cnt, 
	int *if_cnt, char **num_str, int *nl, char *out_fname, int *m_opt)
{
  int status, i;
  char in_fname[255];
  
  status = 1;
  *m_opt = 0;
  *if_cnt = 0;
  *sds_cnt = 0;
  *nl = 0;
  out_fname[0] = '\0';
  
  for (i=1; i<argc; i++)
  {
    if ((is_arg_id(argv[i], "-of=") == 0) || (is_arg_id(argv[i], "-o=") == 0))
      get_arg_val(argv[i], out_fname);
    else if (is_arg_id(argv[i], "-sds=") == 0)
      get_arg_val_arr(argv[i], sds_names, sds_cnt);
    else if (is_arg_id(argv[i], "-sds") == 0) ;

    else if (is_arg_id(argv[i], "-layer=") == 0)
      get_arg_val_arr(argv[i], num_str, nl);
    else if (strcmp(argv[i], "-layer") == 0)
      *nl = 0;     /* MSZ 3/6/2018, for all layers */

    else if (strcmp(argv[i], "-meta") == 0) *m_opt = 1;

    else if (argv[i][0] == '-')
      fprintf(stderr, "Unknown option %s\n", argv[i]);
    else 
      ++*if_cnt;    
  }

  if (out_fname[0] == '\0')
  {
    status = -1;
    fprintf(stderr, "Missing output file name \n");    
  }
  if (*if_cnt == 0) 
  {
    status = -1;
    fprintf(stderr, "Missing input file name \n");    
  }
  if (status != -1)
  {
    if (*sds_cnt == 0)
      fprintf(stderr, "No SDS name input. Reading all SDS . . \n");
    if (*nl == 0)
      fprintf(stderr, "No values input for option -layer. Reading all layers\n");
  }

  return status;
}

int check_param(char *fname, char *sds_name, int res, char **num_str, int nl, char **arg_list, char **SDS_names_num_obs)
{
  /* MSZ 2/16/2018. completely rewrite this function to make it clear and concise */
  int max_nobs;
  int status = 1;
  char tmp_fname[MAX_PATH_LENGTH] = {'\0'};
  char *l2g_storage_format = NULL;
  int32 attr_type, attr_cnt;
  sds_t in_sds_info;

  /* just the real file name to determine if it is l2g lite */
  strcpy(tmp_fname, fname);
  rm_path(tmp_fname);

  if(!res)
  {
     fprintf(stderr, "res number should be greater than 0, exit...\n");
     exit(0);
  }

  max_nobs = get_l2g_max_obs_layers(fname, SDS_names_num_obs[res-1]);
  get_layer_numbers(arg_list, max_nobs, num_str, nl); 

  /* get the l2g storage format */
  in_sds_info.sd_id = in_sds_info.sds_id = -1;
  if(open_sd(fname, &in_sds_info, 'R') == FAIL)
    fprintf(stderr, "Cannot open output HDF file %s\n", fname);
  else
  {
    if( !(l2g_storage_format = (char*)find_get_sd_attr(&in_sds_info, "storage_format", &attr_type, &attr_cnt)) )
      l2g_storage_format = (char*)find_get_sd_attr(&in_sds_info, "storageformat", &attr_type, &attr_cnt);

    if(l2g_storage_format)
    {
      strcpy(arg_list[2], l2g_storage_format);
      if(l2g_storage_format) free(l2g_storage_format);
    }
  }
  close_sd(&in_sds_info, in_sds_info.H5IN);

  return status;
}

int get_sds_names_num_observations(char *fname, char **SDS_names_num_obs)
{
   /* Maosheng Zhao 3/6/2018, SSAI, get the sds names with "num_observations" and also put them in a string array 0 to 3,
    * with 0 1km, 1 500m and 3 250m to correspond res 1, 2, and 4 so that res - 1 matches the correct subscript of 
    * sds names with "num_observations"
    * Fail is the return value is 0.
    * SDS_names_num_obs is 4 strings defined outside of the function */

   char **sds_names = NULL;
   int nsds = 0, nsds_num_observations = 0;
   int i;
   sds_t in_sds_info;

   for(i=0; i < 4; i++) /* corresponding to res 1 to 4 minus 1 */
     SDS_names_num_obs[i][0] = '\0';

   if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
   {
     fprintf(stderr, "Cannot allocate memory for sds_names, exit...\n");
     exit(0);
   }
   else
   {
     if ((nsds = get_sds_names(fname, sds_names)) == 0)
     {
       fprintf(stderr, "No sds found in file %s, exit...\n", fname);
       exit(0);
     }
     else
     {
       int RESs_num = 0;
       for(i = 0, nsds_num_observations = 0; i < nsds; i++)
	  if(strcasestr(sds_names[i], "num_observations") != '\0')
	  {
	    strcpy(in_sds_info.name, sds_names[i]);
	    in_sds_info.sd_id = in_sds_info.sds_id = -1;
	    if(get_sds_info(fname, &in_sds_info) != FAIL)
	    {
	      if(in_sds_info.rank == 2 && in_sds_info.dim_size[0] == in_sds_info.dim_size[1])
	      {
		/* for MODIS 1200, 2400 and 4800 system */
		if(in_sds_info.dim_size[0] % 1200 == 0)  /* for MODIS */
		{
		  if(in_sds_info.dim_size[0] == 1200)
		  {
		    strcpy(SDS_names_num_obs[0], sds_names[i]);
		    nsds_num_observations++;
		  }
		  else if(in_sds_info.dim_size[0] == 2400)
		  {
		    strcpy(SDS_names_num_obs[1], sds_names[i]);
		    nsds_num_observations++;
		  }
		  else if(in_sds_info.dim_size[0] == 4800)
		  {
		    strcpy(SDS_names_num_obs[3], sds_names[i]);
		    nsds_num_observations++;
		  }
		}
		else   /* for non-MODIS system */
		{
		  if((int)(in_sds_info.dim_size[0] * 0.001) == 1)
		  {
		    strcpy(SDS_names_num_obs[0], sds_names[i]);
		    nsds_num_observations++;
		  }
		  else if((int)(in_sds_info.dim_size[0] * 0.001) == 2)
		  {
		    strcpy(SDS_names_num_obs[1], sds_names[i]);
		    nsds_num_observations++;
		  }
		  else if((int)(in_sds_info.dim_size[0] * 0.001) == 4)
		  {
		    strcpy(SDS_names_num_obs[3], sds_names[i]);
		    nsds_num_observations++;
		  }
		}
	      }
	    }
	  }
       if(sds_names) Free2D((void**)sds_names);
     }
   }

   return nsds_num_observations;
}


