/****************************************************************************
!C

!File: read_l2g_lite_lib.c

!Description:
  Supporting routines for read l2g lite compact format files. 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original April/May 2005
    developed from read_l2g_comp.c
    
  Modified Dec 07, 2012 by Xiaoping Zhang
  fixed bug to read M*D09GA sds of number of observations in sub of get_l2g_datac_by_obs.
  
  Modified March, 2013 by Xiaoping Zhang  
  fixed bug to read sds NPP_DSRFLD_L2GD of number of observations in sub of get_l2g_datac_by_obs.	

  Modified Feb, 2018 by Maosheng Zhao, SSAI
  change int m_opt of the argument to int *m_opt for get_l2g_dataf_by_obs() and 
  get_l2g_lite_datac_by_obs() so that *m_opt = 0 when copy_metadata() is done.
  That is we just need to write meta data to the output file once note many times
  
!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include "mfhdf.h"
#include "l2g.h"
#include "str_op.h"
#include "meta.h"
#include "qa_tool.h"
#include "sds_rw.h"
#include "alloc_mem.h"
#include "main_util.h"


void get_l2g_dataf_by_obs(char **arg_list, int res, int32 out_sd_id, int if_cnt, int *m_opt, int H5OUT) 
{
  int c=0;
  int attr_val=0;
  int nrows, ncols;
  int run_cnt, irun;
  int layer_num[MAX_NUM_LAYERS];
  int32 edge1[2], edgef[3];
  int32 start1[2] = {0, 0};
  int32 startf[3] = {0, 0, 0};
  char out_sds_name[MAX_SDS_NAME_LEN];
  sds_t sds1_info, sdsf_info, out_sds_info;
  void *sds_data = NULL;
  int i;
 
  sds1_info.sd_id = sds1_info.sds_id = -1;
  if(open_sd(arg_list[0], &sds1_info, 'R') == FAIL)
  {
    fprintf(stderr, "Cannot open HDF file %s in get_l2g_dataf_by_obs\n", arg_list[0]);
    return;
  }
  sdsf_info.sd_id = sds1_info.sd_id;
  sdsf_info.H5IN = sds1_info.H5IN;

  run_cnt = atoi(arg_list[4]);
  for (irun=0; irun<run_cnt; irun++)
    layer_num[irun] = atoi(arg_list[5+irun]);
  fprintf(stderr, "Reading SDS %s\n", arg_list[3]);
  sprintf(sds1_info.name, "%s_1", arg_list[3]); 
  sprintf(sdsf_info.name, "%s_f", arg_list[3]); 
  sds1_info.sds_id = sdsf_info.sds_id = -1;
  if (get_sds_info((char *)NULL, &sds1_info) == -1)
    close_sds(&sds1_info, sds1_info.H5IN);

  if (open_sds((char *)NULL, &sdsf_info, 'R') == -1)
  {
    close_sds(&sds1_info, sds1_info.H5IN);
    if (sdsf_info.sds_id != -1) 
      close_sds(&sdsf_info, sdsf_info.H5IN);
  }
  nrows = sds1_info.dim_size[0];
  ncols = sds1_info.dim_size[1];
  if ((sds_data = (void *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds1_data in get_l2g_dataf_by_obs\n");
  else
  {
    edge1[0] = nrows;
    edge1[1] = ncols;
    edgef[0] = 1;
    edgef[1] = nrows;
    edgef[2] = ncols;
    for (irun=0; irun<run_cnt; irun++)
    {
      fprintf(stderr, "	Reading Layer %d\n", layer_num[irun] + 1);
      if (if_cnt == 1)
        sprintf(out_sds_name, "%s_layer%d", arg_list[3], layer_num[irun] + 1);
      else
        sprintf(out_sds_name, "%s_layer%d_%s", arg_list[3], layer_num[irun] + 1, arg_list[1]);

      out_sds_info.sd_id = out_sd_id;
      out_sds_info.H5OUT = H5OUT;
      strcpy(out_sds_info.name, out_sds_name);
      out_sds_info.data_type = sds1_info.data_type;
      out_sds_info.rank = sds1_info.rank;
      for(i = 0; i < out_sds_info.rank; i++)
	out_sds_info.dim_size[i] = sds1_info.dim_size[i];

      if (open_sds((char *)NULL, &out_sds_info, 'W') == FAIL)
        fprintf(stderr, "Cannot create the output SDS: %s", out_sds_name);
      else
      {
	write_attr_fval(H5OUT, out_sds_info.sds_id, sds1_info.data_type, c, attr_val);
        if (layer_num[irun] == 0)
        {
	  if (read_sel_sds_data(&sds1_info, start1, edge1, sds_data) == FAIL) 
            fprintf(stderr, "Error reading SDS %s in get_l2g_dataf_by_obs\n", sds1_info.name);
        }
        else
	{
          startf[0] = layer_num[irun]-1;
	  if (read_sel_sds_data(&sdsf_info, startf, edgef, sds_data) == FAIL)
            fprintf(stderr, "Error reading SDS %s in get_l2g_dataf_by_obs\n", sdsf_info.name);
        }
	/*if(set_sel_sds_data(&out_sds_info, start1, edge1, sds_data) == FAIL)*/
	if(set_sel_l2g_sds_2D_data_chunked_gzip(&out_sds_info, start1, edge1, sds_data) == FAIL)
	{
          fprintf(stderr, "Error writing SDS data %s in get_l2g_dataf_by_obs\n", out_sds_name);
	  close_sds(&out_sds_info, out_sds_info.H5OUT);
	}
      }
    } /* for (irun=0; . . . ) */
    free(sds_data);
  }
  close_sds(&sds1_info, sds1_info.H5IN);
  close_sds(&sdsf_info, sdsf_info.H5IN);
  if (*m_opt == 1)
  {
    copy_metadata(sds1_info.H5IN, sds1_info.sd_id, H5OUT, out_sd_id);
    *m_opt = 0;   /* MSZ 2/21/2018, just write meta data once to the output file */
  }
  close_sd(&sds1_info, sds1_info.H5IN);
}

void get_l2g_lite_datac_by_obs(char **arg_list, char *max_nobs_sds_name, int32 out_sd_id, int if_cnt, int *m_opt, int H5OUT) 
{
  int c = 0;
  int status, status1, status2;
  int i, ir, ic, ip;
  int irun, run_cnt;
  int attr_val = 0;
  int skip_obs_ir, skip_obs_ic;
  int nobs, nobs_in_c, nobs_in_c1;
  int nrows, ncols, ic_obs;
  int read_sds1, read_sdsc;
  int layer_num[MAX_NUM_LAYERS];
  int32 edge1[2] = {0, 0};
  int32 edgec[2] = {0, 0};
  int32 start1[2] = {0, 0};
  int32 startc[2] = {0, 0};
  char out_sds_name[MAX_SDS_NAME_LEN];
  sds_t sds_info, sds1_info, sdsc_info, out_sds_info;
  int8 *sds_nobs;
  int32 *sds_add_nobs;   /* MSZ 3/7/2018, update from int16 to int32 for safety */
  void *sds1_data = NULL, *sdsc_data = NULL, *out_data = NULL;

  read_sds1 = read_sdsc = 0;
  run_cnt = atoi(arg_list[4]);
  for (i=0; i<run_cnt; i++)
  {
    layer_num[i] = atoi(arg_list[5+i]);
//    printf("layer_num[%d] = %d\n", i,  layer_num[i]);
    if (layer_num[i] == 0) read_sds1 = 1;
    else if (layer_num[i] > 0) read_sdsc = 1;
  }

  status = 1;
  if (read_sds1 == 1) status1 = 1; 
  else  status1 = -1; 
  if (read_sdsc == 1) status2 = 1;
  else  status2 = -1;  

  sds_info.sd_id = sds_info.sds_id = -1;
  if(open_sd(arg_list[0], &sds_info, 'R') == FAIL)
  {
    fprintf(stderr, "Cannot open HDF file %s in get_l2g_datac_by_obs\n", arg_list[0]);
    return;
  }

  /* read the SDS number of observations */
  strcpy(sds_info.name, max_nobs_sds_name);
  sds_info.H5IN = sds_info.H5IN;
  sds_info.sds_id = -1;
  if (get_sds_info((char *)NULL, &sds_info) == -1)
    status = -1;
  else
  {
    nrows = edge1[0] = sds_info.dim_size[0];
    ncols = edge1[1] = sds_info.dim_size[1];
    if ((sds_nobs = (int8 *)calloc(nrows*ncols, sds_info.data_size)) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for sds_nobs get_l2g_datac_by_obs\n");
      status = -1;
    }
    else if (read_sel_sds_data(&sds_info, start1, edge1, sds_nobs) == FAIL)
    {
      fprintf(stderr, "Cannot read SDS number of observations in get_l2g_datac_by_obs");
      free(sds_nobs);
      status = -1;
    }
    close_sds(&sds_info, sds_info.H5IN);
  }

  /* create SDS sds_add_nobs */
  int ntot_ir_obs = 0;
  if (status != -1) {
    if ((sds_add_nobs = (int32 *)calloc(nrows, sizeof(int32))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for sds_add_nobs get_l2g_datac_by_obs\n");
      status = -1;
    }
    else {
      for (ir=0, ip=0; ir<nrows; ir++) {
        ((int32 *)sds_add_nobs)[ir] = 0;
        for (ic=0; ic<ncols; ic++, ip++) {
	  nobs = ((int8 *)sds_nobs)[ip];
	  if ((nobs != sds_info.fill_val) && (nobs > 1)) {	
	    ((int32 *)sds_add_nobs)[ir] = ((int32 *)sds_add_nobs)[ir] + nobs - 1;
	     ntot_ir_obs = ntot_ir_obs + nobs - 1;
	  }
        }
      }
    }
  }

  if (status != -1)
  {
    fprintf(stderr, "Uncompacting SDS %s\n", arg_list[3]);
  
    /* read the first layer */
    sds1_info.H5IN = sds_info.H5IN;
    sds1_info.sd_id = sds_info.sd_id;
    sds1_info.sds_id = -1;
    sprintf(sds1_info.name, "%s_1", arg_list[3]);
    if (get_sds_info((char *)NULL, &sds1_info) == -1)
      status1 = -1;
    else
    {
      edge1[0] = sds1_info.dim_size[0];
      edge1[1] = sds1_info.dim_size[1];      
      if (read_sds1 == 1) 
      {
	nrows = edge1[0] = sds1_info.dim_size[0];
        ncols = edge1[1] = sds1_info.dim_size[1];
        if ((sds1_data = (int8 *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
        {
          fprintf(stderr, "Cannot allocate memory for sds1_data get_l2g_datac_by_obs\n");
          status1 = -1;
        }
	else if (read_sel_sds_data(&sds1_info, start1, edge1, sds1_data) == FAIL)
        {
          fprintf(stderr, "Cannot read SDS %s in get_l2g_datac_by_obs", sds1_info.name);
          free(sds1_data);
          status1 = -1;
        }
      }
//        SDendaccess(sds1_info.sds_id);
//        sds1_info.sds_id = -1;
    }

    /* read the compact layer */
    if (read_sdsc == 1) 
    {
      sdsc_info.H5IN = sds_info.H5IN;
      sdsc_info.sd_id = sds_info.sd_id;
      sdsc_info.use_L2G_max_nobs = 0;
      sdsc_info.sds_id = -1;
      sprintf(sdsc_info.name, "%s_c", arg_list[3]);
      if (get_sds_info((char *)NULL, &sdsc_info) == FAIL)
        status2 = -1;
      else
      {
        startc[0] = 0;
        edgec[0] = sdsc_info.dim_size[0];
        if ((sdsc_data = (int8 *)calloc(sdsc_info.dim_size[0], sdsc_info.data_size)) == NULL)
        {
          fprintf(stderr, "Cannot allocate memory for sdsc_data get_l2g_datac_by_obs\n");
          status2 = -1;
        }
	else if (read_sel_sds_data(&sdsc_info, startc, edgec, sdsc_data) == FAIL)
        {
          fprintf(stderr, "Cannot read SDS %s in get_l2g_datac_by_obs", sdsc_info.name);
          free(sds1_data);
          status2 = -1;
        }
	close_sds(&sdsc_info, sdsc_info.H5IN);
      }
    }

    if (status != -1)
    {
      if ((out_data = (void *)calloc(nrows*ncols, sds1_info.data_size)) == NULL)
	fprintf(stderr, "Cannot allocate memory for out_data in read_l2g_by_obs\n");
      else 
      {
	for (irun=0; irun<run_cnt; irun++)
	{
          fprintf(stderr, "	Reading layer %d\n", layer_num[irun] + 1);
          if (if_cnt == 1)
            sprintf(out_sds_name, "%s_layer%d", arg_list[3], layer_num[irun] + 1);
          else
            sprintf(out_sds_name, "%s_layer%d_%s", arg_list[3], layer_num[irun] + 1, arg_list[1]);

	  out_sds_info.sd_id = out_sd_id;
	  out_sds_info.H5OUT = H5OUT;
	  strcpy(out_sds_info.name, out_sds_name);
	  out_sds_info.data_type = sds1_info.data_type;
	  out_sds_info.rank = sds1_info.rank;
	  for(i = 0; i < out_sds_info.rank; i++)
	    out_sds_info.dim_size[i] = sds1_info.dim_size[i];

          if (layer_num[irun] == 0) 
          {
	    if (open_sds((char *)NULL, &out_sds_info, 'W') == FAIL)
               fprintf(stderr, "Cannot create the output SDS: %s\n", out_sds_name);              
	    else
	    {	      
	      write_attr_fval(H5OUT, out_sds_info.sds_id, sds1_info.data_type, c, attr_val);
	      /*if(set_sel_sds_data(&out_sds_info, start1, edge1, sds1_data) == FAIL)*/
	      if(set_sel_l2g_sds_2D_data_chunked_gzip(&out_sds_info, start1, edge1, sds1_data) == FAIL)
        	fprintf(stderr, "Cannot write output SDS %s_layer1\n", arg_list[3]);
	    }
	    close_sds(&out_sds_info, out_sds_info.H5OUT);
          }
          else if (status2 != -1)
          {              
	    if (open_sds((char *)NULL, &out_sds_info, 'W') == FAIL)
               fprintf(stderr, "Cannot create the output SDS: %s\n", out_sds_name); 	
	    else
	    {       
	      nobs_in_c = layer_num[irun];
	      nobs_in_c1 = nobs_in_c - 1;
              for (ir=0, ip=0, skip_obs_ir=0; ir<nrows; ir++)
              {
        	for (ic=0, skip_obs_ic=0; ic<ncols; ic++, ip++)
        	{
                  switch(sds1_info.data_type)
                  {
                    case 20: ((int8 *)out_data)[ip] = FILL_VALUE_INT8;
                             break;
                    case 21: ((uint8 *)out_data)[ip] = FILL_VALUE_UINT8;
                             break;
                    case 22: ((int16 *)out_data)[ip] = FILL_VALUE_INT16;
                             break;
                    case 23: ((uint16 *)out_data)[ip] = FILL_VALUE_UINT16;
                             break;
                    case 24: ((int32 *)out_data)[ip] = FILL_VALUE_INT32;
                             break;
                    case 25: ((uint32 *)out_data)[ip] = FILL_VALUE_UINT32;
                             break;
                  }
		  nobs = sds_nobs[ip];
	          if (nobs != sds_info.fill_val) {
		    if (nobs > nobs_in_c) {
                      ic_obs = skip_obs_ir + skip_obs_ic + nobs_in_c1;
		      if(ic_obs < 0)
		      {
			fprintf(stderr, "ic_obs is negative, exit...\n");
			exit(0);
		      }
                      switch(sds1_info.data_type)
                      {
                	case 20: ((int8 *)out_data)[ip] = ((int8 *)sdsc_data)[ic_obs];
                        	 break;
                	case 21: ((uint8 *)out_data)[ip] = ((uint8 *)sdsc_data)[ic_obs];
                        	 break;
                	case 22: ((int16 *)out_data)[ip] = ((int16 *)sdsc_data)[ic_obs];
  	   	        	 break;
                	case 23: ((uint16 *)out_data)[ip] = ((uint16 *)sdsc_data)[ic_obs];
  	   	        	 break;
                	case 24: ((int32 *)out_data)[ip] = ((int32 *)sdsc_data)[ic_obs];
  		        	 break;
                	case 25: ((uint32 *)out_data)[ip] = ((uint32 *)sdsc_data)[ic_obs];
  		        	 break;
                      }
		    }
		    if(nobs >= 1) skip_obs_ic += nobs - 1;     /* MSZ 3/7/2018, add the if() condition to kill the bug */
                  }
        	} /* for ic . . .  */
		skip_obs_ir += sds_add_nobs[ir];
              } /* for ir . . .  */              
	      write_attr_fval(H5OUT, out_sds_info.sds_id, sds1_info.data_type, c, attr_val);
	      /*if(set_sel_sds_data(&out_sds_info, start1, edge1, out_data) == FAIL)*/
	      if(set_sel_l2g_sds_2D_data_chunked_gzip(&out_sds_info, start1, edge1, out_data) == FAIL)
		fprintf(stderr, "Cannot write sds data to output file in get_l2g_datac_by_obs\n");
	    }
	    close_sds(&out_sds_info, out_sds_info.H5OUT);
          } 
	} /* for irun . .  */
	free(out_data);
      }
    } 
    if (status1 != -1) if(sds1_data) free(sds1_data);
    if(read_sds1 == 1) close_sds(&sds1_info, sds1_info.H5IN);  /* MSZ 3/6/2018, add the if() */
 
    if (status2 != -1) if(sdsc_data) free(sdsc_data);   
    if (read_sdsc == 1 && sdsc_info.sds_id != -1) close_sds(&sdsc_info, sdsc_info.H5IN); /* MSZ 3/6/2018, add the if() */
  }  
 
  free(sds_add_nobs);
  free(sds_nobs);
  if (*m_opt == 1 && status != -1)
  {
    copy_metadata(sds_info.H5IN, sds_info.sd_id, H5OUT, out_sd_id);
    *m_opt = 0;  /* MSZ 2/21/2018, just write meta data once to the output file */
  }

  close_sd(&sds_info, sds_info.H5IN);
}

int read_input_sds_names_ress(char *fname, char **in_sds_names, int sds_cnt, char **out_sds_names, int *res)
{
  int H5IN, i, cnt;
  int32 sd_id;
  int nsds = 0;
  sds_t sds_info;
  char tmp_sds_name[MAX_SDS_NAME_LEN];

  if (sds_cnt == 0)
    nsds = get_l2g_sds_names(fname, out_sds_names);
  else  /* user specifies without _1 or _c in the names */
  {
    /* MSZ 2/15/2018, modify to use sds_name_existence() */
    for (i=0, cnt=0; i<sds_cnt; i++)
    {
      sprintf(tmp_sds_name, "%s_1", in_sds_names[i]);
      if(sds_name_existence(fname, tmp_sds_name) == 0)
        fprintf(stderr, "Ignoring invalid SDS name %s\n", in_sds_names[i]);
      else {
        strcpy(out_sds_names[cnt], in_sds_names[i]);
        cnt++;
      }
    }
    
    nsds = cnt;
  }

  /* MSZ 2/15/2018, update to HDF4/HDF5 wrapper functions */
  sds_info.sd_id = -1;
  if(open_sd(fname, &sds_info, 'R') == FAIL)
    fprintf(stderr, "Cannot open the input hdf file %s in read_input_sds_names_ress()\n", fname);
  else {
    H5IN = sds_info.H5IN;
    sd_id = sds_info.sd_id;
    for (i=0; i<nsds; i++) {
      sds_info.H5IN = H5IN;
      sds_info.sd_id = sd_id;
      sds_info.sds_id = -1;
      sprintf(sds_info.name, "%s_1", out_sds_names[i]);
      if (get_sds_info((char *)NULL, &sds_info) != -1) {
#if 0
	if (sds_info.dim_size[0] == 1200) res[i] = 1;               /* 1km */
	else if (sds_info.dim_size[0] == 2400) res[i] = 2;          /* 500m */
	else if (sds_info.dim_size[0] == 4800) res[i] = 4;          /* 250m */
#endif
	if ((int)(sds_info.dim_size[0] * 0.001) == 1) res[i] = 1;               /* 1km */
	else if ((int)(sds_info.dim_size[0] * 0.001) == 2) res[i] = 2;          /* 500m */
	else if ((int)(sds_info.dim_size[0] * 0.001) == 4) res[i] = 4;          /* 250m */

	close_sds(&sds_info, sds_info.H5IN);
      }
    }
    close_sd(&sds_info, sds_info.H5IN);
  }

  return nsds;
}

void get_layer_numbers(char **arg_list, int max_nobs, char **num_str, int nl)
{
  int sds_cnt;
  int i, k, p1, ln;
  int ln1, ln2, len;
  char ln_str[10];

  if (nl == 0)
  {
    sprintf(arg_list[4], "%d", max_nobs);
    for (k=0; k<max_nobs; k++)
      sprintf(arg_list[5+k], "%d", k);
  }
  else
  {
    for (i=0, k=0; i<nl; i++)
    {
      if ((p1 = charpos(num_str[i], '-', 0)) == -1)
      {
        ln = atoi(num_str[i]);
	ln = ln - 1;
        if((ln < max_nobs) && (ln >= 0))
        {
	  sprintf(arg_list[5+k], "%d", ln);
          k++;
        }
        else
          fprintf(stderr, "Ignoring invalid layer %d (max_layers = %d)\n", ln+1, max_nobs);
      }
      else
      {
	len = (int)strlen(num_str[i]);
	strmid(num_str[i], 0, p1, ln_str);
	ln1 = (int)atoi(ln_str);
	ln1 = ln1 - 1;
	strmid(num_str[i], p1+1, len-p1-1, ln_str);
	ln2 = (int)atoi(ln_str);
	ln2 = ln2 - 1;		
	for (ln=ln1; ln<=ln2; ln++)
	{
	  if((ln < max_nobs) && (ln >= 0))
	  {	    
	    sprintf(arg_list[5+k], "%d", ln);
	    k++;
	  }
          else
            fprintf(stderr, "Ignoring invalid layer %d (max_layers = %d)\n", ln+1, max_nobs);
	}
      }
    }
    sprintf(arg_list[4], "%d", k);
  }
}

int get_l2g_max_obs_layers(char *hdf_fname, char *max_obs_sds_name)
{
   /* Maosheng Zhao, SSAI, 2/16/2018 */
   sds_t in_sds_info;
   void *sds_data = NULL;
   int i;
   int32 start[4], edge[4];
   int irow, icol;
   int nrows, ncols;
   int GROWS, last_rows;
   int nlines_a_copy = 20;  /* to speed up */
   int num_layers = -1;     /* fill value of num_observations */
    
   in_sds_info.sd_id = in_sds_info.sds_id = -1;
   if(open_sd(hdf_fname, &in_sds_info, 'R') == FAIL)
     fprintf(stderr, "Cannot open the HDF file %s\n", hdf_fname);
   else
   {
     strcpy(in_sds_info.name, max_obs_sds_name);
     if(get_sds_info((char *)NULL, &in_sds_info) != FAIL)
     {
	for (i=0; i<in_sds_info.rank; i++)
	{
          start[i] = 0;
          edge[i] = in_sds_info.dim_size[i];
	}
	if (in_sds_info.rank != 1) 
	{
	  if(in_sds_info.dim_size[in_sds_info.row_id] < nlines_a_copy) nlines_a_copy = in_sds_info.dim_size[in_sds_info.row_id];
          edge[in_sds_info.row_id] = 1 * nlines_a_copy;
	}

	nrows = (in_sds_info.rank == 1) ? 1 : in_sds_info.dim_size[in_sds_info.row_id];
	ncols = in_sds_info.dim_size[in_sds_info.row_id + 1];
            
	/* MSZ 1/25/2018 */
	GROWS = (int)ceil((double)nrows / (double)nlines_a_copy);
	last_rows = nrows % nlines_a_copy;

	/* MSZ, 1/25/2018, to speedup, copy lines by lines not line by line */
        for (irow=0; irow<GROWS; irow++)
        {
	  start[in_sds_info.row_id] = irow * nlines_a_copy;
	  /* MSZ, 1/25/2018, be cautious about the last group rows because it may be different from nlines_a_copy */
	  if(last_rows > 0 && irow == GROWS - 1)
	    edge[in_sds_info.row_id] = last_rows;

	  if ((sds_data = get_sel_sds_data(&in_sds_info, start, edge)) == NULL)
	    fprintf(stderr, "Error reading lines from %d to %d from SDS %s\n", irow * nlines_a_copy, irow * nlines_a_copy + nlines_a_copy, in_sds_info.name);
	  else
	  {
	    switch(in_sds_info.data_type)
	    {
	      case 20:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((int8 *)sds_data)[icol] > (int8)num_layers)
		    num_layers = (int)((int8 *)sds_data)[icol];
		break;
	      case 21:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((uint8 *)sds_data)[icol] > (uint8)num_layers)
		    num_layers = (int)((uint8 *)sds_data)[icol];
		break;
	      case 22:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((int16 *)sds_data)[icol] > (int16)num_layers)
		    num_layers = (int)((int16 *)sds_data)[icol];
		break;
	      case 23:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((uint16 *)sds_data)[icol] > (uint16)num_layers)
		    num_layers = (int)((uint16 *)sds_data)[icol];
		break;
	      case 24:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((int32 *)sds_data)[icol] > (int32)num_layers)
		    num_layers = (int)((int32 *)sds_data)[icol];
		break;
	      case 25:
		for (icol=0; icol<ncols * edge[in_sds_info.row_id]; icol++)
		  if(((uint32 *)sds_data)[icol] > (uint32)num_layers)
		    num_layers = (int)((uint32 *)sds_data)[icol];
		break;
	      default:
		fprintf(stderr, "The SDS data type should be integer for get_l2g_max_obs_layers().\n");
	    }
	    if(sds_data)
	    {
	      free(sds_data);
	      sds_data = NULL;
	    }
	  }
        }
	close_sds(&in_sds_info, in_sds_info.H5IN);
     }
     close_sd(&in_sds_info, in_sds_info.H5IN);
   }

   return num_layers;
}

