/****************************************************************************
!C

!File: read_meta.c

!Description:
  Contains routines for reading metadata from a hdf file 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:

  Version 1.0 August, 2002
  Version 2.0 March, 2013  
  Updated March, 2013 by Xiaoping Zhang, Sigma Space Corp. 	
  	(to work for MODIS and NPP products, merged read_meta.c and read_npp_meta.c) 
  Updated March, 2016 by Xiaoping Zhang, SSAI. 	
  	modified to get modis/viirs data status.
  Version 3.5.1 Jan, 2018
  Updated Jan., 2018 by Maosheng Zhao, SSAI.
	modified to deal with HDF5 file so that file attrs under root and other
	group objects can be printed out and also deal with cases that some previous
	file attrs in HDF4 can now dataset objects with H5T_STRING data type in copy_metadata()
	in meta.c.
  Modified Jan., 2018 Version 3.5.1 by Maosheng Zhao, SSAI
        1)modify it so that HDF5 can be read
        2)update copy_metadata() so that HDF5 meta data in DATASET object with H5T_STRING can be copied
	  besides these attrs under the group objects 
        3)make it can read attrs under all groups beside the root group 
	For details, please check meta.c

  Modified February, 2020 Version 3.6 by Maosheng Zhao, SSAI
1) updated the tool so that it can handle either ASCII or UTF8 for the strings. meta.c and sds_rw.c.
2) Added -H5meta argument as a switch to printout or not the CoreMetadata 
   and ArchiveMetadata in HDF5 datasets.
3) From line 565 to 614, fprintf() should use stdout not stderr for the three "metadata" datasets
   (CoreMetedata, ArchiveMetedata, and StructMetadata) of HDF5. This causes the failure of reading
   the screen prints to a text file.
4) Developed strpos_wb() in str_op.c and replace all strpos() with trpos_wb() 
   in get_all_metedata() in meta.c so that read_meta works for VNP13 too.
   Maybe UTF8 in VNP13 cause the issue.

	 
!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Yi (Sophyi) Zhang
    LDOPE                             Science Systems and Applications Inc.
    sophyi@ltpmail.gsfc.nasa.gov      NASA/GSFC Code 922
    phone: 301-614-5497               Greenbelt, MD 20771

    David Roy
    LDOPE                             University of Maryland, Department of Geography
    droy@kratmos.gsfc.nasa.gov        NASA/GSFC Code 922 (B32)
    phone: 301-614-5571               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "meta.h"
#include "sds_rw.h"
#include "str_op.h"
#include "qa_tool.h"
#include "alloc_mem.h"
#include "main_util.h"

#define HELP \
"NAME\n" \
"     read_meta - read metadata from any MODIS Land HDF-EOS data product. \n" \
"\n" \
"SYNOPSIS\n" \
"     read_meta -help\n" \
"     read_meta [-core|arch|struct|qa] [-meta=<metadata1>[<,metadata2>...]] [-case] filename(s)\n" \
"\n" \
"\n" \
"DESCRIPTION\n" \
"     Reads a set of specified metadata from MODIS Land HDF-EOS data product and writes the output to \n" \
"     stdout. If no options are specified, all metadata are read. Option -case allows the metadata \n" \
"     names to be case insensitive. \n" \
" \n" \
"     The tool command arguments can be specified in any order.  \n" \
"\n" \
"ARGUMENTS\n" \
"	-help			Display this help message\n" \
"	-core                  	Print core metadata for MODIS\n" \
"	-arch                  	Print archive metadata for MODIS\n" \
"       -struct                 Print struct metadata for MODIS\n" \
"	-qa                    	Print only QA metadata for MODIS\n" \
"	-meta=<metadata list>  	Print the specified metadata. This option is\n" \
"                            	ignored if one of the options above is used.\n" \
"                            	Metadata names are seperated by commas with no space. \n" \
"                               Note that -meta option doesn't work for -struct case. \n" \
"	-case                  	Make specified metadata name(s) case insensitive. \n" \
"                            	This option can be used only with -meta \n" \
"                            	option.   \n" \
"       -h5meta                 This option is used to turn on the printout of \n" \
"                               CoreMetadata and ArchiveMetadata HDF5/NetCDF4 if there are \n" \
"	filename(s)           	One or more input files separatd by space. Files are \n" \
"                              	processed in the order they are listed in the command line. \n" \
"\n" \
"EXAMPLES \n" \
"	read_meta MYD021KM.A2002189.0350.003.2002191000652.hdf \n" \
" \n" \
"	read_meta -arch MOD021KM.A2002189.0350.003.2002191000652.hdf > meta.txt \n" \
" \n" \
"	read_meta -core MYD09GHK.A2002177.h09v05.003.2002180140102.hdf \n" \
" \n" \
"       read_meta -struct MYD09GHK.A2002177.h09v05.003.2002180140102.hdf \n" \
" \n" \
"	read_meta -qa MYDPTHKM.A2002177.h09v05.003.2002180134421.hdf  \n" \
" \n" \
"	read_meta -meta=EQUATORCROSSINGDATE -case \n" \
"		MYD09GHK.A2002177.h09v05.003.2002180140102.hdf  \n" \
" \n" \
"	read_meta -meta=AUTOMATICQUALITYFLAG MOD09GHK.*.hdf \n" \
" \n" \
"	read_meta -meta=DAYNIGHTFLAG, GRINGPOINTLONGITUDE,  \n" \
"	  	GRINGPOINTLATITUDE MYD03.A2002*.hdf \n" \
" \n" \
"	read_meta -meta=DataRows,DataColumns,MaximumObservations NPP*.hdf\n" \
" \n" \
"	read_meta -meta=TileID,EastBoundingCoord,NorthBoundingCoord,WestBoundingCoord,SouthBoundingCoord NPP*.hdf \n" \
" \n" \
"AUTHOR \n" \
"	Code: S. Devadiga, X. Zhang, M. Zhao \n" \
"	Documentation: S.Devadiga and D. Roy \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"Usage:	read_meta [-core|arch|qa] [-meta=<metadata1>[<,metadata2>...]] [-case] filename(s)\n" \
" \n" \
"ARGUMENTS\n" \
"     -help			Display this help message\n" \
"     -core                  	Print core metadata for MODIS\n" \
"     -arch                  	Print archive metadata for MODIS\n" \
"     -struct                 Print struct metadata for MODIS\n" \
"     -qa                    	Print only QA metadata for MODIS\n" \
"     -meta=<metadata list>  	Print the specified metadata. This option is\n" \
"                            	ignored if one of the options above is used.\n" \
"                            	Metadata names are seperated by commas with no space. \n" \
"                               Note that -meta option doesn't work for -struct case. \n" \
"     -case                  	Make specified metadata name(s) case insensitive. \n" \
"                            	This option can be used only with -meta \n" \
"                            	option.   \n" \
"     -h5meta                   Print core metadata and archive metadata in HDF5/NetCDF4 if there are \n" \
"     filename(s)               One or more input files separatd by space. Files are \n" \
"                               processed in the order they are listed in the command line. \n" \
"\n" \
/*************************************************************************************
                            Prototypes.
**************************************************************************************/

int main(int argc, char **argv);
int parse_cmd_read_meta(int argc, char **argv, char **arg_list, 
		        int *arg_cnt, int *case_ch, int *H5meta);
/*************************************************************************************/

int main(int argc, char **argv)
/*
!C************************************************************************************
       
!Description:
  Main is the main program for the read_meta routine. It reads the metadata from a 
  hdf file.

!Input Parameters: (none)

!Output Parameters: 
  (returns)  Completion status: 
             0 - successful completion
	     1 - error exit
    
!Revision History: (none)

!Team-unique Header: (see file prolog)
 
!References and Credits: (see file prolog)

!Design Notes: (none)

!END
**************************************************************************************/

{
  int i, j;
  int case_ch, meta_c;                 /* case option and meta option */
  int arg_cnt;
  int fid, status, meta_cnt, st, meta_cnt_npp[MAX_NUM_META];           
  char *ameta_str = NULL, *cmeta_str = NULL, *smeta_str = NULL;         /* string for archive, core, and struct metadata */
  char **arg_list;                     /* argument list */
  char **meta_val, **meta_name;        /* metadata value and name */
  int32 sd_id, sds_id, dt;
  long attr_len;
  int32 nsds, nattr;
  int32 attr_cnt, attr_index;
  int32 attr_type, attr_size;
  char attr_name[40], *attr_buf;
  char *metadata;    
  char meta_name1[MAX_META_NAME_LEN]; 
  int HDF_ver = 0;
  int H5meta = 0;
  sds_t in_sds_info;

  /* MSZ 1/31/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 1/31/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
   
  /* To handle the case where there is no input file entered. */
  if (argc == 1)
    {    
      status = -1;     
      fprintf(stderr, "Missing input file \n");
      fprintf(stderr, "%s\n", USAGE);
      exit(EXIT_FAILURE);  
    }

  /* help option entered by user */
  if ((argc==2) && (strcmp(argv[1], "-help")==0))
    {
      fprintf(stderr, "%s\n", HELP);
      exit(EXIT_SUCCESS);
    }
  
  if ((argc>=3) && (strcmp(argv[1], "-help")==0))
    {
      fprintf(stderr, "%s\n", HELP);
      exit(EXIT_SUCCESS);
    } 

  /* allocate memory for arg_list */
  if ((arg_list = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for arg_list in read_meta(main)\n");
  if (arg_list != NULL)
  {
    /*fprintf(stdout, "Start processing read_meta\n");*/
    status = parse_cmd_read_meta(argc, argv, arg_list, &arg_cnt, &case_ch, &H5meta);
  }

  if (status == -1) fprintf(stderr, "%s\n", USAGE);
  else if (status != 0)
  { 

    /* allocate memory for meta_name */
    if ((meta_name = (char **)Calloc2D(MAX_NUM_META, MAX_META_NAME_LEN, sizeof(char))) == NULL)
      fprintf(stderr, "Cannot allocate memory for meta_name in read_meta(main)\n");
    /* allocate memory for meta_val */
    if ((meta_val = (char **)Calloc2D(MAX_NUM_META, MAX_META_VAL_LEN, sizeof(char))) == NULL)
      fprintf(stderr, "Cannot allocate memory for meta_val in read_meta(main)\n");
    if ((meta_name != NULL) && (meta_val != NULL))
    {
      for (fid=1; fid<argc; fid++)
      {
        if (argv[fid][0] == '-') continue;
	else 
	{

	  /* MSZ 2/21/2020, for special case of -struct */
	  if (strcmp(arg_list[0], "struct") == 0)
	  {
	     if ((smeta_str = get_attr_metadata(argv[fid], "StructMetadata.0")) != NULL)
	     {
		fprintf(stdout, "StructMetadata.0=\n%s\n", smeta_str);
		if(smeta_str) free(smeta_str);
		exit(0);
	     }
	     else
	     {
		/*fprintf(stderr, "There is no metadata StructMetadata.0 in %s\n", argv[fid]);	*/
		exit(0);
	     }

	  }
	  if ((metadata = get_attr_metadata(argv[fid], "CoreMetadata.0")) != NULL) st = 0;
	  else st = 1;	

	  if(HDF_version_check(argv[fid]) == 5)
	    st = 0;

	  if(HDF_version_check(argv[fid]) == 4 && H5meta)
	    fprintf(stderr, "-H5meta in the argument has no effect because %s is HDF4 not DHF5\n", argv[fid]);
	}

	

	if (st == 0)  /* there is "CoreMetadata" */
	{ 
            /* decide which meta option was selected by user */
            if (strcmp(arg_list[0], "all") == 0) meta_c = 0;
            else if (strcmp(arg_list[0], "qa") == 0) meta_c = 0;
            else if (strcmp(arg_list[0], "core") == 0) meta_c = 1;
            else if (strcmp(arg_list[0], "arch") == 0) meta_c = 2;
            else if (strcmp(arg_list[0], "sel") == 0) meta_c = 3;

	    /*
            fprintf(stderr, "\nReading metadata from hdf file %s \n", argv[fid]);
            fprintf(stdout, "Metadata from hdf file: %s\n", argv[fid]);
            fprintf(stdout, "---------------------------------------------------------------\n");
	    */

	    meta_cnt = 0;

	    switch(meta_c)
            {
	      /* core and archive metadata */
	    case 0: 
	      if(HDF_version_check(argv[fid]) == 4 || (HDF_version_check(argv[fid]) == 5 && H5meta))
	      {
                /* test to retrieve core metadata */ 
	        if ((cmeta_str = get_attr_metadata(argv[fid], "CoreMetadata.0")) != NULL)
		  get_all_metadata(cmeta_str, meta_name, meta_val, &meta_cnt);		      
	        /* test to retrieve Archive metadata */
	        if ((ameta_str = get_attr_metadata(argv[fid], "ArchiveMetadata.0")) != NULL) 
		  get_all_metadata(ameta_str, meta_name, meta_val, &meta_cnt);					

	        if ((cmeta_str != NULL) || (ameta_str != NULL))
		  {
		    if (strcmp(arg_list[0], "all") == 0)
		      {
		        for (i=0; i<meta_cnt; i++)
			  fprintf(stdout, "%s = %s\n", meta_name[i], meta_val[i]);
		      }
		    else if (strcmp(arg_list[0], "qa") == 0)   /* qa metadata option is selected */
		      {
		        for (i=0; i<meta_cnt; i++)
			  if (strpos(meta_name[i], "QUALITY", 0) != -1)
			    fprintf(stdout, "%s = %s\n",  meta_name[i], meta_val[i]);
			  else if (strpos(meta_name[i], "QA", 0) != -1)
			    fprintf(stdout, "%s = %s\n",  meta_name[i], meta_val[i]);
		      }
		  }
	        if (cmeta_str != NULL) free(cmeta_str);
	        if (ameta_str != NULL) free(ameta_str);
	      } /* end if(HDF_version_check(argv[fid]) == 4 || (HDF_version_check(argv[fid]) == 5 && H5meta)) */

	      if(HDF_version_check(argv[fid]) == 5)
	      {
		hid_t h5_id_tmp;
		if((h5_id_tmp = H5Fopen(argv[fid], H5F_ACC_RDONLY, H5P_DEFAULT)) >= 0)
		{
		  if (strcmp(arg_list[0], "all") == 0)
		    H5_print_all_attr_groups(h5_id_tmp, "");
		  else if (strcmp(arg_list[0], "qa") == 0)
		  {
		    H5_print_all_attr_groups(h5_id_tmp, "QUALITY");
		    H5_print_all_attr_groups(h5_id_tmp, "QA");
		  }

		  H5Fclose(h5_id_tmp);
		}
	      }

	      break; 

	    case 1: 
	      /* test to retrieve core metadata */
	      if ((cmeta_str = get_attr_metadata(argv[fid], "CoreMetadata.0")) != NULL)
		{
		  get_all_metadata(cmeta_str, meta_name, meta_val, &meta_cnt);
		  for (i=0; i<meta_cnt; i++)
		    fprintf(stdout, "%s = %s\n", meta_name[i], meta_val[i]);
		  free(cmeta_str);
		}
	      break; 

	    case 2: 
	      /* test to retrieve archive metadata */
	      if ((ameta_str = get_attr_metadata(argv[fid], "ArchiveMetadata.0")) != NULL)
		{
		  get_all_metadata(ameta_str, meta_name, meta_val, &meta_cnt);
		  for (i=0; i<meta_cnt; i++)
		    fprintf(stdout, "%s = %s\n", meta_name[i], meta_val[i]);
		  free(ameta_str);
		}
	      break;

	    case 3: 
	      /* user input metadata name list */
	      cmeta_str = get_attr_metadata(argv[fid], "CoreMetadata.0");
	      ameta_str = get_attr_metadata(argv[fid], "ArchiveMetadata.0");
	      if ((ameta_str != NULL) || (cmeta_str != NULL))
		{
                	for (i=1; i<arg_cnt; i++)
			  {
			    meta_cnt = 0;
			    if (cmeta_str != NULL) 
			      get_sel_metadata(cmeta_str, arg_list[i], meta_val, &meta_cnt, case_ch);
			    if (ameta_str != NULL) 
			      get_sel_metadata(ameta_str, arg_list[i], meta_val, &meta_cnt, case_ch);
			    fprintf(stdout, "%s", arg_list[i]);
			    if (meta_cnt == 0)
			      fprintf(stdout, "\tMetadata not found");
			    else
			      {
				for (j=0; j<meta_cnt; j++)
				  fprintf(stdout, "\t%s", meta_val[j]);
			      }
			    fprintf(stdout, "\n");
			  }
		}
	      if (cmeta_str != NULL) free(cmeta_str);
	      if (ameta_str != NULL) free(ameta_str);
	      break; 
            }
	}  
	else  /* no "CoreMetadata" */
	{
	  for (i=0; i<MAX_NUM_META; i++)    
            meta_cnt_npp[i] = 0;

          /* decide which meta option was selected by user */
          if (strcmp(arg_list[0], "sel") == 0) meta_c = 0;
          else if (strcmp(arg_list[0], "all") == 0) meta_c = 1; 

	  /*
          fprintf(stderr, "\nReading metadata from hdf file %s \n", argv[fid]);
          fprintf(stdout, "Metadata from hdf file of %s\n", argv[fid]);
          fprintf(stdout, "---------------------------------------------------------------\n");        
	   */

	  /* initialize as -1 for sd_id and sds_id */
	  in_sds_info.sd_id = in_sds_info.sds_id = -1;

	  switch(meta_c)
          {
	    /* core and archive metadata */
	    case 0: 	  	  
	      for (i=0; i<arg_cnt-1; i++)
	      {
		strcpy(meta_name[i], arg_list[1+i]);
		strtrim(meta_name[i]);
//		printf("%s\n", meta_name[i]);
	      }

	      if(open_sd(argv[fid], &in_sds_info, 'R') == FAIL)
        	fprintf(stderr, "Cannot open the HDF file %s\n", argv[fid]);
              else
              {        
		fprintf(stderr, "\tWriting metadata to output\n");
		if(in_sds_info.H5IN)
		  /* print out attrs from all groups */
		  H5_print_all_attr_groups(in_sds_info.sd_id, "");
		else  /* for HDF4 */
		{
		  SDfileinfo(in_sds_info.sd_id, &nsds, &nattr);	      
		  for (attr_index=0; attr_index<nattr; attr_index++)
		  {	      		    
		    if (SDattrinfo(in_sds_info.sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
		      fprintf(stdout, "Cannot read information for attribute %d\n", attr_index);
		    else
		    {		    
		      for (i=1; i<arg_cnt; i++)
		      {		      
		        status = strcasecmp(meta_name[i-1], attr_name);	    
		        if (status == 0)
		        {	 	      		      		
			  attr_size = DFKNTsize(attr_type);
			  attr_len = attr_size*attr_cnt;
			  if ((attr_type == DFNT_CHAR) || (attr_type == DFNT_CHAR8))
        		    attr_len++;
			  if ((attr_buf = (void *)calloc(attr_size, attr_len)) == NULL)
        		    fprintf(stdout, "Cannot allocate memory for attr_buf in copy_metadata\n");
			  else
			  {
        		    if (SDreadattr(in_sds_info.sd_id, attr_index, (VOIDP)attr_buf) == FAIL)
        		      fprintf(stdout, "Cannot read %s from input file\n", attr_name);        	      
        		    else 
        		    {
			      switch (attr_type)
			      {			 			        
			        case 4: fprintf(stdout, "%s:\"%s\"\n",attr_name, attr_buf);
				      break; 		
			        case 5: fprintf(stdout, "%s:%f\n", attr_name, ((float32 *)attr_buf)[0]);
				      break; 
			        case 6: fprintf(stdout, "%s:%f\n", attr_name, ((float64 *)attr_buf)[0]);
				      break; 	     
			        case 20: fprintf(stdout, "%s:%d\n", attr_name, ((int8 *)attr_buf)[0]);
				       break; 
			        case 21: fprintf(stdout," %s:%d\n", attr_name, ((uint8 *)attr_buf)[0]);
				       break; 
			        case 22: fprintf(stdout, "%s:%d\n", attr_name, ((int16 *)attr_buf)[0]);
				       break; 
			        case 23: fprintf(stdout, "%s:%d\n", attr_name, ((uint16 *)attr_buf)[0]);
				       break; 
			        case 24: fprintf(stdout, "%s: ", attr_name);
			               if (attr_cnt == 1) fprintf(stdout, "%d\n", ((int32 *)attr_buf)[0]);
				       else
				       {
			        	 for (j=0; j<attr_cnt-1; j++)
					   fprintf(stdout, "%d, ", ((int32 *)attr_buf)[j]);
					 fprintf(stdout, "%d\n", ((int32 *)attr_buf)[attr_cnt-1]); 				  	
				       }			    			    
			   	       break; 			    
			        case 25: fprintf(stdout, "%s:%d\n", attr_name, ((uint32 *)attr_buf)[0]);
				       break; 
			      }			
			      free(attr_buf);
        		    }
			  }	
		        }
		        else meta_cnt_npp[i-1]++;
		      } /* for (i=1; .. */		      
		    } /* else SDattrinfo() */
		  } /* for (attr_index=0; attr_index<nattr; attr_index++) */
		} /* else */ /* for HDF4 */
		for (i=1; i<arg_cnt; i++)
		{	      
		  if (meta_cnt_npp[i-1] == nattr) 
	            fprintf(stderr, "Can not find the meta_name %s in file: %s\n", meta_name[i-1], argv[fid]);
		} 
		close_sd(&in_sds_info, in_sds_info.H5IN);    	    	    	  
	      }
            break;
	    case 1:
	      if(open_sd(argv[fid], &in_sds_info, 'R') == FAIL)
                fprintf(stderr, "Cannot open the HDF file %s\n", argv[fid]);
              else
              {        
		fprintf(stderr, "\tWriting metadata to output\n");	

		if(in_sds_info.H5IN)
		  /* print out attrs from all groups */
		  H5_print_all_attr_groups(in_sds_info.sd_id, "");
		else  /* for HDF4 */
		{
		  SDfileinfo(in_sds_info.sd_id, &nsds, &nattr);
		  for (attr_index=0; attr_index<nattr; attr_index++)
		  {
		    if (SDattrinfo(in_sds_info.sd_id, attr_index, attr_name, &attr_type, &attr_cnt) == FAIL)
		      fprintf(stdout, "Cannot read information for attribute %d\n", attr_index);
		    else
		    {
		      attr_size = DFKNTsize(attr_type);
		      attr_len = attr_size*attr_cnt;
		      if ((attr_type == DFNT_CHAR) || (attr_type == DFNT_CHAR8))
        	        attr_len++;
		      if ((attr_buf = (void *)calloc(attr_size, attr_len)) == NULL)
        	        fprintf(stdout, "Cannot allocate memory for attr_buf in copy_metadata\n");
		      else
		      {
        	        if (SDreadattr(in_sds_info.sd_id, attr_index, (VOIDP)attr_buf) == FAIL)
        		  fprintf(stdout, "Cannot read %s from input file\n", attr_name);
        	        else
        	        {
  /*		      fprintf(stdout, "attr_type=%d attr_size=%d attr_cnt=%d \n",attr_type,attr_size, attr_cnt);*/
			  switch (attr_type)
			  {			 			        
			    case 4: fprintf(stdout, "%s:\"%s\"\n",attr_name, attr_buf);
				  break; 		
			    case 5: fprintf(stdout, "%s:%f\n", attr_name, ((float32 *)attr_buf)[0]);
				  break; 
			    case 6: fprintf(stdout, "%s:%f\n", attr_name, ((float64 *)attr_buf)[0]);
				  break; 	     
			    case 20: fprintf(stdout, "%s:%d\n", attr_name, ((int8 *)attr_buf)[0]);
				   break; 
			    case 21: fprintf(stdout, "%s:%d\n", attr_name, ((uint8 *)attr_buf)[0]);
				   break; 
			    case 22: fprintf(stdout, "%s:%d\n", attr_name, ((int16 *)attr_buf)[0]);
				   break; 
			    case 23: fprintf(stdout, "%s:%d\n", attr_name, ((uint16 *)attr_buf)[0]);
				   break; 
			    case 24: fprintf(stdout, "%s:", attr_name);
			           if (attr_cnt == 1) fprintf(stdout, "%d\n", ((int32 *)attr_buf)[0]);
				   else
				   {
			             for (i=0; i<attr_cnt-1; i++)
				       fprintf(stdout, "%d, ", ((int32 *)attr_buf)[i]);
				     fprintf(stdout, "%d\n", ((int32 *)attr_buf)[attr_cnt-1]); 				  	
				   }
				   break; 
			    case 25: fprintf(stdout, "%s:%d\n", attr_name, ((uint32 *)attr_buf)[0]);
				   break; 
			  }     		      	
			  free(attr_buf);
		        }
		      }
		    }
		  } /* for (attr_index=0; . . .      */	
		} /* else */ /* for HDF4 */
		close_sd(&in_sds_info, in_sds_info.H5IN);
	      }	
	    break;  
          }
	}

	/* MSZ 4/16/2018 for -meta= of both HDF4/HDF5 */
	if( arg_cnt > 1)
	{
	   for(j = 1; j < arg_cnt; j++)
	   {
		if(strstr(arg_list[j], "CoreMetadata") != NULL)
	  	  if((cmeta_str = get_attr_metadata(argv[fid], "CoreMetadata.0")) != NULL)
		  {
	    	    fprintf(stdout, "CoreMetadata.0 = \n%s\n", cmeta_str);
		    if(cmeta_str) free(cmeta_str);
		  }
		
		if(strstr(arg_list[j], "ArchiveMetadata") != NULL)
	  	  if((ameta_str = get_attr_metadata(argv[fid], "ArchiveMetadata.0")) != NULL) 
		  {
	    	    fprintf(stdout, "ArchiveMetadata.0 = \n%s\n", ameta_str);
		    if(ameta_str) free(ameta_str);
		  }

		if(strstr(arg_list[j], "StructMetadata") != NULL)
	  	  if((smeta_str = get_attr_metadata(argv[fid], "StructMetadata.0")) != NULL)
		  {
	    	    fprintf(stdout, "StructMetadata.0 = \n%s\n", smeta_str);
		    if(smeta_str) free(smeta_str);
		  }
	   }
	}
      }/* for fid */  
    } 
    Free2D((void **)meta_val);
    Free2D((void **)meta_name);
  }
  
  Free2D((void **)arg_list);  
  /*fprintf(stderr, "Processing done ! \n");*/
  return 0;
}


int parse_cmd_read_meta(int argc, char **argv, char **arg_list, int *arg_cnt, int *case_ch, int *H5meta)
     
/*
!C*********************************************************************************

!Function: parse_cmd_read_meta
       
!Description:

  Parse the command line arguments into corresponding fields.

!Input Parameters: 
  argc      input argument count
  argv      input argument vector

!Input/output Parameters:

  arg_list  contains a list of user entered metadata names
  arg_cnt   count of arg_list
  case_ch   if "-case" option is specified by user, case_ch is set to 1, all meta 
            data name will be case insensitive and converted to uppercase letters. 
            Otherwise case_ch is set to 0, and meta data name is case sensitive.
Maosheng Zhao, added -struct case to the argument

!Output Parameters:

  status:    1 - successful completion
	     0 - error exit   

!Revision History: (see file prolog) 

!Team-unique Header: (see file prolog)
 
!References and Credits: (see file prolog)

!Design Notes: (none)

!END
*****************************************************************************/

{
  int i;                /* loop variable */
  int m_cnt;            /* count of user input metadata name */
  int status;           /* return status */
  int if_cnt;           /* input file count */
  char r_mode[10];      /* option mode */
  char **meta_names;    /* metadata names entered by user */

  /* initialize variables */ 
  status = 1;           
  m_cnt = 0;          
  if_cnt = 0;
  *case_ch = 0;
  *arg_cnt = 0;
  r_mode[0] = '\0';
  *H5meta = 0;

  /* allocate memory for meta_name */
  if ((meta_names = (char **)Calloc2D(MAX_NUM_META, MAX_META_NAME_LEN, sizeof(char)))
      == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for meta_name in parse_cmd_read_meta\n");
      return 0;
    }
  for (i=1; i<argc; i++)
    {
      /* test to see if case option is selected */
      if ((strcmp(argv[i], "-case") == 0) || (strcmp(argv[i], "-case") == 0)) 
	{
	  *case_ch = 1;
	}
      else if (strcmp(argv[i], "-H5meta") == 0)
      {
	 *H5meta = 1;
      }
      else if (strcmp(argv[i], "-core") == 0)
	/* core metadata selected */
	{
	  if (r_mode[0] != '\0')
	    {
	      if (strcmp(r_mode, "sel") != 0)
		{
		  fprintf(stderr, "Metadata read option already set to %s\n", r_mode);
		}
	      else 
		{ /* if metadata option is set, overwrite it */
		  strcpy(r_mode, "core");
		}
	    }
	  else strcpy(r_mode, "core");
	}
      else if (strcmp(argv[i], "-arch") == 0) 
	/* archive metadata selected */
	{
	  if (r_mode[0] != '\0')
	    {
	      if (strcmp(r_mode, "sel") != 0)
		{
		  fprintf(stderr, "Metadata read option already set to %s\n", r_mode);
		}
	      else 
		{ /* if metadata option is set, overwrite it */
		  strcpy(r_mode, "arch");
		}
	    }
	  else strcpy(r_mode, "arch");
	}
      else if (strcmp(argv[i], "-struct") == 0)    /* MSZ 2/21/2020 */
	/* archive metadata selected */
	{
	  if (r_mode[0] != '\0')
	    {
	      if (strcmp(r_mode, "sel") != 0)
		{
		  fprintf(stderr, "Metadata read option already set to %s\n", r_mode);
		}
	      else 
		{ /* if metadata option is set, overwrite it */
		  strcpy(r_mode, "struct");
		}
	    }
	  else strcpy(r_mode, "struct");
	}
      else if (strcmp(argv[i], "-qa") == 0)
	/* qa metadata selected */
	{
	  if (r_mode[0] != '\0') 
	    {
	      if (strcmp(r_mode, "sel") != 0)
		{
		  fprintf(stderr, "Metadata read option already set to %s\n", r_mode);
		}
	      else 
		{ /* if metadata option is set, overwrite it */
		  strcpy(r_mode, "qa");
		}
	    }
	  else strcpy(r_mode, "qa");
	} 
      else if (is_arg_id(argv[i], "-meta=") == 0)
	/* -meta option is selected, user input a list of metadata name */
	{
	  if(r_mode[0] != '\0')
	    fprintf(stderr, "Metadata read option already set to %s\n", r_mode);
	  else
	    {
	      strcpy(r_mode, "sel");
	      get_arg_val_arr(argv[i], meta_names, &m_cnt);
	    }
	}
      else if (argv[i][0] == '-')
	{
	  if (strcmp(argv[i], "-meta") == 0) 
	  {
	    fprintf(stderr, "Invalid option for -meta option \n");
	    fprintf(stderr, "Usage: -meta=<metadata list> \n"); 
	  } else 
	  {
	    fprintf(stderr, "Unknown option %s\n", argv[i]);
	  }	    
	      }
	  else
	    { 
	      if_cnt++;
	    }
	}
      
      /*printf("\nUser entered %d files for processing \n", if_cnt);*/
      /* test to see if user entered input file */
      if (if_cnt == 0)
	{
	  fprintf(stderr, "Missing input file name \n");
	  status = -1;
	}
      /* test to see if metadata names are supplied by user with -meta option */
      if ((strcmp(r_mode, "sel") == 0) && (m_cnt == 0))
	{
	  fprintf(stderr, "Missing metadata names in the option \n");
	  status = -1;
	} 
      if (status != -1)
	{
	  /* if no options are supplied, process all metadata */
	  if (r_mode[0] == '\0')
	    strcpy(arg_list[0], "all");
	  else 
	    strcpy(arg_list[0], r_mode);
	  if (m_cnt != 0)
	    /* loop through user meta_names list. Stored in meta_names array for later use. */ 
	    for (i=0; i<m_cnt; i++)
	      {
		strcpy(arg_list[1+i], meta_names[i]);
		strtrim(arg_list[1+i]);
	      }
	  *arg_cnt = 1 + m_cnt; 
	}
      Free2D((void **)meta_names);
      return status;
    }
