/****************************************************************************
!C

!File: read_proj_param.c

!Description:
  Read projection parameter from one or more L3 HDF file. 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original March 2001 Version 1.4

  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
        3)update all HDFEOS API to a higher-level APIs to handle 
          both HDFEOS2 and HDFEOS5 in eos_sds_rw.c

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)
    Need to copy HDFEOS structural metadata and bounding box.
    Somehow copy GRING and Bounding box metadata.

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/dir.h>         
#include <math.h>

#include "eos_sds_rw.h"
#include "qa_tool.h"
#include "str_op.h"
#include "sds_rw.h"

#define MAX_GRID_NAME_LEN 40

#define HELP \
"NAME \n" \
"     read_proj_param - Copy all hdfeos metadata to the input L3 HDF files. \n" \
" \n" \
"SYNOPSIS \n" \
"     read_proj_param -help \n" \
"     read_proj_param <input file/directory> \n" \
" \n" \
"DESCRIPTION \n" \
"     Print the projection parameter from the HDF-EOS metadata of the hdfeos file \n" \
"     or all the files in the input directory \n" \
" \n" \
"OPTIONS \n" \
"     -help              Display this help message \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage: read_proj_param <reference L3 tile file/directory> \n" \
" \n" \
"OPTIONS \n" \
"     -help                  Display this help message \n"

void get_hdfeos(char *file_name, grid_mapinfo *h, char *grid_name);
void print_proj_param(grid_mapinfo *h, char *tile_id);
void get_tile_id(char *fname, char *tile_id);

int main(int argc, char **argv)
{
  int iarg;
  DIR *dirp;
  struct direct *dp;  
  grid_mapinfo hdfeos;
  char tile_id[10];
  char in_dir[MAX_PATH_LENGTH];
  char in_fname[MAX_PATH_LENGTH];
  char grid_name[MAX_GRID_NAME_LEN];

  if ((strcmp(argv[1], "-h") == 0) || ( strcmp(argv[1], "-help") == 0))
    fprintf(stderr, "%s\n", HELP);
  else 
  {
    /* MSZ 2/10/2018 silent version mismatch warning standard print out */
    setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

    /* MSZ 2/10/2018 Disable error reporting */
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

    for (iarg=1; iarg<argc; iarg++)
    {
      strcpy(in_fname, argv[iarg]);
      if (HDF_version_check(in_fname))  /* MSZ 4/2/2018, change the previous Hishdf() to the new function */
      {
        if (get_grid_name(in_fname, grid_name) == 1)
        {
          get_tile_id(in_fname, tile_id);
          get_hdfeos(in_fname, &hdfeos, grid_name);
          print_proj_param(&hdfeos, tile_id);
        }
      }
      else
      {
        strcpy(in_dir, in_fname);
        if ((dirp = opendir(in_dir)) != NULL)
        {
          while ((dp = readdir(dirp)) != NULL)
          {
            if ((strcmp(dp->d_name, ".") != 0) && (strcmp(dp->d_name, "..") != 0))
	    {
              sprintf(in_fname, "%s/%s", in_dir, dp->d_name);
              if (Hishdf(in_fname) == TRUE)
	      {
                if (get_grid_name(in_fname, grid_name) == 1)
	        {
        	  get_tile_id(in_fname, tile_id);
                  get_hdfeos(in_fname, &hdfeos, grid_name);
	          print_proj_param(&hdfeos, tile_id);
	        }
	      }
	      else fprintf(stderr, "Not an HDF file %s\n", in_fname);
	    }
          }                     
          closedir(dirp);
        }
        else fprintf(stderr, "Cannot open HDF file or directory  %s\n", in_fname);
      }
    }
  }                      
  return 0;
}

void get_hdfeos(char *file_name, grid_mapinfo *h, char *grid_name)
{
  int i;
  hid_t file_id, grid_id;
  int he5_fmt;
  int HDF_ver = 0;

  HDF_ver = HDF_version_check(file_name);

  if( !(HDF_ver ==4 || HDF_ver == 5) )
  {
    printf("%s is neither HDF4 nor HDF5, exit...\n", file_name);
    exit(0);
  }

  if(HDF_ver == 4) he5_fmt = 0;
  else if(HDF_ver == 5) he5_fmt = 1;

  h->ns = (long)(-1);
  h->nl = (long)(-1);
  h->ulCornerGrid[0] = (float64)(-999.0);
  h->ulCornerGrid[1] = (float64)(-999.0);
  h->lrCornerGrid[0] = (float64)(-999.0);
  h->lrCornerGrid[1] = (float64)(-999.0);
  h->projCode = (int)(-1);
  h->sphereCode = (int)(-1);
  for (i = 0; i < 13; i++)
    h->projParameters[i] = (float64)(-999.0);
  h->zoneCode = (int)(-1);
  h->originCode = (int)(-1);

  if ((file_id = EOS_GDopen(file_name, 'R', he5_fmt)) == FAIL) 
  {
    fprintf(stderr, "Input file %s open failure (EOS_GDopen) in get_hdfeos\n", file_name);
    return;
  }

  if ((grid_id = EOS_GDattach(file_id, grid_name, he5_fmt)) == FAIL)
    fprintf(stderr, "Can't attach to %s (EOS_GDattach) in get_hdfeos\n", grid_name);
  else
  {
    if (EOS_GDgridinfo(grid_id, &h->ns, &h->nl, h->ulCornerGrid, h->lrCornerGrid, he5_fmt) == FAIL)
      fprintf(stderr, "Can't get grid info (EOS_GDgridinfo) in get_hdfeos\n");
    
    if (EOS_GDprojinfo(grid_id, &h->projCode, &h->zoneCode, &h->sphereCode, h->projParameters, he5_fmt) == FAIL)
      fprintf(stderr, "Can't get projection info (EOS_GDprojinfo) in get_hdfeos\n");
    else
      h->projParameters[8] = h->projParameters[8]*4.0;

    if (EOS_GDorigininfo(grid_id, &h->originCode, he5_fmt) == FAIL)
      fprintf(stderr,  "Can't get origin defintion (EOS_GDorigininfo) in get_hdfeos\n");

    if (EOS_GDdetach(grid_id, he5_fmt) == FAIL)
      fprintf(stderr, "Can't detach from grid (EOS_GDdetach) in get_hdfeos\n");
  }

  if (EOS_GDclose(file_id, he5_fmt) == FAIL)
    fprintf(stderr, "Can't close file (EOS_GDclose) in get_hdfeos\n");
}

void print_proj_param(grid_mapinfo *h, char *tile_id)
{
  int i;

  fprintf(stdout, "%s ", tile_id);  
  fprintf(stdout, "%lf %lf ", h->ulCornerGrid[0], h->ulCornerGrid[1]);  
  fprintf(stdout, "%lf %lf ", h->lrCornerGrid[0], h->lrCornerGrid[1]);  
  fprintf(stdout, "%d %d %d %d ", h->projCode, h->sphereCode, h->zoneCode, h->originCode);  
  for (i=0; i<12; i++)
    fprintf(stdout, "%lf ",  h->projParameters[i]);  
  fprintf(stdout, "%lf\n", h->projParameters[12]);
}

void get_tile_id(char *fname, char *tile_id)
{
  int p1, p2;
  int i, len;
  char gid[80];

  tile_id[0] = '\0';
  len = (int)strlen(fname);
  for (i=len-1; i>=0; i--)
    if (fname[i] == '/') break;
  strmid(fname, i+1, len-i-1, gid);
  p1 = charpos(gid, '.', 0);
  if (p1 != -1)
    p2 = charpos(gid, '.', p1+1);
  if ((p1 != -1) && (p2 != -1))
    strmid(gid, p2+1, 6, tile_id);
}


