/****************************************************************************
!C

!File: read_sds_attributes.c

!Description:
  Read attributes of one or more SDS of MODIS Land HDF-EOS data product
  
!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:

  Version 1.0    April 5, 2003

  Modified April, 2018 Version 3.5.1 by Maosheng Zhao, SSAI
        1)modify it so that HDF5 can be read
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	3)develope H5_SDattrinfo() and H5_SDreadattr() to mimic HDF4 API for SDattr

  Modified February, 2020 Version 3.6 by Maosheng Zhao, SSAI
1) updated the tool so that it can handle either ASCII or UTF8 for the strings. meta.c and sds_rw.c.

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Yi Zhang 
    LDOPE                             Science Systems and Applications Inc.
    sophyi@ltpmail.gsfc.nasa.gov      NASA/GSFC Code 922
    phone: 301-614-5497               Greenbelt, MD 20771

    David Roy 
    LDOPE                             University of Maryland 
    droy@kratmos.gsfc.nasa.gov	      NASA/GSFC Code 922
    phone: 301-614-5571               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "meta.h"
#include "l2g.h"

#define MAX_NUM_VALUES 16000

#define HELP \
"NAME \n" \
"    read_sds_attributes - read attributes of one or more SDSs of input\n"\
"    MODIS Land HDF-EOS data product \n" \
" \n" \
" \n" \
"SYNOPSIS \n" \
"    read_sds_attributes -help [filename]\n" \
"    read_sds_attributes [-sds=<sds_name>] filename \n" \
" \n" \
"DESCRIPTION \n" \
"    Read attributes of one or more SDSs of input MODIS Land HDF-EOS data \n" \
"    product and output the result to stdout. The SDS attributes include: \n" \
"    fill values, units, scaling and offsets values, SDS long name, etc. \n" \
" \n" \
"OPTIONS \n" \
"    -help                    Display this help message, If the input\n"\
"                             filename is specified with this option, then\n"\
"                             the names of all the SDSs in the file are\n"\
"                             displayed. \n" \
"    -sds=<SDS_list>          List of SDS to process. SDS names are\n"\
"                             separated by commas with no space. By default\n"\
"                             attributes for all SDSs in the input file are\n"\
"                             output. \n" \
"    filename                 Input filename. \n" \
" \n" \
"Examples: \n" \
"    read_sds_attributes \n" \
"        -sds='500m Surface Reflectance Band 1,\n"\
"              500m Surface Reflectance Band 3,\n"\
"              500m Surface Reflectance Band 4'\n"\
"              MOD09.A2002123.0040.003.2002125174437.hdf  \n" \
"    read_sds_attributes MOD35_L2.A2002123.0040.003.2002124023706.hdf \n" \
"\n"\
"AUTHOR \n" \
"    Code: S. Devadiga and Yi Zhang \n" \
"    Documentation: S. Devadiga and D. Roy \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:	\n" \
"    read_sds_attributes -help [filename]\n" \
"    read_sds_attributes [-sds=<sds_name>] filename \n" \
"OPTIONS \n" \
"    -help                    Display this help message, If the input\n"\
"                             filename is specified with this option, then\n"\
"                             the names of all the SDSs in the file are\n"\
"                             displayed. \n" \
"    -sds=<SDS_list>          List of SDS to process. SDS names are\n"\
"                             separated by commas with no space. By default\n"\
"                             attributes for all SDSs in the input file are\n"\
"                             output. \n" \
"    filename                 Input filename. \n" \
" \n"

void read_sds_attr(char *fname, char **sds_names, int32 nsds); 
void print_sds_values(sds_t *sds_info, int **sds_values, float **sds_fvalues, int *num_val);
int parse_cmd_read_sds_attr(int argc, char **argv, char *in_fname, char **sds_names, int* sds_cnt);

int main(int argc, char **argv)
/*
!C*********************************************************************************

!read_sds_attr
       
!Description:
   read_sds_attributes display the attribute name, type and value of SDS(s) to 
   the screen.

!Revision History: (see file prolog) 

!Team-unique Header: (see file prolog)
 
!References and Credits: (see file prolog)

!Design Notes: (none)

!END
*****************************************************************************/
{
   char **sds_names_exclusion = NULL, **sds_names = NULL;
 
   int i, k, status = 0;
   int isds, sds_cnt;
   int32 sd_id, sds_id, msds, nattr, dim_size[4], dt;
   char in_fname[MAX_PATH_LENGTH];
   char name[MAX_SDS_NAME_LEN];
   char sds_name[MAX_SDS_NAME_LEN];
   char dim_str[MAX_STR_LEN];
	
   if (argc == 1)
   {
      status = -1;
      fprintf(stderr, "Missing input file \n");
      fprintf(stderr, "%s\n", USAGE);
      exit(EXIT_FAILURE);
   }   
  
   /* display help */
   if ((argc == 2) && ( strcmp(argv[1], "-help") == 0 || strcmp(argv[1], "-h") == 0 ))
   {
     	fprintf(stderr, "%s\n", HELP);
     	exit(0);
   }

   /* MSZ 4/6/2018 silent version mismatch warning standard print out */
   setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

   /* MSZ 4/6/2018 Disable error reporting */
   H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  
   /*  Display SDS names of input HDF file */
  
   if ((argc>=3) && ( strcmp(argv[1], "-help") == 0 || strcmp(argv[1], "-h") == 0 ))
   {
     /* Print all the SDS names in the input file */
     for (i=2; i<argc; i++)
       if (argv[i][0] != '-')
	  display_sds_info_of_file(argv[i]);
	
     exit(EXIT_SUCCESS);
   }
  
   if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
   {
    	fprintf(stderr, "Cannot allocate memory for sds_names in sds_range: main()\n");
    	exit(EXIT_FAILURE);
   }
   
   if (sds_names != NULL)
   {
        sds_cnt = 0;
    	status = parse_cmd_read_sds_attr(argc, argv, in_fname, sds_names, &sds_cnt);

	if (sds_cnt != 0)
	{
            msds = sds_cnt;
	}
        else
	{
	  if ((sds_names_exclusion = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	  {
	    fprintf(stderr, "Cannot allocate memory for sds_names_exclusion in main(), exit... \n");
	    exit(0);
	  }

	  strcpy(sds_names_exclusion[0], "StructMetadata.0");
	  strcpy(sds_names_exclusion[1], "CoreMetadata.0");
	  strcpy(sds_names_exclusion[2], "ArchiveMetadata.0");

	  msds = get_sds_names_exclusions(in_fname, sds_names, sds_names_exclusion, 3);

	  if(sds_names_exclusion) Free2D((void**)sds_names_exclusion);
	}

        if (status == -1)
        {
    	    fprintf(stderr, "%s", USAGE);
	    exit(EXIT_FAILURE);
        }
        else if (status != 0)
        {
	    read_sds_attr(in_fname, sds_names, msds);
        }
   }
   if(sds_names) Free2D((void **)sds_names);
   fprintf(stderr, "Processing done ! \n");
   /*return status;*/
}
 
   
void read_sds_attr(char *fname, char **sds_names, int32 nsds)

/*
!C*********************************************************************************

!Function: read_sds_attr
       
!Description:
   function read_sds_attr display the attribute name, type and value to 
   the screen.

!Input Parameters:

  fname     Input HDF file name.
  sds_names User input sds_name. 
  sds_cnt   Number of input sds_name.

!Revision History: (see file prolog) 

!Team-unique Header: (see file prolog)
 
!References and Credits: (see file prolog)

!Design Notes: (none)

!END
*****************************************************************************/

{
  int isds, j, m;
  int32 k, attr_size;
  int32 attr_type, attr_cnt;
  sds_t *sds_info;
  void *attr_buf = NULL;
  char sds_name[MAX_SDS_NAME_LEN];
  char attr_name[MAX_SDS_NAME_LEN];
  char tempstr[256];
  char path_str[256];
  int HDF_ver = 0;

  /* allocate memory for sds_info */
  if ((sds_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
    	fprintf(stderr, "Cannot allocate memory for sds_info in read_sds_attr()\n");
  if (sds_info != NULL)
  {
    HDF_ver = HDF_version_check(fname);
    if(!HDF_ver)
    {
      fprintf(stderr, "The input %s is neither HDF4 nor HDF5, exit...\n", fname);
      exit(0);
    }

    for (isds=0; isds<nsds; isds++)
    {
      if(isds == 0) sds_info[0].sd_id = -1;
      else 
      {
	sds_info[isds].sd_id = sds_info[0].sd_id;
	sds_info[isds].H5IN = sds_info[0].H5IN;
      }

      sds_info[isds].sds_id = -1;
      strcpy(sds_info[isds].name, sds_names[isds]);

      if (get_sds_info(fname, &sds_info[isds]) != FAIL)
      {
	fprintf(stdout, "===============================================================================\n");
	fprintf(stdout, "SDS : %s \n", sds_names[isds]);
	fprintf(stdout, "  %-21s%-10s%16s\n", "Attribute", "Data Type", "Value");
	fprintf(stdout, "----------------------+---------+----------------------------------------------\n");
	for(k=0; k < sds_info[isds].nattr; k++) /* k is the index of the attribute */
	{
	  if(attr_buf)
	  {
	    free(attr_buf);
	    attr_buf = NULL;
	  }
	  /* 1, read each attribute */
	  /* get each attrbute's name, type, count, and its values dependant on the HDF version */
	  if(HDF_ver == 4)
	  {
	    if (SDattrinfo(sds_info[isds].sds_id, k, attr_name, &attr_type, &attr_cnt) != FAIL)
	    { 
	      attr_size = DFKNTsize(attr_type);
              /* If data type is a string, add an extra space for null char */
              if( attr_type == 4 ) {
                  attr_buf = (void *)calloc(attr_cnt + 1, attr_size);
              } else {
                  attr_buf = (void *)calloc(attr_cnt, attr_size);
              }
	      if (attr_buf == NULL)
              {
		  fprintf(stderr, "Cannot allocate memory for attr_buf in read_sds_attr(), exit...\n");
		  exit(EXIT_FAILURE);		 
	      }

	      if (SDreadattr(sds_info[isds].sds_id, k, attr_buf) == FAIL)
	      {
		fprintf(stdout, "%-23s", attr_name);
		switch(attr_type){
		  case 4:
			fprintf(stdout, "%-10s", "CHAR");
			break;
		  case 3:
			fprintf(stdout, "%-10s", "UCHAR");
			break;
		  case 20:
			fprintf(stdout, "%-10s", "INT8");
			break;
		  case 21:
			fprintf(stdout, "%-10s", "UINT8");
			break;
		  case 22:
			fprintf(stdout, "%-10s", "INT16");
			break;
		  case 23:
			fprintf(stdout, "%-10s", "UINT16");
			break;
		  case 24:
			fprintf(stdout, "%-10s", "INT32");
			break;
		  case 25:
			fprintf(stdout, "%-10s", "UINT32");
			break;
		  case 5:
			fprintf(stdout, "%-10s", "FLOAT32");
			break;
		  case 6:
			fprintf(stdout, "%-10s", "FLOAT64");
			break;
		}
		fprintf(stdout, "%-10s\n", "unknown");
		if(attr_buf)
		{
		  free(attr_buf);
		  attr_buf = NULL;
		}
	      }
	    }
	  }
	  else if(HDF_ver == 5)
	  {
	    strcpy(path_str, ".");
	    if (H5_SDattrinfo(sds_info[isds].sds_id, path_str, k, attr_name, &attr_type, &attr_cnt) != FAIL)
	    { 
	      if(attr_type == 99 || attr_cnt == 0)
	      {
	         attr_buf = NULL;
		 fprintf(stdout, "%-23s", attr_name);
		 fprintf(stdout, "%-10s", "unknown");
		 fprintf(stdout, "%-10s\n", "unknown");
	      }
	      else
	      {
	        attr_size = DFKNTsize(attr_type);
                /* If data type is a string, add an extra space for null char */
                if(attr_type == 4) {
                  attr_buf = (void *)calloc(attr_cnt + 1, 1);
                } else {
                  attr_buf = (void *)calloc(attr_cnt, attr_size);
                }
	        if (attr_buf == NULL)
                {
		  fprintf(stderr, "Cannot allocate memory for attr_buf in read_sds_attr(), exit...\n");
		  exit(EXIT_FAILURE);		 
	        }
		strcpy(path_str, ".");
	        if (H5_SDreadattr(sds_info[isds].sds_id, path_str, k, attr_cnt, attr_buf) == FAIL)
	        {
		   fprintf(stdout, "%-23s", attr_name);
		   switch(attr_type){
		     case 4:
			fprintf(stdout, "%-10s", "CHAR");
			break;
		     case 3:
			fprintf(stdout, "%-10s", "UCHAR");
			break;
		     case 20:
			fprintf(stdout, "%-10s", "INT8");
			break;
		     case 21:
			fprintf(stdout, "%-10s", "UINT8");
			break;
		     case 22:
			fprintf(stdout, "%-10s", "INT16");
			break;
		     case 23:
			fprintf(stdout, "%-10s", "UINT16");
			break;
		     case 24:
			fprintf(stdout, "%-10s", "INT32");
			break;
		     case 25:
			fprintf(stdout, "%-10s", "UINT32");
			break;
		     case 5:
			fprintf(stdout, "%-10s", "FLOAT32");
			break;
		     case 6:
			fprintf(stdout, "%-10s", "FLOAT64");
			break;
		   }

		   fprintf(stdout, "%-10s\n", "unknown");

		   if(attr_buf)
		   {
		     free(attr_buf);
		     attr_buf = NULL;
		   }
	        } /* fail to read attr */
	      } /* else with valid H4 type and non 0 cnt */
	    }
	  }
	  else
	  {
	     fprintf(stderr, "file %s is not HDF4, HDF5 or NetCDF4 format, exit...\n", fname);
	     exit(0);
	  }

	  /* 2, print out each attribute */
	  if(attr_buf)
	  {
	    fprintf(stdout, "%-23s", attr_name);
	    switch(attr_type)
	    {
	      case  4:
                ((char *)attr_buf)[attr_cnt] = 0;
	        if (attr_cnt == 1 )
		  {
		    fprintf(stdout, "%-10s", "CHAR");
		    fprintf(stdout, "%-10s\n", (char *)attr_buf);
		  }
	        else
		  {
                    int hasSpecialFormatting = 0;
                    char *tptr = (char *)attr_buf;
                    for( m = 0; m < attr_cnt; m++, tptr++ ) {
                      if( *tptr == '\n' || *tptr == '\r' ) {
                        hasSpecialFormatting = 1;
                        break;
                    }
                  }

                  fprintf(stdout, "%-10s", "STRING");
		  if (attr_cnt < 255 && ! hasSpecialFormatting )
		    {
                      int first = 1;
                      tptr = (char *)attr_buf;
                      while( *tptr != 0 ) {
                        int len = (int)strlen( tptr );
                        if( len > 47 ) {
                          int tlen;
                          int found = 0;
                          len = 47;
                          for( tlen = len; tlen > 0; tlen-- ) {
                             if( *(tptr + tlen) == ' ' ) {
                                found = 1;
                                break;
                               }
                             }
                          if( found ) len = tlen;
                        }
                        strmid(tptr, 0, len, tempstr);
                        if( first ) {
                          fprintf( stdout, "%-s\n", tempstr);
                          first = 0;
                        } else {
                          fprintf(stdout,"%-33s%-s\n"," ",tempstr);
                        }
                        tptr += len;
                        if( *tptr == ' ' )
                             tptr++;
                      }
		    }
		  else 
		    {
		       fprintf(stdout, "\n");
		       fprintf(stdout, "Value =");
		       fprintf(stdout, "%s\n", (char *)attr_buf);
		    }
		}
		break;		      
	      case  5: 
		fprintf(stdout, "%-10s", "FLOAT32");
		fprintf(stdout, "%f", ((float32 *)attr_buf)[0]);
		for (j=1; j<attr_cnt; j++)
		{
		  fprintf(stdout, ", %f", ((float32 *)attr_buf)[j]);
		}
		fprintf(stdout,"\n");
		break;				        
	      case  6:   	
		fprintf(stdout, "%-10s", "FLOAT64");
		fprintf(stdout, "%f", ((float64 *)attr_buf)[0]);
		for (j=1; j<attr_cnt; j++)
		{
		  fprintf(stdout, ", %f", ((float64 *)attr_buf)[j]);
		}
		fprintf(stdout,"\n");
		break;
	      case 20: 
		fprintf(stdout, "%-10s", "INT8");
		fprintf(stdout, "%d", ((int8 *)attr_buf)[0]);
		for (j=1; j<attr_cnt; j++)
		{
		  fprintf(stdout, ", %d", ((int8 *)attr_buf)[j]);
		}
		fprintf(stdout,"\n");
		break;
	      case 21:
		fprintf(stdout, "%-10s", "UINT8");
		fprintf(stdout, "%d", ((uint8 *)attr_buf)[0]);
		for (j=1; j<attr_cnt; j++)
		{
		  fprintf(stdout, ", %d", ((uint8 *)attr_buf)[j]);
		}
		fprintf(stdout,"\n");
		break;
	      case 22: 
		fprintf(stdout, "%-10s", "INT16");
		fprintf(stdout, "%d", ((int16 *)attr_buf)[0]);
		for (j=1; j<attr_cnt; j++)
		{
		  fprintf(stdout, ", %d", ((int16 *)attr_buf)[j]);
		}
		fprintf(stdout,"\n");
		break;
	      case 23: 
		fprintf(stdout, "%-10s", "UINT16");
		fprintf(stdout, "%d", ((uint16 *)attr_buf)[0]);
		for (j=1; j<attr_cnt; j++)
		{
		  fprintf(stdout, ", %d", ((uint16 *)attr_buf)[j]);
		}
		fprintf(stdout,"\n");
		break;
	      case 24:
		fprintf(stdout, "%-10s", "INT32");
		fprintf(stdout, LONG_INT_FMT, ((int32 *)attr_buf)[0]);
		for (j=1; j<attr_cnt; j++)
		{
		  fprintf(stdout, ", " LONG_INT_FMT, ((int32 *)attr_buf)[j]);
		}
		fprintf(stdout,"\n");
		break;
	      case 25:
		fprintf(stdout, "%-10s", "UINT32");
		fprintf(stdout, LONG_INT_FMT, ((uint32 *)attr_buf)[0]);
		for (j=1; j<attr_cnt; j++)
		{
		  fprintf(stdout, ", " LONG_INT_FMT, ((uint32 *)attr_buf)[j]);
		}
		fprintf(stdout,"\n");
		break;
	      default: 
		fprintf(stdout, "%-10s\n", "Unknown");
		break;
	    } /* switch(attr_type) */

	    if(attr_buf)
	    {
	      free(attr_buf);
	      attr_buf = NULL;
	    }
	  } /* if(attr_cnt > 0) */

	} /* for(k=0; k < sds_info[isds].nattr; k++) */

	close_sds(&sds_info[isds], sds_info[isds].H5IN);
      } /* if (get_sds_info...)  */
    } /* for (isds=0; . . ) */
    close_sd(&sds_info[0], sds_info[0].H5IN);
  } /* if (sds_info != NULL) */	        
}

int parse_cmd_read_sds_attr(int argc, char **argv, char *in_fname, char **sds_names, int* sds_cnt)
/*
!C*********************************************************************************

!Function: parse_cmd_read_sds_attr
       
!Description:
   function parse_cmd_read_sds_attr parse the argument string to the corresponding
   input filename (in_fname), output filename (out_fname) and SDS name
   (sds_name).
   It print out a error message if there the input file, output file or 
   SDS name is missing. 

!Input Parameters: 
  argc      input argument count
  argv      input argument vector

!Input/output Parameters:

  in_fname  Input HDF file name.
  sds_names User input sds_name. 
  sds_cnt   Number of input sds_name.

!Output Parameters:

  ret   :    1  - successful completion
	     -1 - error exit   

!Revision History: (see file prolog) 

!Team-unique Header: (see file prolog)
 
!References and Credits: (see file prolog)

!Design Notes: (none)

!END
*****************************************************************************/
{
  int i;
  int ret;            /* return value */
  
  in_fname[0] = '\0';
  
  for (i=1, ret=1; i<argc; i++)
    {
      if (is_arg_id(argv[i], "-sds") == 0) 
	{
	  *sds_cnt = 0;
	  get_arg_val_arr(argv[i], sds_names, sds_cnt);
	}
      else if (argv[i][0] == '-') fprintf(stderr, "Ignoring invalid option %s\n", argv[i]);
      else 
	{
	  strcpy(in_fname, argv[i]);
	}
    }
  if (strlen(in_fname) <= 0) 
    { 
      fprintf(stderr, "Missing input file \n"); 
      ret = -1; 
    }
  return ret;
}


