/****************************************************************************
!C

!File: reduce_sds.c

!Description:
  Contains routines for sds reduction 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original October 2002. Version 1.0
  Modified July 2011 by Xiaoping Zhang, Sigma Space Corp.
    added -fill_value option to input arguments.
  Modified in October, 2013, Version 3.0 by Xiaoping Zhang, Sigma Space Corp.
	added help with input filename option on command line, so if the input 
	filename is specified with the help, then the names of all the SDSs in
	the file are displayed.	
  Modified in August, 2014, Version 3.0 by Xiaoping Zhang, Sigma Space Corp.
	changed attr_name string length from 40 to 100 in copy_metadata of meta.c. 
  Modified in February, 2016, Version 3.2 by Xiaoping Zhang, SSAI.
	applied scale and offset to output average(min, max) in float data type 
	if the input is integer SDS with scale and offset.	
	changed the members of sds_t: scalef to fscale, offsetf to foffset, 
	scalef_type to fscale_type
  Modified in January, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
	1)modify it so that HDF5 can be read.
	2)update all HDF4 API SDxxx() in the code to wrapped functions 
	  to deal with HDF4/HDF5 in sds_rw.c

  Modified in Feb., 2020, Version 3.6 by Maosheng Zhao, SSAI.
	1) Correct the issue of adjusting user-specified -rf by commenting out the processes to chang the -rf.
	2) speedup all 4 reduce methods to make both reading and writing in chunks not line by line
	3) make all the output HDF4 with the internal compression
	4) add Qingsong's change so that reduce methods can be specified with -sds
	5) correct bug for m_res after founding that the m_res for the last column is wrong when
	   testing it on the first SDS of MOD09.A2018193.1140.006.2018195024337.hdf,
	   "1km Atmospheric Optical Depth Band 1"
	6) adopt the changes of reduce_an_sds_by_cnt() by Sudipta for adding vsds and mval to correct
	   wrong backgroud fill value.

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Xiaoping Zhang
    LDOPE	SSAI
    NASA's GSFC Code 619
    Greenbelt, MD 20771
    phone: 301-614-6709
    xiaoping.zhang-1@nasa.gov

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "hdf5.h"
#include "mfhdf.h"
#include "qa_tool.h"
#include "str_op.h"
#include "sds_rw.h"
#include "meta.h"
#include "main_util.h"
#include "alloc_mem.h"
#define MAX_VAL_INIT -99999

#define HELP \
"NAME \n" \
"     reduce_sds - reduce the spatial dimensions of one or more SDSs of a Land HDF-EOS data product. \n" \
" \n" \
"SYNOPSIS \n" \
"     reduce_sds -help [filename] \n" \
"     reduce_sds -rf=<reduction factor> -sub|avg|cnt|cl -of=output_file \n" \
"                [-sds=SDSname1[,SDSname2...]] [-bit=[<bit range>]<opr><value>] filename \n" \
"                [-min] [-max] [-std] [-num] [-meta] [-float] [-fill_value=<fill value>]\n" \
" \n" \
"     echo -rf=<reduction factor> -sub|avg|cnt|cl -of=output_file -fill_value=<fill value>\n" \
"          [-sds=SDSname1[,SDSname2...]] [-bit=[<bit range>]<opr><value>] filename \n" \
"          [-min] [-max] [-std] [-num] [-meta] [-float] | reduce_sds \n" \
" or  reduce_sds < argument_file \n" \
" \n" \
"DESCRIPTION \n" \
"     The spatial dimensions of input SDS(s) may be reduced using one of four \n" \
"     different methods. The reduction factor (rf) must be a non-zero positive \n" \
"     integer. The output SDS x dimension will be ((x div rf) + (x mod rf)) and \n" \
"     similarly the y dimension will be ((y div rf) + (y mod rf)). If the input SDS list \n" \
"     contains SDSs with different spatial dimensions the reduction factor will be \n" \
"     applied to the SDS with the smallest spatial dimension and the other \n" \
"     SDS(s) will be reduced to have the same output dimension.  \n" \
"      \n" \
"     All SDS fill values are ignored. \n" \
"      \n" \
"     This tool may be used to reduce data volumes, and to enable analysis of \n" \
"     the different MODLAND product spatial resolutions (250m, 500m, 1km), or \n" \
"     to enable quick comparison with other coarser spatial resolution data sets. \n" \
"      \n" \
"     This tool compliments the tool enlarge_sds. \n" \
"      \n" \
"     This tool supports 2D/3D/4D SDS(s). \n" \
"      \n" \
"     The tool command arguments can be specified in any order. \n" \
" \n" \
"OPTIONS \n" \
"     -help            Display this help message \n" \
"     -rf=factor       Specify the reduction factor (must be integer) \n" \
"     -sub|avg|cnt|cl  Specify the reduction method (subsampling|average|pixel count|class) \n" \
"     -of=filename     Output file \n" \
"     -sds=<SDS list>  List of SDS to reduce (separated by commas). To reduce \n" \
"                      a specific layer in a 3D/4D SDS specify the layer numbers \n" \
"                      (1-based) as dot extension in the SDS name: sds_name.n.m \n" \
"                         Ex: EV_250m_RefSB.1, Surf_Refl.2.1 \n" \
"                      Default is all SDS and all layers. \n" \
"     -bit=[[<bit range>][<operator>]<bit or SDS value>, \n" \
"           [<bit range>][<operator>]<bit or SDS value>, . . . ] \n" \
"                      Bit range and value in decimal separated by the comparison \n" \
"                      operator. If bit range is not specified considers the value as \n" \
"                      the SDS value. Option value part is repeated to specify \n" \
"                      additional bit ranges and values. Multiple option will result \n" \
"                      in separate output SDS for each of the input option. This option \n" \
"                      is applicable only for the -cnt option and must be specified. \n" \
"                      Valid comparison operators are: ==, <, >, <=, >=, != \n" \
"     -std             Compute the standard deviation of each averaged subwindow \n" \
"                      (for -avg option only) \n" \
"     -min             Compute the minimum of each averaged subwindow \n" \
"                      (for -avg option only) \n" \
"     -max             Compute the maximum of each averaged subwindow \n" \
"                      (for -avg option only) \n" \
"     -num             Compute the number of averaged pixels in each subwindow \n" \
"                      (for -avg option only) \n" \
"     -meta            Copy metadata from input file to output \n" \
"     -float           Output average value SDS in float data type. Default is \n" \
"                      input data type. (only applicable with -avg option) \n" \
"     -fill_value      Specify the fill value in case of no fill value of attribute or \n" \
"     		       the fill value of attribute is different than the fill value in data set. \n" \
" \n" \
"Examples: \n" \
"	reduce_sds -sds=\"250m Surface Reflectance Band 1\" -sub -rf=10 -of=red10_sds.hdf ./DATA/MOD09/MOD09.A1996214.0811.002.hdf \n" \
"	reduce_sds -sds=\"250m Surface Reflectance Band 1\" -avg -min -max -std -num -rf=10 -of=red10_sds.hdf ./DATA/MOD09/MOD09.A1996214.0811.002.hdf \n" \
"	reduce_sds -cnt -bit='0-3>=7' -meta -sds='fire mask' -rf=5 -of=fire_mask_crs.hdf VNP14IMG.A2020030.1112.001.2020030190902.nc \n" \
"	reduce_sds -cnt -bit=\"0-3<=2,0-3==3,0-3==4,0-3==5,0-3==6,0-3==7\" -sds=\"fire mask\" -rf=5 -of=fire_cnt.hdf ./DATA/SAMPLES/MOD14.A1996213.1024.002.hdf \n" \
"	reduce_sds -sds=Cloud_Mask -sub -rf=5 -of=cloud_sub.hdf ./DATA/MOD35/1999053/MOD35_L2.A1999053.0720.002.1999218204033.hdf \n" \
"	reduce_sds -sds=Cloud_Mask.1-2 -avg -rf=5 -of=cloud_sub.hdf ./DATA/MOD35/1999053/MOD35_L2.A1999053.0720.002.1999218204033.hdf \n" \
"	reduce_sds -cnt -bit=\"0-3<=2,0-3==3,0-3==4,0-3==5,0-3==6,0-3==7\" -meta -sds=\"most confident detected fire\" -rf=5 -of=fire.hdf ./DATA/MOD14A1/MOD14A1.A1999053.h08v06.002.1999250122510.hdf \n" \
"	reduce_sds -sds=\"sur_refl_b01\" -sub -rf=5 -of=red5_NPP_D8SRF1KM_L3D.hdf /LANDQA/f003/xzhang/QA_TOOLS/NPP/data_in/NPP_D8SRF1KM_L3D.A2003193.h17v07.P1_00602.2009252202610.hdf -fill_value=\"-5\" \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:      reduce_sds -rf=<reduction factor> -sub|avg|cnt|cl -of=output_file \n" \
"                [-sds=SDSname1[,SDSname2...]] [-bit=[<bit range>]<opr><value>] filename \n" \
"                [-min] [-max] [-std] [-num] [-meta] \n" \
" \n" \
"OPTIONS \n" \
"     -help            Display this help message \n" \
"     -rf=factor       Specify the reduction factor (must be integer) \n" \
"     -sub|avg|cnt|cl  Specify the reduction method (subsampling|average|pixel count|class) \n" \
"     -of=filename     Output file \n" \
"     -sds=<SDS list>  List of SDS to reduce (separated by commas). To reduce \n" \
"                      a specific layer in a 3D/4D SDS specify the layer numbers \n" \
"                      (1-based) as dot extension in the SDS name: sds_name.n.m \n" \
"                         Ex: EV_250m_RefSB.1, Surf_Refl.2.1 \n" \
"                      Reduce method can be assigned separately for each of the SDS, and it \n" \
"                      will override the overall reduce method. \n" \
"                         Ex: EV_250m_RefSB.1:sub, Surf_Refl.2.1:avg \n" \
"                      Note: Geolocation SDS will ignore reduce method options and is always reduced by -sub \n" \
"                      Default is all SDS and all layers. \n" \
"     -bit=[[<bit range>][<operator>]<bit or SDS value>, \n" \
"           [<bit range>][<operator>]<bit or SDS value>, . . . ] \n" \
"                      Bit range and value in decimal separated by the comparison \n" \
"                      operator. If bit range is not specified considers the value as \n" \
"                      the SDS value. Option value part is repeated to specify \n" \
"                      additional bit ranges and values. Multiple option will result \n" \
"                      in separate output SDS for each of the input option. This option \n" \
"                      is applicable only for the -cnt option and must be specified. \n" \
"                      Valid comparison operators are: ==, <, >, <=, >=, != \n" \
"     -std             Compute the standard deviation of each averaged subwindow \n" \
"                      (N/A if subsampling method) \n" \
"     -min             Compute the minimum of each averaged subwindow \n" \
"                      (N/A if subsampling method) \n" \
"     -max             Compute the maximum of each averaged subwindow \n" \
"                      (N/A if subsampling method) \n" \
"     -num             Compute the number of averaged pixels in each subwindow \n" \
"                      (N/A if subsampling method) \n" \
"     -meta            Copy metadata from input file to output \n" \
"     -float           Output average value SDS in float data type. Default is \n" \
"                      input data type. (only applicable with -avg option) \n"	\
"     -fill_value      Specify the fill value in case of no fill value of attribute or \n" \
"     		       the fill value of attribute is different than the fill value in data set. \n" 

static unsigned int BIT[] = {
  0x1,        0x3,        0x7,        0xf,
  0x1f,       0x3f,       0x7f,       0xff,
  0x1ff,      0x3ff,      0x7ff,      0xfff,
  0x1fff,     0x3fff,     0x7fff,     0xffff,
  0x1ffff,    0x3ffff,    0x7ffff,    0xfffff,
  0x1fffff,   0x3fffff,   0x7fffff,   0xffffff,
  0x1ffffff,  0x3ffffff,  0x7ffffff,  0xfffffff,
  0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff
};

/* Qingsong added **methods to parse_cmd_reduce_sds & reduce_mult_sds */
int parse_cmd_reduce_sds(int argc, char **argv, char **sds_names, int *sds_cnt, char **arg_list, char **methods);
void reduce_mult_sds(char **sds_names, int sds_cnt, int32 out_sd_id, char **arg_list, char **methods);
int get_bit_opt(char *cmp_str, int **bn_arr, int *bn_val, int *bn_opr, int *bn_cnt);
void reduce_an_sds_by_sub(sds_t *in_sds_info, int32 out_sd_id, int res);
void reduce_an_sds_by_avg(sds_t *in_sds_info, int32 out_sd_id, int res, int *out_flag, char avg_dt);
void reduce_an_sds_by_cnt(sds_t *in_sds_info, int32 out_sd_id, int res, char *cmp_str);
void reduce_an_sds_by_class(sds_t *in_sds_info, int32 out_sd_id, int res);

int main(int argc, char *argv[])
/****************************************************************************
!C

!Description:
  Function main() for sds reduction 

!Input Parameters:
  Command line arguments (See help for details)

!Input/Output Parameters: (none)

!Output Parameters:
  Exit status 0 on completion, and a completion message.

!Revision History:
  Original October 2002. Version 1.0

!Team-unique Header:
  See file prologue.

!Developer(s):

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/
{
  int i, k, fid, status;
  int nsds, isds;
  sds_t out_sds_info;
  int32 sd_id, sds_id, dt;
  int32 msds, nattr, rank, dim_size[4];
  char **arg_list, **sds_names;
  /* Qingsong */
  char **methods = NULL;

  char dim_str[MAX_STR_LEN], in_fname[MAX_PATH_LENGTH];
  char name[MAX_SDS_NAME_LEN], sds_name[MAX_SDS_NAME_LEN];

  /* MSZ 1/25/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 1/25/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if (argc == 1 || ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0))))
  {
    /* Print command line help when -help option is input */
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  else
  {
    /* Print all SDS names when -help option is input with the input filename */
    for (i=0, status=0; i<argc; i++)
      if ((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "-help") == 0)) status = i;
    if (status == 0)
    {
      /* Print all the SDS names in the input file */
      for (i=1; i<argc; i++)
        if (argv[i][0] != '-')
	{
	  display_sds_info_of_file(argv[i]);
	  break;
	}
    }
  }

  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names in read_l2g(main)\n");
  /* Qingsong */
  if ((methods = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for methods in read_l2g(main)\n");
  if ((arg_list = (char **)Calloc2D(12, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for arg_list in read_l2g(main)\n");
  if ((sds_names != NULL) && (arg_list != NULL))
  {
    fprintf(stdout, "Started process reduce_sds . . . . \n");

    /* Parse command arguments */
    status = parse_cmd_reduce_sds(argc, argv, sds_names, &nsds, arg_list, methods);
    if (status == -1)
      fprintf(stderr, "%s\n", USAGE);
    else if (status != 0)
    {
      out_sds_info.H5IN = 0;         /* nothing to do with the input format */
      out_sds_info.H5OUT = 0;        /* so far, just output in HDF4 */
      out_sds_info.sd_id = out_sds_info.sds_id = -1;

	/* open output HDF file */
      if(open_sd(arg_list[0], &out_sds_info, 'W') == FAIL)
        fprintf(stderr, "Cannot create the output file %s\n", arg_list[0]);
      else
      {
	strcpy(in_fname, arg_list[10]);
	fprintf(stderr, "Processing input HDF file %s\n", in_fname);

	/* reformat SDS names with proper name extension for layer numbers */
	update_nd_sdsnames(sds_names, &nsds, in_fname);

	/* reduce all the input SDS */
        reduce_mult_sds(sds_names, nsds, out_sds_info.sd_id, arg_list, methods); 

	/* close the output HDF file */
	close_sd(&out_sds_info, out_sds_info.H5OUT);
      }
    }
    Free2D((void **)arg_list);
    Free2D((void **)sds_names);
    /* Qingsong */
    Free2D((void **)methods);
  }
  fprintf(stdout, "Done Processing\n");
  return 0;
}

int parse_cmd_reduce_sds(int argc, char **argv, char **sds_names, int *sds_cnt, char **arg_list, char **methods)
/****************************************************************************
!C

!Description:
  Parse the command line arguments.

!Input Parameters:
  argc:	number of arguments in the command line.
  argv: string array containing the input arguments.

!Input/Output Parameters: (none)

!Output Parameters:
  sds_names: string array containing the input SDS names.
  sds_cnt: number of SDS input
  arg_list: return argument list.
  returns 1 on success and -1 on error.

!Revision History:
  Original October 2002. Version 1.0

!Team-unique Header:
  See file prologue.

!Developer(s):

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
***************************************************************************/
{
  int i;
  int if_cnt;
  int status;

  status = 1;
  if_cnt = *sds_cnt = 0;
  for (i=0; i<12; i++)
    arg_list[i][0] = '\0';

  /* parse arguments and argument values */
  for (i=1; i<argc; i++)
  {
    if (is_arg_id(argv[i], "-sds=") == 0) 
      get_arg_val_arr(argv[i], sds_names, sds_cnt);
    else if (is_arg_id(argv[i], "-sds") == 0) ;
    else if ((is_arg_id(argv[i], "-of=") == 0) || (is_arg_id(argv[i], "-o=") == 0))
      get_arg_val(argv[i], arg_list[0]);
    else if ((is_arg_id(argv[i], "-rf=") == 0) || (is_arg_id(argv[i], "-r=") == 0))
      get_arg_val(argv[i], arg_list[1]);
    else if ((strcmp(argv[i], "-sub") == 0) || (strcmp(argv[i], "-avg") == 0) || 
		(strcmp(argv[i], "-cnt") == 0) || (strcmp(argv[i], "-cl") == 0))
      strcpy(arg_list[2], argv[i]);
    else if (strcmp(argv[i], "-min") == 0) strcpy(arg_list[3], "y");
    else if (strcmp(argv[i], "-max") == 0) strcpy(arg_list[4], "y");
    else if (strcmp(argv[i], "-std") == 0) strcpy(arg_list[5], "y");
    else if (strcmp(argv[i], "-num") == 0) strcpy(arg_list[6], "y");
    else if (is_arg_id(argv[i], "-bit=") == 0) get_arg_val(argv[i], arg_list[7]);
    else if (strcmp(argv[i], "-float") == 0) strcpy(arg_list[8], "y");
    else if (strcmp(argv[i], "-meta") == 0) strcpy(arg_list[9], "y");
    else if (is_arg_id(argv[i], "-fill_value=") == 0) get_arg_val(argv[i], arg_list[11]);
    else if (argv[i][0] == '-') fprintf(stderr, "Unknonwn option %s\n", argv[i]);
    else {
      if (if_cnt == 1)
	fprintf(stdout, "Only one input file is accepted, input file %s is ignored.\n", argv[i]);
      else {
	strcpy(arg_list[10], argv[i]);
        ++if_cnt;
      }
    }
  }

  /* Qingsong */
  int nmethod = 0;
  for(i=0; i<*sds_cnt; i++){
    int len = strlen(sds_names[i]);
    int j, p=0;
    for(j=0; j<len; j++){
      if(sds_names[i][j] == ':'){
        int k;
        methods[i][p++] = '-';
        for(k=j+1; k<len; k++){
          methods[i][p++] = sds_names[i][k];
        }
        methods[i][p] = '\0';
        sds_names[i][j] = '\0';
        nmethod ++;
        break;
      }
    }
    if(p <= 1){
      if (arg_list[2][0] == '\0') {
        status = -1;
        fprintf(stderr, "No reduction option sub|avg|cnt specified for sds %s\n", sds_names[i]);
      }
      else{
        strcpy(methods[i], arg_list[2]);
      }
    }
  }

  /* check for required arguments */
  if (if_cnt == 0)
  {
    status = -1;
    fprintf(stderr, "No input filename specified\n");
  }
  /* Qingsong check i==2 later separately */
  //for (i=0; i<3; i++)
  for (i=0; i<2; i++)
    if (arg_list[i][0] == '\0') 
    {
      status = -1;
      if (i == 0) fprintf(stderr, "No output filename specified\n");
      else if (i == 1) fprintf(stderr, "No reduction factor specified\n");
      else if (i == 2) fprintf(stderr, "No reduction option sub|avg|cnt specified\n");
    }
  
  /* Qingsong */
  if(arg_list[2][0] == '\0' && nmethod != *sds_cnt){
    status = -1;
    fprintf(stderr, "No reduction option sub|avg|cnt specified\n");
  }

  /* if required arguments are missing update the status to failure */
  if (status != -1)
  {
    if (strcmp(arg_list[2], "-cnt") == 0)
    {
      if (strlen(arg_list[7]) <= 0) 
      {
	fprintf(stderr, "Missing -bn option for sds reduce by cnt\n");
        status = -1;
      }
    }
    else if (strlen(arg_list[7]) != 0)
      fprintf(stderr, "Ignoring -val option for reduce option %s\n", arg_list[2]);
  }
  if ((status != -1) && (atoi(arg_list[1]) <= 0))
  {
    fprintf(stderr, "Invalid value for reduction factor \n");
    status = 0;
  }

  /* if SDS names not input set to default all SDS */
  if ((status > 0) && (*sds_cnt == 0))
  {
    fprintf(stderr, "No SDS name input. Reading all SDS . . \n");
    *sds_cnt = 1;
    strcpy(sds_names[0], "all");
  }
  return status;
}

void reduce_mult_sds(char **sds_names, int sds_cnt, int32 out_sd_id, char **arg_list, char **methods) 
/****************************************************************************
!C

!Description:
  Routine to initiate reduce sds based on user options for all the input SDSs.

!Input Parameters:
  sds_names: string array containing the input SDS names.
  sds_cnt: number of SDS input
  out_sd_id: ouput SDS id.
  arg_list: argument list containing paramters.

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original October 2002. Version 1.0

!Team-unique Header:
  See file prologue.

!Developer(s):

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
***************************************************************************/
{
  int rank, out_flag[5];
  int i, isds, in_res, res;
  int xdim_min, rf[MAX_NUM_SDS];
  sds_t in_sds_info;
  char infile[MAX_PATH_LENGTH];
  void  *attr_val;
  int32 attr_type, attr_cnt; 
   
  /* set input SDS names */
  strcpy(infile, arg_list[10]);
  if ((sds_cnt == 1) && (strcmp(sds_names[0], "all") == 0))
  {
    sds_cnt = get_sds_names(infile, sds_names);
    if (sds_cnt == 0)
    {
      fprintf(stderr, "No SDS found in %s\n", infile);
      return;
    }
    /* Qingsong */
    for(i=0; i<sds_cnt; i++){
	strcpy(methods[i], arg_list[2]);
    }
  }

  /* initialize as -1 for sd_id and sds_id */
  in_sds_info.sd_id = in_sds_info.sds_id = -1;

  /* open the input HDF file */
  if(open_sd(infile, &in_sds_info, 'R') == FAIL)
  {
    fprintf(stderr, "Cannot open input HDF file in reduce_mult_sds\n");
    return;
  }

  /* open all input SDS and obtain the dimension information */
  for (isds=0; isds<sds_cnt; isds++)
  {
    strcpy(in_sds_info.name, sds_names[isds]);
    in_sds_info.sds_id = -1;
    if (get_sds_info((char *)NULL, &in_sds_info) != -1)
    {
      rank = in_sds_info.rank;
      rf[isds] = (in_sds_info.dim_size[rank-1] < in_sds_info.dim_size[0]) ?
			in_sds_info.dim_size[0] : in_sds_info.dim_size[rank-2];

      /* diagnosis */
      /*printf("ZMS0 rf[%d]=%d\n", isds, rf[isds]);*/
      close_sds(&in_sds_info, in_sds_info.H5IN);
    }
    else rf[isds] = -1;
  }

  /* MSZ 1/9/2020, don't adjust user-specified -rf or in_res */
  in_res = (int)atoi(arg_list[1]);
  /*
  xdim_min = 9999;
  in_res = (int)atoi(arg_list[1]);
  for (isds=0; isds<sds_cnt; isds++)
    if ((rf[isds] != -1) && (rf[isds] < xdim_min))
      xdim_min = rf[isds];
  for (isds=0; isds<sds_cnt; isds++)
  {
    if (rf[isds] != -1)
      rf[isds] = rf[isds]/xdim_min;
  }
  */

  /* process each SDS according to input option */
  for (isds=0; isds<sds_cnt; isds++)
  {
    /* MSZ 1/9/2020, don't adjust user-specified -rf or in_res */
    /*res = in_res*rf[isds];*/
    res = in_res;

    /* diagnosis */
    /*printf("ZMS2 in_res=%d rf[%d]=%d res=%d\n", in_res, isds, rf[isds], res);*/
    strcpy(in_sds_info.name, sds_names[isds]);
    fprintf(stderr, "\tProcessing SDS %s\n", in_sds_info.name);
    in_sds_info.sds_id = -1;
    if (get_sds_info((char *)NULL, &in_sds_info) == -1)
    {
      fprintf(stderr, "Ignoring invalid input sds %s\n", in_sds_info.name);
      if (in_sds_info.sds_id == FAIL)
	close_sds(&in_sds_info, in_sds_info.H5IN);
      continue;
    }    
/*  printf("arg_list[11]=%s\n",arg_list[11]);*/
    if (arg_list[11][0] != '\0')       
      switch(in_sds_info.data_type)
      {
	case  5: in_sds_info.fill_fval = (float32)atof(arg_list[11]); break;
	case  6: in_sds_info.fill_fval = (float64)atof(arg_list[11]); break;
	case 20: in_sds_info.fill_val = (int8)atoi(arg_list[11]); break;
	case 21: in_sds_info.fill_val = (uint8)atoi(arg_list[11]); break;
	case 22: in_sds_info.fill_val = (int16)atoi(arg_list[11]); break;
	case 23: in_sds_info.fill_val = (uint16)atoi(arg_list[11]); break;
	case 24: in_sds_info.fill_val = (int32)atoi(arg_list[11]); break;
	case 25: in_sds_info.fill_val = (uint32)atoi(arg_list[11]); break;
      }   

    strcpy(in_sds_info.name, sds_names[isds]);		/*  added on 05/03/2011	*/

    /* Qingsong lat/lon always use -sub method */
    if(strcmp(sds_names[isds], "Latitude") == 0 ||
       strcmp(sds_names[isds], "latitude") == 0 ||
       strcmp(sds_names[isds], "Longitude") == 0 ||
       strcmp(sds_names[isds], "longitude") == 0){
       reduce_an_sds_by_sub(&in_sds_info, out_sd_id, res);
    }
    /* Qingsong check methods instead arg_list[2] */
    else if (strcmp(methods[isds], "-sub") == 0){
      fprintf(stdout, "Reducing sds %s with method -sub . . . . \n", sds_names[isds]);
      reduce_an_sds_by_sub(&in_sds_info, out_sd_id, res);
    }
    else if (strcmp(methods[isds], "-cnt") == 0){
      fprintf(stdout, "Reducing sds %s with method -cnt . . . . \n", sds_names[isds]);
      reduce_an_sds_by_cnt(&in_sds_info, out_sd_id, res, arg_list[7]);
    }
    else if (strcmp(methods[isds], "-cl") == 0){
      fprintf(stdout, "Reducing sds %s with method -cl . . . . \n", sds_names[isds]);
      reduce_an_sds_by_class(&in_sds_info, out_sd_id, res);
    }
    else if (strcmp(methods[isds], "-avg") == 0)  
    {
      out_flag[0] = 1;
      for (i=3; i<7; i++)
        out_flag[i-2] = (arg_list[i][0] == 'y') ? 1 : 0;
      fprintf(stdout, "Reducing sds %s with method -avg . . . . \n", sds_names[isds]);
      reduce_an_sds_by_avg(&in_sds_info, out_sd_id, res, out_flag, arg_list[8][0]);
    }
    close_sds(&in_sds_info, in_sds_info.H5IN);
  }

  /* copy metadata from input to output */
  if (arg_list[9][0] == 'y') 
  {
    int H5OUT = 0;
    copy_metadata(in_sds_info.H5IN, in_sds_info.sd_id, H5OUT, out_sd_id);
  }
  /* close the input HDF file */
  close_sd(&in_sds_info, in_sds_info.H5IN);
}

void reduce_an_sds_by_sub(sds_t *in_sds_info, int32 out_sd_id, int res)
/****************************************************************************
!C

!Description:
  Routine to reduce an sds by sub-sampling.

!Input Parameters:
  in_sds_info: structure containing input SDS information.
  out_sd_id: ouput SDS id.
  res: reduction factor.

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original October 2002. Version 1.0

!Team-unique Header:
  See file prologue.

!Developer(s):

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Maosheng Zhao
    Jan, 2020, mask input and output in chunk style and also the output has internal compression


!Design Notes: (none)

!END
****************************************************************************/
{
  int	i, j, k, n, m; 
  int   st_c, st1_c, offset;
  int   ic, ic1, ic2, rank;
  int   res2, res_by_2; 
  int   ndata_line, ndata_out;
  int   ndata_sm_in, ndata_sm_out;
  int	irow, nrow, ncol, out_ncol;
  int32 attr_type, attr_cnt;
  int32 in_edge[4] = {0, 0, 0, 0};
  int32 out_edge[4] = {0, 0, 0, 0};
  int32 in_start[4] = {0, 0, 0, 0};
  int32 out_start[4] = {0, 0, 0, 0};
  sds_t out_sds_info;
  char sds_name[MAX_SDS_NAME_LEN];
  void  *attr_val, *data_line = NULL, *data_out_chunk = NULL;
  int nattrs=22;
  float fill_val_flt;

  /* MSZ 1/16/2020 */
  int jobs_num = 30;
  int n_chunks, ndata_out_chunk;
  int I_ROW, i_chunk, i_chunk_row_out;
  int chunk_rows_in, chunk_rows_out, last_chunk_rows_in;
  int chunk_lines_in, line_start_in;
  size_t ndata_in_chunk;
  void *data_in_chunk = NULL;

  /* MSZ 1/17/2020, for the chunked output */
  int out_nrow, last_chunk_rows_out;
  int chunk_lines_out;
  int shift_data_out_bsq, shift_data_out_bip;

  /* MSZ 1/17/2020, for output with the internal compression */
  int comp_chunk_rows;
  
  /* create output SDS with proper dimension and other attributes */
  if ((attr_val = (void *)get_sds_attr(in_sds_info, "_FillValue", &attr_type, &attr_cnt)) == NULL) 
  {
    fprintf(stderr, "Attribute "_FillValue" not defined for output.\n");
  } 
  
  switch (in_sds_info->data_type)
  {
    case 5:  fill_val_flt = (float)in_sds_info->fill_fval; break;
    case 6:  fill_val_flt = (float)in_sds_info->fill_fval; break;
    case 20: fill_val_flt = (float)in_sds_info->fill_val;  break;
    case 21: fill_val_flt = (float)in_sds_info->fill_val;  break;
    case 22: fill_val_flt = (float)in_sds_info->fill_val;  break;
    case 23: fill_val_flt = (float)in_sds_info->fill_val;  break;
    case 24: fill_val_flt = (float)in_sds_info->fill_val;  break;
    case 25: fill_val_flt = (float)in_sds_info->fill_val;  break;
  }  

  rank = in_sds_info->rank;
  out_sds_info.H5IN = 0;  /* nothing to do with the input format */
  out_sds_info.H5OUT= 0;  /* so far, output to HDF4 */
  out_sds_info.sds_id = -1;
  out_sds_info.sd_id = out_sd_id;
  out_sds_info.data_size = in_sds_info->data_size;
  out_sds_info.data_type = in_sds_info->data_type;
  out_sds_info.interleave=in_sds_info->interleave;
  

  if (charpos(in_sds_info->name, '.', 0) != -1)
  {
    out_sds_info.rank = 2;
    if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
    {
      out_sds_info.dim_size[0] = (int)ceil((float)in_sds_info->dim_size[0]/res);
      out_sds_info.dim_size[1] = (int)ceil((float)in_sds_info->dim_size[1]/res);

      /* MSZ 1/17/2020, speedup */
      out_nrow = out_sds_info.dim_size[0];
    }
    else
    {
      out_sds_info.dim_size[0] = (int)ceil((float)in_sds_info->dim_size[rank-2]/res);
      out_sds_info.dim_size[1] = (int)ceil((float)in_sds_info->dim_size[rank-1]/res);

      /* MSZ 1/17/2020, speedup */
      out_nrow = out_sds_info.dim_size[0];
    }
  }                                           
  else
  {
    out_sds_info.rank = in_sds_info->rank;
    for (i=0; i<rank; i++)
      out_sds_info.dim_size[i] = in_sds_info->dim_size[i];
    if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))          /* BIP */
    {
      out_sds_info.dim_size[0] = (int)ceil((float)in_sds_info->dim_size[0]/res);
      out_sds_info.dim_size[1] = (int)ceil((float)in_sds_info->dim_size[1]/res);

      /* MSZ 1/17/2020, speedup */
      out_nrow = out_sds_info.dim_size[0];
    }
    else                                                                                   /* BSQ */
    {
      out_sds_info.dim_size[rank-2] = (int)ceil((float)in_sds_info->dim_size[rank-2]/res);
      out_sds_info.dim_size[rank-1] = (int)ceil((float)in_sds_info->dim_size[rank-1]/res);

      /* MSZ 1/17/2020, speedup */
      out_nrow = out_sds_info.dim_size[rank-2];
    }
  }
  
  /* get input SDS name and layer options */
  /* get the number of observations in the input and output */

  get_sdsname_dim(in_sds_info->name, sds_name, &n, &m);
  if (rank == 2)
  {
    nrow = in_sds_info->dim_size[0];
    ncol = ndata_line = in_sds_info->dim_size[1];
    out_ncol = ndata_out = out_sds_info.dim_size[1];
  }                                                                            
  else if (rank > 2)
  {
    if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
    {
      nrow = in_sds_info->dim_size[0];
      ncol = ndata_line = in_sds_info->dim_size[1];
      out_ncol = out_sds_info.dim_size[1];
      for (i=2; i<rank; i++)
        ndata_line *= in_sds_info->dim_size[i];
      ndata_out = (int)ceil((float)in_sds_info->dim_size[1]/res);
      if ((n == -1) && (m == -1))
        for (i=2; i<rank; i++)
          ndata_out *= in_sds_info->dim_size[i];
    }
    else
    {
      nrow = in_sds_info->dim_size[rank-2];
      ncol = ndata_line = in_sds_info->dim_size[rank-1];
      out_ncol = ((n == -1)&&(m == -1)) ? out_sds_info.dim_size[rank-1] : out_sds_info.dim_size[1];
      for (i=0; i<rank-2; i++)
        ndata_line *= in_sds_info->dim_size[i];
      ndata_out = (int)ceil((float)in_sds_info->dim_size[rank-1]/res);
      if ((n == -1) && (m == -1))
        for (i=0; i<rank-2; i++)
          ndata_out *= in_sds_info->dim_size[i];
    }
  }

  /**************************************************************************/
  /* set output compression's chunk */
  comp_chunk_rows = 30;
  if(comp_chunk_rows > out_nrow) comp_chunk_rows = out_nrow;     /* to fix comp_chunk_rows */

  /* MSZ 1/16/2020, get chunk_rows_in and chunk_rows_out (times of res) */
  chunk_rows_in = (int)ceil((double)nrow / jobs_num);

  if(chunk_rows_in > 0)
    chunk_rows_out = (int)ceil((double)chunk_rows_in / res);

  if(chunk_rows_out < 1) chunk_rows_out = 1;

  /* correct chunk_rows_in so that it reads chunk in chunk_rows_out times of res.
   * This also links chunk_rows_out with chunk_rows_in */
  chunk_rows_in = res * chunk_rows_out;

  /* get n_chunks and rows of the last chunk in */
  /* Note that n_chunks applies to both input chunks and output chunks,
   * i.e, both input and output have the same number of chunks */
  n_chunks = (int)ceil((float)nrow / (float)(chunk_rows_in));

  /* number of rows for the last chunk of input and output */
  last_chunk_rows_in  = nrow % chunk_rows_in;
  last_chunk_rows_out = out_nrow % chunk_rows_out;

  /* number of elements for ndata_in_chunk and chunked ndata_out_chunk */
  ndata_in_chunk  = (size_t)ndata_line * (size_t)chunk_rows_in;
  ndata_out_chunk = ndata_out * chunk_rows_out;
  /**************************************************************************/

  if ((data_in_chunk = (void *)calloc(ndata_in_chunk, in_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_in_chunk in reduce_an_sds_by_sub\n");
  if ((data_line = (void *)calloc(ndata_line, in_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_line in reduce_an_sds_by_sub\n");
  if ((data_out_chunk = (void *)calloc(ndata_out_chunk, out_sds_info.data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_out_chunk in reduce_an_sds_by_sub\n");

  if ( (data_in_chunk != NULL) && (data_line != NULL) && (data_out_chunk != NULL))
  {

    sprintf(out_sds_info.name, "%s_sub", in_sds_info->name);

    /* if (open_sds((char *)NULL, &out_sds_info, 'W') == -1) */
    if (open_sds_compression((char *)NULL, &out_sds_info, comp_chunk_rows, 'W') == -1)
    {
      if (attr_val != NULL) free(attr_val);
      return;
    }
    else 
      write_attr_fvalf(out_sds_info.H5OUT, out_sds_info.sds_id, in_sds_info->data_type, 1, (double)fill_val_flt);

  if (attr_val != NULL) free(attr_val);
    if (rank == 2)
    {
      in_edge[0] = 1 * chunk_rows_in;
      out_edge[0] = 1 * chunk_rows_out;
      in_edge[1] = in_sds_info->dim_size[1];
      out_edge[1] = out_sds_info.dim_size[1];
    }                       
    else if (rank > 2)
    {
      /* initialize in and out edge values */
      for (i=0; i<rank; i++)
        in_edge[i] = out_edge[i] = in_sds_info->dim_size[i];

      if ((n == -1) && (m == -1))
      {
        if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
        {
          in_edge[0]= 1 * chunk_rows_in;
	  out_edge[0] = 1 * chunk_rows_out;
          out_edge[1] = out_sds_info.dim_size[1];
        }
        else
        {
          in_edge[rank-2] = 1 * chunk_rows_in;
	  out_edge[rank-2] = 1 * chunk_rows_out;
          out_edge[rank-1] = out_sds_info.dim_size[rank-1];
        }
      }
      else
      {
 	/* now it is important to have the right rank for the output because of chunk output */

        out_edge[0] = 1 * chunk_rows_out;
        out_edge[1] = out_sds_info.dim_size[1];

        if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
          in_edge[0] = 1 * chunk_rows_in;
        else
          in_edge[rank-2] = 1 * chunk_rows_in;
      }
    }                                

    /* compute start of the requested SDS layer and offset of each
       observations */

    compute_sds_start_offset(in_sds_info, n, m, &st_c, &offset);   

    /* compute the number of output data samples per line */

    if (out_sds_info.interleave == 1)                       /* BSQ */
    {
      ndata_sm_out = 1;
      if ((n == -1) && (m == -1))
      {
        for (i=0; i<rank-2; i++)
	  ndata_sm_out *= out_sds_info.dim_size[i];
      }
    }
    else                                                    /* BIP */
    {
      ndata_sm_in = 1;
      for (i=2; i<rank; i++)
	ndata_sm_in *= in_sds_info->dim_size[i]; 
      ndata_sm_out = ((n == -1) && (m == -1)) ? ndata_sm_in : 1;
    }
   
    /* process the requested layer of the input SDS */
 
    res_by_2 = res/2;
    if ((rank>2) && (n == -1) && (m == -1) && (in_sds_info->dim_size[rank-1]>in_sds_info->dim_size[0]))
      out_start[rank-2] = 0;
    else
      out_start[0] = 0;

    /* process one line per res lines of input */
    /* process only one sample per res samples of input samples */
    
    write_all_sds_attr(in_sds_info->H5IN, in_sds_info->sds_id, out_sds_info.H5OUT, out_sds_info.sds_id, nattrs);

    i_chunk = 0;
    for (I_ROW = 0; I_ROW < nrow; I_ROW += chunk_rows_in)
    {
      /* read the whole chunk of the input data */
      if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank- 1]))  /* BIP */
        in_start[0] = I_ROW;
      else in_start[rank-2] = I_ROW;                                                   /* BSQ */

      /* correct chunk_lines_in, chunk_lines_out, in_edge and out_edge of the row dimension for the last chunk */
      chunk_lines_in = chunk_rows_in;
      chunk_lines_out = chunk_rows_out;

      if(i_chunk == n_chunks - 1)  /* last chunk */
      {
	/* only when the remainder is not zero */
	if(last_chunk_rows_in) chunk_lines_in  = last_chunk_rows_in;
	if(last_chunk_rows_out) chunk_lines_out = last_chunk_rows_out;

	if (rank == 2)
	{
	  in_edge[0] = chunk_lines_in;
	  out_edge[0] = chunk_lines_out;
	}
	else if (rank > 2)
	{
	   if ((n == -1) && (m == -1))
	   {
	     if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])      /* BIP */
	     {
	       in_edge[0] = chunk_lines_in;
	       out_edge[0] = chunk_lines_out;
	     }
	     else                                                               /* BSQ */
	     {
	       in_edge[rank-2] = chunk_lines_in;
	       out_edge[rank-2] = chunk_lines_out;
	     }
	   }
	   else
	   {
	     if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
		in_edge[0] = chunk_lines_in;
	     else
		in_edge[rank-2] = chunk_lines_in;

	     /* now it is important to have the right rank for the output because of chunk output */
	     if(out_sds_info.rank == 2)
	     {
	     	out_edge[0] = chunk_lines_out;
	     	out_edge[1] = out_sds_info.dim_size[1];
	     }
	     else
	     {
		if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
		   out_edge[0] = chunk_lines_out;
		else
		   out_edge[rank-2] = chunk_lines_out;
	     }

	   }
	}
      }

      /* read the whole chunk data */
      if (read_sel_sds_data(in_sds_info, in_start, in_edge, data_in_chunk) == FAIL)
      {
	fprintf(stderr, "Error reading input HDF file in reduce_sds\n");
	break;
      }

      /* process each sample line for the chunk with the out rows of last chunk being considered, chunk_lines_out */
      for (i_chunk_row_out=0; i_chunk_row_out < chunk_lines_out; i_chunk_row_out++) 
      {
	irow = I_ROW + i_chunk_row_out * res;

	/* get data_line for the line */
        /* res2 = irow + res_by_2; */
	res2 = i_chunk_row_out * res + res_by_2;

	/* now the start for each res is i_chunk_row_out * res, and the below is correct */
	line_start_in = (res2 < chunk_lines_in) ? res2 : i_chunk_row_out * res + (chunk_lines_in - i_chunk_row_out * res)/2;

	/* get just one line sample data from the read chunk */
	get_lines_from_chunk(data_in_chunk, in_sds_info->rank, in_edge, in_sds_info->interleave, in_sds_info->data_size, line_start_in, 1, data_line);

        if (out_sds_info.interleave == 1)                            /* BSQ */
        {
	  st1_c = st_c;
	  /*for (k=0, ic=0; k<ndata_sm_out; k++, st1_c += ncol)*/
	  for (k=0; k<ndata_sm_out; k++, st1_c += ncol)
          {
	    /* MSZ 01/17/2020, add the shift_data_out_bsq for BSQ chunked output data_out,
 	     * but here last chunk should be considered because out_edge accounts for that
 	     * so chunk_lines_out not chunk_rows_out should be used. */
	    shift_data_out_bsq = i_chunk_row_out * out_ncol + k * out_ncol * chunk_lines_out;

            /*for (i=0, ic1=0; i<out_ncol; i++, ic1 += res, ic++)*/
	    for (i=0, ic1=0, ic=0; i<out_ncol; i++, ic1 += res, ic++)
            {
	      res2 = ic1 + res_by_2;
	      j = (res2 < ncol) ? st1_c + res2 : st1_c + ic1 + (ncol - ic1)/2;
	      switch(in_sds_info->data_type)
	      {
	        case  5: ((float32 *)data_out_chunk)[ic + shift_data_out_bsq] = ((float32 *)data_line)[j]; break;
	        case  6: ((float64 *)data_out_chunk)[ic + shift_data_out_bsq] = ((float64 *)data_line)[j]; break;
	        case 20: ((int8 *)data_out_chunk)[ic + shift_data_out_bsq] = ((int8 *)data_line)[j]; break;
	        case 21: ((uint8 *)data_out_chunk)[ic + shift_data_out_bsq] = ((uint8 *)data_line)[j]; break;
	        case 22: ((int16 *)data_out_chunk)[ic + shift_data_out_bsq] = ((int16 *)data_line)[j]; break;
	        case 23: ((uint16 *)data_out_chunk)[ic + shift_data_out_bsq] = ((uint16 *)data_line)[j]; break;
	        case 24: ((int32 *)data_out_chunk)[ic + shift_data_out_bsq] = ((int32 *)data_line)[j]; break;
	        case 25: ((uint32 *)data_out_chunk)[ic + shift_data_out_bsq] = ((uint32 *)data_line)[j]; break;
	      }
	    } /* for (i=0; . . .) */
	  } /* for (k=0; . . .) */
        }
        else                                                         /* BIP */
        {
	  /* MSZ 01/17/2020, add the shift_data_out for BIP chunked output data_out,
 	   * but here last chunk should be considered because out_edge accounts for that, here for
 	   * BIP, Y is controled by i_chunk_row_out only */
	  if(out_sds_info.rank == 2)
	     shift_data_out_bip = out_ncol * i_chunk_row_out;
	  else
	     shift_data_out_bip = out_ncol * i_chunk_row_out * in_sds_info->non_xy_dims_size;

          for (i=0, ic1=0, ic=0; i<out_ncol; i++, ic1 += res)
          {
	    res2 = ic1 + res_by_2;
	    ic2 = (res2 < ncol) ? res2 : ic1 + (ncol - ic1)/2;
	    j = st_c + ic2*ndata_sm_in; 
	    for (k=0; k<ndata_sm_out; k++, j++, ic++)
	    {
              switch(in_sds_info->data_type)
              {
	        case  5: ((float32 *)data_out_chunk)[ic + shift_data_out_bip] = ((float32 *)data_line)[j]; break;
	        case  6: ((float64 *)data_out_chunk)[ic + shift_data_out_bip] = ((float64 *)data_line)[j]; break;
                case 20: ((int8 *)data_out_chunk)[ic + shift_data_out_bip] = ((int8 *)data_line)[j]; break;
                case 21: ((uint8 *)data_out_chunk)[ic + shift_data_out_bip] = ((uint8 *)data_line)[j]; break;
                case 22: ((int16 *)data_out_chunk)[ic + shift_data_out_bip] = ((int16 *)data_line)[j]; break;
                case 23: ((uint16 *)data_out_chunk)[ic + shift_data_out_bip] = ((uint16 *)data_line)[j]; break;
                case 24: ((int32 *)data_out_chunk)[ic + shift_data_out_bip] = ((int32 *)data_line)[j]; break;
                case 25: ((uint32 *)data_out_chunk)[ic + shift_data_out_bip] = ((uint32 *)data_line)[j]; break;
              }
	    } /* for (k=0; . . ) */
          } /* for (i=0; . . .) */
        }

      } /* for (i_chunk_row_out=0; . . )	*/

      /* write chunk sample data out, the start won't consider the last chunk effect */
      if ((rank>2) && (n == -1) && (m == -1) && (in_sds_info->dim_size[rank-1]>in_sds_info->dim_size[0]))
	out_start[rank-2] = i_chunk * chunk_rows_out;
      else
	out_start[0] = i_chunk * chunk_rows_out;

      /*if(set_sel_sds_data(&out_sds_info, out_start, out_edge, data_out_chunk) == FAIL)*/
      if(set_sel_sds_data_compression(&out_sds_info, out_start, out_edge, comp_chunk_rows, data_out_chunk) == FAIL)
      {
	fprintf(stderr, "Error writing output to the HDF file in reduce_sds\n");
 	break;
      }
	
      i_chunk++;

    } /* for (I_ROW = 0; ...) */
  }
  close_sds(&out_sds_info, out_sds_info.H5OUT);
  if (data_in_chunk != NULL) free(data_in_chunk);
  if (data_line != NULL) free(data_line);
  if (data_out_chunk != NULL) free(data_out_chunk);
}

void reduce_an_sds_by_avg(sds_t *in_sds_info, int32 out_sd_id, int res, int *out_flag, char avg_dt)
/****************************************************************************
!C

!Description:
  Routine to reduce an sds by local averaging.

!Input Parameters:
  in_sds_info: structure containing input SDS information.
  out_sd_id: ouput SDS id.
  res: reduction factor.
  out_flag: optional output flag (min, max, avg, std_dev, num)
	    (0: no, 1: yes for each output)
  avg_dt: output data type for average data out.
	    (0: integer data type, 1: float data type)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original October 2002. Version 1.0

!Team-unique Header:
  See file prologue.

!Developer(s):

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Maosheng Zhao
    Jan, 2020, mask input and output in chunk style and also the output has internal compression

!Design Notes: (none)

!END
****************************************************************************/
{ 
  int status = 1;
  int n, m, k;
  int n_res, m_res;
  int ic, ic1, ic2;
  int rank, out_rank;
  int out_dim_size[4];
  int c, i, j, ii, jj;
  int ndata_in_res, ndata_out;
  int npix_in_avg;
/*  int fill_val;*/
  int ndata_sm_in, ndata_sm_out;
  int st_c, st_c1, sh_sm, offset;
  int irow, icol, jcol, ncol, out_nrow, out_ncol;
  int32 attr_type1;
  int32 in_edge[4] = {0, 0, 0, 0};
  int32 out_edge[4] = {0, 0, 0, 0};
  int32 in_start[4] = {0, 0, 0, 0};
  int32 out_start[4] = {0, 0, 0, 0};
  sds_t out_sds_info[5];
  double sum2;
  float *sig_buf = NULL, avg, value, sum, min, max;
  void *avg_buf = NULL, *num_buf = NULL, *min_buf = NULL, *max_buf = NULL, *data_in_res = NULL;
  char sds_name[MAX_SDS_NAME_LEN];
  void *fill_val = NULL;

  /* MSZ 1/16/2020 */
  int jobs_num = 30;
  int nrow, ndata_line, n_chunks, ndata_out_chunk;
  int I_ROW, i_chunk, i_chunk_row_out;
  int chunk_rows_in, chunk_rows_out, last_chunk_rows_in;
  int chunk_lines_in, line_start_in;
  size_t ndata_in_chunk;
  void *data_in_chunk = NULL;

  /* MSZ 1/17/2020, for the chunked output */
  int last_chunk_rows_out;
  int chunk_lines_out;
  int shift_data_out_bsq, shift_data_out_bip;

  /* MSZ 1/17/2020, for output with the internal compression */
  int comp_chunk_rows;
  
  /* set the output SDS attributes */
  /* set the number of input and output samples per line */
  printf("in_sds_info->scale = %d\n",in_sds_info->scale);
  printf("in_sds_info->fscale = %f\n",in_sds_info->fscale);
  rank = in_sds_info->rank;
  
  switch (in_sds_info->data_type)
  {
    case 5: if ((fill_val = (float32 *)calloc(1, sizeof(float32))) != NULL)
	       ((float32 *)fill_val)[0] = (float32)in_sds_info->fill_fval;
	    else status = -1;   
	    break; 
    case 6: if ((fill_val = (float64 *)calloc(1, sizeof(float64))) != NULL)
	       ((float64 *)fill_val)[0] = (float64)in_sds_info->fill_fval;
	     break; 	     
    case 20: if ((fill_val = (int8 *)calloc(1, sizeof(int8))) != NULL)
	       ((int8 *)fill_val)[0] = (int8)in_sds_info->fill_val;
	     else status = -1;
	     break; 
    case 21: if ((fill_val = (uint8 *)calloc(1, sizeof(uint8))) != NULL)
	       ((uint8 *)fill_val)[0] = (uint8)in_sds_info->fill_val;
	     else status = -1;
	     break; 
    case 22: if ((fill_val = (int16 *)calloc(1, sizeof(int16))) != NULL)
	       ((int16 *)fill_val)[0] = (int16)in_sds_info->fill_val;
	     else status = -1;
	     break; 
    case 23: if ((fill_val = (uint16 *)calloc(1, sizeof(uint16))) != NULL)
	       ((uint16 *)fill_val)[0] = (uint16)in_sds_info->fill_val;
	     else status = -1;
	     break; 
    case 24: if ((fill_val = (int32 *)calloc(1, sizeof(int32))) != NULL)
	       ((int32 *)fill_val)[0] = (int32)in_sds_info->fill_val;
	     else status = -1;
	     break; 
    case 25: if ((fill_val = (uint32 *)calloc(1, sizeof(uint32))) != NULL)
	       ((uint32 *)fill_val)[0] = (uint32)in_sds_info->fill_val;
	     else status = -1;
	     break; 
  }        

  get_sdsname_dim(in_sds_info->name, sds_name, &n, &m);
  out_rank = ((n == -1) && (m == -1)) ? rank : 2;
  if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))   /* BIP */
  {
    /* MSZ 1/30/2020 */    
    nrow = in_sds_info->dim_size[0];

    out_nrow = out_dim_size[0] = (int)ceil((float)in_sds_info->dim_size[0]/res);
    out_ncol = out_dim_size[1] = (int)ceil((float)in_sds_info->dim_size[1]/res);
    for (i=2; i<out_rank; i++)
      out_dim_size[i] = in_sds_info->dim_size[i];
    ncol = ndata_line = in_sds_info->dim_size[1];
    for (i=2; i<rank; i++) 
      ndata_line *= in_sds_info->dim_size[i];
    ndata_out = out_dim_size[1];
    for (i=2; i<out_rank; i++) 
      ndata_out *= out_dim_size[i];
  }
  else                                                                             /* BSQ */
  {
    /* MSZ 1/30/2020 */
    nrow = in_sds_info->dim_size[rank-2];

    if (out_rank == 2)
    {
      out_nrow = out_dim_size[0] = (int)ceil((float)in_sds_info->dim_size[rank-2]/res);
      out_ncol = out_dim_size[1] = (int)ceil((float)in_sds_info->dim_size[rank-1]/res);
      ndata_out = out_dim_size[1];
    }
    else
    {
      out_nrow = out_dim_size[rank-2] = (int)ceil((float)in_sds_info->dim_size[rank-2]/res);
      out_ncol = out_dim_size[rank-1] = (int)ceil((float)in_sds_info->dim_size[rank-1]/res);
      for (i=0; i<rank-2; i++) 
        out_dim_size[i] = in_sds_info->dim_size[i];
      ndata_out = out_dim_size[rank-1];
      for (i=0; i<out_rank-2; i++) 
        ndata_out *= out_dim_size[i];
    }
    ncol = ndata_line = in_sds_info->dim_size[rank-1];
    for (i=0; i<rank-2; i++) 
      ndata_line *= in_sds_info->dim_size[i];
  }
  ndata_in_res = ndata_line * res;

  /* MSZ 1/30/2020, to enhance the performacne so that I/O chunk by chunk */
  /**************************************************************************/
  /* set output compression's chunk */
  comp_chunk_rows = 30;
  if(comp_chunk_rows > out_nrow) comp_chunk_rows = out_nrow;     /* to fix comp_chunk_rows */

  /* MSZ 1/16/2020, get chunk_rows_in and chunk_rows_out (times of res) */
  chunk_rows_in = (int)ceil((double)nrow / jobs_num);

  if(chunk_rows_in > 0)
    chunk_rows_out = (int)ceil((double)chunk_rows_in / res);

  if(chunk_rows_out < 1) chunk_rows_out = 1;

  /* correct chunk_rows_in so that it reads chunk in chunk_rows_out times of res.
 *    * This also links chunk_rows_out with chunk_rows_in */
  chunk_rows_in = res * chunk_rows_out;

  /* get n_chunks and rows of the last chunk in */
  /* Note that n_chunks applies to both input chunks and output chunks,
 *    * i.e, both input and output have the same number of chunks */
  n_chunks = (int)ceil((float)nrow / (float)(chunk_rows_in));

  /* number of rows for the last chunk of input and output */
  last_chunk_rows_in  = nrow % chunk_rows_in;
  last_chunk_rows_out = out_nrow % chunk_rows_out;

  /* number of elements for ndata_in_chunk and chunked ndata_out_chunk */
  ndata_in_chunk  = (size_t)ndata_line * (size_t)chunk_rows_in;
  ndata_out_chunk = ndata_out * chunk_rows_out;
  /**************************************************************************/

  /* set the SDS data types for optional output SDS */

  for (i=0; i<5; i++)
    if (out_flag[i] == 1)
    {
      out_sds_info[i].H5IN = 0;    /* nothing to do with input format */
      out_sds_info[i].H5OUT = 0;   /* so far, output to HDF4 */
      out_sds_info[i].sds_id = -1;
      out_sds_info[i].sd_id = out_sd_id;
      out_sds_info[i].rank = out_rank;
      for (j=0; j<out_rank; j++)
        out_sds_info[i].dim_size[j] = out_dim_size[j];
      switch(i)
      {
	case 0: sprintf(out_sds_info[i].name, "%s_avg", in_sds_info->name);
	        out_sds_info[i].data_type = (avg_dt == 'y') ? DFNT_FLOAT32 : in_sds_info->data_type;
	        out_sds_info[i].data_size = DFKNTsize(out_sds_info[i].data_type);
                if ((avg_buf = (void *)calloc(ndata_out_chunk, out_sds_info[i].data_size)) == NULL)
		{
      		  fprintf(stderr, "Cannot allocate memory for avg_buf in reduce_an_sds_by_avg\n");
		  status = -1;
		}
	    	break;
	case 1: sprintf(out_sds_info[i].name, "%s_min", in_sds_info->name);
		out_sds_info[i].data_type = (avg_dt == 'y') ? DFNT_FLOAT32 : in_sds_info->data_type;
	        out_sds_info[i].data_size = DFKNTsize(out_sds_info[i].data_type);
                if ((min_buf = (void *)calloc(ndata_out_chunk, out_sds_info[i].data_size)) == NULL)
		{
      		  fprintf(stderr, "Cannot allocate memory for min_buf in reduce_an_sds_by_avg\n");
		  status = -1;
		}
		break;
	case 2: sprintf(out_sds_info[i].name, "%s_max", in_sds_info->name);
		out_sds_info[i].data_type = (avg_dt == 'y') ? DFNT_FLOAT32 : in_sds_info->data_type;
	        out_sds_info[i].data_size = DFKNTsize(out_sds_info[i].data_type);
                if ((max_buf = (void *)calloc(ndata_out_chunk, out_sds_info[i].data_size)) == NULL)
		{
      		  fprintf(stderr, "Cannot allocate memory for max_buf in reduce_an_sds_by_avg\n");
		  status = -1;
		}
		break;
	case 3: sprintf(out_sds_info[i].name, "%s_sig", in_sds_info->name);
	        out_sds_info[i].data_type = DFNT_FLOAT32;
	        out_sds_info[i].data_size = DFKNTsize(out_sds_info[i].data_type);
                if ((sig_buf = (void *)calloc(ndata_out_chunk, out_sds_info[i].data_size)) == NULL)
		{
      		  fprintf(stderr, "Cannot allocate memory for sig_buf in reduce_an_sds_by_avg\n");
		  status = -1;
		}
		break;
	case 4: sprintf(out_sds_info[i].name, "%s_num", in_sds_info->name);
	        out_sds_info[i].data_type = (res*res < 256) ? DFNT_INT8 : DFNT_INT16;
	        out_sds_info[i].data_size = DFKNTsize(out_sds_info[i].data_type);
                if ((num_buf = (void *)calloc(ndata_out_chunk, out_sds_info[i].data_size)) == NULL)
		{
      		  fprintf(stderr, "Cannot allocate memory for num_buf in reduce_an_sds_by_avg\n");
		  status = -1;
		}
		break;
      }
      if (open_sds_compression((char *)NULL, &out_sds_info[i], comp_chunk_rows, 'W') == -1) status = -1;
    } /* if (out_flag . . . 	*/

  /* compute start and offset for the selected layer of the input SDS */

  compute_sds_start_offset(in_sds_info, n, m, &st_c, &offset);   

  /* compute the number of output samples per line */
  if (in_sds_info->interleave)
  {
    ndata_sm_out = 1;
    if ((n == -1) && (m == -1))
    {
      for (i=0; i<rank-2; i++)
        ndata_sm_out *= out_sds_info[0].dim_size[i];
    }
    else
    {
      if (m == -1)
	st_c = n*res*ncol;
      else
        st_c = res*ncol*in_sds_info->dim_size[0]*in_sds_info->dim_size[1];
    }
  }
  else
  {
    ndata_sm_in = 1;
    for (i=2; i<rank; i++)
      ndata_sm_in *= in_sds_info->dim_size[i]; 
    ndata_sm_out = ((n == -1) && (m == -1)) ? ndata_sm_in : 1;
  }

  if ((data_in_chunk = (void *)calloc(ndata_in_chunk, in_sds_info->data_size)) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for data_in_chunk in reduce_an_sds_by_avg\n");
    status = -1;
  }

  if ((data_in_res = (void *)calloc(ndata_in_res, in_sds_info->data_size)) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for data_in_res in reduce_an_sds_by_avg\n");
    status = -1;
  }

  if (data_in_res != NULL && data_in_chunk != NULL)
  {
    if (rank == 2)
    {
      in_edge[0] = chunk_rows_in; in_edge[1] = in_sds_info->dim_size[1];
      out_edge[0] = chunk_rows_out; out_edge[1] = out_sds_info[0].dim_size[1];
    }
    else if (rank > 2)
    {
      for (i=0; i<rank; i++) in_edge[i] = in_sds_info->dim_size[i];
      for (i=0; i<out_rank; i++) out_edge[i] = out_sds_info[0].dim_size[i];
      if ((n == -1) && (m == -1))
      {
        if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
	{
	  in_edge[0] = chunk_rows_in; out_edge[0] = chunk_rows_out;
	}
	else 
	{
	  in_edge[rank-2] = chunk_rows_in; out_edge[rank-2] = chunk_rows_out;
	}
      }
      else
      {
        out_edge[0] = chunk_rows_out;
        if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]) in_edge[0] = chunk_rows_in;
        else in_edge[rank-2] = chunk_rows_in;
      }
    }

    /* reads res input lines and average res x res samples */
    /* output one line per res input lines. */
    /* if required create min, max, std, and npix SDS */

    /* MSZ 1/31/2020, chunk by chunk not line by line */
    i_chunk = 0;
    for (I_ROW = 0; I_ROW < nrow; I_ROW += chunk_rows_in)
    {
      /* read the whole chunk of the input data */
      if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank- 1]))  /* BIP */
	in_start[0] = I_ROW;
      else
	in_start[rank-2] = I_ROW;

      /* correct chunk_lines_in, chunk_lines_out, in_edge and out_edge of the row dimension for the last chunk */
      chunk_lines_in = chunk_rows_in;
      chunk_lines_out = chunk_rows_out;

      if(i_chunk == n_chunks - 1)   /* last chunk */
      {
	/* only when the remainder is not zero */
	if(last_chunk_rows_in) chunk_lines_in  = last_chunk_rows_in;
	if(last_chunk_rows_out) chunk_lines_out = last_chunk_rows_out;

        if (rank == 2)
        {
          in_edge[0] = chunk_lines_in; in_edge[1] = in_sds_info->dim_size[1];
          out_edge[0] = chunk_lines_out; out_edge[1] = out_sds_info[0].dim_size[1];
        }
        else if (rank > 2)
        {
          for (i=0; i<rank; i++) in_edge[i] = in_sds_info->dim_size[i];
          for (i=0; i<out_rank; i++) out_edge[i] = out_sds_info[0].dim_size[i];
          if ((n == -1) && (m == -1))
          {
            if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
	    {
	      in_edge[0] = chunk_lines_in; out_edge[0] = chunk_lines_out;
	    }
	    else 
	    {
	      in_edge[rank-2] = chunk_lines_in; out_edge[rank-2] = chunk_lines_out;
	    }
          }
          else
          {
            out_edge[0] = chunk_lines_out;
            if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]) in_edge[0] = chunk_lines_in;
            else in_edge[rank-2] = chunk_lines_in;
          }
        }
      }

      if( read_sel_sds_data(in_sds_info, in_start, in_edge, data_in_chunk) == FAIL)
      {
  	fprintf(stderr, "Error reading data line from SDS %s in reduce_an_sds_by_avg\n", in_sds_info->name);
        break;
      }

      /* process each output line for the chunk with the out rows of last chunk being considered, chunk_lines_out */
      for (i_chunk_row_out=0; i_chunk_row_out<chunk_lines_out; i_chunk_row_out++)
      {
	n_res = res;
	/* correct the n_res for the last output row of the last chunk
 	 * it is the left rows from the input for the last row */
	if(i_chunk == n_chunks - 1 && i_chunk_row_out == chunk_lines_out - 1)
	{
	   if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
	     n_res = in_sds_info->dim_size[0] - (I_ROW + res * i_chunk_row_out);
	   else
	     n_res = in_sds_info->dim_size[rank-2] - (I_ROW + res * i_chunk_row_out);
	}

	/* now the start for each res is i_chunk_row_out * res, and the below is correct */
	line_start_in = i_chunk_row_out * res;

	/* get the data_in_res */
        get_lines_from_chunk(data_in_chunk, in_sds_info->rank, in_edge, in_sds_info->interleave, in_sds_info->data_size, line_start_in, n_res, data_in_res);

	/*if(i_chunk == 0 && i_chunk_row_out == 0)
	   printf("ZMS interleave = %d\n", in_sds_info->interleave);*/

	if (in_sds_info->interleave == 1)                             /* BSQ */
	{
          if ((n == -1) && (m == -1)) offset = ncol;
          else
          {
	    if (m == -1) offset = ncol*n_res;
	    else offset = ncol*(in_sds_info->dim_size[0]*in_sds_info->dim_size[1])*res;
          }
	  sh_sm = n_res*ncol;
	  st_c1 = st_c;
	  /*for (k=0, ic=0; k<ndata_sm_out; k++, st_c1 += sh_sm)*/
	  for (k=0; k<ndata_sm_out; k++, st_c1 += sh_sm)
	  {
	    /* MSZ 01/17/2020, add the shift_data_out_bsq for BSQ chunked output data_out,
	     * but here last chunk should be considered because out_edge accounts for that
	     * so chunk_lines_out not chunk_rows_out should be used. */
            shift_data_out_bsq = i_chunk_row_out * out_ncol + k * out_ncol * chunk_lines_out;

	    /*for (ic1=0, ic2=0; ic1<out_ncol; ic1++, ic2 += res, ic++)*/
	    for (ic=0, ic1=0, ic2=0; ic1<out_ncol; ic1++, ic2 += res, ic++)
	    {
	      sum = 0.0;
	      sum2 = 0.0;
	      min = 120000;
	      max = -120000;
	      npix_in_avg = 0;
	      /* MSZ 2/3/2020, correct the bug of m_res for last out row */
	      /*m_res = (ic2 < ncol) ? res : (ncol - ic2 + res);*/
	      m_res = res;
	      if(ic1 == out_ncol - 1 && ncol % res)
		m_res = ncol % res;

	      for (ii=0, icol=ic2; ii<n_res; ii++, icol += ncol)     /* ii is Y dim widow */
	      {
	        jcol = st_c1 + icol;
	        for (jj=0; jj<m_res; jj++, jcol++)                   /* jj is X dim window */
	        {
                  switch (in_sds_info->data_type)
                  {
                    case  5: value = ((float32 *)data_in_res)[jcol];
                	     if (value > ((float32 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       sum += value;
                	       if ((out_flag[1] == 1) && (min > value)) min = value;
                	       if ((out_flag[2] == 1) && (max < value)) max = value;
                	       if (out_flag[3] == 1) sum2 += value*value;
                	     }		     
		             break;
                    case  6: value = ((float64 *)data_in_res)[jcol]; 
                	     if (value > ((float64 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       sum += value;
                	       if ((out_flag[1] == 1) && (min > value)) min = value;
                	       if ((out_flag[2] == 1) && (max < value)) max = value;
                	       if (out_flag[3] == 1) sum2 += value*value;
                	     }			    	     
			     break;
                    case 20: value = ((int8 *)data_in_res)[jcol]; 
                	     if (value != ((int8 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       sum += value;
                	       if ((out_flag[1] == 1) && (min > value)) min = value;
                	       if ((out_flag[2] == 1) && (max < value)) max = value;
                	       if (out_flag[3] == 1) sum2 += value*value;
                	     }			             
			     break;
                    case 21: value = ((uint8 *)data_in_res)[jcol]; 
                	     if (value != ((uint8 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       sum += value;
                	       if ((out_flag[1] == 1) && (min > value)) min = value;
                	       if ((out_flag[2] == 1) && (max < value)) max = value;
                	       if (out_flag[3] == 1) sum2 += value*value;
                	     }		             
			     break;
                    case 22: value = ((int16 *)data_in_res)[jcol]; 
                             /*if(i_chunk == 0 && i_chunk_row_out == 0 && ic1 == 225)
				printf("ZMS11 %f %d\n", value, in_sds_info->fill_val);*/
		    	     if (in_sds_info->fill_val < 0) {
                	       if (value >= ((int16 *)fill_val)[0])
                	       {
	        		 npix_in_avg++;
				 if (in_sds_info->fscale != 0) {				 
			           if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;				  				 
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;
				 }
				 else {
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;			 
				 }
                	       }
			     }	
			     else {
	    	       	       if (value < ((int16 *)fill_val)[0])
                	       {
	        		 npix_in_avg++;
				 if (in_sds_info->fscale != 0) {				 
			           if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;				  				 
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;
				 }
				 else {
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;			 
				 }
                	       }		     
			     }	             
			     break;
                    case 23: value = ((uint16 *)data_in_res)[jcol]; 		    		    
               	     	     if (value < ((uint16 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       if (in_sds_info->fscale != 0) {				 
			         if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;	
				 sum += value;
                		 if ((out_flag[1] == 1) && (min > value)) min = value;
                		 if ((out_flag[2] == 1) && (max < value)) max = value;
                		 if (out_flag[3] == 1) sum2 += value*value;
			       }
			       else {
				 sum += value;
                		 if ((out_flag[1] == 1) && (min > value)) min = value;
                		 if ((out_flag[2] == 1) && (max < value)) max = value;
                		 if (out_flag[3] == 1) sum2 += value*value;			       
			       }
                	     }		    
		    	     break;
                    case 24: value = ((int32 *)data_in_res)[jcol]; 
		             if (in_sds_info->fill_val < 0) {
                	       if (value >= ((int32 *)fill_val)[0])
                	       {
	        		 npix_in_avg++;
				 if (in_sds_info->fscale != 0) {				 
			           if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;	
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;
				 }
				 else {
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;				 
				 }
                	       }	
			     }
			     else {
                	       if (value < ((int32 *)fill_val)[0])
                	       {
	        		 npix_in_avg++;
				 if (in_sds_info->fscale != 0) {				 
			           if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;	
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;
				 }
				 else {
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;				 
				 }
                	       }			     
			     }	    
		             break;
                    case 25: value = (float)((uint32 *)data_in_res)[jcol]; 
                	     if (value < ((uint32 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       if (in_sds_info->fscale != 0) {				 
				 if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;	
				 sum += value;
                		 if ((out_flag[1] == 1) && (min > value)) min = value;
                		 if ((out_flag[2] == 1) && (max < value)) max = value;
                		 if (out_flag[3] == 1) sum2 += value*value;
			       }
			       else {
				 sum += value;
                		 if ((out_flag[1] == 1) && (min > value)) min = value;
                		 if ((out_flag[2] == 1) && (max < value)) max = value;
                		 if (out_flag[3] == 1) sum2 += value*value;			       
			       }
                	     }		    
		             break;
                  }   	     	  
	        }  /* for (jj=0; . . . ) */
	      }  /* for (ii=0; . . .) */

              switch (in_sds_info->data_type)
              {
                case  5: avg = (npix_in_avg == 0) ? ((float32 *)fill_val)[0] : sum/npix_in_avg; break;	     
                case  6: avg = (npix_in_avg == 0) ? ((float64 *)fill_val)[0] : sum/npix_in_avg; break;
                case 20: avg = (npix_in_avg == 0) ? ((int8 *)fill_val)[0] : sum/npix_in_avg; break;
                case 21: avg = (npix_in_avg == 0) ? ((uint8 *)fill_val)[0] : sum/npix_in_avg; break;
                case 22: avg = (npix_in_avg == 0) ? ((int16 *)fill_val)[0] : sum/npix_in_avg; break;
                case 23: avg = (npix_in_avg == 0) ? ((uint16 *)fill_val)[0] : sum/npix_in_avg; break;
                case 24: avg = (npix_in_avg == 0) ? ((int32 *)fill_val)[0] : sum/npix_in_avg; break;
		case 25: avg = (npix_in_avg == 0) ? ((uint32 *)fill_val)[0] : sum/npix_in_avg; break;
              }   	     	        

	      /*if(i_chunk == 0 && i_chunk_row_out == 0 && ic1 == 225)
		printf("ZMS22 %d %d %d %f\n", value, in_sds_info->fill_val, npix_in_avg, sum);*/

	      if (avg_dt == 'y') {
	        if (npix_in_avg == 0) {
		  if (in_sds_info->data_type == 5) 
		    ((float32 *)avg_buf)[ic + shift_data_out_bsq] = in_sds_info->fill_fval;
		  else 
		    ((float32 *)avg_buf)[ic + shift_data_out_bsq] = (float32)FILL_VALUE_INT32;
		} 
		else {
		  ((float32 *)avg_buf)[ic + shift_data_out_bsq] = sum/npix_in_avg;
		} 
	      }
	      else
	      {
                switch (in_sds_info->data_type)
                {
                  case  5: ((float32 *)avg_buf)[ic + shift_data_out_bsq] = (float32)avg; break;
                  case  6: ((float64 *)avg_buf)[ic + shift_data_out_bsq] = (float64)avg; break;
                  case 20: ((int8 *)avg_buf)[ic + shift_data_out_bsq] = (int8)(avg + 0.5); break;
                  case 21: ((uint8 *)avg_buf)[ic + shift_data_out_bsq] = (uint8)(avg + 0.5); break;
                  case 22: ((int16 *)avg_buf)[ic + shift_data_out_bsq] = (int16)(avg + 0.5); break;
                  case 23: ((uint16 *)avg_buf)[ic + shift_data_out_bsq] = (uint16)(avg + 0.5); break;
                  case 24: ((int32 *)avg_buf)[ic + shift_data_out_bsq] = (int32)(avg + 0.5); break;
                  case 25: ((uint32 *)avg_buf)[ic + shift_data_out_bsq] = (uint32)(avg + 0.5); break;
                }
	      }
	      
              if (out_flag[1] == 1)
              {
	        if (avg_dt == 'y') {
		  if (npix_in_avg == 0) {
		    if (in_sds_info->data_type == 5) 		   
		      ((float32 *)min_buf)[ic + shift_data_out_bsq] = in_sds_info->fill_fval;
		    else ((float32 *)min_buf)[ic + shift_data_out_bsq] = (float32)FILL_VALUE_INT32;
		  }
		  else ((float32 *)min_buf)[ic + shift_data_out_bsq] = (float32)min;
		}
		else 
		{
                  switch (in_sds_info->data_type)
                  {
                    case  5: ((float32 *)min_buf)[ic + shift_data_out_bsq] = (float32)min; break;
                    case  6: ((float64 *)min_buf)[ic + shift_data_out_bsq] = (float64)min; break;
                    case 20: ((int8 *)min_buf)[ic + shift_data_out_bsq] = (int8)min; break;
                    case 21: ((uint8 *)min_buf)[ic + shift_data_out_bsq] = (uint8)min; break;
                    case 22: ((int16 *)min_buf)[ic + shift_data_out_bsq] = (int16)min; break;
                    case 23: ((uint16 *)min_buf)[ic + shift_data_out_bsq] = (uint16)min; break;
                    case 24: ((int32 *)min_buf)[ic + shift_data_out_bsq] = (int32)min; break;
                    case 25: ((uint32 *)min_buf)[ic + shift_data_out_bsq] = (uint32)min; break;
                  }
		}
              }
	      
              if (out_flag[2] == 1)
              {
	        if (avg_dt == 'y') {
		  if (npix_in_avg == 0) {
		    if (in_sds_info->data_type == 5) 		   
		      ((float32 *)max_buf)[ic + shift_data_out_bsq] = in_sds_info->fill_fval;
		    else ((float32 *)max_buf)[ic + shift_data_out_bsq] = (float32)FILL_VALUE_INT32;
		  }
		  else ((float32 *)max_buf)[ic + shift_data_out_bsq] = (float32)max;
		}
		else 
		{
                  switch (in_sds_info->data_type)
                  {
                    case  5: ((float32 *)max_buf)[ic + shift_data_out_bsq] = (float32)max; break;
                    case  6: ((float64 *)max_buf)[ic + shift_data_out_bsq] = (float64)max; break;
                    case 20: ((int8 *)max_buf)[ic + shift_data_out_bsq] = (int8)max; break;
                    case 21: ((uint8 *)max_buf)[ic + shift_data_out_bsq] = (uint8)max; break;
                    case 22: ((int16 *)max_buf)[ic + shift_data_out_bsq] = (int16)max; break;
                    case 23: ((uint16 *)max_buf)[ic + shift_data_out_bsq] = (uint16)max; break;
                    case 24: ((int32 *)max_buf)[ic + shift_data_out_bsq] = (int32)max; break;
                    case 25: ((uint32 *)max_buf)[ic + shift_data_out_bsq] = (uint32)max; break;
                  }
		}	        
              }
	      
              if (out_flag[3] == 1)
		sig_buf[ic + shift_data_out_bsq] = (npix_in_avg == 0) ? (float32)FILL_VALUE_INT32 : sqrt(sum2/npix_in_avg - avg*avg);
              if (out_flag[4] == 1)
              {
                if (out_sds_info[4].data_type == DFNT_INT8)
                  ((int8 *)num_buf)[ic + shift_data_out_bsq] = (int8)npix_in_avg;
                else if (out_sds_info[4].data_type == DFNT_INT16)
                  ((int16 *)num_buf)[ic + shift_data_out_bsq] = (int16)npix_in_avg;
              }
	    }  /* for (ic1=0; . . .) */
	  }  /* for (k=0; . . .) */
	}  /* if (in_sds_info->interleave == 1)   */
	else                                                         /* BIP */
	{
          /* MSZ 01/17/2020, add the shift_data_out for BIP chunked output data_out,
           * but here last chunk should be considered because out_edge accounts for that, here for
           * BIP, Y is controled by i_chunk_row_out only */
          if(out_rank == 2)
             shift_data_out_bip = out_ncol * i_chunk_row_out;
          else
             shift_data_out_bip = out_ncol * i_chunk_row_out * in_sds_info->non_xy_dims_size;

	  sh_sm = ncol*ndata_sm_in;
	  for (ic1=0, ic2=0, ic=0; ic1<out_ncol; ic1++, ic2 += res)
	  {
	    st_c1 = st_c + ic1*res*ndata_sm_in;
	    /* MSZ 2/3/2020, correct the bug of m_res for last out row */
	    /*m_res = (ic2 < ncol) ? res : (ncol - ic2 + res);*/
	    m_res = res;
	    if(ic1 == out_ncol - 1 && ncol % res)
		m_res = ncol % res;

	    for (k=0; k<ndata_sm_out; k++, ic++, st_c1++)
	    {
	      sum = 0.0;
	      sum2 = 0.0;
	      min = 120000;
	      max = -120000;
	      npix_in_avg = 0;
	      for (ii=0; ii<n_res; ii++)
	      {
		icol = st_c1 + ii*sh_sm;
	        for (jj=0, jcol=icol; jj<m_res; jj++, jcol += ndata_sm_in)
	        {
                  switch (in_sds_info->data_type)
                  {
                    case  5: value = ((float32 *)data_in_res)[jcol];
                	     if (value > ((float32 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       sum += value;
                	       if ((out_flag[1] == 1) && (min > value)) min = value;
                	       if ((out_flag[2] == 1) && (max < value)) max = value;
                	       if (out_flag[3] == 1) sum2 += value*value;
                	     }		     
		             break;
                    case  6: value = ((float64 *)data_in_res)[jcol]; 
                	     if (value > ((float64 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       sum += value;
                	       if ((out_flag[1] == 1) && (min > value)) min = value;
                	       if ((out_flag[2] == 1) && (max < value)) max = value;
                	       if (out_flag[3] == 1) sum2 += value*value;
                	     }			    	     
			     break;
                    case 20: value = ((int8 *)data_in_res)[jcol]; 
                	     if (value != ((int8 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       sum += value;
                	       if ((out_flag[1] == 1) && (min > value)) min = value;
                	       if ((out_flag[2] == 1) && (max < value)) max = value;
                	       if (out_flag[3] == 1) sum2 += value*value;
                	     }			             
			     break;
                    case 21: value = ((uint8 *)data_in_res)[jcol]; 
                	     if (value != ((uint8 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       sum += value;
                	       if ((out_flag[1] == 1) && (min > value)) min = value;
                	       if ((out_flag[2] == 1) && (max < value)) max = value;
                	       if (out_flag[3] == 1) sum2 += value*value;
                	     }		             
			     break;
                    case 22: value = ((int16 *)data_in_res)[jcol]; 
			     /*if(i_chunk == 0 && i_chunk_row_out == 0 && ic1 == 224)
				printf("ZMS value= %d\n", value);*/

		    	     if (in_sds_info->fill_val < 0) {
                	       if (value >= ((int16 *)fill_val)[0])
                	       {
	        		 npix_in_avg++;
				 if (in_sds_info->fscale != 0) {				 
			           if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;				  				 
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;
				 }
				 else {
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;			 
				 }
                	       }
			     }	
			     else {
	    	       	       if (value < ((int16 *)fill_val)[0])
                	       {
	        		 npix_in_avg++;
				 if (in_sds_info->fscale != 0) {				 
			           if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;				  				 
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;
				 }
				 else {
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;			 
				 }
                	       }		     
			     }	             
			     break;
                    case 23: value = ((uint16 *)data_in_res)[jcol]; 		    		    
               	     	     if (value < ((uint16 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       if (in_sds_info->fscale != 0) {				 
			         if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;	
				 sum += value;
                		 if ((out_flag[1] == 1) && (min > value)) min = value;
                		 if ((out_flag[2] == 1) && (max < value)) max = value;
                		 if (out_flag[3] == 1) sum2 += value*value;
			       }
			       else {
				 sum += value;
                		 if ((out_flag[1] == 1) && (min > value)) min = value;
                		 if ((out_flag[2] == 1) && (max < value)) max = value;
                		 if (out_flag[3] == 1) sum2 += value*value;			       
			       }
                	     }		    
		    	     break;
                    case 24: value = ((int32 *)data_in_res)[jcol]; 
		             if (in_sds_info->fill_val < 0) {
                	       if (value >= ((int32 *)fill_val)[0])
                	       {
	        		 npix_in_avg++;
				 if (in_sds_info->fscale != 0) {				 
			           if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;	
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;
				 }
				 else {
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;				 
				 }
                	       }	
			     }
			     else {
                	       if (value < ((int32 *)fill_val)[0])
                	       {
	        		 npix_in_avg++;
				 if (in_sds_info->fscale != 0) {				 
			           if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;	
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;
				 }
				 else {
				   sum += value;
                		   if ((out_flag[1] == 1) && (min > value)) min = value;
                		   if ((out_flag[2] == 1) && (max < value)) max = value;
                		   if (out_flag[3] == 1) sum2 += value*value;				 
				 }
                	       }			     
			     }	    
		             break;
                    case 25: value = (float)((uint32 *)data_in_res)[jcol]; 
                	     if (value < ((uint32 *)fill_val)[0])
                	     {
	        	       npix_in_avg++;
			       if (in_sds_info->fscale != 0) {				 
				 if (avg_dt == 'y') value = value * in_sds_info->fscale + in_sds_info->foffset;	
				 sum += value;
                		 if ((out_flag[1] == 1) && (min > value)) min = value;
                		 if ((out_flag[2] == 1) && (max < value)) max = value;
                		 if (out_flag[3] == 1) sum2 += value*value;
			       }
			       else {
				 sum += value;
                		 if ((out_flag[1] == 1) && (min > value)) min = value;
                		 if ((out_flag[2] == 1) && (max < value)) max = value;
                		 if (out_flag[3] == 1) sum2 += value*value;			       
			       }
                	     }		    
		             break;
                  }   	     	  			          
	        }  /* for (jj=0; . . . ) */
	      }  /* for (ii=0; . . .) */
      
              switch (in_sds_info->data_type)
              {
                case  5: avg = (npix_in_avg == 0) ? ((float32 *)fill_val)[0] : sum/npix_in_avg; break;	     
                case  6: avg = (npix_in_avg == 0) ? ((float64 *)fill_val)[0] : sum/npix_in_avg; break;
                case 20: avg = (npix_in_avg == 0) ? ((int8 *)fill_val)[0] : sum/npix_in_avg; break;
                case 21: avg = (npix_in_avg == 0) ? ((uint8 *)fill_val)[0] : sum/npix_in_avg; break;
                case 22: avg = (npix_in_avg == 0) ? ((int16 *)fill_val)[0] : sum/npix_in_avg; break;
                case 23: avg = (npix_in_avg == 0) ? ((uint16 *)fill_val)[0] : sum/npix_in_avg; break;
                case 24: avg = (npix_in_avg == 0) ? ((int32 *)fill_val)[0] : sum/npix_in_avg; break;
		case 25: avg = (npix_in_avg == 0) ? ((uint32 *)fill_val)[0] : sum/npix_in_avg; break;
              }   	     	        
	      if (avg_dt == 'y') {
	        if (npix_in_avg == 0) {
		  if (in_sds_info->data_type == 5) 
		    ((float32 *)avg_buf)[ic + shift_data_out_bip] = in_sds_info->fill_fval;
		  else 
		    ((float32 *)avg_buf)[ic + shift_data_out_bip] = (float32)FILL_VALUE_INT32;
		} 
		else {
		  ((float32 *)avg_buf)[ic + shift_data_out_bip] = sum/npix_in_avg;
		} 
	      }
	      else
	      {
                switch (in_sds_info->data_type)
                {
                  case  5: ((float32 *)avg_buf)[ic + shift_data_out_bip] = (float32)avg; break;
                  case  6: ((float64 *)avg_buf)[ic + shift_data_out_bip] = (float64)avg; break;
                  case 20: ((int8 *)avg_buf)[ic + shift_data_out_bip] = (int8)(avg + 0.5); break;
                  case 21: ((uint8 *)avg_buf)[ic + shift_data_out_bip] = (uint8)(avg + 0.5); break;
                  case 22: ((int16 *)avg_buf)[ic + shift_data_out_bip] = (int16)(avg + 0.5); break;
                  case 23: ((uint16 *)avg_buf)[ic + shift_data_out_bip] = (uint16)(avg + 0.5); break;
                  case 24: ((int32 *)avg_buf)[ic + shift_data_out_bip] = (int32)(avg + 0.5); break;
                  case 25: ((uint32 *)avg_buf)[ic + shift_data_out_bip] = (uint32)(avg + 0.5); break;
                }
	      }
	      
              if (out_flag[1] == 1)
              {
	        if (avg_dt == 'y') {
		  if (npix_in_avg == 0) {
		    if (in_sds_info->data_type == 5) 		   
		      ((float32 *)min_buf)[ic + shift_data_out_bip] = in_sds_info->fill_fval;
		    else ((float32 *)min_buf)[ic + shift_data_out_bip] = (float32)FILL_VALUE_INT32;
		  }
		  else ((float32 *)min_buf)[ic + shift_data_out_bip] = (float32)min;
		}
		else 
		{
                  switch (in_sds_info->data_type)
                  {
                    case  5: ((float32 *)min_buf)[ic + shift_data_out_bip] = (float32)min; break;
                    case  6: ((float64 *)min_buf)[ic + shift_data_out_bip] = (float64)min; break;
                    case 20: ((int8 *)min_buf)[ic + shift_data_out_bip] = (int8)min; break;
                    case 21: ((uint8 *)min_buf)[ic + shift_data_out_bip] = (uint8)min; break;
                    case 22: ((int16 *)min_buf)[ic + shift_data_out_bip] = (int16)min; break;
                    case 23: ((uint16 *)min_buf)[ic + shift_data_out_bip] = (uint16)min; break;
                    case 24: ((int32 *)min_buf)[ic + shift_data_out_bip] = (int32)min; break;
                    case 25: ((uint32 *)min_buf)[ic + shift_data_out_bip] = (uint32)min; break;
                  }
		}
              }
	      
              if (out_flag[2] == 1)
              {
	        if (avg_dt == 'y') {
		  if (npix_in_avg == 0) {
		    if (in_sds_info->data_type == 5) 		   
		      ((float32 *)max_buf)[ic + shift_data_out_bip] = in_sds_info->fill_fval;
		    else ((float32 *)max_buf)[ic + shift_data_out_bip] = (float32)FILL_VALUE_INT32;
		  }
		  else ((float32 *)max_buf)[ic + shift_data_out_bip] = (float32)max;
		}
		else 
		{
                  switch (in_sds_info->data_type)
                  {
                    case  5: ((float32 *)max_buf)[ic + shift_data_out_bip] = (float32)max; break;
                    case  6: ((float64 *)max_buf)[ic + shift_data_out_bip] = (float64)max; break;
                    case 20: ((int8 *)max_buf)[ic + shift_data_out_bip] = (int8)max; break;
                    case 21: ((uint8 *)max_buf)[ic + shift_data_out_bip] = (uint8)max; break;
                    case 22: ((int16 *)max_buf)[ic + shift_data_out_bip] = (int16)max; break;
                    case 23: ((uint16 *)max_buf)[ic + shift_data_out_bip] = (uint16)max; break;
                    case 24: ((int32 *)max_buf)[ic + shift_data_out_bip] = (int32)max; break;
                    case 25: ((uint32 *)max_buf)[ic + shift_data_out_bip] = (uint32)max; break;
                  }
		}	        
              }
	      
              if (out_flag[3] == 1)
		sig_buf[ic + shift_data_out_bip] = (npix_in_avg == 0) ? (float32)FILL_VALUE_INT32 : sqrt(sum2/npix_in_avg - avg*avg);
              if (out_flag[4] == 1)
              {
                if (out_sds_info[4].data_type == DFNT_INT8)
                  ((int8 *)num_buf)[ic + shift_data_out_bip] = (int8)npix_in_avg;
                else if (out_sds_info[4].data_type == DFNT_INT16)
                  ((int16 *)num_buf)[ic + shift_data_out_bip] = (int16)npix_in_avg;
              }      
	    }  /* for (k=0; . . .) */
	  }  /* for (ic1=0; . . .) */
	} /* else  . . .  */
      }  /* for (i_chunk_row_out=0; i_chunk_row_out<out_nrow; i_chunk_row_out++) */ 	        

      /* write chunk sample data out, the start won't consider the last chunk effect */
      if ((rank>2) && (n == -1) && (m == -1) && (in_sds_info->dim_size[rank-1]>in_sds_info->dim_size[0]))
        out_start[rank-2] = i_chunk * chunk_rows_out;
      else
        out_start[0] = i_chunk * chunk_rows_out;

      if(set_sel_sds_data_compression(&out_sds_info[0], out_start, out_edge, comp_chunk_rows, avg_buf) == FAIL)
      {
	fprintf(stderr, "Error writing line of SDS to HDF file in reduce_an_sds_by_avg\n");
	break;
      }
      if (out_flag[1] == 1)
      {
	if(set_sel_sds_data_compression(&out_sds_info[1], out_start, out_edge, comp_chunk_rows, min_buf) == FAIL)
	{
	  fprintf(stderr, "Error writing line of SDS to HDF file in reduce_an_sds_by_avg\n");
	  break;
	}
      }
      if (out_flag[2] == 1)
      {
	if(set_sel_sds_data_compression(&out_sds_info[2], out_start, out_edge, comp_chunk_rows, max_buf) == FAIL)
	{
	  fprintf(stderr, "Error writing line of SDS to HDF file in reduce_an_sds_by_avg\n");
	  break;
	}
      }
      if (out_flag[3] == 1)
      {
	if(set_sel_sds_data_compression(&out_sds_info[3], out_start, out_edge, comp_chunk_rows, sig_buf) == FAIL) 
	{
	  fprintf(stderr, "Error writing line of SDS to HDF file in reduce_an_sds_by_avg\n");
	  break;
	}
      }
      if (out_flag[4] == 1)
      {
	if(set_sel_sds_data_compression(&out_sds_info[4], out_start, out_edge, comp_chunk_rows, num_buf) == FAIL)
	{
	  fprintf(stderr, "Error writing line of SDS to HDF file in reduce_an_sds_by_avg\n");
	  break;
	}
      }

      i_chunk++;

    }  /* for (I_ROW = 0; I_ROW < nrow; I_ROW += chunk_rows_in) */
    for (i=0; i<5; i++)
      if (out_flag[i] == 1)
      {
	switch(i)
	{
      	  case 0: attr_type1 = (avg_dt == 'y') ? DFNT_FLOAT32 : in_sds_info->data_type; c = 0; break;
	  case 1: attr_type1 = (avg_dt == 'y') ? DFNT_FLOAT32 : in_sds_info->data_type; c = 0; break;
	  case 2: attr_type1 = (avg_dt == 'y') ? DFNT_FLOAT32 : in_sds_info->data_type; c = 0; break;
	  case 3: attr_type1 = DFNT_FLOAT32; c = 0; break;	  
	  case 4: attr_type1 = DFNT_INT8; c = 0; break;  
	}
	write_attr_fval(out_sds_info[i].H5OUT, out_sds_info[i].sds_id, attr_type1, c, ((uint16 *)fill_val)[0]);	
      }
  }  /* if ((data_in_res != NULL) . . . 	*/
  close_sds(in_sds_info, in_sds_info->H5IN);
  close_sds(&out_sds_info[0], out_sds_info[0].H5OUT);
  if(fill_val) free(fill_val);	
  if(data_in_res) free(data_in_res);
  if(data_in_chunk) free(data_in_chunk);
  if(avg_buf) free(avg_buf);
  if (out_flag[1] == 1) { free(min_buf); close_sds(&out_sds_info[1], out_sds_info[1].H5OUT); }
  if (out_flag[2] == 1) { free(max_buf); close_sds(&out_sds_info[2], out_sds_info[2].H5OUT); }
  if (out_flag[3] == 1) { free(sig_buf); close_sds(&out_sds_info[3], out_sds_info[3].H5OUT); }
  if (out_flag[4] == 1) { free(num_buf); close_sds(&out_sds_info[4], out_sds_info[4].H5OUT); } 
}


void reduce_an_sds_by_cnt(sds_t *in_sds_info, int32 out_sd_id, int res, char *cmp_str)
/****************************************************************************
!C

!Description:
  Routine to reduce an sds by outputing the count of the number of pixels
  that satisfy input data value constraint.

!Input Parameters:
  in_sds_info: structure containing input SDS information.
  out_sd_id: ouput SDS id.
  res: reduction factor.
  cmp_str: string containing the data value constraint.

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original October 2002. Version 1.0

!Team-unique Header:
  See file prologue.

!Developer(s):

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Maosheng Zhao
    Jan, 2020, mask input and output in chunk style and also the output has internal compression

!Design Note(s): (none)

!End
******************************************************************************/
{ 
  int **bn_arr;
  int n, m;
  int status = 1;
  int st_c, st_c1;
  int ic, ic1, ic2;
  int n_res, m_res;
  int sh_sm, offset;
  int rank, out_rank;
  int out_dim_size[4];
  int i, j, k, ii, jj;
  int cnt[MAX_NUM_SDS];
  int isds, nsds, opr_id;
  int out_nrow, out_ncol;
  int ndata_in_res, ndata_out;
  int bn_val[MAX_NUM_BITS];
  int bn_cnt[MAX_NUM_BITS];
  int bn_opr[MAX_NUM_BITS];
  int ncol, irow, icol, jcol;
  int value, fill_val=-888, sds_val, vsds[MAX_NUM_SDS], mval[MAX_NUM_SDS];
  int ndata_sm_in, ndata_sm_out;
  int32 attr_type, attr_cnt;
  int32 in_edge[4] = {0, 0, 0, 0};
  int32 out_edge[4] = {0, 0, 0, 0};
  int32 in_start[4] = {0, 0, 0, 0};
  int32 out_start[4] = {0, 0, 0, 0};
  sds_t *out_sds_info;
  char sds_name[MAX_SDS_NAME_LEN];
  void *attr_val, *data_in_res, *data_out[MAX_NUM_SDS] = {NULL};
  char *opr[] = {"<=", ">=", "==", "!=", "<", ">"};

  /* MSZ 1/16/2020 */
  int jobs_num = 30;
  int nrow, ndata_line, n_chunks, ndata_out_chunk;
  int I_ROW, i_chunk, i_chunk_row_out;
  int chunk_rows_in, chunk_rows_out, last_chunk_rows_in;
  int chunk_lines_in, line_start_in;
  size_t ndata_in_chunk;
  void *data_in_chunk = NULL;

  /* MSZ 1/17/2020, for the chunked output */
  int last_chunk_rows_out;
  int chunk_lines_out;
  int shift_data_out_bsq, shift_data_out_bip;

  /* MSZ 1/17/2020, for output with the internal compression */
  int comp_chunk_rows;

  if ((bn_arr = (int **)Calloc2D(MAX_NUM_BITS, 16, sizeof(int))) == NULL)
    fprintf(stderr, "Cannot allocate memory for bn_arr in reduce_sds_by_cnt\n");

  /* get the number of ouptut SDS based on the bit constraint */
  /* also get the bit number array and associated value */

  if ((nsds = get_bit_opt(cmp_str, bn_arr, bn_val, bn_opr, bn_cnt)) == 0)
    fprintf(stderr, "-bn option value %s is incorrect or empty\n", cmp_str);
  if (nsds != 0)
    if ((out_sds_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL)
      fprintf(stderr, "Cannot allocate memory for out_sds_info in reduce_sds_by_cnt\n");

  /* set output SDS attributes */
  /* compute number of input and output samples per line */

  if ((out_sds_info != NULL) && (bn_arr != NULL) && (nsds != 0))
  {
    switch(in_sds_info->data_type)
    {      
      case 20: fill_val = (int8)in_sds_info->fill_val; break;
      case 21: fill_val = (uint8)in_sds_info->fill_val; break;
      case 22: fill_val = (int16)in_sds_info->fill_val; break;
      case 23: fill_val = (uint16)in_sds_info->fill_val; break;
      case 24: fill_val = (int32)in_sds_info->fill_val; break;
      case 25: fill_val = (uint32)in_sds_info->fill_val; break;      
    }          
    
    if ((attr_val = get_sds_attr(in_sds_info, "_FillValue", &attr_type, &attr_cnt)) == NULL)
    {
      fprintf(stderr, "Attribute fill value not defined for output.\n");
    }
    else
    {
      if (fill_val == -888)
	switch(attr_type)
	{
          case 20: fill_val = ((int8 *)attr_val)[0]; break;
          case 21: fill_val = ((uint8 *)attr_val)[0]; break;
          case 22: fill_val = ((int16 *)attr_val)[0]; break;
          case 23: fill_val = ((uint16 *)attr_val)[0]; break;
          case 24: fill_val = ((int32 *)attr_val)[0]; break;
          case 25: fill_val = ((uint32 *)attr_val)[0]; break;
	}
    }
    rank = in_sds_info->rank;
    get_sdsname_dim(in_sds_info->name, sds_name, &n, &m);
    out_rank = ((n == -1) && (m == -1)) ? rank : 2;
    if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
    {
      /* MSZ 1/30/2020 */
      nrow = in_sds_info->dim_size[0];

      out_nrow = out_dim_size[0] = (int)ceil((float)in_sds_info->dim_size[0]/res);
      out_ncol = out_dim_size[1] = (int)ceil((float)in_sds_info->dim_size[1]/res);
      for (i=2; i<out_rank; i++)
        out_dim_size[i] = in_sds_info->dim_size[i];
      ncol = ndata_line = in_sds_info->dim_size[1];
      for (i=2; i<rank; i++)
        ndata_line *= in_sds_info->dim_size[i];
      ndata_out = out_dim_size[1];
      for (i=2; i<out_rank; i++)
        ndata_out *= out_dim_size[i];
    }
    else
    {

      /* MSZ 1/30/2020 */
      nrow = in_sds_info->dim_size[rank-2];

      if (out_rank == 2)
      {
        out_nrow = out_dim_size[0] = (int)ceil((float)in_sds_info->dim_size[rank-2]/res);
        out_ncol = out_dim_size[1] = (int)ceil((float)in_sds_info->dim_size[rank-1]/res);
        ndata_out = out_dim_size[1];
      }
      else
      {
        out_nrow = out_dim_size[rank-2] = (int)ceil((float)in_sds_info->dim_size[rank-2]/res);
        out_ncol = out_dim_size[rank-1] = (int)ceil((float)in_sds_info->dim_size[rank-1]/res);
        for (i=0; i<rank-2; i++)
          out_dim_size[i] = in_sds_info->dim_size[i];
        ndata_out = out_dim_size[rank-1];
        for (i=0; i<out_rank-2; i++)
          ndata_out *= out_dim_size[i];
      }
      ncol = ndata_line = in_sds_info->dim_size[rank-1];
      for (i=0; i<rank-2; i++)
        ndata_line *= in_sds_info->dim_size[i];
    }

    ndata_in_res = ndata_line * res;

    /* MSZ 1/30/2020, to enhance the performacne so that I/O chunk by chunk */
    /**************************************************************************/
    /* set output compression's chunk */
    comp_chunk_rows = 30;
    if(comp_chunk_rows > out_nrow) comp_chunk_rows = out_nrow;     /* to fix comp_chunk_rows */

    /* MSZ 1/16/2020, get chunk_rows_in and chunk_rows_out (times of res) */
    chunk_rows_in = (int)ceil((double)nrow / jobs_num);

    if(chunk_rows_in > 0)
      chunk_rows_out = (int)ceil((double)chunk_rows_in / res);

    if(chunk_rows_out < 1) chunk_rows_out = 1;

    /* correct chunk_rows_in so that it reads chunk in chunk_rows_out times of res.
     * This also links chunk_rows_out with chunk_rows_in */
    chunk_rows_in = res * chunk_rows_out;

    /* get n_chunks and rows of the last chunk in */
    /* Note that n_chunks applies to both input chunks and output chunks,
     * i.e, both input and output have the same number of chunks */
    n_chunks = (int)ceil((float)nrow / (float)(chunk_rows_in));

    /* number of rows for the last chunk of input and output */
    last_chunk_rows_in  = nrow % chunk_rows_in;
    last_chunk_rows_out = out_nrow % chunk_rows_out;

    /* number of elements for ndata_in_chunk and chunked ndata_out_chunk */
    ndata_in_chunk  = (size_t)ndata_line * (size_t)chunk_rows_in;
    ndata_out_chunk = ndata_out * chunk_rows_out;
    /**************************************************************************/

    /* set the output SDS name and open the output SDS */

    for (isds=0; isds<nsds; isds++)
    {
      out_sds_info[isds].H5IN = 0;     /* nothing to do with the input file format */
      out_sds_info[isds].H5OUT = 0;    /* so far, output to HDF4 */
      out_sds_info[isds].sds_id = -1;
      out_sds_info[isds].sd_id = out_sd_id;
      out_sds_info[isds].rank = out_rank;
      for (j=0; j<out_rank; j++)
        out_sds_info[isds].dim_size[j] = out_dim_size[j];
      opr_id = bn_opr[isds];
      if (bn_cnt[isds] == 0)
	sprintf(out_sds_info[isds].name, "%s_cnt_sds%s%d", in_sds_info->name, opr[opr_id], 
		bn_val[isds]);
      else if (bn_cnt[isds] == 1) 
	sprintf(out_sds_info[isds].name, "%s_cnt_bit%d%s%d", in_sds_info->name, 
		bn_arr[isds][0], opr[opr_id], bn_val[isds]);
      else
	sprintf(out_sds_info[isds].name, "%s_cnt_bits%d-%d%s%d", in_sds_info->name, 
		bn_arr[isds][0], bn_arr[isds][0]+bn_cnt[isds]-1, opr[opr_id], bn_val[isds]);
      out_sds_info[isds].data_type = (res <= 16) ? DFNT_UINT8 : DFNT_UINT16;
      out_sds_info[isds].data_size = DFKNTsize(out_sds_info[isds].data_type);
      /*if (open_sds((char *)NULL, &out_sds_info[isds], 'W') != -1)*/
      if (open_sds_compression((char *)NULL, &out_sds_info[isds], comp_chunk_rows, 'W') != -1)
      {
        if (res <= 16) write_attr_fval(out_sds_info[isds].H5OUT, out_sds_info[isds].sds_id, DFNT_UINT8, 0, fill_val);
        else write_attr_fval(out_sds_info[isds].H5OUT, out_sds_info[isds].sds_id, DFNT_UINT16, 0, fill_val);
      }
    } /* for (isds=0; . . . ) */

    /* compute the start and offset of requested layer */

    compute_sds_start_offset(in_sds_info, n, m, &st_c, &offset);
    if (in_sds_info->interleave == 1)
    {
      ndata_sm_out = 1;
      if ((n == -1) && (m == -1))
      {
        for (i=0; i<rank-2; i++)
          ndata_sm_out *= out_sds_info[0].dim_size[i];
      }
      else
      {
        if (m == -1)
          st_c = n*res*ncol;
        else
          st_c = res*ncol*in_sds_info->dim_size[0]*in_sds_info->dim_size[1];
      }
    }
    else
    {
      ndata_sm_in = 1;
      for (i=2; i<rank; i++)
        ndata_sm_in *= in_sds_info->dim_size[i];
      ndata_sm_out = ((n == -1) && (m == -1)) ? ndata_sm_in : 1;
    }

    if ((data_in_chunk = (void *)calloc(ndata_in_chunk, in_sds_info->data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for data_in_chunk in reduce_an_sds_by_cnt\n");

    if ((data_in_res = (void *)calloc(ndata_in_res, in_sds_info->data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for data_in_res in reduce_an_sds_by_cnt\n");

    for (isds=0; isds<nsds; isds++)
      if ((data_out[isds] = (void *)calloc(ndata_out_chunk, out_sds_info[isds].data_size)) == NULL)
      {
        fprintf(stderr, "Cannot allocate memory for data_out in reduce_an_sds_by_cnt\n");
	status = -1;
      }

    if (data_in_chunk != NULL && data_in_res != NULL && (status != -1))
    {
      if (rank == 2)
      {
        in_edge[0] = chunk_rows_in; in_edge[1] = in_sds_info->dim_size[1];
        out_edge[0] = chunk_rows_out; out_edge[1] = out_sds_info[0].dim_size[1];
      }
      else if (rank > 2)
      {
        for (i=0; i<rank; i++) in_edge[i] = in_sds_info->dim_size[i];
        for (i=0; i<out_rank; i++) out_edge[i] = out_sds_info[0].dim_size[i];
        if ((n == -1) && (m == -1))
        {
          if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
	  {
	    in_edge[0] = chunk_rows_in; out_edge[0] = chunk_rows_out;
	  }
	  else 
	  {
	    in_edge[rank-2] = chunk_rows_in; out_edge[rank-2] = chunk_rows_out;
	  }
        }
        else
        {
          out_edge[0] = chunk_rows_out;
          if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]) in_edge[0] = chunk_rows_in;
          else in_edge[rank-2] = chunk_rows_in;
        }
      }

      /* read res lines from input and output one sample per res x res input */
      /* output one line per res line of input data */

      i_chunk = 0;
      for (I_ROW = 0; I_ROW < nrow; I_ROW += chunk_rows_in)
      {
        /* read the whole chunk of the input data */
        if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank- 1]))  /* BIP */
	  in_start[0] = I_ROW;
        else
	  in_start[rank-2] = I_ROW;

        /* correct chunk_lines_in, chunk_lines_out, in_edge and out_edge of the row dimension for the last chunk */
        chunk_lines_in = chunk_rows_in;
        chunk_lines_out = chunk_rows_out;

        if(i_chunk == n_chunks - 1)   /* last chunk */
        {
	  /* only when the remainder is not zero */
	  if(last_chunk_rows_in) chunk_lines_in  = last_chunk_rows_in;
	  if(last_chunk_rows_out) chunk_lines_out = last_chunk_rows_out;

          if (rank == 2)
          {
            in_edge[0] = chunk_lines_in; in_edge[1] = in_sds_info->dim_size[1];
            out_edge[0] = chunk_lines_out; out_edge[1] = out_sds_info[0].dim_size[1];
          }
          else if (rank > 2)
          {
            for (i=0; i<rank; i++) in_edge[i] = in_sds_info->dim_size[i];
            for (i=0; i<out_rank; i++) out_edge[i] = out_sds_info[0].dim_size[i];
            if ((n == -1) && (m == -1))
            {
              if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
	      {
	        in_edge[0] = chunk_lines_in; out_edge[0] = chunk_lines_out;
	      }
	      else 
	      {
	        in_edge[rank-2] = chunk_lines_in; out_edge[rank-2] = chunk_lines_out;
	      }
            }
            else
            {
              out_edge[0] = chunk_lines_out;
              if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]) in_edge[0] = chunk_lines_in;
              else in_edge[rank-2] = chunk_lines_in;
            }
          }
        }
 
	if(read_sel_sds_data(in_sds_info, in_start, in_edge, data_in_chunk) == FAIL) 
        {
          fprintf(stderr, "Error reading data line from SDS %s in reduce_an_sds_by_avg\n", in_sds_info->name);
          break;
        }

	for (i_chunk_row_out=0; i_chunk_row_out<chunk_lines_out; i_chunk_row_out++)
	{
	  n_res = res;
	  /* correct the n_res for the last output row of the last chunk
 	   * it is the left rows from the input for the last row */
	  if(i_chunk == n_chunks - 1 && i_chunk_row_out == chunk_lines_out - 1)
	  {
	    if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
	      n_res = in_sds_info->dim_size[0] - (I_ROW + res * i_chunk_row_out);
	    else
	      n_res = in_sds_info->dim_size[rank-2] - (I_ROW + res * i_chunk_row_out);
	  }

	  /* now the start for each res is i_chunk_row_out * res, and the below is correct */
	  line_start_in = i_chunk_row_out * res;

	  /* get the data_in_res */
          get_lines_from_chunk(data_in_chunk, in_sds_info->rank, in_edge, in_sds_info->interleave, in_sds_info->data_size, line_start_in, n_res, data_in_res);

          if (in_sds_info->interleave == 1)
          {
            if ((n == -1) && (m == -1))
              offset = ncol;
            else
            {
              if (m == -1)
                offset = ncol*n_res;
              else
                offset = ncol*(in_sds_info->dim_size[0]*in_sds_info->dim_size[1])*res;
            }
            sh_sm = n_res*ncol;
            st_c1 = st_c;
            /*for (k=0, ic=0; k<ndata_sm_out; k++, st_c1 += sh_sm)*/
	    for (k=0; k<ndata_sm_out; k++, st_c1 += sh_sm)
            {
	      /* MSZ 01/17/2020, add the shift_data_out_bsq for BSQ chunked output data_out,
	       * but here last chunk should be considered because out_edge accounts for that
	       * so chunk_lines_out not chunk_rows_out should be used. */
              shift_data_out_bsq = i_chunk_row_out * out_ncol + k * out_ncol * chunk_lines_out;

              /*for (ic1=0, ic2=0; ic1<out_ncol; ic1++, ic2 += res, ic++) */
	      for (ic=0, ic1=0, ic2=0; ic1<out_ncol; ic1++, ic2 += res, ic++)
              {
		for (isds=0; isds<nsds; isds++)
		{
		  cnt[isds] = 0;
		  vsds[isds] = 0;                /* SSarkar */
		  mval[isds] = MAX_VAL_INIT;     /* SSarkar */
		}
		  /* MSZ 2/3/2020, correct the bug of m_res for last out row */
		  /*m_res = (ic2 < ncol) ? res : (ncol - ic2 + res);*/
		  m_res = res;
		  if(ic1 == out_ncol - 1 && ncol % res)
		    m_res = ncol % res;

                for (ii=0, icol=ic2; ii<n_res; ii++, icol += ncol)
	        {
                  jcol = st_c1 + icol;
                  for (jj=0; jj<m_res; jj++, jcol++)
		  {
		    switch (in_sds_info->data_type)
                    {
                      case 20: value = ((int8 *)data_in_res)[jcol]; break;
                      case 21: value = ((uint8 *)data_in_res)[jcol]; break;
                      case 22: value = ((int16 *)data_in_res)[jcol]; break;
                      case 23: value = ((uint16 *)data_in_res)[jcol]; break;
                      case 24: value = ((int32 *)data_in_res)[jcol]; break;
                      case 25: value = ((uint32 *)data_in_res)[jcol]; break;
		    }	
		    if (ii == n_res/2 && jj == m_res/2)  /* SSarkar */
		    {
		      for (isds=0; isds<nsds; isds++)
			vsds[isds] = value;
		    }

		    if (value != fill_val)
		      for (isds=0; isds<nsds; isds++)
		      {
			sds_val = value;
		        if (bn_cnt[isds] != 0)
		        {
			  if (bn_arr[isds][0] > 0) 
			    sds_val = sds_val >> bn_arr[isds][0];
		          sds_val &= BIT[bn_cnt[isds]-1];
		        }
		        switch(bn_opr[isds])
		        {
		          case 0: if (sds_val <= bn_val[isds])
				  {
				    cnt[isds]++;
				    mval[isds] = MAX(sds_val,mval[isds]);
				  }
				  break;
		          case 1: if (sds_val >= bn_val[isds])
				  {
				    cnt[isds]++;
				    mval[isds] = MAX(sds_val,mval[isds]);
				  }
				  break;
		          case 2: if (sds_val == bn_val[isds])
				  {
				    cnt[isds]++;
				    mval[isds] = sds_val;
				  }
				  break;
		          case 3: if (sds_val != bn_val[isds])
				  {
				    cnt[isds]++;
				    mval[isds] = MAX(sds_val,mval[isds]);
				  }
				  break;
		          case 4: if (sds_val < bn_val[isds])
				  {
				    cnt[isds]++;
				    mval[isds] = MAX(sds_val,mval[isds]);
				  }
				  break;
		          case 5: if (sds_val > bn_val[isds])
				  {
				    cnt[isds]++;
				    mval[isds] = MAX(sds_val,mval[isds]);
				  }
				  break;
		        }
		      }  /* for (isds=0; . . . ) */
		  }  /* for (jj=0; . . .) */
		}  /* for (ii=0; . . . ) */
		for (isds=0; isds<nsds; isds++)
		{
	           if (cnt[isds] > 0) 
		  {
		    if (res <= 16) ((uint8 *)data_out[isds])[ic + shift_data_out_bsq] = mval[isds];
	            else ((uint16 *)data_out[isds])[ic + shift_data_out_bsq] = mval[isds];		    		    
		  }
          	  else
		  {	    
		    switch(out_sds_info[isds].data_type)
		    {
		      case  5: ((float32 *)data_out[isds])[ic + shift_data_out_bsq] = (float32)(vsds[isds]); break;
		      case  6: ((float64 *)data_out[isds])[ic + shift_data_out_bsq] = (float64)vsds[isds]; break;
		      case 20: ((int8 *)data_out[isds])[ic + shift_data_out_bsq] = (int8)vsds[isds]; break;
		      case 21: ((uint8 *)data_out[isds])[ic + shift_data_out_bsq] = (uint8)vsds[isds]; break;
		      case 22: ((int16 *)data_out[isds])[ic + shift_data_out_bsq] = (int16)vsds[isds]; break;
		      case 23: ((uint16 *)data_out[isds])[ic + shift_data_out_bsq] = (uint16)vsds[isds]; break;
		      case 24: ((int32 *)data_out[isds])[ic + shift_data_out_bsq] = (int32)vsds[isds]; break;
		      case 25: ((uint32 *)data_out[isds])[ic + shift_data_out_bsq] = (uint32)vsds[isds]; break;
		    }    
		  }
		}
	      }  /* for (ic1=0; . . .)  */
	    }  /* for (k=0; . . . . ) 	*/
	  } 
	  else
	  {
            /* MSZ 01/17/2020, add the shift_data_out for BIP chunked output data_out,
             * but here last chunk should be considered because out_edge accounts for that, here for
             * BIP, Y is controled by i_chunk_row_out only */
            if(out_rank == 2)
              shift_data_out_bip = out_ncol * i_chunk_row_out;
            else
              shift_data_out_bip = out_ncol * i_chunk_row_out * in_sds_info->non_xy_dims_size;

            sh_sm = ncol*ndata_sm_in;
            for (ic1=0, ic2=0, ic=0; ic1<out_ncol; ic1++, ic2 += res)
            {
              st_c1 = st_c + ic1*res*ndata_sm_in;
	      /* MSZ 2/3/2020, correct the bug of m_res for last out row */
	      /*m_res = (ic2 < ncol) ? res : (ncol - ic2 + res);*/
	      m_res = res;
	      if(ic1 == out_ncol - 1 && ncol % res)
		m_res = ncol % res;

              for (k=0; k<ndata_sm_out; k++, ic++, st_c1++)
              {
		for (isds=0; isds<nsds; isds++)
		{
		  cnt[isds] = 0;
		  vsds[isds] = 0;                 /* SSarkar */
		  mval[isds] = MAX_VAL_INIT;      /* SSarkar */
		}
                for (ii=0; ii<n_res; ii++)
                {
                  icol = st_c1 + ii*sh_sm;
                  for (jj=0, jcol=icol; jj<m_res; jj++, jcol += ndata_sm_in)
                  {
                    switch (in_sds_info->data_type)
                    {
                      case 20: value = ((int8 *)data_in_res)[jcol]; break;
                      case 21: value = ((uint8 *)data_in_res)[jcol]; break;
                      case 22: value = ((int16 *)data_in_res)[jcol]; break;
                      case 23: value = ((uint16 *)data_in_res)[jcol]; break;
                      case 24: value = ((int32 *)data_in_res)[jcol]; break;
                      case 25: value = ((uint32 *)data_in_res)[jcol]; break;
                    }
		    if (ii == n_res/2 && jj == m_res/2)  /* SSarkar */
		    {
		      for (isds=0; isds<nsds; isds++)
			vsds[isds] = value;
		    }
                    if (value != fill_val)
                      for (isds=0; isds<nsds; isds++)
                      {
                        sds_val = value;
                        if (bn_cnt[isds] != 0)
                        {
                          if (bn_arr[isds][0] > 0)
                            sds_val = sds_val >> bn_arr[isds][0];
                          sds_val &= BIT[bn_cnt[isds]-1];
                        }
                        switch(bn_opr[isds])
                        {
                            case 0: if (sds_val <= bn_val[isds])
                            {
                              cnt[isds]++;
                              mval[isds] = MAX(sds_val,mval[isds]);
                            }
                            break;
                            case 1: if (sds_val >= bn_val[isds])
                            {
                              cnt[isds]++;
                              mval[isds] = MAX(sds_val,mval[isds]);
                            }
                            break;
                            case 2: if (sds_val == bn_val[isds])
                            {
                              cnt[isds]++;
                              mval[isds] = sds_val;
                            }
                            break;
                            case 3: if (sds_val != bn_val[isds])
                            {
                              cnt[isds]++;
                              mval[isds] = MAX(sds_val,mval[isds]);
                            }
                            break;
                            case 4: if (sds_val < bn_val[isds])
                            {
                              cnt[isds]++;
                              mval[isds] = MAX(sds_val,mval[isds]);
                            }
                            break;
                            case 5: if (sds_val > bn_val[isds])
                            {
                              cnt[isds]++;
                              mval[isds] = MAX(sds_val,mval[isds]);
                            }
                            break;
                        }
                      } /* for (isds=0; . . . ) */
                  }  /* for (jj=0; . . . ) */
                }  /* for (ii=0; . . .) */
		for (isds=0; isds<nsds; isds++)
		{
	          if (cnt[isds] > 0) 
            	  {
		    if (res <= 16) ((uint8 *)data_out[isds])[ic + shift_data_out_bip] = mval[isds];
		    else ((uint16 *)data_out[isds])[ic + shift_data_out_bip] = mval[isds];		    		    
		  }
		  else
		  {	    
		    switch(out_sds_info[isds].data_type)
		    {
		      case  5: ((float32 *)data_out[isds])[ic + shift_data_out_bip] = (float32)(vsds[isds]); break;
		      case  6: ((float64 *)data_out[isds])[ic + shift_data_out_bip] = (float64)vsds[isds]; break;
                      case 20: ((int8 *)data_out[isds])[ic + shift_data_out_bip] = (int8)vsds[isds]; break;
                      case 21: ((uint8 *)data_out[isds])[ic + shift_data_out_bip] = (uint8)vsds[isds]; break;
                      case 22: ((int16 *)data_out[isds])[ic + shift_data_out_bip] = (int16)vsds[isds]; break;
                      case 23: ((uint16 *)data_out[isds])[ic + shift_data_out_bip] = (uint16)vsds[isds]; break;
                      case 24: ((int32 *)data_out[isds])[ic + shift_data_out_bip] = (int32)vsds[isds]; break;
                      case 25: ((uint32 *)data_out[isds])[ic + shift_data_out_bip] = (uint32)vsds[isds]; break;		      
                    } 		    
		   }		
		}
	      }  /* for (k=0; . . . ) */
	    }  /* for (ic1=0; . .  ) */
	  } /* else . . . .  */
        }  /* for (i_chunk_row_out=0; i_chunk_row_out<chunk_lines_out; i_chunk_row_out++)*/ 

        /* write chunk sample data out, the start won't consider the last chunk effect */
        if ((rank>2) && (n == -1) && (m == -1) && (in_sds_info->dim_size[rank-1]>in_sds_info->dim_size[0]))
          out_start[rank-2] = i_chunk * chunk_rows_out;
        else
          out_start[0] = i_chunk * chunk_rows_out;

	for (isds=0; isds<nsds; isds++)
	  if(set_sel_sds_data_compression(&out_sds_info[isds], out_start, out_edge, comp_chunk_rows, data_out[isds]) == FAIL)
	  {
	    fprintf(stderr, "Error writing line of SDS to file in reduce_an_sds_by_cnt\n");
            break;
	  }

        i_chunk++;

      } /* for (I_ROW = 0; I_ROW < nrow; I_ROW += chunk_rows_in) */
      free(data_in_chunk);
      free(data_in_res);
      for (isds=0; isds<nsds; isds++)
      {
        if(data_out[isds]) free(data_out[isds]);
      }
    }  /* if (data_in_chunk != NULL && data_in_res != NULL && (status != -1))*/
    free(attr_val);
    close_sds(in_sds_info, in_sds_info->H5IN);
    for (isds=0; isds<nsds; isds++)
      if (out_sds_info[isds].sds_id != -1) close_sds(&out_sds_info[isds], out_sds_info[isds].H5OUT);
  }  
  free(out_sds_info);
  return;
}

int get_bit_opt(char *cmp_str, int **bn_arr, int *bn_val, int *bn_opr, int *bn_cnt)
/****************************************************************************
!C

!Description:
  Obtain the bit array and the corresponding bit data value for use in 
  reduce an sds by cnt option.

!Input Parameters:
  cmp_str: input bit number and bit value option string 

!Input/Output Parameters: (none)

!Output Parameters: (none)
  bn_arr: 2-d array containing the start and end bit numbers of
	  each bit range array.
  bn_val: bit range value for each of the range in bn_arr
  bn_opr: bit relational operation for each bn_arr and bn_val
  bn_cnt: number of bit relational constraints 

!Revision History:
  Original October 2002. Version 1.0

!Team-unique Header:
  See file prologue.

!Developer(s):

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Note(s): (none)

!End
******************************************************************************/
{
  int p1, p2, p3;
  int i, j, k, j1; 
  int len, len1, cnt;
  int ibit, st_bit, end_bit;
  char val_str[10], num_str[10], bn_str[20];
  char *opr[] = {"<=", ">=", "==", "!=", "<", ">"};

  len = (int)strlen(cmp_str);
  p1 = p2 = k = 0;
  while (p2 != -1)
  {
    p2 = charpos(cmp_str, ',', p1);
    if (p2 != -1) strmid(cmp_str, p1, p2-p1, bn_str);
    else strmid(cmp_str, p1, len-p1, bn_str);
    for (i=0; i<6; i++)
      if ((p3 = strpos(bn_str, opr[i], 0)) != -1) break;
    if (i >= 6)
      fprintf(stderr, "Ignoring invalid value %s in -bit option \n", bn_str);
    else
    {
      bn_opr[k] = i;
      cnt = 0;
      if (p3 != 0)
      {
        for (j=0; j<p3; j++) 
          if (bn_str[j] == '-') break; 
	  else num_str[j] = bn_str[j];
        num_str[j] = '\0';
	st_bit = (int)atoi(num_str);
        if (j >= p3) end_bit = st_bit;
	else
        {
	  j++;
	  for (j1=j; j1<p3; j1++)
	    num_str[j1-j] = bn_str[j1];
	  num_str[j1-j] = '\0';
	  end_bit = (int)atoi(num_str);
        }
	for (ibit=st_bit; ibit<=end_bit; ibit++, cnt++)
	  bn_arr[k][cnt] = ibit;
      }
      bn_cnt[k] = cnt;
      len1 = (int)strlen(bn_str);
      p3 = (i > 3) ? (p3 + 1) : (p3 + 2);
      strmid(bn_str, p3, len1-p3, val_str);
      bn_val[k] = (int)atoi(val_str);
      k++;
    }
    p1 = p2 + 1;
  }
  return k;
}

void reduce_an_sds_by_class(sds_t *in_sds_info, int32 out_sd_id, int res)
/****************************************************************************
!C

!Description:
  Routine to reduce an sds by classifying the local area to one of the set
  of classes in the input.

!Input Parameters:
  in_sds_info: structure containing input SDS information.
  out_sd_id: ouput SDS id.
  res: reduction factor.

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original October 2002. Version 1.0

!Team-unique Header:
  See file prologue.

!Developer(s):

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

    Maosheng Zhao
    Jan, 2020, mask input and output in chunk style and also the output has internal compression

!Design Note(s): (none)

!End
******************************************************************************/
{ 
  int n, m;
  int st_c, st_c1;
  int ic, ic1, ic2;
  int n_res, m_res;
  int sh_sm, offset;
  int rank, out_rank;
  int out_dim_size[4];
  int i, j, k, ii, jj;
  int value, fill_val=-888;
  int out_nrow, out_ncol;
  int ndata_in_res, ndata_out;
  int ncol, irow, icol, jcol;
  int ndata_sm_in, ndata_sm_out;
  int32 attr_type, attr_cnt;
  int32 in_edge[4] = {0, 0, 0, 0};
  int32 out_edge[4] = {0, 0, 0, 0};
  int32 in_start[4] = {0, 0, 0, 0};
  int32 out_start[4] = {0, 0, 0, 0};
  sds_t out_sds_info;
  char sds_name[MAX_SDS_NAME_LEN];
  void *attr_val = NULL, *data_in_res = NULL, *data_out = NULL;
  int cl_num, class_cnt[MAX_NUM_CLASS], max_cl_id;

  /* MSZ 1/16/2020 */
  int jobs_num = 30;
  int nrow, ndata_line, n_chunks, ndata_out_chunk;
  int I_ROW, i_chunk, i_chunk_row_out;
  int chunk_rows_in, chunk_rows_out, last_chunk_rows_in;
  int chunk_lines_in, line_start_in;
  size_t ndata_in_chunk;
  void *data_in_chunk = NULL;

  /* MSZ 1/17/2020, for the chunked output */
  int last_chunk_rows_out;
  int chunk_lines_out;
  int shift_data_out_bsq, shift_data_out_bip;

  /* MSZ 1/17/2020, for output with the internal compression */
  int comp_chunk_rows;

  switch(in_sds_info->data_type)
  {      
    case 20: fill_val = (int8)in_sds_info->fill_val; break;
    case 21: fill_val = (uint8)in_sds_info->fill_val; break;
    case 22: fill_val = (int16)in_sds_info->fill_val; break;
    case 23: fill_val = (uint16)in_sds_info->fill_val; break;
    case 24: fill_val = (int32)in_sds_info->fill_val; break;
    case 25: fill_val = (uint32)in_sds_info->fill_val; break;      
  }  
  
  if ((attr_val = get_sds_attr(in_sds_info, "_FillValue", &attr_type, &attr_cnt)) == NULL)
    fprintf(stderr, "Attribute fill value not defined for output.\n");  
  else
  {
    if (fill_val == -888)
      switch(attr_type)
      {
	case 20: fill_val = ((int8 *)attr_val)[0]; break;
	case 21: fill_val = ((uint8 *)attr_val)[0]; break;
	case 22: fill_val = ((int16 *)attr_val)[0]; break;
	case 23: fill_val = ((uint16 *)attr_val)[0]; break;
	case 24: fill_val = ((int32 *)attr_val)[0]; break;
	case 25: fill_val = ((uint32 *)attr_val)[0]; break;
      }
  } 
  
  /* set the output SDS attributes */
  /* compute number of input and output samples per output line */

  rank = in_sds_info->rank;
  get_sdsname_dim(in_sds_info->name, sds_name, &n, &m);
  out_rank = ((n == -1) && (m == -1)) ? rank : 2;
  if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
  {
    /* MSZ 1/30/2020 */
    nrow = in_sds_info->dim_size[0];

    out_nrow = out_dim_size[0] = (int)ceil((float)in_sds_info->dim_size[0]/res);
    out_ncol = out_dim_size[1] = (int)ceil((float)in_sds_info->dim_size[1]/res);
    for (i=2; i<out_rank; i++)
      out_dim_size[i] = in_sds_info->dim_size[i];
    ncol = ndata_line = in_sds_info->dim_size[1];
    for (i=2; i<rank; i++)
      ndata_line *= in_sds_info->dim_size[i];
    ndata_out = out_dim_size[1];
    for (i=2; i<out_rank; i++)
      ndata_out *= out_dim_size[i];
  }
  else
  {
    /* MSZ 1/30/2020 */
    nrow = in_sds_info->dim_size[rank-2];

    if (out_rank == 2)
    {
      out_nrow = out_dim_size[0] = (int)ceil((float)in_sds_info->dim_size[rank-2]/res);
      out_ncol = out_dim_size[1] = (int)ceil((float)in_sds_info->dim_size[rank-1]/res);
      ndata_out = out_dim_size[1];
    }
    else
    {
      out_nrow = out_dim_size[rank-2] = (int)ceil((float)in_sds_info->dim_size[rank-2]/res);
      out_ncol = out_dim_size[rank-1] = (int)ceil((float)in_sds_info->dim_size[rank-1]/res);
      for (i=0; i<rank-2; i++)
        out_dim_size[i] = in_sds_info->dim_size[i];
      ndata_out = out_dim_size[rank-1];
      for (i=0; i<out_rank-2; i++)
        ndata_out *= out_dim_size[i];
    }
    ncol = ndata_line = in_sds_info->dim_size[rank-1];
    for (i=0; i<rank-2; i++)
      ndata_line *= in_sds_info->dim_size[i];
  }

  ndata_in_res = ndata_line * res;

  /* MSZ 1/30/2020, to enhance the performacne so that I/O chunk by chunk */
  /**************************************************************************/
  /* set output compression's chunk */
  comp_chunk_rows = 30;
  if(comp_chunk_rows > out_nrow) comp_chunk_rows = out_nrow;     /* to fix comp_chunk_rows */

  /* MSZ 1/16/2020, get chunk_rows_in and chunk_rows_out (times of res) */
  chunk_rows_in = (int)ceil((double)nrow / jobs_num);

  if(chunk_rows_in > 0)
    chunk_rows_out = (int)ceil((double)chunk_rows_in / res);

  if(chunk_rows_out < 1) chunk_rows_out = 1;

  /* correct chunk_rows_in so that it reads chunk in chunk_rows_out times of res.
 *  *    * This also links chunk_rows_out with chunk_rows_in */
  chunk_rows_in = res * chunk_rows_out;

  /* get n_chunks and rows of the last chunk in */
  /* Note that n_chunks applies to both input chunks and output chunks,
 *  *    * i.e, both input and output have the same number of chunks */
  n_chunks = (int)ceil((float)nrow / (float)(chunk_rows_in));

  /* number of rows for the last chunk of input and output */
  last_chunk_rows_in  = nrow % chunk_rows_in;
  last_chunk_rows_out = out_nrow % chunk_rows_out;

  /* number of elements for ndata_in_chunk and chunked ndata_out_chunk */
  ndata_in_chunk  = (size_t)ndata_line * (size_t)chunk_rows_in;
  ndata_out_chunk = ndata_out * chunk_rows_out;
  /**************************************************************************/

  /* open output SDS */
  out_sds_info.H5IN = 0;    /* nothing to do with input file format */
  out_sds_info.H5OUT= 0;    /* so far, output to HDF4 */
  out_sds_info.sds_id = -1;
  out_sds_info.sd_id = out_sd_id;
  out_sds_info.rank = out_rank;
  for (j=0; j<out_rank; j++)
    out_sds_info.dim_size[j] = out_dim_size[j];
  sprintf(out_sds_info.name, "%s", in_sds_info->name);
  out_sds_info.data_type = in_sds_info->data_type; 
  out_sds_info.data_size = DFKNTsize(out_sds_info.data_type);
  /*if (open_sds((char *)NULL, &out_sds_info, 'W') != -1)*/
  if (open_sds_compression((char *)NULL, &out_sds_info, comp_chunk_rows, 'W') != -1)
    write_attr_fval(out_sds_info.H5OUT, out_sds_info.sds_id, in_sds_info->data_type, 1, fill_val);

  /* compute start and offset of requested layer */

  compute_sds_start_offset(in_sds_info, n, m, &st_c, &offset);
  if (in_sds_info->interleave == 1)
  {
    ndata_sm_out = 1;
    if ((n == -1) && (m == -1))
    {
      for (i=0; i<rank-2; i++)
        ndata_sm_out *= out_sds_info.dim_size[i];
    }
    else
    {
      if (m == -1) st_c = n*res*ncol;
      else st_c = res*ncol*in_sds_info->dim_size[0]*in_sds_info->dim_size[1];
    }
  }
  else
  {
    ndata_sm_in = 1;
    for (i=2; i<rank; i++)
      ndata_sm_in *= in_sds_info->dim_size[i];
    ndata_sm_out = ((n == -1) && (m == -1)) ? ndata_sm_in : 1;
  }

  if ((data_in_chunk = (void *)calloc(ndata_in_chunk, in_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_in_chunk in reduce_an_sds_by_cnt\n");

  if ((data_in_res = (void *)calloc(ndata_in_res, in_sds_info->data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_in_res in reduce_an_sds_by_cnt\n");

  if ((data_out = (void *)calloc(ndata_out_chunk, out_sds_info.data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for data_out in reduce_an_sds_by_cnt\n");

  if (data_in_chunk != NULL && data_in_res != NULL && data_out != NULL)
  {
    if (rank == 2)
    {
      in_edge[0] = chunk_rows_in; in_edge[1] = in_sds_info->dim_size[1];
      out_edge[0] = chunk_rows_out; out_edge[1] = out_sds_info.dim_size[1];
    }
    else if (rank > 2)
    {
      for (i=0; i<rank; i++) in_edge[i] = in_sds_info->dim_size[i];
      for (i=0; i<out_rank; i++) out_edge[i] = out_sds_info.dim_size[i];
      if ((n == -1) && (m == -1))
      {
        if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
	{
	  in_edge[0] = chunk_rows_in; out_edge[0] = chunk_rows_out;
	}
	else 
	{
	  in_edge[rank-2] = chunk_rows_in; out_edge[rank-2] = chunk_rows_out;
	}
      }
      else
      {
        out_edge[0] = chunk_rows_out;
        if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]) in_edge[0] = chunk_rows_in;
        else in_edge[rank-2] = chunk_rows_in;
      }
    }

    /* read res lines of input at a time. Process res x res */
    /* block and ouput one line per res line of input */
    i_chunk = 0;
    for (I_ROW = 0; I_ROW < nrow; I_ROW += chunk_rows_in)
    {
      /* read the whole chunk of the input data */
      if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank- 1]))  /* BIP */
	in_start[0] = I_ROW;
      else
	in_start[rank-2] = I_ROW;

      /* correct chunk_lines_in, chunk_lines_out, in_edge and out_edge of the row dimension for the last chunk */
      chunk_lines_in = chunk_rows_in;
      chunk_lines_out = chunk_rows_out;

      if(i_chunk == n_chunks - 1)   /* last chunk */
      {
	/* only when the remainder is not zero */
	if(last_chunk_rows_in) chunk_lines_in  = last_chunk_rows_in;
	if(last_chunk_rows_out) chunk_lines_out = last_chunk_rows_out;

        if (rank == 2)
        {
          in_edge[0] = chunk_lines_in; in_edge[1] = in_sds_info->dim_size[1];
          out_edge[0] = chunk_lines_out; out_edge[1] = out_sds_info.dim_size[1];
        }
        else if (rank > 2)
        {
          for (i=0; i<rank; i++) in_edge[i] = in_sds_info->dim_size[i];
          for (i=0; i<out_rank; i++) out_edge[i] = out_sds_info.dim_size[i];
          if ((n == -1) && (m == -1))
          {
            if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1])
	    {
	      in_edge[0] = chunk_lines_in; out_edge[0] = chunk_lines_out;
	    }
	    else 
	    {
	      in_edge[rank-2] = chunk_lines_in; out_edge[rank-2] = chunk_lines_out;
	    }
          }
          else
          {
            out_edge[0] = chunk_lines_out;
            if (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]) in_edge[0] = chunk_lines_in;
            else in_edge[rank-2] = chunk_lines_in;
          }
        }
      }

      if( read_sel_sds_data(in_sds_info, in_start, in_edge, data_in_chunk) == FAIL)
      {
        fprintf(stderr, "Error reading data line from SDS %s in reduce_an_sds_by_avg\n", in_sds_info->name);
        break;
      }

      for (i_chunk_row_out=0; i_chunk_row_out<chunk_lines_out; i_chunk_row_out++)
      {
	n_res = res;
	/* correct the n_res for the last output row of the last chunk
 	 * it is the left rows from the input for the last row */
	if(i_chunk == n_chunks - 1 && i_chunk_row_out == chunk_lines_out - 1)
	{
	   if ((rank == 2) || (in_sds_info->dim_size[0] > in_sds_info->dim_size[rank-1]))
	     n_res = in_sds_info->dim_size[0] - (I_ROW + res * i_chunk_row_out);
	   else
	     n_res = in_sds_info->dim_size[rank-2] - (I_ROW + res * i_chunk_row_out);
	}

	/* now the start for each res is i_chunk_row_out * res, and the below is correct */
	line_start_in = i_chunk_row_out * res;

	/* get the data_in_res */
        get_lines_from_chunk(data_in_chunk, in_sds_info->rank, in_edge, in_sds_info->interleave, in_sds_info->data_size, line_start_in, n_res, data_in_res);

        if (in_sds_info->interleave == 1)
        {
          if ((n == -1) && (m == -1)) offset = ncol;
          else
          {
            if (m == -1) offset = ncol*n_res;
            else offset = ncol*(in_sds_info->dim_size[0]*in_sds_info->dim_size[1])*res;
          }
          sh_sm = n_res*ncol;
          st_c1 = st_c;
          /*for (k=0, ic=0; k<ndata_sm_out; k++, st_c1 += sh_sm)*/
	  for (k=0; k<ndata_sm_out; k++, st_c1 += sh_sm)
          {
	    /* MSZ 01/17/2020, add the shift_data_out_bsq for BSQ chunked output data_out,
	     * but here last chunk should be considered because out_edge accounts for that
	     * so chunk_lines_out not chunk_rows_out should be used. */
            shift_data_out_bsq = i_chunk_row_out * out_ncol + k * out_ncol * chunk_lines_out;

            /*for (ic1=0, ic2=0; ic1<out_ncol; ic1++, ic2 += res, ic++)*/
	    for (ic=0, ic1=0, ic2=0; ic1<out_ncol; ic1++, ic2 += res, ic++)
            {
	      for (cl_num=0; cl_num<MAX_NUM_CLASS; cl_num++)
		class_cnt[cl_num] = 0;
	      /* MSZ 2/3/2020, correct the bug of m_res for last out row */
	      /*m_res = (ic2 < ncol) ? res : (ncol - ic2 + res);*/
	      m_res = res;
	      if(ic1 == out_ncol - 1 && ncol % res)
		m_res = ncol % res;

              for (ii=0, icol=ic2; ii<n_res; ii++, icol += ncol)
	      {
                jcol = st_c1 + icol;
                for (jj=0; jj<m_res; jj++, jcol++)
		{
		  switch (in_sds_info->data_type)
                  {
                    case 20: value = ((int8 *)data_in_res)[jcol]; break;
                    case 21: value = ((uint8 *)data_in_res)[jcol]; break;
                    case 22: value = ((int16 *)data_in_res)[jcol]; break;
                    case 23: value = ((uint16 *)data_in_res)[jcol]; break;
                    case 24: value = ((int32 *)data_in_res)[jcol]; break;
                    case 25: value = ((uint32 *)data_in_res)[jcol]; break;
		  }	
		  if ((value != fill_val) && (value < MAX_NUM_CLASS))
		    ++class_cnt[value];
		}  /* for (jj=0; . . . ) */
	      }  /* for (ii=0; . . . ) */
	      for (cl_num=0, max_cl_id=0; cl_num<MAX_NUM_CLASS; cl_num++)
	        if (class_cnt[cl_num] > class_cnt[max_cl_id]) 
		  max_cl_id = cl_num;
	      if (class_cnt[max_cl_id] == 0) max_cl_id = fill_val;
	      switch (out_sds_info.data_type)
              {
                case 20: ((int8 *)data_out)[ic + shift_data_out_bsq] = max_cl_id; break;
                case 21: ((uint8 *)data_out)[ic + shift_data_out_bsq] = max_cl_id; break;
                case 22: ((int16 *)data_out)[ic + shift_data_out_bsq] = max_cl_id; break;
                case 23: ((uint16 *)data_out)[ic + shift_data_out_bsq] = max_cl_id; break;
                case 24: ((int32 *)data_out)[ic + shift_data_out_bsq] = max_cl_id; break;
                case 25: ((uint32 *)data_out)[ic + shift_data_out_bsq] = max_cl_id; break;
	      }	
	    }  /* for (ic1=0; . . .)  */
	  }  /* for (k=0; . . . . ) 	*/
	} 
	else
	{
          /* MSZ 01/17/2020, add the shift_data_out for BIP chunked output data_out,
           * but here last chunk should be considered because out_edge accounts for that, here for
           * BIP, Y is controled by i_chunk_row_out only */
          if(out_rank == 2)
             shift_data_out_bip = out_ncol * i_chunk_row_out;
          else
             shift_data_out_bip = out_ncol * i_chunk_row_out * in_sds_info->non_xy_dims_size;

          sh_sm = ncol*ndata_sm_in;
          for (ic1=0, ic2=0, ic=0; ic1<out_ncol; ic1++, ic2 += res)
          {
            st_c1 = st_c + ic1*res*ndata_sm_in;
	    /* MSZ 2/3/2020, correct the bug of m_res for last out row */
	    /*m_res = (ic2 < ncol) ? res : (ncol - ic2 + res);*/
	    m_res = res;
	    if(ic1 == out_ncol - 1 && ncol % res)
	      m_res = ncol % res;

            for (k=0; k<ndata_sm_out; k++, ic++, st_c1++)
            {
	      for (cl_num=0; cl_num<MAX_NUM_CLASS; cl_num++)
		class_cnt[cl_num] = 0;
              for (ii=0; ii<n_res; ii++)
              {
                icol = st_c1 + ii*sh_sm;
                for (jj=0, jcol=icol; jj<m_res; jj++, jcol += ndata_sm_in)
                {
                  switch (in_sds_info->data_type)
                  {
                    case 20: value = ((int8 *)data_in_res)[jcol]; break;
                    case 21: value = ((uint8 *)data_in_res)[jcol]; break;
                    case 22: value = ((int16 *)data_in_res)[jcol]; break;
                    case 23: value = ((uint16 *)data_in_res)[jcol]; break;
                    case 24: value = ((int32 *)data_in_res)[jcol]; break;
                    case 25: value = ((uint32 *)data_in_res)[jcol]; break;
                  }
                  if ((value != fill_val) && (value < MAX_NUM_CLASS))
		    ++class_cnt[value];
                }  /* for (jj=0; . . . ) */
              }  /* for (ii=0; . . .) */
              for (cl_num=0, max_cl_id=0; cl_num<MAX_NUM_CLASS; cl_num++)
                if (class_cnt[cl_num] > class_cnt[max_cl_id])
                  max_cl_id = cl_num;
	      if (class_cnt[max_cl_id] == 0) max_cl_id = fill_val;
              switch (out_sds_info.data_type)
              {
                case 20: ((int8 *)data_out)[ic + shift_data_out_bip] = max_cl_id; break;
                case 21: ((uint8 *)data_out)[ic + shift_data_out_bip] = max_cl_id; break;
                case 22: ((int16 *)data_out)[ic + shift_data_out_bip] = max_cl_id; break;
                case 23: ((uint16 *)data_out)[ic + shift_data_out_bip] = max_cl_id; break;
                case 24: ((int32 *)data_out)[ic + shift_data_out_bip] = max_cl_id; break;
                case 25: ((uint32 *)data_out)[ic + shift_data_out_bip] = max_cl_id; break;
              }
	    }  /* for (k=0; . . . ) */
	  }  /* for (ic1=0; . .  ) */
	} /* else . . . .  */
      } /* for (i_chunk_row_out=0; i_chunk_row_out<chunk_lines_out; i_chunk_row_out++) */

      /* write chunk sample data out, the start won't consider the last chunk effect */
      if ((out_rank>2) && (n == -1) && (m == -1) && (in_sds_info->dim_size[rank-1]>in_sds_info->dim_size[0]))
	out_start[rank-2] = i_chunk * chunk_rows_out;
      else
	out_start[0] = i_chunk * chunk_rows_out;

      if(set_sel_sds_data_compression(&out_sds_info, out_start, out_edge, comp_chunk_rows, data_out) == FAIL)
      { 
	fprintf(stderr, "Error writing output to the HDF file in reduce_sds\n");
	break;
      }

      i_chunk++;

    }  /* for (I_ROW = 0; I_ROW < nrow; I_ROW += chunk_rows_in)	*/ 
    free(data_in_chunk);
    free(data_in_res);
    free(data_out);
  }  /* if ((data_in_res != NULL) . . . 	*/
  free(attr_val);
  close_sds(in_sds_info, in_sds_info->H5IN);
  close_sds(&out_sds_info, out_sds_info.H5OUT);
}  


