/****************************************************************************
!C

!File: reduce_sds_rank.c

!Description:
  Contains routines for converting a mutlidimensional SDS into 
  many SDS of two dimension

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original June 1998 Version 1.1
  Modified January 1999 Version 1.1 (fixed bugs)
  Modified September 1999 (included -meta option)
  Modified May 2000 (dot extension identifying dimensions in SDS names are
			removed to support projection tool)
  Modified March 2005 - Fixed bugs.
  Modified May 2011 by Xiaoping zhang, Sigma Space Corp.
  		      - Fixed bugs for dimension name with bsq is 1.
  Modified July 2011 by Xiaoping zhang, Sigma Space Corp.
  		      - added -fill_value option to input arguments. .		      
  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c

  Modified in Oct. 2018, Version 6 by Maosheng Zhao, SSAI.
	1) speedup in a way to read lines by lines not line by line
	2) add the internal compression for the output when all is switched

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "meta.h"
#include "math.h"

#define HELP \
"NAME \n" \
"     reduce_sds_rank - convert one or more 3D/4D SDS from a MODIS Land \n" \
"	                HDF-EOS data product to many 2D SDSs. \n" \
" \n" \
"SYNOPSIS \n" \
"     reduce_sds_rank -help \n" \
"     reduce_sd_rank [-sds=[SDSname [-dim=dimstr [dim=dimstr. . ]]] \n" \
"			-sds=[SDSname [-dim=dimstr [dim=dimstr. . ]]] . . ] \n" \
"			-all -meta -fill_value=<fillVal> -of=output_file filename \n" \
" \n" \
"     echo [-sds=[SDSname [-dim=dimstr [dim=dimstr. . ]]] \n" \
"                       -sds=[SDSname [-dim=dimstr [dim=dimstr. . ]]] . . ] \n" \
"                       -all -meta -fill_value=<fillVal> -of=output_file filename | reduce_sds_rank \n" \
" or  reduce_sds_rank < argument_file \n" \
" \n" \
"DESCRIPTION \n" \
"    Several MODIS Land HDF-EOS data products (e.g., MOD43, MYD43) and\n"\
"    related MODIS products (e.g., MOD35) contain multidimensional SDSs.\n"\
"    This tool converts one or more multidimensional (3D or 4D) SDS to a\n"\
"    series of 2D HDF SDSs. Specific SDS layers may be selected using the\n"\
"    -sds and -dim options for each input 3D or 4D SDS.\n" \
" \n" \
"    The output file SDS names reflect the input SDS name, the dimension\n"\
"    name and the dimension element numbers. For example: \n" \
"        BRDF_Albedo_Parameters.Num_Land_Bands_Plus3_3.Num_Parameters_1 \n" \
"        (parameter 1 for land band 3 of SDS BRDF_Albedo_Parameters in\n"\
"        MOD43B1. Note that BRDF_Albedo_Parameters is the input SDS\n"\
"        name, Num_Land_Bands_Plus3 and Num_Parameters are the 3rd and\n"\
"        the 4th dimension names) \n" \
"        Surface_Refl.Num_Obs_Max_1.Num_Land_Bands_2 (1st observation of\n"\
"        land band 2 of SDS Surface_Refl in MODAGAGG.  Note that\n"\
"        Surface_Refl is the input SDS name, Num_Obs_Max and Num_Land_Bands\n"\
"        are the 3rd and the 4th dimension names)\n"\
"        Angles.Num_Obs_Max_2.Num_Angles_3 (2nd observation of the 3rd angle\n"\
"        component of SDS Angles in MODAGAGG.  Note that Angles is the input\n"\
"        SDS name, Num_Obs_Max and Num_Angles are the 3rd and the 4th\n"\
"        dimension names) \n" \
" \n" \
"    This tool supports 3D and 4D SDS(s). \n" \
" \n" \
"    The tool command arguments can be specified in any order. \n" \
" \n" \
"OPTIONS \n" \
"     -help            Display this help message \n" \
"     -sds=<SDS name>  Name of SDS to reduce. Default is all SDS \n" \
"     -dim=<dimension name and value> \n" \
"		       Name or dimension number of the 3rd or any higher dimensions \n" \
"		       and set of values separated by comma. To reduce additional \n" \
"		       dimension repeat the option. Default reduces all of the 3rd \n" \
"		       and higher dimensions. This option when specified is assumed \n" \
"		       to be applicable only to the SDS listed under the option \n" \
"		       -sds option prior to this -dim option\n" \
"     -all	       Create an additional output sds containing all the 2D sds in \n" \
"		       a row-column order \n" \
"     -of=filename     Output file \n" \
"     -meta            Copy metadata from input file to output \n" \
"     -fill_value=fillVal  User inputed fill value \n" \
" \n" \
"Examples: \n" \
"	reduce_sds_rank -sds=Surface_Refl -dim=Num_Obs_Max,1,2 -dim=Num_Land_Bands,1-6 -all -fill_value=-1000 -of=agg_res.hdf ./DATA/MODAGG/MODAGAGG.A1996214.h12v04.002.hdf \n" \
"	reduce_sds_rank -sds=Angles -dim=Num_Obs_Max,1,2 -dim=Num_Angles,1-3 -of=agg_res.hdf ./DATA/MODAGG/MODAGAGG.A1996214.h12v04.002.hdf \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:		reduce_sd_rank [-sds=[SDSname [-dim=dimstr [dim=dimstr. . ]]] \n" \
"                       -sds=[SDSname [-dim=dimstr [dim=dimstr. . ]]] . . ] \n" \
"                       -all -meta -of=output_file filename \n" \
" \n" \
"OPTIONS \n" \
"     -help            Display this help message \n" \
"     -dim=<dimension name and value> \n" \
"                      Name or dimension number of the 3rd or any higher dimensions \n" \
"                      and set of values separated by comma. To reduce additional \n" \
"                      dimension repeat the option. Default reduces all of the 3rd \n" \
"                      and higher dimensions. This option when specified is assumed \n" \
"                      to be applicable only to the SDS listed under the option \n" \
"                      -sds option prior to this -dim option\n" \
"     -all	       Create an additional output sds containing all the 2D sds in \n" \
"		       a row-column order \n" \
"     -of=filename     Output file \n" \
"     -meta            Copy metadata from input file to output \n" \
"     -fill_value=fillVal  User inputed fill value \n" \
" \n"

int parse_cmd_reduce_sds_rank(int argc, char **argv, int *if_cnt, char **sds_names, int *sds_cnt, 
	char ***dim_ids_nums, int *dim_ids_cnt, char *out_fname, int *out_all, int *m_opt, char *fillVal);
int get_out_dim_id(char **dim_names, int32 *dim_sz, int rank, char **dim_ids, 
	int dim_cnt, int **out_dim_ids, int *out_dim_ids_cnt);
void cvrt_rank_m2(sds_t *first_in_sds_info, sds_t *first_out_sds_info, char *sds_name, char **dim_ids_nums,
	int dim_ids_cnt, int if_cnt, char *tmp_str, int out_all, char *fillVal);

int main(int argc, char *argv[])
{
  int i;
  int fid;
  int isds;
  int m_opt;
  int status;
  int if_cnt;
  int sds_cnt;
  int out_all;
  int is_all_sds;
  int stdin_argc;
  char **sds_names;
  char **stdin_argv;
  char ***dim_ids_nums;
  int dim_ids_cnt[MAX_NUM_SDS];
  char tmp_str[MAX_PATH_LENGTH];
  char in_fname[MAX_PATH_LENGTH];
  char out_fname[MAX_PATH_LENGTH];
  char fillVal[10];
  sds_t in_sds_info;
  char temp_sds_names[MAX_STR_LEN];
  sds_t out_sds_info;

  if (argc == 1)
    {
      status = -1;
      fprintf(stderr, "Missing input file \n");
      fprintf(stderr, "%s\n", USAGE);
      exit(EXIT_FAILURE);
    }

  if ((argc==2) && (strcmp(argv[1], "-help")==0))
    {
      fprintf(stderr, "%s\n", HELP);
      exit(EXIT_SUCCESS);
    }                
  
  if ((argc > 3) && (strcmp(argv[1], "-help")==0))
    {
      fprintf(stderr, "%s\n", USAGE);
      exit(EXIT_SUCCESS);
    }                
  
  /* MSZ 1/31/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 1/31/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  /* Display SDS names of input HDF file */ 
  if ((argc==3) && (strcmp(argv[1], "-help")==0))
    {
      display_sds_info_of_file(argv[2]);
      exit(EXIT_SUCCESS);
    } 

  /* initialize dim_ids_cnt */
  
  for (i=0; i<MAX_NUM_SDS; i++)
    {
      dim_ids_cnt[i] = 0;
    }
    
  if ((dim_ids_nums = (char ***)Calloc3D(MAX_NUM_SDS, MAX_NUM_DIM, MAX_DIM_NAME_LEN, sizeof(char))) == NULL)  
    fprintf(stderr, "Cannot allocate memory for dim_ids_nums in reduce_sds_rank(main)\n");
  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL) 
    fprintf(stderr, "Cannot allocate memory for sds_names in reduce_sds_rank(main)\n");

  if ((sds_names != NULL) && (dim_ids_nums != NULL))
  {
    if (argc == 1)
      status = parse_cmd_reduce_sds_rank(stdin_argc, stdin_argv, &if_cnt, sds_names, 
		&sds_cnt, dim_ids_nums, dim_ids_cnt, out_fname, &out_all, &m_opt, fillVal);
    else
      status = parse_cmd_reduce_sds_rank(argc, argv, &if_cnt, sds_names, 
		&sds_cnt, dim_ids_nums, dim_ids_cnt, out_fname, &out_all, &m_opt, fillVal);
    if (status == -1)
      fprintf(stderr, "%s\n", USAGE);
    else if (status != 0)
    {
      out_sds_info.H5IN = 0;     /* nothing to do with the input file format */ 
      out_sds_info.H5OUT = 0;    /* so far, output to HDF4 */
      out_sds_info.sd_id = out_sds_info.sds_id = -1;

      if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
	fprintf(stderr, "Cannot create output hdf file %s\n", out_fname);
      else 
      {
        if ((sds_cnt == 1) && (strcmp(sds_names[0], "all") == 0)) is_all_sds = 1;
	else is_all_sds = 0;
	if (argc != 1) stdin_argc = argc;
	for (fid=1; fid<stdin_argc; fid++)
	{
	  if (argc == 1) strcpy(in_fname, stdin_argv[fid]);
	  else strcpy(in_fname, argv[fid]);
	  if (in_fname[0] == '-') continue;
          if (is_all_sds == 1)
	    sds_cnt = get_sds_names(in_fname, sds_names);
	  if (sds_cnt != 0)
	  {
	    in_sds_info.sd_id = in_sds_info.sds_id = -1;

	    if(open_sd(in_fname, &in_sds_info, 'R') == FAIL)
	      fprintf(stderr, "Cannot open input HDF file %s\n", in_fname);
	    else
	    {
	      fprintf(stderr, "\nProcessing input file: %s\n", in_fname);
	      fprintf(stderr, "------------------------------------------------------------\n");

	      strcpy(tmp_str, in_fname);
 	      rm_path(tmp_str);
	      tmp_str[24] = '\0'; 
	      for (isds=0; isds<sds_cnt; isds++)
	      {
	        cvrt_rank_m2(&in_sds_info, &out_sds_info, sds_names[isds], dim_ids_nums[isds], 
			dim_ids_cnt[isds], if_cnt, tmp_str, out_all, fillVal);
	      }
	      if (m_opt == 1)
	        copy_metadata(in_sds_info.H5IN, in_sds_info.sd_id, out_sds_info.H5OUT, out_sds_info.sd_id);
	      close_sd(&in_sds_info, in_sds_info.H5IN);
	    }
	  }
	} /* for (fid=0; . . . */
	close_sd(&out_sds_info, out_sds_info.H5OUT);
      }
    }
  }
  if(sds_names) Free2D((void **)sds_names);
  if (argc == 1) if(stdin_argv) Free2D((void **)stdin_argv);
  if(dim_ids_nums) Free3D((void ***)dim_ids_nums);
  return 0;
}

int parse_cmd_reduce_sds_rank(int argc, char **argv, int *if_cnt, char **sds_names, int *sds_cnt, 
	char ***dim_ids_nums, int *dim_ids_cnt, char *out_fname, int *out_all, int *m_opt, char *fillVal)
{
  int i;
  int status;
  int dim_cnt;

  status = 1;
  out_fname[0] = '\0';
  fillVal[0] = '\0'; 
  *m_opt = *if_cnt = 0;
  *sds_cnt = *out_all = 0;
  i = 1;
  while (i<argc)
  {
    if (is_arg_id(argv[i], "-sds=") == 0)
    {
      get_arg_val(argv[i], sds_names[*sds_cnt]);
      i++;
      dim_cnt = 0;
      while ((i<argc) && (((is_arg_id(argv[i], "-dim") == 0) || (is_arg_id(argv[i], "-d") == 0))))
      {
        get_arg_val(argv[i], dim_ids_nums[*sds_cnt][dim_cnt]);
	dim_cnt++;
	i++;
      }
      i--;
      dim_ids_cnt[*sds_cnt] = dim_cnt;
      ++*sds_cnt;
    }
    else if (strcmp(argv[i], "-sds") == 0) ;
    else if ((is_arg_id(argv[i], "-of=") == 0) || (is_arg_id(argv[i], "-o=") == 0))
      get_arg_val(argv[i], out_fname);
    else if (strcmp(argv[i], "-all") == 0) *out_all = 1;
    else if (strcmp(argv[i], "-meta") == 0) *m_opt = 1;
    else if (is_arg_id(argv[i], "-fill_value=") == 0) get_arg_val(argv[i], fillVal);
    else if (argv[i][0] == '-')
      fprintf(stderr, "Unknown option %s\n", argv[i]);
    else ++*if_cnt;
    i++;
  }
  if ((*if_cnt == 0) || (out_fname[0] == '\0')) status = -1;
  if (*if_cnt == 0) fprintf(stderr, "Missing input filename\n");
  if (out_fname[0] == '\0') fprintf(stderr, "Missing output filename\n");
  if ((status != -1) && (*sds_cnt == 0))
  {
    fprintf(stderr, "No SDS name input. Reading all SDS . . \n");
    *sds_cnt = 1;
    strcpy(sds_names[0], "all");
  }
  return status;
}

void cvrt_rank_m2(sds_t *in_sds_info, sds_t *out_sds_info, char *sds_name, char **dim_ids_nums,
		  int dim_ids_cnt, int if_cnt, char *tmp_str, int out_all, char *fillVal)
/*
!C******************************************************************************
    
!Function: cvrt_rank_m2
        
!Description:
  This routines converts the input SDS to 2D SDSs. Also, if the out_all is specifed, 
  a mosaic of all the 2D SDSs from the same input SDS is created in the output file.
    
!Input Parameters: 
  in_sd_id        input file descriptor.
  out_sd_id       output file descriptor.
  sds_names       user input SDS names.
  dim_ids_nums    dimension element numbers.
  dim_ids_cnt     dimension element counts.
  out_all         if the out_all is specifed, a mosaic of all the 2D SDSs from 
                  the same input SDS is created in the output file.
    
!Revision History: (see file prolog) 
    
!Team-unique Header: (see file prolog)
    
!References and Credits: (see file prolog)
    
!Design Notes: (none)
    
!END
********************************************************************************/
{
  int offset_is;
  int i, j1, j2;
  int is, ilines;
  int offset = 0, p1;
  int tot_num_vals;
  int idim, isds, nsds;
  int32 num_data;
  int32 data_size;
  int32 edge1[2] = {0, 0};
  int32 start1[2] = {0, 0};
  int32 edgef[2] = {0, 0};
  int32 startf[2] = {0, 0};
  int32 edge[4] = {0, 0, 0, 0};
  int32 start[4] = {0, 0, 0, 0};
  char tmp_name[MAX_NAME_LENGTH];
  sds_t *out_sds1_info;
  char **dim_names, **short_dim_names;
  void *sds_data_in;
  void **sds_data_out;
  int **out_dim_ids;
  int out_dim_ids_cnt[MAX_NUM_DIM];
  char tmp_dname1[MAX_DIM_NAME_LEN];
  char tmp_dname2[MAX_DIM_NAME_LEN];

  /* MSZ 10/26/2018, speedup */
  int jobs_num = 20;
  int nlines_a_copy;
  int GROWS, last_rows;
  int rec_line_len;
  HDF_CHUNK_DEF c_def;
  int32 flag;

  in_sds_info->sds_id = -1;
  strcpy(in_sds_info->name, sds_name);
  fprintf(stderr, "	Processing SDS: %s\n", sds_name);
  
  /* initalize array */
  for (i=0; i< MAX_NUM_DIM; i++)
    out_dim_ids_cnt[i] = 0;
  
  if (get_sds_info((char *)NULL, in_sds_info) != FAIL)
  {    
    if (in_sds_info->data_type == 5 || in_sds_info->data_type == 6)
      printf("in_sds_info->fill_fval in file=%lf\n",in_sds_info->fill_fval); 
    else 
      printf("in_sds_info->fill_val in file=%d\n",in_sds_info->fill_val); 	
    if (fillVal[0] != '\0')   
    {     	
      switch(in_sds_info->data_type)
      {
	case  5: in_sds_info->fill_fval = (float32)atof(fillVal); break;
	case  6: in_sds_info->fill_fval = (float64)atof(fillVal); break;
	case 20: in_sds_info->fill_val = (int8)atoi(fillVal); break;
	case 21: in_sds_info->fill_val = (uint8)atoi(fillVal); break;
	case 22: in_sds_info->fill_val = (int16)atoi(fillVal); break;
	case 23: in_sds_info->fill_val = (uint16)atoi(fillVal); break;
	case 24: in_sds_info->fill_val = (int32)atoi(fillVal); break;
	case 25: in_sds_info->fill_val = (uint32)atoi(fillVal); break;
      } 
    }
    data_size = in_sds_info->data_size;
    if ((dim_names = (char **)Calloc2D(in_sds_info->rank, MAX_NAME_LENGTH, sizeof(char))) 
	  == NULL)
	fprintf(stderr, "Cannot allocate memory for dim_names\n");
      
    if ((short_dim_names = (char **)Calloc2D(in_sds_info->rank, MAX_NAME_LENGTH, sizeof(char))) 
	  == NULL)
	fprintf(stderr, "Cannot allocate memory for short_dim_names\n");
      
    if ((out_dim_ids = (int **)Calloc2D(MAX_NUM_DIM, MAX_NUM_DIM, sizeof(int))) == NULL)
	fprintf(stderr, "Cannot allocate memory for out_dim_ids\n");
      
    if ((dim_names == NULL) || (short_dim_names == NULL) || (out_dim_ids == NULL))
    {
      if(dim_names) Free2D((void **)dim_names);
      if(short_dim_names) Free2D((void **)short_dim_names);
      if(out_dim_ids) Free2D((void **)out_dim_ids);
      return;
    }
    else 
      get_sds_dim_name(in_sds_info, dim_names, short_dim_names);
            
    if (get_out_dim_id(short_dim_names, in_sds_info->dim_size, in_sds_info->rank, dim_ids_nums, 
	 dim_ids_cnt, out_dim_ids, out_dim_ids_cnt) == FAIL)
    {
      if(dim_names) Free2D((void **)dim_names);
      return;
    }

    /* works only when rank > 1 */
    if (in_sds_info->rank > 1) 
    {
	/* get num_data */
	if (in_sds_info->interleave == 0)  /* BIP */
	  for (i=2,nsds=1,num_data=in_sds_info->dim_size[1]; i<in_sds_info->rank; i++)
	  {
	    nsds *= out_dim_ids_cnt[i-2];
	    num_data *= in_sds_info->dim_size[i];
	  }
	else   /* BSQ */
	  for (i=0,nsds=1,num_data=in_sds_info->dim_size[in_sds_info->rank - 1]; i<in_sds_info->rank - 2; i++)
	  {
	    nsds *= out_dim_ids_cnt[i];
	    num_data *= in_sds_info->dim_size[i];
	  }
	  
	/* allocate memory for variables out_sds1_info, sds_data_in and sds_data_out which holds 
	     the data for input sds and output sds */
	  
	if ((out_sds1_info = (sds_t *)calloc(nsds, sizeof(sds_t))) == NULL) 
	  fprintf(stderr, "Cannot allocate memory for out_sds1_info\n");
	else
	  for (isds=0; isds<nsds; isds++)
	  {
	    out_sds1_info[isds].H5IN = 0;     /* nothing to do with the input file format */
	    out_sds1_info[isds].H5OUT = 0;    /* so far, output to HDF4 */
	    out_sds1_info[isds].sd_id = out_sds_info->sd_id;
	    out_sds1_info[isds].sds_id = -1;
	  }

	/* very important values for speedup */
	/*###################################################################*/
	if (in_sds_info->interleave == 0)  /* BIP */
	{
	  /* MSZ 10/26/2018 speedup */
	  nlines_a_copy = (int)floor(in_sds_info->dim_size[0] / jobs_num + 0.5);

	  /* correct nlines_a_copy if necessary */
	  if(in_sds_info->dim_size[0] < nlines_a_copy)
	    nlines_a_copy = in_sds_info->dim_size[0];

	  GROWS = (int)ceil((double)in_sds_info->dim_size[0] / (double)nlines_a_copy);
	  last_rows = in_sds_info->dim_size[0] % nlines_a_copy;

	  /*if ((sds_data_out = (void **)Calloc2D(nsds, in_sds_info->dim_size[1], data_size)) == NULL)*/
	  if ((sds_data_out = (void **)Calloc2D(nsds, in_sds_info->dim_size[1] * nlines_a_copy, data_size)) == NULL)
	    fprintf(stderr, "Cannot allocate memory sds_data_out\n");
	}
	else    /* BSQ */
	{
	  /* MSZ 10/26/2018 speedup */
	  nlines_a_copy = (int)floor(in_sds_info->dim_size[in_sds_info->rank - 2] / jobs_num + 0.5);

	  /* correct nlines_a_copy if necessary */
	  if(in_sds_info->dim_size[in_sds_info->rank - 2] < nlines_a_copy)
	    nlines_a_copy = in_sds_info->dim_size[in_sds_info->rank - 2];

	  GROWS = (int)ceil((double)in_sds_info->dim_size[in_sds_info->rank - 2] / (double)nlines_a_copy);
	  last_rows = in_sds_info->dim_size[in_sds_info->rank - 2] % nlines_a_copy;

	  /*if ((sds_data_out = (void **)Calloc2D(nsds, in_sds_info->dim_size[in_sds_info->rank - 1], data_size)) == NULL)*/
	  if ((sds_data_out = (void **)Calloc2D(nsds, in_sds_info->dim_size[in_sds_info->rank - 1] * nlines_a_copy, data_size)) == NULL)
	    fprintf(stderr, "Cannot allocate memory sds_data_out\n");
	}

	/* MSZ 10/26/2018 speedup */
	/*if ((sds_data_in = (void *)calloc(num_data, data_size)) == NULL)*/
	if ((sds_data_in = (void *)calloc(num_data * nlines_a_copy, data_size)) == NULL)
	  fprintf(stderr, "Cannot allocate memory for sds_data_in\n");
	/*###################################################################*/

	/* read and output based on BSP or BIP band interleave */
	if ((dim_names != NULL) && (sds_data_in != NULL) && (sds_data_out != NULL) && (out_sds1_info != NULL))
	{
	  /* BIP begin $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ */
	  if (in_sds_info->interleave == 0)
	  {
	    for (isds=0; isds<nsds; isds++)
	    {
	      out_sds1_info[isds].rank = 2;
	      out_sds1_info[isds].data_type = in_sds_info->data_type;
	      if (in_sds_info->data_type == 5 || in_sds_info->data_type == 6)
	        out_sds1_info[isds].fill_fval = in_sds_info->fill_fval;
	      else 
	        out_sds1_info[isds].fill_val = in_sds_info->fill_val;		      
		      
	      out_sds1_info[isds].dim_size[0] = in_sds_info->dim_size[0];
	      out_sds1_info[isds].dim_size[1] = in_sds_info->dim_size[1];
	      for (idim=2; idim<in_sds_info->rank; idim++)
	      {
		p1 = charpos(dim_names[idim], ':', 0);
		if (p1 != -1)
		{
		  if (idim == 2) strmid(dim_names[idim], 0, p1, tmp_dname1);
		  else if (idim == 3) strmid(dim_names[idim], 0, p1, tmp_dname2);
		}
		else 
		{
		  if (idim == 2) strcpy(tmp_dname1, dim_names[idim]);
		  else if (idim == 3) strcpy(tmp_dname2, dim_names[idim]);
		}
	      } /* for (idim = 2; .... */
		      
	      if (in_sds_info->rank == 3)
		sprintf(tmp_name, "%s_%s%d", sds_name, tmp_dname1, out_dim_ids[0][isds+1] + 1); 
	      else if (in_sds_info->rank == 4)
	      {
		j1 = isds/out_dim_ids_cnt[1];
		j2 = isds%out_dim_ids_cnt[1];
		sprintf(tmp_name, "%s_%s%d_%s%d", sds_name, tmp_dname1, out_dim_ids[0][j1+1]+1, 
		  tmp_dname2, out_dim_ids[1][j2+1] + 1); 
	      }
		      
	      if (if_cnt == 1)
		strcpy(out_sds1_info[isds].name, tmp_name);
	      else
		sprintf(out_sds1_info[isds].name, "%s_%s", tmp_name, tmp_str);

	      if (open_sds((char *)NULL, &out_sds1_info[isds], 'W') != FAIL)
	      {

		/* for HDF4 internal compression of 2D output data */
		/* Set up the compression parameters */
		c_def.chunk_lengths[0] = c_def.comp.chunk_lengths[0] = nlines_a_copy;  /* Rows */
		c_def.chunk_lengths[1] = c_def.comp.chunk_lengths[1] = out_sds1_info[isds].dim_size[1]; /* Columns */
		c_def.comp.comp_type = COMP_CODE_DEFLATE;     /* GZIP */
		c_def.comp.cinfo.deflate.level = 6;           /* level, valid range from 1 to 9 for HDF-EOS internal gzip compression */

		if (SDsetchunk(out_sds1_info[isds].sds_id, c_def, HDF_CHUNK | HDF_COMP) == FAIL)
		  /*fprintf(stderr,"cannot set BIP SDchunk for out_sds1_info[%d].sds_id=%d sds_name=%s\n", isds, out_sds1_info[isds].sds_id, out_sds1_info[isds].name)*/;

		if (out_sds1_info[isds].data_type == 5 || out_sds1_info[isds].data_type == 6)
		  write_attr_fvalf(out_sds1_info[isds].H5OUT, out_sds1_info[isds].sds_id, out_sds1_info[isds].data_type, 1, out_sds1_info[isds].fill_fval);
		else
		  write_attr_fval(out_sds1_info[isds].H5OUT, out_sds1_info[isds].sds_id, out_sds1_info[isds].data_type, 1, out_sds1_info[isds].fill_val);
	      }	
	    } /* for (isds=0; isds<nsds; isds++) */
	  }
	  /* BIP end $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ */

	  /* BSQ begin $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ */
	  else
	  {  
	    for (isds=0; isds<nsds; isds++)
	    {
	      out_sds1_info[isds].rank = 2;
	      if (in_sds_info->data_type == 5 || in_sds_info->data_type == 6)
	        out_sds1_info[isds].fill_fval = in_sds_info->fill_fval;
	      else 
	        out_sds1_info[isds].fill_val = in_sds_info->fill_val;
	      out_sds1_info[isds].data_type = in_sds_info->data_type;
	      
	      out_sds1_info[isds].dim_size[0] = in_sds_info->dim_size[in_sds_info->rank - 2 ];
	      out_sds1_info[isds].dim_size[1] = in_sds_info->dim_size[in_sds_info->rank - 1];
			      
	      for (idim = 0; idim < in_sds_info->rank -2; idim++)
	      {
		p1 = charpos(dim_names[idim], ':', 0);
		if (p1 != -1)
		{ 
		  if (idim == 0) strmid(dim_names[idim], 0, p1, tmp_dname1);
		  else if (idim == 1) strmid(dim_names[idim], 0, p1, tmp_dname2);
		}
		else 
		{
		  if (idim == 0) strcpy(tmp_dname1, dim_names[idim]);
		  else if (idim == 1) strcpy(tmp_dname2, dim_names[idim]);
		}
	      } /* for (idim = 0; .... */
		      
	      if (in_sds_info->rank == 3)
		sprintf(tmp_name, "%s_%s%d", sds_name, tmp_dname1, out_dim_ids[0][isds+1] + 1); 
	      else if (in_sds_info->rank == 4)
	      {
		j1 = isds/out_dim_ids_cnt[1];
		j2 = isds%out_dim_ids_cnt[1];
		sprintf(tmp_name, "%s_%s%d_%s%d", sds_name, tmp_dname1, out_dim_ids[0][j1+1]+1, 
		  tmp_dname2, out_dim_ids[1][j2+1] + 1); 
	      }

	      if (if_cnt == 1)
		strcpy(out_sds1_info[isds].name, tmp_name);
	      else
		sprintf(out_sds1_info[isds].name, "%s_%s", tmp_name, tmp_str);			      
	      		      
	      if (in_sds_info->rank == 2)
		strcpy(out_sds1_info[isds].name, in_sds_info->name);

	      if (open_sds((char *)NULL, &out_sds1_info[isds], 'W') != FAIL)
	      {
		/* for HDF4 internal compression of 2D output data */
		/* Set up the compression parameters */
		c_def.chunk_lengths[0] = c_def.comp.chunk_lengths[0] = nlines_a_copy;  /* Rows */
		c_def.chunk_lengths[1] = c_def.comp.chunk_lengths[1] = out_sds1_info[isds].dim_size[1]; /* Columns */
		c_def.comp.comp_type = COMP_CODE_DEFLATE;     /* GZIP */
		c_def.comp.cinfo.deflate.level = 6;           /* level, valid range from 1 to 9 for HDF-EOS internal gzip compression */

		if (SDsetchunk(out_sds1_info[isds].sds_id, c_def, HDF_CHUNK | HDF_COMP) == FAIL)
		  /*fprintf(stderr,"cannot set BSQ SDchunk for out_sds1_info[%d].sds_id=%d sds_name=%s\n", isds, out_sds1_info[isds].sds_id, out_sds1_info[isds].name)*/;
		if (out_sds1_info[isds].data_type == 5 || out_sds1_info[isds].data_type == 6)
		  write_attr_fvalf(out_sds1_info[isds].H5OUT, out_sds1_info[isds].sds_id, out_sds1_info[isds].data_type, 1, out_sds1_info[isds].fill_fval);
		else
		  write_attr_fval(out_sds1_info[isds].H5OUT, out_sds1_info[isds].sds_id, out_sds1_info[isds].data_type, 1, out_sds1_info[isds].fill_val);
	      }	
	    } /* for (isds=0; ....... */	    
	  }
	  /* BSQ end $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ */

	  if (out_all == 1)
	  {
	    out_sds_info->rank = 2;
	    if (in_sds_info->data_type == 5 || in_sds_info->data_type == 6)
	      out_sds_info->fill_fval = in_sds_info->fill_fval;
	    else 
	      out_sds_info->fill_val = in_sds_info->fill_val;		  
	    out_sds_info->data_type = in_sds_info->data_type;
	    sprintf(out_sds_info->name, "%s_%s", sds_name, "all");
		  
	    if (in_sds_info->interleave == 0)  /* BIP */
	    {
	      if (in_sds_info->rank == 3)
	      {
		out_sds_info->dim_size[0] = in_sds_info->dim_size[0];
		out_sds_info->dim_size[1] = in_sds_info->dim_size[1]*out_dim_ids_cnt[0];
	      }
	      else if (in_sds_info->rank == 4)
	      {
		out_sds_info->dim_size[0] = in_sds_info->dim_size[0]*out_dim_ids_cnt[0];
		out_sds_info->dim_size[1] = in_sds_info->dim_size[1]*out_dim_ids_cnt[1];
	      }
	      
	      if (open_sds((char *)NULL, out_sds_info, 'W') != FAIL)
	      {
		/* MSZ 10/31/2018. note that internal compression has been set */
		if (out_sds_info->data_type == 5 || out_sds_info->data_type == 6)
		  write_attr_fvalf(out_sds_info->H5OUT, out_sds_info->sds_id, out_sds_info->data_type, 1, out_sds_info->fill_fval);
		else
		  write_attr_fval(out_sds_info->H5OUT, out_sds_info->sds_id, out_sds_info->data_type, 1, out_sds_info->fill_val); 
	      }
	    }
	    else /* BSQ */
	    {
	      if (in_sds_info->rank ==2)
	      {
		out_sds_info->dim_size[0] = in_sds_info->dim_size[0];
		out_sds_info->dim_size[1] = in_sds_info->dim_size[1];
	      }
	      if (in_sds_info->rank == 3)
	      {
		out_sds_info->dim_size[0] = in_sds_info->dim_size[in_sds_info->rank -2];
		out_sds_info->dim_size[1] = in_sds_info->dim_size[in_sds_info->rank -1]*out_dim_ids_cnt[0];
	      }
	      else if (in_sds_info->rank == 4)
	      {
		out_sds_info->dim_size[0] = in_sds_info->dim_size[in_sds_info->rank -2]*out_dim_ids_cnt[0];
		out_sds_info->dim_size[1] = in_sds_info->dim_size[in_sds_info->rank -1]*out_dim_ids_cnt[1];
	      }
		
	      if (open_sds((char *)NULL, out_sds_info, 'W') != FAIL)
	      {
		/* MSZ 10/31/2018. note that internal compression has been set */
		if (out_sds_info->data_type == 5 || out_sds_info->data_type == 6)
		  write_attr_fvalf(out_sds_info->H5OUT, out_sds_info->sds_id, out_sds_info->data_type, 1, out_sds_info->fill_fval);
		else
		  write_attr_fval(out_sds_info->H5OUT, out_sds_info->sds_id, out_sds_info->data_type, 1, out_sds_info->fill_val); 
	      }
	    }
		  
	  } /* if (out_all == 1) */

	  /* read and write multilines */
	  /* BIP begin ################################################# */
	  if (in_sds_info->interleave == 0)  /* BIP */
	  {
	    for (i=1; i<in_sds_info->rank; i++)
	      edge[i] = in_sds_info->dim_size[i];

	    edge1[1] = in_sds_info->dim_size[1];
	    edgef[1] = in_sds_info->dim_size[1];

	    if (in_sds_info->rank == 3)
	      tot_num_vals = in_sds_info->dim_size[2];
	    else
	      tot_num_vals = in_sds_info->dim_size[2]*in_sds_info->dim_size[3];

	    /* MSZ 10/26/2018, speedup */		    
	    /*for (ilines=0; ilines<in_sds_info->dim_size[0]; ilines++)*/
	    for (ilines=0; ilines<GROWS; ilines++)
	    { 
	      /*start[0] = start1[0] = (int32)ilines;*/
	      start[0] = start1[0] = ilines * nlines_a_copy;

	      /* MSZ, 10/26/2018, be cautious about the last group rows because it may be different from nlines_a_copy */
	      edge[0] = edge1[0] = edgef[0] = nlines_a_copy;
	      if(last_rows > 0 && ilines == GROWS - 1)
	        edge[0] = edge1[0] = edgef[0] = last_rows;

	      if(read_sel_sds_data(in_sds_info, start, edge, sds_data_in) != FAIL)
	      {
	        /* MSZ, 10/26/2018, speedup */
	        /*for (is=0; is<in_sds_info->dim_size[1]; is++)*/
		rec_line_len = in_sds_info->dim_size[1] * edge[0];
	        for (is=0; is<rec_line_len; is++)
	        {
		  offset_is = is*tot_num_vals;
		  for (isds=0; isds<nsds; isds++)
		  {
		    if (in_sds_info->rank == 3)
		      offset = offset_is + out_dim_ids[0][1+isds];
		    else if (in_sds_info->rank == 4)
		    {
		      j1 = isds/out_dim_ids_cnt[1];
		      j2 = isds%out_dim_ids_cnt[1];
		      offset = offset_is + j1*in_sds_info->dim_size[3] + out_dim_ids[1][j2+1];
		    }
		    switch(in_sds_info->data_type)
		    {
		      case 5:
			((float32 *)sds_data_out[isds])[is] =  ((float32 *)sds_data_in)[offset];
			break;
		      case 6:
			((float64 *)sds_data_out[isds])[is] =  ((float64 *)sds_data_in)[offset];
			break;			      
		      case 20: ((int8 *)sds_data_out[isds])[is] = ((int8 *)sds_data_in)[offset];
			break;
		      case 21: ((uint8 *)sds_data_out[isds])[is] = ((uint8 *)sds_data_in)[offset];
			break;
		      case 22: ((int16 *)sds_data_out[isds])[is] = ((int16 *)sds_data_in)[offset];
			break;
		      case 23: ((uint16 *)sds_data_out[isds])[is] = ((uint16 *)sds_data_in)[offset];
			break;
		      case 24: ((int32 *)sds_data_out[isds])[is] = ((int32 *)sds_data_in)[offset];
			break;
		      case 25: ((uint32 *)sds_data_out[isds])[is] = ((uint32 *)sds_data_in)[offset];
			break;
		      default: fprintf(stdout,
                       	 "HDF datatype " LONG_INT_FMT " not supported", 
                      	 in_sds_info->data_type);

		    } /* switch(in_sds_info->data_type) */
		  } /* for (isds=0; . . . */
	        } /* for (is=0; . . . */
	        for (isds=0; isds<nsds; isds++)
	        {
		  if (set_sel_sds_data(&out_sds1_info[isds], start1, edge1, sds_data_out[isds]) == FAIL)
		    fprintf(stderr, "Error writing data line for %s\n", out_sds1_info[isds].name);
		  if (out_all == 1)
		  {
		    if (in_sds_info->rank == 3)
		    {
		      /* MSZ 10/29/2018 */
		      /*startf[0] = ilines;*/
		      startf[0] = ilines * nlines_a_copy;
		      startf[1] = in_sds_info->dim_size[1]*isds;
		    }
		    else
		    {
		      j1 = isds/out_dim_ids_cnt[1];
	 	      j2 = isds%out_dim_ids_cnt[1];
		      /* MSZ 10/29/2018 */
		      /*startf[0] = ilines + (j1*in_sds_info->dim_size[0]);*/
		      startf[0] = ilines * nlines_a_copy + (j1*in_sds_info->dim_size[0]);
		      startf[1] = in_sds_info->dim_size[1]*j2;
		    }
		    if (set_sel_sds_data(out_sds_info, start, edge, sds_data_out[isds]) == FAIL)
		      fprintf(stderr, "Error writing data line for %s\n", out_sds_info->name);
		  }
	        } /* for isds =0; . . . */
	      } /* if (read_sel_sds_data . . . */
	    } /* for (ilines=0; . . . . . */
	  } /* if (in_sds_info->interleave == 0) */
	  /* BIP end ################################################# */

	  /* BSQ begin ################################################# */
	  else   /* (in_sds_info->interleave == 1), BSQ */ 
	  { 
	    /* MSZ 10/26/2018 */
	    /* edge1[0] = edgef[0] = 1 */
	    edge1[0] = edgef[0] = nlines_a_copy;

	    for (i=0; i<in_sds_info->rank; i++) 
	      edge[i] = in_sds_info->dim_size[i]; 

	    edge[in_sds_info->rank -2] = nlines_a_copy;

	    edge1[1] = in_sds_info->dim_size[in_sds_info->rank - 1];
	    edgef[1] = in_sds_info->dim_size[in_sds_info->rank - 1];

	    /*for (ilines=0; ilines<in_sds_info->dim_size[in_sds_info->rank - 2]; ilines++)*/
	    for (ilines=0; ilines<GROWS; ilines++)
	    { 
	      /* MSZ 10/26/2018 */
	      /*start1[0] = (int32)ilines; 
	        start[in_sds_info->rank -2] = (int32)ilines; */
	      start1[0] = (int32)ilines * nlines_a_copy;
	      start[in_sds_info->rank -2] = (int32)ilines * nlines_a_copy;

	      /* MSZ, 10/26/2018, be cautious about the last group rows because it may be different from nlines_a_copy */
              if(last_rows > 0 && ilines == GROWS - 1)
	      {
		edge1[0] = edgef[0] = last_rows;
	        edge[in_sds_info->rank -2] = last_rows;
	      }

	      if(read_sel_sds_data(in_sds_info, start, edge, sds_data_in) != FAIL)
	      { 	
		/* MSZ 10/26/2018 */
	        /*for (is=0; is<in_sds_info->dim_size[in_sds_info->rank - 1]; is++)*/
		rec_line_len = in_sds_info->dim_size[in_sds_info->rank - 1] * edge[in_sds_info->rank -2];
	        for (is=0; is<rec_line_len; is++)
	        { 
		  offset_is = is;
		  for (isds=0; isds<nsds; isds++)
		  {
		    if (in_sds_info->rank == 2)
		      offset = offset_is;
		    if (in_sds_info->rank == 3)
		      /* MSZ 10/26/2018 */
		      /*
		      offset = offset_is + 
	      	      out_dim_ids[0][1+isds]*in_sds_info->dim_size[in_sds_info->rank-1];
		      */
		      offset = offset_is +
		      out_dim_ids[0][1+isds]*rec_line_len;
		    if (in_sds_info->rank == 4)
		    {
		      j1 = isds/out_dim_ids_cnt[1];
		      j2 = isds%out_dim_ids_cnt[1];
		      /* MSZ 10/26/2018 */
		      /*
		      offset = offset_is + j1*in_sds_info->dim_size[1] + 
			out_dim_ids[1][j2+1]*in_sds_info->dim_size[in_sds_info->rank - 1];
		      */
		      offset = offset_is + j1*in_sds_info->dim_size[1] +
		         out_dim_ids[1][j2+1]*rec_line_len;
		    }

		    switch(in_sds_info->data_type)
		    {
		      case 5:
		        ((float32 *)sds_data_out[isds])[is] =  ((float32 *)sds_data_in)[offset];
		        break;
		      case 6:
		        ((float64 *)sds_data_out[isds])[is] =  ((float64 *)sds_data_in)[offset];
		        break;
		      case 20: 
		        ((int8 *)sds_data_out[isds])[is] = ((int8 *)sds_data_in)[offset];
		        break;
		      case 21: 
		        ((uint8 *)sds_data_out[isds])[is] = ((uint8 *)sds_data_in)[offset];
		        break;
		      case 22: 
		        ((int16 *)sds_data_out[isds])[is] = ((int16 *)sds_data_in)[offset];
		        break;
		      case 23: 
		        ((uint16 *)sds_data_out[isds])[is] = ((uint16 *)sds_data_in)[offset];
		        break;
		      case 24: 
		        ((int32 *)sds_data_out[isds])[is] = ((int32 *)sds_data_in)[offset];
		        break;
		      case 25: 
		        ((uint32 *)sds_data_out[isds])[is] = ((uint32 *)sds_data_in)[offset];
		        break;
		      default: fprintf(stdout,
                      	 "HDF datatype " LONG_INT_FMT " not supported", 
                       	 in_sds_info->data_type);

		    } /* switch(in_sds_info->data_type) */
		  } /* for (isds=0; . . . */
	        } /* for (is=0; . . . */

	        for (isds=0; isds<nsds; isds++)
	        {	
		  if (set_sel_sds_data(&out_sds1_info[isds], start1, edge1, sds_data_out[isds]) == FAIL) 
		    fprintf(stderr, "Error writing data line for %s\n", out_sds1_info[isds].name);
		  if (out_all == 1)
	  	  {
		    if (in_sds_info->rank == 2)
		    {
		      /* MSZ 10/29/2018 */
		      /*startf[0] = (int32)ilines;*/
		      startf[0] = (int32)ilines * nlines_a_copy;
		      startf[1] = 0;
		    }
		    else 
		    {
		      if (in_sds_info->rank == 3)
		      {
		        /* MSZ 10/29/2018 */
		        /*startf[0] = (int32)ilines;*/
		        startf[0] = (int32)ilines * nlines_a_copy;
		        startf[1] = in_sds_info->dim_size[in_sds_info->rank - 1]*isds; 
		      }
		      else
		      {
		        j1 = isds/out_dim_ids_cnt[1];
		        j2 = isds%out_dim_ids_cnt[1];
		        /* MSZ 10/29/2018 */
		        /*startf[0] = ilines + (j1*in_sds_info->dim_size[in_sds_info->rank - 2]);*/ 
		        startf[0] = ilines * nlines_a_copy + (j1*in_sds_info->dim_size[in_sds_info->rank - 2]);
		        startf[1] = in_sds_info->dim_size[in_sds_info->rank - 1]*j2;
		      }
		    } /* else */

		    if (set_sel_sds_data(out_sds_info, startf, edgef, sds_data_out[isds]) == FAIL)
		      fprintf(stderr, "Error writing data line for %s\n", out_sds_info->name);
		  } /* if (out_all == 1) */
	        } /* for (isds =0; . . . */
	      } /* if (read_sel_sds_data . . . */
	    } /* for (ilines=0; . . . . . */
	  } /* else (in_sds_info->interleave == 1) */
	  /* BSQ end ################################################# */
	  for (isds=0; isds<nsds; isds++)
	    close_sds(&out_sds1_info[isds], out_sds1_info[isds].H5OUT);
	  if (out_all == 1) close_sds(out_sds_info, out_sds_info->H5OUT);

	}  /* if (dim_names != . . . .   */ 
	if(dim_names) Free2D((void **)dim_names);
	if(sds_data_out) Free2D((void **)sds_data_out);
	if(out_dim_ids) Free2D((void **)out_dim_ids);
	if (sds_data_in) free(sds_data_in);
	if (out_sds1_info) free(out_sds1_info);
    }/* if (in_sds_info->rank > 1) */
  } /* if (get_sds_info((char *)NULL, in_sds_info) != FAIL) */
  close_sds(in_sds_info, in_sds_info->H5IN);
}

int get_out_dim_id(char **dim_names, int32 *dim_size, int rank, char **dim_ids, 
		   int dim_cnt, int **out_dim_ids, int *out_dim_ids_cnt)
/*
!C******************************************************************************
    
!Function: get_out_dim_id
        
!Description:
  Get the output dimension id.
    
!Input Parameters: 
  dim_names       Dimension names
  dim_size        Dimension size of each dimension.
  rank            The rank of the input SDS.
  dim_ids         Selected dimension name.
  dim_cnt         Count of selected dimension number.
  out_dim_ids     Output dimension name.
  out_dim_ids_cnt Count of each output dimension.
                                  
!Output Parameters: 
  1               On succeed.
  -1              On failure.
    
!Revision History: (see file prolog) 
    
!Team-unique Header: (see file prolog)
    
!References and Credits: (see file prolog)
    
!Design Notes: (none)
    
!END
********************************************************************************/
{
  int ir, isz, bsq;
  int cnt, len;
  int idim, icnt;
  int p1, p2, p3;
  int **sel_dim_ids;
  int one_based_num, num1, num2;
  char num_str[10];
  char num1_str[10];
  char num2_str[10];
  char sel_dim_name[50];
  int sel_dim_cnt[MAX_NUM_DIM];
  int sel_dim_name_is_valid ;
  char tmp[100];
  
  bsq = ((rank == 2) || (dim_size[0] < dim_size[rank - 1])) ? 1 : 0;   

  if (dim_cnt == 0)
    {
      if (bsq == 0)
	{
	  for (ir=2; ir<rank; ir++)
	    {
	      out_dim_ids[ir-2][0] = ir;
	      for (isz=0; isz<dim_size[ir]; isz++)
		{
		  out_dim_ids[ir-2][1+isz] = isz; 
		}
	      out_dim_ids_cnt[ir-2] = dim_size[ir];
	    }
	}
      else
	{
	  for (ir=0; ir<rank-2; ir++)
	    {
	      out_dim_ids[ir][0] = ir;
	      for (isz=0; isz<dim_size[ir]; isz++)
		{
		  out_dim_ids[ir][1+isz] = isz; 
		}
	      out_dim_ids_cnt[ir] = dim_size[ir];
	    }
	}
    }	  
  else
  {
    if ((sel_dim_ids = (int **)Calloc2D(MAX_NUM_DIM, MAX_DIM_SIZE, sizeof(int))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for sel_dim_ids in get_out_dim_id\n"); 
      return -1;
    }
    for (idim=0; idim < dim_cnt; idim++)
    {
      p1 = cnt = 0;
      sel_dim_cnt[idim] = -1;
      sel_dim_ids[idim][0] = -1;
      len = (int)strlen(dim_ids[idim]);
      p2 = charpos(dim_ids[idim], ',', p1);
//      printf("dim_ids[%d]=%s\n", idim,dim_ids[idim]);      
      if (p2 == -1)
        strmid(dim_ids[idim], 0, len, sel_dim_name);
      else
        strmid(dim_ids[idim], 0, p2, sel_dim_name);

      p1 = charpos(sel_dim_name, ':', 0);	
      if (p1 != -1)
      {
	strmid(sel_dim_name, 0, p1, tmp);
	strcpy(sel_dim_name, tmp);
      }	      
      	
//      printf("sel_dim_name=%s\n",	 sel_dim_name);      
      sel_dim_name_is_valid = 0;
      for (ir=0; ir<rank; ir++)
	{ 
//          printf("dim_names[%d]=%s\n",	ir,  dim_names[ir]);
	  if (strstr(dim_names[ir],sel_dim_name) != NULL) 
	    {	      
	      sel_dim_name_is_valid = 1;
	      break;
	    }
	} /* for (ir=0;......) */
      
      if (sel_dim_name_is_valid == 0)
	{
	  fprintf(stderr, "input dim name %s is invalid. Program exit.\n", sel_dim_name);
	  exit(EXIT_FAILURE);
	}
      else
	{
	    sel_dim_ids[idim][cnt] = ir;
	}

      if (sel_dim_ids[idim][cnt] != -1)
	{
	  if (p2 == -1)
	    {
	      for (isz=0; isz<dim_size[ir]; isz++)
		sel_dim_ids[idim][++cnt] = isz; 
	      sel_dim_cnt[idim] = dim_size[idim];
	    }
	  else
	    {
	      while (p2 != -1)
		{
		  p1 = p2+1;
		  p2 = charpos(dim_ids[idim], ',', p1);
		  p3 = charpos(dim_ids[idim], '-', p1);
		  if ((p3 == -1) || ((p2 != -1) && (p2 < p3)))   /* user use ',' between dimension 
								    selection */
		    {
		      if (p2 == -1)
			strmid(dim_ids[idim], p1, len-p1, num_str);
		      else
			strmid(dim_ids[idim], p1, p2-p1, num_str);

		      one_based_num = atoi(num_str);

		      if ((one_based_num -1 >= dim_size[ir]) || ( one_based_num -1 < 0 ))
			fprintf(stderr, "Ignoring the invalid dimension size %d\n", one_based_num);
		      else sel_dim_ids[idim][++cnt] = one_based_num - 1;
		    }
		  else if ((p2 == -1) || (p2 > p3))           /* user use '-' between dimension 
								     selection */
		    {
		      strmid(dim_ids[idim], p1, p3-p1, num1_str);
		      if (p2 == -1)
			strmid(dim_ids[idim], p3+1, len-p3-1, num2_str);
		      else
			strmid(dim_ids[idim], p3+1, p2-p3-1, num2_str);
		      num1 = atoi(num1_str);
		      num2 = atoi(num2_str);
		      
		      for (one_based_num=num1; one_based_num<=num2; one_based_num++)
			{
			  if ((one_based_num - 1 >= dim_size[ir]) || (one_based_num - 1 < 0))
				fprintf(stderr, "Ignoring the invalid dimension size %d\n", one_based_num);
			  else sel_dim_ids[idim][++cnt] = one_based_num - 1;
			}
		    }
		}
	      sel_dim_cnt[idim] = cnt;
	    }
	} /* if sel_dim_nums. . . . */
    }   /* for (idim=0; . . . */

    if (bsq == 0)
      {
	for (ir=2, cnt=0; ir<rank; ir++, cnt++)
	  {
	    out_dim_ids[cnt][0] = ir;
	    for (idim=0; idim<dim_cnt; idim++)
	      if (sel_dim_ids[idim][0] == ir)
		{
		  for (icnt=1; icnt<=sel_dim_cnt[idim]; icnt++)
		    out_dim_ids[cnt][icnt] = sel_dim_ids[idim][icnt];
		  out_dim_ids_cnt[cnt] = sel_dim_cnt[idim];
		  break;
		}
	    if (idim == dim_cnt)
	      {
		for (icnt=0; icnt<dim_size[ir]; icnt++)
		  out_dim_ids[cnt][1+icnt] = icnt;
		out_dim_ids_cnt[cnt] = dim_size[ir];
	      }
	  }	/* for (ir=2 . . .  */
      }
    else 
      {
	for (ir=0, cnt=0; ir<rank-2; ir++, cnt++)
	  {
	    out_dim_ids[cnt][0] = ir;
	    for (idim=0; idim<dim_cnt; idim++)
	      if (sel_dim_ids[idim][0] == ir)
		{
		  for (icnt=1; icnt<=sel_dim_cnt[idim]; icnt++)
		    out_dim_ids[cnt][icnt] = sel_dim_ids[idim][icnt];
		  out_dim_ids_cnt[cnt] = sel_dim_cnt[idim];
		  break;
		}
	    if (idim == dim_cnt)
	      {
		for (icnt=0; icnt<dim_size[ir]; icnt++)
		  out_dim_ids[cnt][1+icnt] = icnt;
		out_dim_ids_cnt[cnt] = dim_size[ir];
	      }
	  }	/* for (ir=0 . . .  */	
      }

    if(sel_dim_ids) Free2D((void **)sel_dim_ids);
  }  /* else . . . */
  return 1;
}


