/****************************************************************************
!C

!File: sds2gif_v3.2.c

!Description:
  This file contains the routines for converting given an SDS of an HDF file 
	to a GIF file.

!Input Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original August 1999. QA Tool Version 1.2
  Modified March 2000 (modified to handle float data type)
  v3.3 binary - Modified February 2009 (modified the range_min and range_max vectors to float
		in the v3.2 code

  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	3)add the '.num' to -sds=argument so that user can specify the layer for 3D data set.

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes:
  This program uses the GIF library in gifwr.c

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "str_op.h"
#include "alloc_mem.h"
#include "qa_tool.h"
#include "meta.h"
#include "main_util.h"
#include "gif.h"

typedef unsigned char uchar;

#define MAX_NUM_RES_VAL 10

#define HELP \
"NAME \n" \
"     sds2gif - Convert an SDS in an HDF file to GIF \n" \
" \n" \
"SYNOPSIS \n" \
"     sds2gif -help \n" \
"     sds2gif -sds=SDSname [-clr=<color table>] -of=<out filename> [-scale[=min[,max]]] \n" \
"             [-res=<value list>] filename \n" \
" \n" \
"     echo -sds=SDSname [-clr=<color table>] -of=<out filename> [-scale[=min[,max]]] \n" \
"          [-res=<value list] filename | sds2gif \n" \
" \n" \
"DESCRIPTION \n" \
"     Convert user specified SDS in an HDF file to GIF. If color output is required input \n" \
"     color table, else output will be grayscale. Data will be scaled if it is not in the \n" \
"     range 0 to 255. To force scaling anyway enter -scale option and provide optional range \n" \
"     of input to be scaled. When range is specified all values less than min is set to min \n" \
"     all values greater than max is set to max. \n" \
"     Color table can also contain a set of ranges of input values and associated color mapping (5 columns). \n" \
"     In this case scale option is invalid or ignored. First two columns are considred as input range values. \n" \
" \n" \
"OPTIONS \n" \
"     -help               Display this help message \n" \
"     -sds=<SDS name>     SDS to be convereted to GIF, Only 2D SDS is allowed. \n" \
"     -of=<filename>      Output filename \n" \
"     -clr=<filename>     Input optional color table filename \n" \
"     -scale[=min[,max]]  Scale input. Minimum and maximum value are optional. \n" \
"                         This option is ignored if color table set of range values\n" \
"     -res=<value list>   List of SDS values (separated by comma) not to be scaled. \n" \
" \n" \
"Example: \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:      sds2gif -sds=SDSname [-clr=<color table>] -of=<out filename> \n" \
"            [-scale[=min[,max]]] [-res=<value list>] filename \n" \
" \n" \
"OPTIONS \n" \
"     -help               Display this help message \n" \
"     -sds=<SDS name>     SDS to be convereted to GIF, Only 2D SDS is allowed. \n" \
"     -of=<filename>      Output filename \n" \
"     -clr=<filename>     Input optional color table filename \n" \
"     -scale[=min[,max]]  Scale input. Minimum and maximum value are optional. \n" \
"     -res=<value list>   List of SDS values (separated by comma) not to be scaled. \n"

int parse_cmd_sds2gif(int argc, char **argv, char *in_fname, char *out_fname, char *clr_fname, 
		char *sds_name, int *sc, int *min, int *max, float *res_list);
int read_gif_clr_table(char *clr_fname, uchar *red, uchar *green, uchar *blue, float *min, float *max, int *n_clr);
void scale8(void *data_in, int32 data_type, int nrow, int ncol, uchar *img, int sc, int min,
		int max, float *res_list, float *range_min, float *range_max, int n_clr);

int main(int argc, char **argv)
{
  FILE *fp;
  void *data_in;
  sds_t sds_info;
  int sc, gray_scale;
  float clr_scale_min[256];
  float clr_scale_max[256];
  int p1, len, rank;
  int min, max, nrow, ncol, n_clr;
  int layer_num, layer;
  float res_list[MAX_NUM_RES_VAL];
  int32 edge[4] = {0, 0, 0};
  int32 start[4] = {0, 0, 0};
  uchar *tmp_img;
  uchar red[256]={0}, green[256]={0}, blue[256]={0};
  char num_str[5];
  char in_fname[MAX_PATH_LENGTH], out_fname[MAX_PATH_LENGTH];
  char clr_fname[MAX_PATH_LENGTH], sds_name[MAX_SDS_NAME_LEN];

  if ((argc == 2) && ((strcmp(argv[1], "-h") == 0) || (strcmp(argv[1], "-help") == 0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  if (parse_cmd_sds2gif(argc, argv, in_fname, out_fname, clr_fname, sds_name, &sc, &min, &max, res_list) == -1)
    fprintf(stderr, "%s\n", USAGE);
  else
  {
    /* MSZ 2/1/2018 silent version mismatch warning standard print out */
    setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

    /* MSZ 2/1/2018 Disable error reporting */
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

    gray_scale = read_gif_clr_table(clr_fname, red, green, blue, clr_scale_min, clr_scale_max, &n_clr);

    sds_info.sd_id = sds_info.sds_id = -1;

    strcpy(sds_info.name, sds_name);
    if (get_sds_info(in_fname, &sds_info) != FAIL)
    {
      /* MSZ 2/23/2018, add the below in response to the request from our team */
      if(sds_info.rank == 1)
      {
	fprintf(stderr, "sds2gif just deal with >=2-D SDS whereas %s has %d dimensions, eixt...\n", sds_info.name, sds_info.rank);
	exit(0);
      }
      else if(sds_info.rank == 2)
      {
	nrow = sds_info.dim_size[0];
	ncol = sds_info.dim_size[1];
	edge[0] = nrow; edge[1] = ncol;
      }
      else if(sds_info.rank == 3)
      {
        p1 = charpos(sds_name, '.', 0);
        if (p1 != -1) {
          len = strlen(sds_name);
          p1++;
          len -= p1;
          strmid(sds_name, p1, len, num_str);
          layer_num = atoi(num_str);
        }
	printf("layer_num = %d\n", layer_num);

	if(sds_info.interleave == 1) /* bsq */
 	{
	   nrow = sds_info.dim_size[1];
	   ncol = sds_info.dim_size[2];
	   layer = layer_num - 1;    /* 0 based */
	   start[0] = layer;
	   edge[0] = 1;
	   edge[1] = nrow;
	   edge[2] = ncol;
	}
	else    /* bip */
	{
	   nrow = sds_info.dim_size[0];
	   ncol = sds_info.dim_size[1];
	   layer = layer_num - 1;    /* 0 based */
	   start[2] = layer;
	   edge[0] = nrow;
	   edge[1] = ncol;
	   edge[2] = 1;
	}
      }
      else
      {
        fprintf(stderr, "%s has more than 3D dimensions and the code needs to be enhanced, exit...\n", sds_info.name);
        exit(0);
      }

      if ((tmp_img = (uchar *)calloc(nrow*ncol, sizeof(uchar))) == NULL)
        fprintf(stderr, "Cannot allocate memory for tmp_img in sds2gif: main()\n");
      if (tmp_img != NULL)
      {
	if ((data_in = get_sel_sds_data(&sds_info, start, edge)) == NULL)
	  fprintf(stderr, "Cannot read data for SDS %s\n", sds_info.name);
	else
	{
          if (gray_scale == 5) sc = 2;
	  scale8(data_in, sds_info.data_type, nrow, ncol, tmp_img, sc, min, max, res_list, clr_scale_min, clr_scale_max, n_clr);

	  if ((fp = fopen(out_fname, "w")) == NULL)
	    fprintf(stderr, "Cannot create output file: %s\n", out_fname);
	  else
	  {
	    WriteGIF(fp, tmp_img, ncol, nrow, red, green, blue, 256, gray_scale, "");
	    fclose(fp);
	  }
	}

	if(data_in) free(data_in);
	if(tmp_img) free(tmp_img);
      } 
      close_sds(&sds_info, sds_info.H5IN);
      close_sds(&sds_info, sds_info.H5IN);
    }
  }
  return 0;
}

int parse_cmd_sds2gif(int argc, char **argv, char *in_fname, char *out_fname, char *clr_fname, 
		char *sds_name, int *sc, int *min, int *max, float *res_list)
{
  int st = 1;
  int i, j, k;
  int num[2], len, nres;
  char **res_str, nstr[10], sc_str[20];

  nres = 0;
  num[0] = num[1] = -1;
  nstr[0] = sc_str[0] = '\0';
  in_fname[0] = out_fname[0] = '\0';
  clr_fname[0] = sds_name[0] = '\0';
  if ((res_str = (char **)Calloc2D(MAX_NUM_RES_VAL, 10, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for res_str in parse_cmd_sds2gif\n");
  for (i=1, *sc = 0; i<argc; i++)
  {
    if (is_arg_id(argv[i], "-sds") == 0) get_arg_val(argv[i], sds_name);
    else if (is_arg_id(argv[i], "-res") == 0) get_arg_val_arr(argv[i], res_str, &nres);
    else if (is_arg_id(argv[i], "-of") == 0) get_arg_val(argv[i], out_fname);
    else if (is_arg_id(argv[i], "-clr") == 0) get_arg_val(argv[i], clr_fname);
    else if (is_arg_id(argv[i], "-scale") == 0) { *sc = 1; get_arg_val(argv[i], sc_str); }
    else if (strcmp(argv[i], "-scale") == 0) *sc = 1;
    else if (argv[i][0] == '-') fprintf(stderr, "Ignoring invalid option %s\n", argv[i]);
    else 
    {
      if (strlen(in_fname) > 0)
	fprintf(stderr, "Input ignored: %s\n", argv[i]);
      else strcpy(in_fname, argv[i]);
    }
  }
  if (*sc == 1)
  {
    len = (int)strlen(sc_str);
    i = j = k = 0;
    while (i < len)
    {
      if (sc_str[i] == ',')
      {
	nstr[j] = '\0';
        if (j > 0) num[k] = (int)atoi(nstr);
	j = 0; i++; k++;
	if (k > 1) break;
      }
      else nstr[j++] = sc_str[i++];
    }
    nstr[j] = '\0';
    if (j > 0) num[k] = (int)atoi(nstr);
  }
  *min = num[0]; *max = num[1];
  if (strlen(in_fname) <= 0)
    { fprintf(stderr, "Missing input file \n"); st = -1; }
  if (strlen(out_fname) <= 0)
    { fprintf(stderr, "Missing output file \n"); st = -1; }
  if (strlen(sds_name) <= 0)
    { fprintf(stderr, "Missing input SDS name \n"); st = -1; }
  if (st == 1)
  {
    res_list[0] = nres;
    if (nres > 0)
      for (j=0, i=1; j<nres; j++, i++)
	res_list[i] = (float)atof(res_str[j]);
  }
  Free2D((void **)res_str);
  return st;
}

int read_gif_clr_table(char *clr_fname, uchar *red, uchar *green, uchar *blue, float *min, float *max, int *n_clr)
{
  FILE *fp;
  int i = 0;
  int nw = 0;
  int state = 0;
  int r, g, b;
  int gif_clr, len;
  float m1, m2;
  char s[255];

  gif_clr = -1;
  if (strlen(clr_fname) > 0)
  {
    if ((fp = fopen(clr_fname, "r")) != NULL)
    {
      if ((len = getlineL(fp, s)) > 0)
      {
        while (i <= len) 
	{
          if ((s[i] == ' ') || (s[i] == '\t')) 
	    state = 1;
	  else if (state == 1) {
	    state = 0;
	    ++nw;
	  }
          ++i;
        }
      }
      if (i != 0) ++nw;
      fclose(fp);
    }
    else fprintf(stderr, "Cannot open the color table: %s\n", clr_fname);
    if ((nw == 3) || (nw == 5))
    {
      if ((fp = fopen(clr_fname, "r")) != NULL)
      {
        i = 0;
        gif_clr = nw;
        if (nw == 3)
        {
          while (fscanf(fp, "%d %d %d", &r, &g, &b) == 3)
          {
            red[i]=(uchar)r;
            green[i]=(uchar)g;
            blue[i]=(uchar)b;
            i++;
          }
        }
        else
        {
          while (fscanf(fp, "%f %f %d %d %d", &m1, &m2, &r, &g, &b) == 5)
          {
	    min[i] = m1;
	    max[i] = m2;
            red[i]=(uchar)r;
            green[i]=(uchar)g;
            blue[i]=(uchar)b;
            i++;
          } 
        }
        *n_clr = i;
        fclose(fp);
      }
      else fprintf(stderr, "Cannot open the color table: %s\n", clr_fname);
    }
    else fprintf(stderr, "Data format error in color table: %s\n", clr_fname);
  }
  if (gif_clr == -1)
  {
    gif_clr = 1;
    fprintf(stderr, "Output is a grayscale image\n");
    for (i=0; i<256; i++)
    {
      red[i]=(uchar) i;
      green[i]=(uchar) i;
      blue[i]=(uchar) i;
    }
  }
  return gif_clr;
}

void scale8(void *data_in, int32 data_type, int nrow, int ncol, uchar *img, int sc, int min,
		int max, float *res_list, float *range_min, float *range_max, int n_clr)
{
  int i, ii, j, k, pt;
  int sc_val, nres;
  float min1, max1;
  float max_gray, val;

  if (sc == 0)
  {
    for (i=0; i<nrow; i++)
    {
      pt = i*ncol;
      for (j=0; j<ncol; j++, pt++)
      {
        switch(data_type)
        {
          case 5 : val = ((float32 *)data_in)[pt]; break;
          case 20: val = ((int8 *)data_in)[pt]; break;
          case 21: val = ((uint8 *)data_in)[pt]; break;
          case 22: val = ((int16 *)data_in)[pt]; break;
          case 23: val = ((uint16 *)data_in)[pt]; break;
          case 24: val = ((int32 *)data_in)[pt]; break;
          case 25: val = ((uint32 *)data_in)[pt]; break;
        }
        img[pt] = (uchar)val;
      }
    }
  }
  else if (sc == 1)
  {
    if ((min == -1) || (max == -1))
    {
      min1 = 9999999.0;
      max1 = -9999999.0;
      for (i=0; i<nrow; i++)
      {
        pt = i*ncol;
        for (j=0; j<ncol; j++, pt++)
        {
          switch(data_type)
          {
	    case 5 : val = ((float32 *)data_in)[pt]; break;
	    case 20: val = ((int8 *)data_in)[pt]; break;
	    case 21: val = ((uint8 *)data_in)[pt]; break;
	    case 22: val = ((int16 *)data_in)[pt]; break;
	    case 23: val = ((uint16 *)data_in)[pt]; break;
	    case 24: val = ((int32 *)data_in)[pt]; break;
	    case 25: val = ((uint32 *)data_in)[pt]; break;
          }
          if (val < min1) min1 = val;
          if (val > max1) max1 = val;
        }
      }
      if (min == -1) min = min1;
      if (max == -1) max = max1;
    }
    nres = (int)res_list[0];
    max_gray = (nres == 0) ? 255.0 : 255.0 - nres;
    for (i=0; i<nrow; i++)
    {
      pt = i*ncol;
      for (j=0; j<ncol; j++, pt++)
      {
        switch(data_type)
        {
	  case 5 : val = ((float32 *)data_in)[pt]; break;
          case 20: val = ((int8 *)data_in)[pt]; break;
          case 21: val = ((uint8 *)data_in)[pt]; break;
          case 22: val = ((int16 *)data_in)[pt]; break;
          case 23: val = ((uint16 *)data_in)[pt]; break;
          case 24: val = ((int32 *)data_in)[pt]; break;
          case 25: val = ((uint32 *)data_in)[pt]; break;
        }
	if (nres == 0)
	{
          if (val < min) val = min;
          if (val > max) val = max;
          sc_val = (int)((float)(val-min)*max_gray/(max-min));
	}
	else
	{
	  for (k=0; k<nres; k++)
	    if (val == res_list[k+1])
	    {
	      sc_val = 255 - k;
	      break;
	    }
	  if (k >= nres)
	  {
            if (val < min) val = min;
            if (val > max) val = max;
	    sc_val = (int)((float)(val-min)*max_gray/(max-min));
	  }
	}
        img[pt] = (uchar)sc_val;
      }
    } 
  }
  else if (sc == 2)
  {
    nres = (int)res_list[0];
    for (i=0; i<nrow; i++)
    {
      pt = i*ncol;
      for (j=0; j<ncol; j++, pt++)
      {
        switch(data_type)
        {
          case 5 : val = ((float32 *)data_in)[pt]; break;
          case 20: val = ((int8 *)data_in)[pt]; break;
          case 21: val = ((uint8 *)data_in)[pt]; break;
          case 22: val = ((int16 *)data_in)[pt]; break;
          case 23: val = ((uint16 *)data_in)[pt]; break;
          case 24: val = ((int32 *)data_in)[pt]; break;
          case 25: val = ((uint32 *)data_in)[pt]; break;
        }
        for (k=0; k<n_clr; k++)
        {
          if ((val >= range_min[k]) && (val <= range_max[k]))
          {
            img[pt] = (uchar)k;
            break;
          }
        }
        if (nres != 0)
        {
          k = n_clr - nres;
          for (ii=0; ii<nres; ++ii, ++k)
          {
            if ((val >= range_min[k]) && (val <= range_max[k]))
	    {
              img[pt] = (uchar)k;
	      break;
	    }
          }
        }
      }
    }
  }
}


