/****************************************************************************
!C

!File: sds2jpg.c

!Description:
  This file contains the routines for converting given one or three SDS of
an HDF file to a JPEG file.

!Input Parameters: (none) 

!Output Parameters: (none)

!Revision History:
  Original March 1999. QA Tool Version 1.2
  Modified September 1999. To accomodate jack's projection script requirements.
  Modified January 2000, Accept user input for SDS back ground color and SDS number for
			 Solar Zenith angle
  Modified April 2000, Add option to input smoothing factor 
  Modified May 21 2003 - For MOD09, Use the command line as in the original sds2jpg. The 
			-R -G -B options should be same as in the original sds2jpg, but 
			the code will ignore it and use the log stretching (5.5 - 9.0). 
			Code will function as the original sds2jpg for all other ESDTs
  Modified Sept 21 2009
		        The tool will use the default log scale or the linear scale 
                        from the individual RGB option 	
			The code will work for the selected 2D slice of the 3D dataset
  Modified June 10 2011 from above version of Sept 21 2009 by Xiaoping Zhang, Sigma Space Corp.
		        The code will handle the floating data type to input SDS data. 
  Modified Nov 22 2011  from above version of Sept 21 2009 by Sadashiva Devadiga, Sigma Space Corp.	
  			added -esdt option for MODISL1B, MODISL2, NPPL1B, NPPL2 to handle different log scale.
  Modified July 2014 	by Xiaoping Zhang, Sigma Space Corp.			
			merged version of June 10 2011 with version of Nov 22 2011.			
  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	3)update the 3D so that it can handle both BSQ and BIP format data

  Modified February, 2020 Version 3.6 by Maosheng Zhao, SSAI
  1) Modified in Dec. 11, 2018. A bug is identified which treating some SDS name pattern like (name.number_and_more).
  For pattern like name.number, it is MCD43 or NVP43 or the resulting SDS from reduce_sds_rank. But some resulting HDF
  has similar pattern but the last characer is not number. To rule out this case, sds2jpg changed
  from
  if((p1 = charpos(rgb_sds[i].name, '.', 0)) != -1)
  to
  if((p1 = charpos(rgb_sds[i].name, '.', 0)) != -1 && rgb_sds[i].name[strlen(rgb_sds[i].name)] >= '0' && rgb_sds[i].name[strlen(rgb_sds[i].name)] <= '9')
  2) October, 2018 Speed up the process by reading and output in lines by lines not line by line.
	Test shows that the speed is about 18(16) times faster than line by line though
	the job is devided into 10(20) times not nrows.Z

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes:
  This program uses the JPEG library stored in the directory
/home/devadiga/free_ware/jpeg-6b/jpeglib.h. Note that the JPEG library 
available on TLCF machines is a older one and will not work with this code.

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <setjmp.h>

#include "jpeglib.h"
#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "qa_tool.h"
#include "str_op.h"
#include "alloc_mem.h"
#include "main_util.h"
#include "qa_jpeg.h"

typedef struct {
 uint8 domain[3];
 uint8 outline[3];
 uint8 grid12[3];
 uint8 grid3[3];
 uint8 map[3];
 uint8 sds[3];
 uint8 night[3];
} FILL_COLOR;

typedef struct {
  int domain;
  int outline;
  int grid[4];
  int map[8];
} OFFSET;

/* Fill value for night */

#define NIGHT_VAL 10000

/*************	    Offset values for domain, map and grid     **************/
#define DOMAIN_OFFSET -50
#define OUTLINE_OFFSET -60
#define GRID_OFFSETS {-20, -20, -30, -40}
#define MAP_OFFSETS {-100, -100, -100, -110, -110, -110, -120, -130}
/****************************************************************************/

/*****      Color Codes for domain, grid, map and fill values     ***********/
#define DOMAIN_COLOR {255, 255, 255}
#define OUTLINE_COLOR {0, 0, 0}
#define GRID12_COLOR {0, 0, 0}
#define GRID3_COLOR {255, 0, 0}
#define MAP_COLOR {255, 255, 0}
#define SDS_COLOR  {35, 35, 115}
#define NIGHT_COLOR  {160, 32, 240}
/***************************************************************************/

#define PI 3.141592653589793
#define RAD (PI / 180.0)
#define SIG_IMG_NROWS 25
#define SIG_IMG_NCOLS 180 

#define HELP \
"NAME \n" \
"       sds2jpg - Convert SDS(s) of an HDF file to a JPEG \n" \
" \n" \
"SYNOPSIS \n" \
"       sds2jpg -help \n" \
"       sds2jpg -of=<filename> [-qa=<quality>] -R=<sds_id>[,<range_string>] \n" \
"               -G=<sds_id>[,<range_string>] -B=<sds_id>[,<range_string>] [-log] [-off] [-esdt=<ESDTname>]\n" \
"               -gray=<sds_id>[,<range_string>] [-bg=r,g,b] [-SZ=<sds_index>] filename \n" \
"       echo    [-qa=<quality>] -R=<sds_id>[,<range_string>] -G=<sds_id>[,<range_string>]\n" \
"               -B=<sds_id>[,<range_string>] [-log] [-off] -gray=<sds_id>[,<range_string>]\n" \
"               [-SZ=<sds_index>] [-bg=r,g,b] -of=<filename> \n" \
"               filename | sds2jpg \n" \
"       or sds2jpg < argument_file \n" \
" \n " \
"DESCRIPTION \n" \
"       Convert the input SDS(s) of an HDF file into a JPEG. User can either \n" \
"       specify three SDSs for color or single SDS for grayscale output. This tool \n" \
"       is designed to use predetermined colors for a set of fill values used by the \n" \
"       projection tool e. g. grid lines, maps etc. To turn off this use the option -off \n" \
"       Projection specific fill values are assumed to be present in al the three \n" \
"       input bands to use the predetermined colors. The input SDS are scaled using \n" \
"       the SDS specified in the -SZ option when specified. \n" \
" \n " \
"OPTIONS \n" \
"       -help                        Display this help message \n" \
"       -of=filename                 Output filename \n" \
"       -qa=quality                  Ouput JPEG quality (scale of 0 - 100, default 75)\n" \
"       -sf=smoothing factor         Ouput JPEG smoothing factor (scale of 0 - 100, default 0)\n" \
"       -R=sds_id[,range_string]     SDS index (1-based) or name, and SDS value range for RED band\n" \
"                                    Range of SDS values to be stretched can be specified by \n" \
"                                    providing min and max of SDS values for linear stretching \n" \
"                                    or by specifying series of SDS values and output values for\n" \
"                                    piecewise linear stretching of input SDS. \n" \
"                                    Format for range string: \n" \
"                                      Xmin,Xmax \n" \
"                                      X0,DN0,X1,DN1,X2,DN2 . . .\n" \
"                                    Examples:\n" \
"                                      -R=sur_refl_b03,0,1500 \n" \
"                                      -R=sur_refl_b03,0,0,1000,125,2000,200,3000,255 \n" \
"                                    (default for range_string is valid_range of sds.) \n" \
"                                    (single value entry is considered max, and min set to 0.)\n" \
"                                    The input SDS can be 2D or a 2D layer of a 3D SDS. The 2D layer \n" \
"                                    number is input dot extension in the SDS name \n" \
"                                    e.g -R=surf_refl_b01.1 or -R=1.1 \n" \
"                                    sds_index and layer numbers are 1-based numbers \n" \
"       -G=sds_id[,range_string]     SDS index (1-based) or name, and SDS value range for GREEN band\n" \
"                                    (default for range_string is valid_range of sds.) \n" \
"                                    (single value entry is considered max, and min set to 0.)\n" \
"       -B=sds_id[,range_string]     SDS index (1-based) or name, and SDS value range for BLUE band\n" \
"                                    (default for range_string is valid_range of sds.) \n" \
"                                    (single value entry is considered max, and min set to 0.)\n" \
"       -gray=sds_id[,range_string]  SDS index (1-based) or name, and SDS value range for gray output\n" \
"                                    (default for range_string is valid_range of sds.) \n" \
"                                    (single value entry is considered max, and min set to 0.)\n" \
"       -bg=r,g,b                    RGB color code for SDS background color inside the domain. \n" \
"	                             default is (35, 35, 115) \n" \
"       -SZ=sds_index                SDS index of SDS Solar Zenith angle for scaling input SDS. \n" \
"       -off                         Turn off predetermined projection specific colors \n" \
"       -log                         Use log scaling for all the input bands. Ignore \n" \
"                                    value input to range_string in the option -R, -G, and -B \n" \
"       -esdt                        ESDT of the product \n" \
"       			     only for MODISL1B, MODISL2, NPPL1B, NPPL2\n" \
" \n" \
"Example: \n" \
"	sds2jpg -qa=75 -R=\"EV_250_Aggr5km_RefSB_Band1 (20 km),0,4000\" -G=\"EV_500_Aggr5km_RefSB_Band4 (20 km),0,6500\" -B=\"EV_500_Aggr5km_RefSB_Band3 (20 km),0,6500\" -of=test.jpg /DATA/MOD02CRS/MOD02CRS.B1B3B4.GLB.PROJ.hdf \n\n" \
"	sds2jpg -qa=75 -R=\"EV_250_Aggr5km_RefSB_Band1 (20 km)\" -G=\"EV_500_Aggr5km_RefSB_Band4 (20 km)\" -B=\"EV_500_Aggr5km_RefSB_Band3 (20 km)\" -of=test.jpg /DATA/MOD02CRS/MOD02CRS.B1B3B4.GLB.PROJ.hdf \n\n" \
"	sds2jpg -qa=75 -R=1,0,4000 -G=2,0,6500 -B=3,0,6500 -of=test.jpg /DATA/MOD02CRS/MOD02CRS.B1B3B4.GLB.PROJ.hdf \n" \
"	sds2jpg -qa=75 -R=1,0,0,1200,200,9999,255 -G=2,0,0,1200,200,9999,255 -B=3,0,0,1500,200,9999,255 -of=MOD09CRS_TEST_2P_GLB.jpg MOD09CRS_TEST_GLB.hdf \n" \
"	sds2jpg -qa=75 -R=i1 -G=i2 -B=i3 -log -esdt=NPPL2 -of=NPP_SRFLIIP_L2.A2014190.1445.P1_03001.2014190193209.jpg NPP_SRFLIIP_L2.A2014190.1445.P1_03001.2014190193209.hdf \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage: sds2jpeg -of=<filename> [-qa=<quality>] -R=<sds_id>[,range_string] -G=<sds_id>[,range_string] \n" \
"\t-B=<sds_id>[,range_string] -gray=<sds_id>[,range_string] [-bg=r,g,b] [-SZ=<sds_index>] [-log] -esdt=ESDT filename \n" \
"OPTIONS \n" \
"       -help                        Display this help message \n" \
"       -of=filename                 Output filename \n" \
"       -qa=quality                  Ouput JPEG quality (scale of 0 - 100, default 75)\n" \
"       -sf=smoothing factor         Ouput JPEG smoothing factor (scale of 0 - 100, default 0)\n" \
"       -R=sds_id[,range_string]     SDS index (1-based) or name, and SDS value range for RED band\n" \
"                                    Range of SDS values to be stretched can be specified by \n" \
"                                    providing min and max of SDS values for linear stretching \n" \
"                                    or by specifying series of SDS values and output values for\n" \
"                                    piecewise linear stretching of input SDS. \n" \
"                                    Format for range string: \n" \
"                                      Xmin,Xmax \n" \
"                                      X0,DN0,X1,DN1,X2,DN2 . . .\n" \
"                                    Examples:\n" \
"                                      -R=sur_refl_b03,0,1500 \n" \
"                                      -R=sur_refl_b03,0,0,1000,125,2000,200,3000,255 \n" \
"                                    (default for range_string is valid_range of sds.) \n" \
"                                    (single value entry is considered max, and min set to 0.)\n" \
"                                    The input SDS can be 2D or a 2D layer of a 3D SDS. The 2D layer \n" \
"                                    number is input dot extension in the SDS name \n" \
"                                    e.g -R=surf_refl_b01.1 or -R=1.1 \n" \
"       -G=sds_id[,range_string]     SDS index (1-based) or name, and SDS value range for GREEN band\n" \
"                                    (default for range_string is valid_range of sds.) \n" \
"                                    (single value entry is considered max, and min set to 0.)\n" \
"       -B=sds_id[,range_string]     SDS index (1-based) or name, and SDS value range for BLUE band\n" \
"                                    (default for range_string is valid_range of sds.) \n" \
"                                    (single value entry is considered max, and min set to 0.)\n" \
"       -gray=sds_id[,range_string]  SDS index (1-based) or name, and SDS value range for gray output\n" \
"                                    (default for range_string is valid_range of sds.) \n" \
"                                    (single value entry is considered max, and min set to 0.)\n" \
"       -bg=r,g,b                    RGB color code for SDS background color inside the domain. \n" \
"	                             default is (35, 35, 115) \n" \
"       -SZ=sds_index                SDS index of SDS Solar Zenith angle for scaling input SDS. \n" \
"       -off                         Turn off predetermined projection specific colors \n" \
"       -log                         Use log scaling for all the input bands. Ignore \n" \
"                                    value input to range_string in the option -R, -G, and -B \n" \
"       -esdt                        ESDT of the product \n" \
"       			     only for MODISL1B, MODISL2, NPPL1B, NPPL2\n" \
" \n"

int parse_cmd_sds2jpg(char **argv, int argc, char *in_fname, char *out_fname, jp_band *rgb_sds, 
	int *bg_clr, int *qa, int *sf, int *gray, int *off, int *log_st, int *sz, char *sz_sds_name, char *esdt);
int sds2jpg(char *in_fname, char *out_fname, jp_band *rgb_sds, int *bg_clr, int quality, int sm_fact,
	int gray, int off, int log_st, int sz, char *sz_sds_name, char *esdt);
int get_val_range(sds_t *sds_info, int *min, int *max);
int get_rgb_param(char **rgb_param, int pcnt, jp_band *rgb_sds, int ch, char **sds_names, int nsds);

int main(int argc, char **argv)
{
  jp_band rgb_sds[3];
  int quality, gray, log_st;
  int bg_clr[3], st, off, sz, s_fact;
  char sz_sds_name[MAX_SDS_NAME_LEN];
  char in_fname[MAX_PATH_LENGTH], out_fname[MAX_PATH_LENGTH];
  char esdt[10];

  /* MSZ 2/1/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 2/1/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  st = parse_cmd_sds2jpg(argv, argc, in_fname, out_fname, rgb_sds, bg_clr, &quality, &s_fact, &gray, &off, 
			&log_st, &sz, sz_sds_name, esdt);
  if (st == -1)
  {
    fprintf(stderr, "%s\n", USAGE);
    exit(1);
  }
  else if (st != 0) 
    sds2jpg(in_fname, out_fname, rgb_sds, bg_clr, quality, s_fact, gray, off, log_st, sz, sz_sds_name, esdt);

  return 0;
}

int parse_cmd_sds2jpg(char **argv, int argc, char *in_fname, char *out_fname, jp_band *rgb_sds, 
	int *bg_clr, int *qa, int *sf, int *gray, int *off, int *log_st, int *sz, char *sz_sds_name, char *esdt)
{
  int i, st;
  int nsds, len, p1, p2;
  int rgb_pcnt[3], gray_pcnt;
  char ***rgb_param, **gray_param, **sds_names;
  static char *RGB_NAME[] = {"RED", "GREEN", "BLUE"};
  char sf_str[5], sz_str[5], qa_str[10], clr_str[10], bg_str[40];

  st = 1;
  esdt[0] = '\0';
  *sf = *off = *gray = *log_st = 0;
  qa_str[0] = out_fname[0] = '\0';
  sz_str[0] = sf_str[0] = bg_str[0] = '\0';
  for (i=0; i<3; i++)
    rgb_pcnt[i] = 0;
  for (i=0; i<3; i++)
    bg_clr[i] = -1;
  gray_pcnt = 0;

  if ((rgb_param = (char ***)Calloc3D(3, 11, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for rgb_param in sds2jpg: parse_cmd_sds2jpg()\n");
  if ((gray_param = (char **)Calloc2D(11, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for gray_param in sds2jpg: parse_cmd_sds2jpg()\n");
  if ((rgb_param != NULL) && (gray_param != NULL))
  {
    for (i=0; i<3; i++)
      rgb_param[i][0][0] = '\0';
    for (i=0; i<11; i++)
      gray_param[i][0] = '\0';
    for (i=1; i<argc; i++)
    {
      if (is_arg_id(argv[i], "-of") == 0) get_arg_val(argv[i], out_fname);
      else if (is_arg_id(argv[i], "-esdt") == 0) get_arg_val(argv[i], esdt);
      else if (is_arg_id(argv[i], "-R") == 0) get_arg_val_arr(argv[i], rgb_param[0], &rgb_pcnt[0]);
      else if (is_arg_id(argv[i], "-G") == 0) get_arg_val_arr(argv[i], rgb_param[1], &rgb_pcnt[1]);
      else if (is_arg_id(argv[i], "-B") == 0) get_arg_val_arr(argv[i], rgb_param[2], &rgb_pcnt[2]);
      else if (is_arg_id(argv[i], "-gray") == 0) get_arg_val_arr(argv[i], gray_param, &gray_pcnt);
      else if (is_arg_id(argv[i], "-qa") == 0) get_arg_val(argv[i], qa_str);
      else if (is_arg_id(argv[i], "-bg") == 0)  get_arg_val(argv[i], bg_str);
      else if (is_arg_id(argv[i], "-sf") == 0)  get_arg_val(argv[i], sf_str);
      else if ((is_arg_id(argv[i], "-sz") == 0) || (is_arg_id(argv[i], "-SZ") == 0)) get_arg_val(argv[i], sz_str);
      else if (strcmp(argv[i], "-off") == 0) *off = 1;
      else if (strcmp(argv[i], "-log") == 0) *log_st = 1;
      else if (argv[i][0] == '-') fprintf(stderr, "unknown option %s\n", argv[i]);
      else strcpy(in_fname, argv[i]);
    }
    if ((len = (int)strlen(bg_str)) > 0)
    {
      p1 = charpos(bg_str, ',', 0);
      if (p1 != -1)
      {
	if ((p2 = charpos(bg_str, ',', 0)) != -1)
	{
	  strmid(bg_str, 0, p1, clr_str);
	  bg_clr[0] = (int)atoi(clr_str);
	  strmid(bg_str, p1+1, p2-p1-1, clr_str);
	  bg_clr[1] = (int)atoi(clr_str);
	  strmid(bg_str, p2+1, len-p2-1, clr_str);
	  bg_clr[2] = (int)atoi(clr_str);
	}
      }
    }
    for (i=0; i<3; i++)
      if (rgb_param[i][0][0] == '\0') 
      {
        st = -1;
        fprintf(stderr, "Missing value for %s in RGB option\n", RGB_NAME[i]);
      }
    if ((out_fname[0] == '\0') || (in_fname[0] == '\0')) st = -1;
    if (out_fname[0] == '\0') fprintf(stderr, "Missing output filename . . .\n");
    if (in_fname[0] == '\0') fprintf(stderr, "Missing input filename . . .\n");
    if (st != -1)
    {
      *qa = (strlen(qa_str) > 0) ? (int)atoi(qa_str) : 75;
      *sf = (strlen(sf_str) > 0) ? (int)atoi(sf_str) : 0;
      if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
      {
        fprintf(stderr, "Cannot allocate memory for sds_names in parse_cmd_sds2jpg\n");
        st = 0;
      }
      else
      {
        if ((nsds = get_sds_names(in_fname, sds_names)) == 0)
        {
          fprintf(stderr, "No sds found in file %s\n", in_fname);
          st = 0;
        }
        else
        {
          *sz = (strlen(sz_str) > 0) ? (int)atoi(sz_str) - 1 : -1;
	  if (*sz != -1)
            strcpy(sz_sds_name, sds_names[*sz]);
          if (strlen(gray_param[0]) > 0)
	  {
	    for (i=0, *gray=1; i<3; i++)
              if ((st = get_rgb_param(gray_param, gray_pcnt, &rgb_sds[i], i, sds_names, nsds)) == -1)
	        break;
	  }
          else
	  {
            for (i=0, *gray=0; i<3; i++)
              if ((st = get_rgb_param(rgb_param[i], rgb_pcnt[i], &rgb_sds[i], i, sds_names, nsds)) == -1)
	        break;
	  }
        }
        Free2D((void **)sds_names);
      }
    }
  }
  else st = 0;
  if (rgb_param != NULL) Free3D((void ***)rgb_param);
  if (gray_param != NULL) Free2D((void **)gray_param);
  return st;
}

int sds2jpg(char *in_fname, char *out_fname, jp_band *rgb_sds, int *bg_clr, int quality, int sm_fact, int gray, 
	int off, int log_st, int sz, char *sz_sds_name, char *esdt)
{
  double ang;
  FILE *out_fid;
  int16 *sz_buf;
  void **in_buf;
  uint8 **tmp_buf;
  float s_fact[3][5], ln_val, tmp_val;
  int p1, len, rank;
  int layer_num[4];   /* for 3D */
  int st1, st2, st3, st4, mod09_st;
  int nrows, ncols;
  int grid3_val, map_val;
  int ir, ic, ic1, val[3];
  int i, j, k, m, st, nb, pcnt;
  int domain_val, outline_val;
  int grid12_val, night_val;
  int32 start[4], edge[4];
  sds_t sds_info[3], sds_sz_info;
  char num_str[5];
  int HDF_ver = 0;

  /* MSZ 10/24/2018, set the number of lines to be processed for each loop */
  int res_last, jcol_shift;
  int jobs_num = 20;
  int nlines_per_output_last;
  int nlines_per_output;
  int lines_num, line;
  int GROWS, last_rows, grow;

  OFFSET offset = { DOMAIN_OFFSET, OUTLINE_OFFSET, GRID_OFFSETS, MAP_OFFSETS };
  FILL_COLOR fill_color = { DOMAIN_COLOR, OUTLINE_COLOR, GRID12_COLOR, GRID3_COLOR, MAP_COLOR, SDS_COLOR, NIGHT_COLOR };
  /* MSZ 10/30/2018, speedup */
  /* JSAMPROW row_pointer[1];*/
  JSAMPROW *rows_pointers = NULL;
  struct jpeg_error_mgr jerr;
  struct jpeg_compress_struct cinfo;

  if ((out_fid = fopen(out_fname, "wb")) == NULL)
  {
    fprintf(stderr, "Cannot open output file %s\n", out_fname);
    return -1;
  }

  if ((bg_clr[0] != -1) && (bg_clr[1] != -1) && (bg_clr[2] != -1))
    for (i=0; i<3; i++)
      fill_color.sds[i] = bg_clr[i];

  nb = (gray == 1) ? 3 : 3;		/* work around for gray */

  HDF_ver = HDF_version_check(in_fname);
  for (i=0; i<nb; i++)
  {
    if(HDF_ver == 4) sds_info[i].H5IN = 0;
    else if(HDF_ver == 5) sds_info[i].H5IN = 1;
    else
    {
	fprintf(stderr, "%s is neither HDF4 or HDF5.\n", in_fname);
	return -1;
    }
    sds_info[i].sd_id = sds_info[i].sds_id = -1;
  }
  
  for (i=0, st=1; i<nb; i++)
  {
    sds_info[i].sd_id = (i == 0) ? -1 : sds_info[0].sd_id;

    /* MSZ 2/26/2018 remove the user-added suffix '.num' if there are */
    /* MSZ 12/11/2018, remove the cases that the last character is not number */
    /* MSZ 1/29/2019, correct the bug so that it is [strlen(rgb_sds[i].name) - 1] not [strlen(rgb_sds[i].name)] */
    if((p1 = charpos(rgb_sds[i].name, '.', 0)) != -1 && rgb_sds[i].name[strlen(rgb_sds[i].name) - 1] >= '0' && rgb_sds[i].name[strlen(rgb_sds[i].name) - 1] <= '9')
    {
      /* MSZ 2/26/2018, copy the rgb_sds[i].name to sds_info[i].name 
       * but remove the user-added '.num' for sds_info[i].name
       * and also get the layer_num[i] */
      strmid(rgb_sds[i].name, 0, p1, sds_info[i].name);

      p1 = charpos(rgb_sds[i].name, '.', 0);
      if (p1 != -1)
      {
        len = strlen(rgb_sds[i].name);
	p1++;
        len -= p1;
        strmid(rgb_sds[i].name, p1, len, num_str);
	layer_num[i] = atoi(num_str);
      }
    }
    else
    {
      strcpy(sds_info[i].name, rgb_sds[i].name);
      layer_num[0] = layer_num[1] = layer_num[2] = layer_num[3] = 1;   /* just 1 layer */
    }

    if ((st = get_sds_info(in_fname, &sds_info[i])) == -1)
    { 
      for (j=0; j<=i; j++)
        if (sds_info[j].sds_id != -1) close_sds(&sds_info[i], sds_info[i].H5IN);
      if (sds_info[0].sd_id != -1) close_sd(&sds_info[0], sds_info[0].H5IN);
    }
    if (st == -1) 
    {
      fprintf(stderr, "Cannot continue hdf to JPEG conversion. \n");
      break;
    }
  } /* for (i=0; . . . . . . */
  if ((st == 1) && (sz != -1))
  {
    sds_sz_info.H5IN = sds_info[0].H5IN;
    sds_sz_info.sd_id = sds_info[0].sd_id;
    sds_sz_info.sds_id = -1;
    strcpy(sds_sz_info.name, sz_sds_name);
    get_sds_info((char *)NULL, &sds_sz_info);
  }

  if (st != -1)
    for (i=0; i<nb; i++)
    {
      rgb_sds[i].f_val = sds_info[i].fill_val;
      if (rgb_sds[i].cnt == 0)
      {
        rgb_sds[i].cnt = 1;
        rgb_sds[i].dn[0] = 0; rgb_sds[i].dn[1] = 255;
	st = get_val_range(&sds_info[i], &rgb_sds[i].x[0], &rgb_sds[i].x[1]);
      }
    }

  if (st != -1)
  {
    /* MSZ 2/26/2018, add both cases of BSQ or BIP not just BSQ before */
    /* for 2D and 3D in band sequential or bip interleave */
    rank = sds_info[0].rank;
    if(rank == 1)
    {
      fprintf(stderr, "The rank of the data is 1, and we don't deal with 2D data, exit...\n");
      exit(0);
    }
    else if(rank == 2)
    {
      nrows = sds_info[0].dim_size[0];
      ncols = sds_info[0].dim_size[1];
    }
    else if(rank == 3)
    {
      if(sds_info[0].interleave == 1)      /* bsq */
      {
	nrows = sds_info[0].dim_size[1];
	ncols = sds_info[0].dim_size[2];
      }
      else                                 /* bip */
      {
	nrows = sds_info[0].dim_size[0];
	ncols = sds_info[0].dim_size[1];
      }
    }
    else
    {
      fprintf(stderr, "The rank of the data is >3D, this version of the tool doesn't deal with >3D data, exit...\n");
      exit(0);
    }

    /* MSZ 10/31/2018, speedup */
    /***************/
    nlines_per_output = (int)floor(nrows / jobs_num + 0.5);

    /* correct nlines_per_output if necessary */
    if(nrows < nlines_per_output) nlines_per_output = nrows;
    /***************/

    /* MSZ 10/31/2018, speedup, use group rows or GROWS */
    GROWS = (int)ceil((double)nrows / (double)nlines_per_output);
    last_rows = nrows % nlines_per_output;

    /* MSZ 10/31/2018, speedup */
    if ((rows_pointers = (JSAMPROW *)calloc(nlines_per_output, sizeof(JSAMPROW))) == NULL)
      fprintf(stderr, "Cannot allocate memory for sz_buf in sds2jpg\n");

  
    /* MSZ 10/31/2018, speedup */  
    /*if ((in_buf = (void **)Calloc2D(nb, ncols, sds_info[0].data_size)) == NULL)*/
    if ((in_buf = (void **)Calloc2D(nb, ncols * nlines_per_output, sds_info[0].data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for in_buf in sds2jpg\n");
    if (sz != -1)
    {
      /* MSZ 10/31/2018, speedup */
      /*if ((sz_buf = (int16 *)calloc(ncols, sizeof(int16))) == NULL)*/
      if ((sz_buf = (int16 *)calloc(ncols * nlines_per_output, sizeof(int16))) == NULL)
        fprintf(stderr, "Cannot allocate memory for sz_buf in sds2jpg\n");
    }
    /* MSZ 10/31/2018, speedup */
    /*if ((tmp_buf = (unsigned char *)calloc(3*ncols, sizeof(unsigned char))) == NULL)*/
    if ((tmp_buf = (unsigned char **)Calloc2D(nlines_per_output, 3*ncols, sizeof(unsigned char))) == NULL)
      fprintf(stderr, "Cannot allocate memory for tmp_buf in sds2jpg\n");

    if ((in_buf != NULL) && (tmp_buf != NULL))
    {
      domain_val = rgb_sds[0].f_val + offset.domain;
      outline_val = rgb_sds[0].f_val + offset.outline;
      grid12_val = rgb_sds[0].f_val + offset.grid[0];
      grid3_val = rgb_sds[0].f_val + offset.grid[2];
      map_val = rgb_sds[0].f_val + offset.map[0];
      night_val = NIGHT_VAL;
      
      fprintf(stderr, "Initializing the JPEG object and starting the JPEG compressor\n");
      cinfo.err = jpeg_std_error(&jerr);
      jpeg_create_compress(&cinfo);
      jpeg_stdio_dest(&cinfo, out_fid);
      cinfo.image_width = ncols;
      cinfo.image_height = nrows;
      cinfo.input_components = nb;
      cinfo.in_color_space = JCS_RGB;
      cinfo.smoothing_factor = sm_fact;
      jpeg_set_defaults(&cinfo);
      jpeg_set_quality(&cinfo, quality, TRUE);
      jpeg_start_compress(&cinfo, TRUE);

      fprintf(stderr, "Reading and processing rows for JPEG conversion \n");
      for (i=0; i<nb; i++)
      {
        pcnt = rgb_sds[i].cnt;
        for (j=0; j<pcnt; j++)
          s_fact[i][j] = (float)(rgb_sds[i].dn[j+1] - rgb_sds[i].dn[j])/
			 (float)(rgb_sds[i].x[j+1] - rgb_sds[i].x[j]);
      }

      /* MSZ 10/31/2018, speedup */
      /* for (ir=0; ir<nrows; ir++) */
      for (grow=0; grow<GROWS; grow++)
      {

	/* update the start for the each line */
        if (rank == 2) {
	  start[0] = grow * nlines_per_output;
          start[1] = 0;

          edge[0] = nlines_per_output;
	  if(last_rows > 0 && grow == GROWS - 1)
	    edge[0] = last_rows;

          edge[1] = ncols;
        }
        else if (rank == 3) {
          start[0] = start[1] = start[2] = 0;  /* initialized to 0 and to be updated below */

	  if(sds_info[0].interleave == 1)   /* bsq */
	  {
	    start[1] = grow * nlines_per_output;

            edge[0] = 1;

            edge[1] = nlines_per_output;
	    if(last_rows > 0 && grow == GROWS - 1)
	      edge[1] = last_rows;

            edge[2] = ncols;
	  }
	  else                              /* bip */
	  {
	    start[0] = grow * nlines_per_output;

	    edge[0] = nlines_per_output;
	    if(last_rows > 0 && grow == GROWS - 1)
		edge[0] = last_rows;

	    edge[1] = ncols;
	    edge[2] = 1;
	  }
        }

        for (i=0; i<nb; i++)
	{
	  /* MSZ 2/26/2018, for the possible input 3D data */
	  if(rank == 3)
	  {
	    if(sds_info[0].interleave == 1)     /* bsq */
	      start[0] = layer_num[i] - 1;      /* update and it is 0-based */
	    else                                /* bip */
	      start[2] = layer_num[i] - 1;      /* update and it is 0-based */
	  }

	  if (read_sel_sds_data(&sds_info[i], start, edge, in_buf[i]) == FAIL)
	  {
	    fprintf(stderr, "Error reading data row %d of SDS %s \n", grow * nlines_per_output, sds_info[i].name);
 	    break;
 	  }
	}

	if ((sz != -1) && (sds_sz_info.sds_id != FAIL))
	{
	  if (read_sel_sds_data(&sds_sz_info, start, edge, sz_buf) == FAIL)
	    fprintf(stderr, "Error reading data row %d of SDS %s \n", grow * nlines_per_output, sds_sz_info.name);
	}
	if (i < nb) break;

	/* for (ic=0, ic1=0; ic<ncols; ic++, ic1 += nb) */
        for (ic=0, ic1=0; ic<ncols * nlines_per_output; ic++, ic1 += nb)
        {
	  line = (int)(ic / ncols);

	  /* MSZ 10/31/2018, critical to reinitialize ic1 to be 0 */
	  if(ic % ncols == 0)  ic1 = 0;

	  switch(sds_info[0].data_type)
	  {
            case 3 : for (k=0; k<3; k++) val[k] = ((uchar8 **)in_buf)[k][ic]; break;
            case 4 : for (k=0; k<3; k++) val[k] = ((char8 **)in_buf)[k][ic]; break;
	    case 5 : for (k=0; k<3; k++) val[k] = (int)((float32 **)in_buf)[k][ic]; break;	    
            case 20: for (k=0; k<3; k++) val[k] = ((int8 **)in_buf)[k][ic]; break;
            case 21: for (k=0; k<3; k++) val[k] = ((uint8 **)in_buf)[k][ic]; break;
            case 22: for (k=0; k<3; k++) val[k] = ((int16 **)in_buf)[k][ic]; break;
            case 23: for (k=0; k<3; k++) val[k] = ((uint16 **)in_buf)[k][ic]; break;
            case 24: for (k=0; k<3; k++) val[k] = ((int32 **)in_buf)[k][ic]; break;
            case 25: for (k=0; k<3; k++) val[k] = ((uint32 **)in_buf)[k][ic]; break;
	  }
	  if (sz != -1) ang = sz_buf[ic];

	  if (off == 0)
	  {
	    if ((val[0] == domain_val) && (val[1]  == domain_val) && (val[2] == domain_val))
	      for (k=0; k<nb; k++) tmp_buf[line][ic1+k] = fill_color.domain[k];
	    else if ((val[0] == outline_val) && (val[1]  == outline_val) && (val[2] == outline_val))
	      for (k=0; k<nb; k++) tmp_buf[line][ic1+k] = fill_color.outline[k];
	    else if ((val[0] == grid12_val) && (val[1] == grid12_val) && (val[2] == grid12_val))
	      for (k=0; k<nb; k++) tmp_buf[line][ic1+k] = fill_color.grid12[k];
	    else if ((val[0] == grid3_val) && (val[1] == grid3_val) && (val[2] == grid3_val))
	      for (k=0; k<nb; k++) tmp_buf[line][ic1+k] = fill_color.grid3[k];
	    else if ((val[0] == map_val) && (val[1] == map_val) && (val[2] == map_val))
	      for (k=0; k<nb; k++) tmp_buf[line][ic1+k] = fill_color.map[k];
	    else if ((val[0] == rgb_sds[0].f_val) && (val[1] == rgb_sds[1].f_val) 
						    && (val[2] == rgb_sds[2].f_val)) 
	      for (k=0; k<nb; k++) tmp_buf[line][ic1+k] = fill_color.sds[k];
	    else if ((val[0] == night_val) && (val[1] == night_val) && (val[2] == night_val))
	      for (k=0; k<nb; k++) tmp_buf[line][ic1+k] = fill_color.night[k];
            else for (k=0; k<nb; k++)
            {
	      pcnt = rgb_sds[k].cnt;
	      if (log_st == 1) {
		if (val[k] <= 0) val[k] = 0;
		else {
		  ln_val = (float)log((float)val[k]);
		  if ((strcmp(esdt, "NPPL2") == 0) || (strcmp(esdt, "MODISL2") == 0)) {
		    if (ln_val <= 5.5) val[k] = 0;
		    else if (ln_val >= 9.0) val[k] = 255;
		    else {
		      tmp_val = 255.0 * (ln_val - 5.5) / (9.0 - 5.5);
		      val[k] = (int)ceilf(tmp_val);
		    }
		  }
		  else if ((strcmp(esdt, "NPPL1B") == 0) || (strcmp(esdt, "MODISL1B") == 0)) {
		    if (ln_val <= 7.5) val[k] = 0;
		    else if (ln_val >= 10.0) val[k] = 255;
		    else {
		      tmp_val = 255.0 * (ln_val - 7.5) / (10.0 - 7.5);
		      val[k] = (int)ceilf(tmp_val);
		    }
		  }
		  else {
		    if (ln_val <= 5.5) val[k] = 0;
		    else if (ln_val >= 9.0) val[k] = 255;
		    else {
		      tmp_val = 255.0 * (ln_val - 5.5) / (9.0 - 5.5);
		      val[k] = (int)ceilf(tmp_val);
		    }		  
		  }			  
		} 
	      }
	      else if (sz == -1)
	      {
	        if (val[k] < rgb_sds[k].x[0]) val[k] = rgb_sds[k].x[0];
	        else if (val[k] > rgb_sds[k].x[pcnt]) val[k] = rgb_sds[k].x[pcnt];
	      }
	      else 
	      {
	        if (ang != -32767.0)
	        {
	          if (fabs(ang) < 8900.0)
	            val[k] = (int)((double)val[k]/cos(ang*0.01*RAD));
	          else val[k] = rgb_sds[k].x[pcnt];
	          if (val[k] < rgb_sds[k].x[0]) val[k] = rgb_sds[k].x[0];
	          else if (val[k] > rgb_sds[k].x[pcnt]) val[k] = rgb_sds[k].x[pcnt];
	        }
	        else
	        {
	          if (val[k] < rgb_sds[k].x[0]) val[k] = rgb_sds[k].x[0];
	          else if (val[k] > rgb_sds[k].x[pcnt]) val[k] = rgb_sds[k].x[pcnt];
	        }
	      }
	      if (log_st == 0) {
	        for (m=0; m<pcnt; m++)
		{
	          if ((val[k] >= rgb_sds[k].x[m]) && (val[k] <= rgb_sds[k].x[m+1]))
	          {
	            tmp_buf[line][ic1+k] = (uint8)(rgb_sds[k].dn[m] + (val[k] - rgb_sds[k].x[m])*s_fact[k][m]);
	            break;
	          }
		}
	      }
	      else tmp_buf[line][ic1+k] = (uint8)val[k];
	    } /* for (k=0; . . . ) */
	  } /* if (off = 0) . . . */
	  else 
	  {
	    for (k=0; k<nb; k++)
	    {
	      pcnt = rgb_sds[k].cnt;
              if (log_st == 1) {
                if (val[k] <= 0) val[k] = 0;
                else {
                  ln_val = (float)log(val[k]);
                  if (ln_val <= 5.5) val[k] = 0;
                  else if (ln_val >= 9.0) val[k] = 255;
                  else { 
		    tmp_val = 255.0 * (ln_val - 5.5) / (9.0 - 5.5);
		    val[k] = (int)ceilf(tmp_val);
		  }
                }
              }
	      else if (sz == -1)
	      {
	        if (val[k] < rgb_sds[k].x[0]) val[k] = rgb_sds[k].x[0];
	        else if (val[k] > rgb_sds[k].x[pcnt]) val[k] = rgb_sds[k].x[pcnt];
	      }
	      else 
	      {
	        if (ang != -32767.0)
	        {
	          if (fabs(ang) < 8900.0)
	            val[k] = (int)((double)val[k]/cos(ang*0.01*RAD));
	          else val[k] = rgb_sds[k].x[pcnt];
	          if (val[k] < rgb_sds[k].x[0]) val[k] = rgb_sds[k].x[0];
	          else if (val[k] > rgb_sds[k].x[pcnt]) val[k] = rgb_sds[k].x[pcnt];
	        }
	        else
	        {
	          if (val[k] < rgb_sds[k].x[0]) val[k] = rgb_sds[k].x[0];
	          else if (val[k] > rgb_sds[k].x[pcnt]) val[k] = rgb_sds[k].x[pcnt];
	        }
	      }
	      if (log_st == 0) {
	        for (m=0; m<pcnt; m++) {
	          if ((val[k] >= rgb_sds[k].x[m]) && (val[k] < rgb_sds[k].x[m+1]))
	          {
	            tmp_buf[line][ic1+k] = (uint8)(rgb_sds[k].dn[m] + (val[k] - rgb_sds[k].x[m])*s_fact[k][m]);
	            break;
	          }
		}
	      }
	      else tmp_buf[line][ic1+k] = (uint8)val[k];
	    } /* for (k=0; . . .) */
	  } /* else */
        } /* for (ic=0; . . .) */

	/* MSZ 10/31/2018, write lines not single line for each loop */
	/*row_pointer[0] = tmp_buf;*/
	/* (void) jpeg_write_scanlines(&cinfo, row_pointer, 1); */

	/* MSZ 10/31/2018, note that the last chunk may have different lines than nlines_per_output */
	lines_num = nlines_per_output;
	if(last_rows > 0 && grow == GROWS - 1)
	   lines_num = last_rows;

	/* MSZ 10/31/2018, get rows_pointers */
	for(ir = 0; ir < lines_num; ir++)
	  rows_pointers[ir] = tmp_buf[ir];

	/* MSZ 10/31/2018, write multilines by single line */
	(void) jpeg_write_scanlines(&cinfo, rows_pointers, lines_num);

      } /* for (grow=0; . . . . . */

      jpeg_finish_compress(&cinfo);
      jpeg_destroy_compress(&cinfo);
      fprintf(stderr, "JPEG conversion completed. Output JPEG file is %s\n", out_fname);
    } /* if ((in_buf != . . . . */
    Free2D((void **)in_buf);
    if (sz != -1) free(sz_buf);
    if (tmp_buf != NULL) Free2D((void **)tmp_buf);
  } /* if (st != -1) . . . */ 
  fclose(out_fid);
  return st;
}

int get_val_range(sds_t *sds_info, int *min, int *max)
{
  void *data;
  int32 edge[2] = {0, 0};
  int32 start[2] = {0, 0};
  int ir, ic, nr, nc, val;
  int i, fval_st, fill_val;
  int map_offset[8] = {-20, -30, -40, -50, -100, -110, -120, -130};

  /* MSZ 10/24/2018, set the number of lines to be processed for each loop */
  int res_last, jcol_shift;
  int jobs_num = 8;
  int nlines_per_output_last;
  int nlines_per_output;
  int GROWS, last_rows, grow;

  fill_val = sds_info->fill_val;

  for (i=0; i<8; i++)
    map_offset[i] += fill_val;
  nr = sds_info->dim_size[0];
  nc = sds_info->dim_size[1];

  /* MSZ 10/31/2018, speedup */
  /***************/
  nlines_per_output = (int)floor(nr / jobs_num + 0.5);

  /* correct nlines_per_output if necessary */
  if(nr < nlines_per_output) nlines_per_output = nr;
  /***************/

  /* MSZ 10/31/2018, speedup, use group rows or GROWS */
  GROWS = (int)ceil((double)nr / (double)nlines_per_output);
  last_rows = nr % nlines_per_output;

  /* MSZ 10/31/2018, speedup */
  /*if ((data = (void *)calloc(nc, sds_info->data_size)) == NULL)*/
  if ((data = (void *)calloc(nc * nlines_per_output, sds_info->data_size)) == NULL)
  {
    fprintf(stderr, "Cannot allocate memory for data in get_val_range\n");
    return -1;
  }

  *min = 99999;
  *max = -99999;
  edge[0] = 1;
  start[1] = 0;
  edge[1] = nc;
  /* MSZ 10/31/2018, speedup */
  /*for (ir=0; ir<nr; ir++) */
  for (grow=0; grow<GROWS; grow++)
  {
    /* start[0] = ir; */
    start[0] = grow * nlines_per_output;

    /* MSZ, 10/31/2018, be cautious about the last group rows because it may be different from nlines_per_output */
     edge[0] = nlines_per_output;
    if(last_rows > 0 && grow == GROWS - 1)
      edge[0] = last_rows;

    if (read_sel_sds_data(sds_info, start, edge, data) == FAIL)
    {
      fprintf(stderr, "Error reading data row %d of SDS %s \n", start[0], sds_info->name);
      break;
    }
    /* MSZ 10/31/2018, speedup */
    /*for (ic=0; ic<nc; ic++)*/
    for (ic=0; ic<nc * edge[0]; ic++)
    {
      switch(sds_info->data_type)
      {
        case 3 : val = ((uchar8 *)data)[ic]; break;
        case 4 : val = ((char8 *)data)[ic]; break;
	case 5 : val = (int)((float32 *)data)[ic]; break;
        case 20: val = ((int8 *)data)[ic]; break;
        case 21: val = ((uint8 *)data)[ic]; break;
        case 22: val = ((int16 *)data)[ic]; break;
        case 23: val = ((uint16 *)data)[ic]; break;
        case 24: val = ((int32 *)data)[ic]; break;
        case 25: val = ((uint32 *)data)[ic]; break;
      }
      if (val != fill_val)
      {
	for (i=0, fval_st=0; i<8; i++)
	  if (val == map_offset[i]) { fval_st = 1; break; }
	if (fval_st == 0)
	{
	  if (val > *max) *max = val;
          else if (val < *min) *min = val;
	}
      }
    }
  }
  free(data);
  return 1;
}

int get_rgb_param(char **rgb_param, int pcnt, jp_band *rgb_sds, int ch, char **sds_names, int nsds)
{
  int p1, len; 
  int i, ii, isds, sds_index;
  char id_str[10];
  static char *RGB_NAME[] = {"RED", "GREEN", "BLUE"};

  if (strlen(rgb_param[0]) <= 0)
  {
    fprintf(stderr, "No sds_name specified for %s band\n",  RGB_NAME[ch]);
    return -1;
  }
  else
  {
    /* SDS name can contain the dot extension identifying the 3rd layer */
    for (isds=0; isds<nsds; isds++)
      if (strstr(rgb_param[0], sds_names[isds]) != NULL) {
        strcpy(rgb_sds->name, rgb_param[0]);
	break;
      }
    /* If SDS could not be identified convert this to integer index */
    /* integer index could also contain dot extension to identify the 3rd layer */
    if (isds >= nsds) {
      if ((p1 = charpos(rgb_param[0], '.', 0)) != -1) {
        strmid(rgb_param[0], 0, p1, id_str);
        sds_index = atoi(id_str) - 1;
	len = (int)strlen(rgb_param[0]);
        p1++;
        len -= p1;
        strmid(rgb_param[0], p1, len, id_str);
        p1 = (int)atoi(id_str); 
        sprintf(rgb_sds->name, "%s.%d",  sds_names[sds_index], p1);
      }
      else {
        sds_index = atoi(rgb_param[0]) - 1;
        strcpy(rgb_sds->name, sds_names[sds_index]);
      }
    }
  }
  if (pcnt == 1)
  {
    rgb_sds->cnt = 0;
    fprintf(stderr, "   Range of values set to SDS valid range \n");
  }
  else if (pcnt <= 3)
  {
    rgb_sds->cnt = 1;
    rgb_sds->dn[0] = 0; rgb_sds->dn[1] = 255;
    rgb_sds->x[0] = (pcnt == 2) ? 0 : (int)atoi(rgb_param[1]);
    rgb_sds->x[1] = (pcnt == 2) ? (int)atoi(rgb_param[1]) : (int)atoi(rgb_param[2]);
    if (pcnt == 2) fprintf(stderr, "Minimum value set to 0\n");
  }
  else
  {
    rgb_sds->cnt = (pcnt - 1)/2 - 1;
    for (i=0; i<=rgb_sds->cnt; i++)
    {
      ii = 2*i;
      rgb_sds->x[i] = (int)atoi(rgb_param[ii + 1]);
      rgb_sds->dn[i] = (int)atoi(rgb_param[ii + 2]);
    }
  }
  rgb_sds->f_val = -1;
  return 1;
}


