/****************************************************************************
!C

!File: sds_hist.c

!Description:
   Compute histogram of SDS values in an SDS of an HDF file.  

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original June 2001. QA Tool Version 1.4
  Changed version v3.0 April 2013
  Modified July 2014 version 3.1 by Xiaoping Zhang
  	added a function of add_to_npp_hist_from_row to handle NPP fill value issue.

  Modified in April, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	3)remove add_to_npp_hist_from_row() but just use
	   add_to_hist_from_row() for both HDF4/HDF5

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "sds_rw.h"
#include "qa_tool.h"
#include "main_util.h"
#include "alloc_mem.h"
#include "str_op.h"
#include "l2g.h"


#define MAX_NUM_INT8 256
#define MAX_NUM_INT16 65536 
#define MAX_NUM_INT32 4294967296 

#define HELP \
"NAME \n" \
"       sds_hist - Compute histogram of SDS values. \n" \
" \n" \
"SYNOPSIS \n" \
"       sds_hist [-help] \n" \
"       sds_hist [-sds=<sds_name>] [-layer] [-range=<min,max>] filename \n" \
" \n" \
"       echo [-sds=sds_name] [-layer] [-range=<min,max>] filename | sds_hist \n" \
" \n" \
"DESCRIPTION \n" \
"       Compute histogram of data values in an SDS. Tool accepts 2D/3D/4D SDS. \n" \
"In case of 3D and 4D SDS user can request the detail for every SDS layer/slice. \n" \
"Output is directed to stdout. Output contains SDS name with dimension size and \n" \
"fill value followed by list of SDS values. In case of 3D and 4D SDSs fill value \n" \
"and list of SDS values are listed for each layer. SDS layers are identified by the \n" \
"dimension index of that layer. \n" \
" \n" \
"OPTIONS \n" \
"       -help                  Display this help message \n" \
"       -sds=<sds_name>        SDS names separated by comma. (default: all SDS) \n" \
"       -layer                 Compute histogram for every layer separately \n" \
"       -range=<min,max>       Histogram range (minimum and maximum values). \n" \
"                              Default is set to valid range of the SDS. Fill value is \n" \
"                              counted separately. If valid range attribute is not \n" \
"                              available the range of the SDS data type is used as the \n" \
"                              the limit. The range value for various data type is shown \n" \
"                              below \n" \
"                                INT8:  (-128, 127)     UINT8: (0, 255)\n" \
"                                INT16: (-32768, 32767) UINT16: (0, 65535) \n" \
"                                INT32: (-2147483648, 2147483647) UINT32: (0, 4294967295) \n" \
"                                FLOAT32: UNDEFINED. \n" \
"                              For float data type the histogram is computed after converting\n" \
"                              the float values to their closest integer. \n" \
"	filename               One or more input filenames \n" \
" \n" \
"Examples: \n" \
"\tsds_hist -sds=SurfReflect_M1 VNP09A1.A2016289.h11v08.001.2017189045223.h5 \n" \
"\tsds_hist -layer -sds=BRDF_Albedo_Parameters_M1 -range=0,10000 VNP43M1.A2016195.h09v05.001.2017223180734.h5\n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:		sds_hist [-sds=<sds name>] [-layer] [-range=<min,max>] filename\n" \
" \n" \
"OPTIONS \n" \
"       -help                  Display this help message \n" \
"       -sds=<sds_name>        SDS names separated by comma. (default: all SDS) \n" \
"       -layer                 Compute histogram for every layer separately\n" \
"       -range=<min,max>       Histogram range (minimum and maximum values). \n" \
"                              Default is set to valid range of the SDS. Fill value is \n" \
"                              counted separately. If valid range attribute is not \n" \
"                              available the range of the SDS data type is used as the \n" \
"                              the limit. The range value for various data type is shown \n" \
"                              below \n" \
"                                INT8:  (-128, 127)     UINT8: (0, 255)\n" \
"                                INT16: (-32768, 32767) UINT16: (0, 65535) \n" \
"                                INT32: (-2147483648, 2147483647) UINT32: (0, 4294967295) \n" \
"                                FLOAT32: UNDEFINED. \n" \
"	filename               One or more input filenames \n" \
" \n"

int parse_cmd_sds_hist(int argc, char **argv, int *nd_info, char **sds_names, int *nsds, int *hist_range, int *fcnt);
void compute_sds_hist(char *fname, char **sds_names, int nsds, int nd_info, int *hist_range);
void add_to_hist_from_row(sds_t *sds_info, void *data_in, int ncol, int st_c, int offset, int *hist_cnt, int *fill_cnt, int *hist_range);
void print_sds_hist(sds_t *sds_info, int **hist_cnt, int *fill_cnt, int n_layer, int *hist_range);

int int8_range[2] = {-128, 127};
int uint8_range[2] = {0, 255};
int int16_range[2] = {-32768, 32767};
int uint16_range[2] = {0, 65535};
int32 int32_range[2] = {-2147483648, 2147483647};
uint32 uint32_range[2] = {0, 4294967295};

int main(int argc, char **argv)
{
  char **sds_names = NULL, **sds_names_exclusion = NULL;
  int hist_range[2];
  int nsds, sds_cnt;
  int i, fcnt, nd_info;

  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }

  /* MSZ 5/5/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 5/5/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

  if ((sds_names = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_names in sds_range: main()\n");
  else
  {
    if (parse_cmd_sds_hist(argc, argv, &nd_info, sds_names, &nsds, hist_range, &fcnt) == -1)
      fprintf(stderr, "%s\n", USAGE);
    else
    {
      for (i=1; i<argc; i++)
      {
        if (argv[i][0] != '-')
        {
	  fprintf(stdout, "Reading from %s\n", argv[i]);
	  if (nsds == 0) 
	  {
	    /* for HDF5 input, excluded StructMetadata.0 for sds_names */
	    if ((sds_names_exclusion = (char **)Calloc2D(MAX_NUM_SDS, MAX_SDS_NAME_LEN, sizeof(char))) == NULL)
	    {
	      fprintf(stderr, "Cannot allocate memory for sds_names_exclusion in main(), exit... \n");
	      exit(0);
	    }

	    strcpy(sds_names_exclusion[0], "StructMetadata.0");
	    strcpy(sds_names_exclusion[1], "CoreMetadata.0");
	    strcpy(sds_names_exclusion[2], "ArchiveMetadata.0");

	    sds_cnt = get_sds_names_exclusions(argv[i], sds_names, sds_names_exclusion, 3);

	    if(sds_names_exclusion) Free2D((void**)sds_names_exclusion);
	  }
          else sds_cnt = nsds;  
	  compute_sds_hist(argv[i], sds_names, sds_cnt, nd_info, hist_range); 
        }
      }
    }
    if(sds_names) Free2D((void **)sds_names);
  }
  return 0;
}

int parse_cmd_sds_hist(int argc, char **argv, int *nd_info, char **sds_names, int *nsds, int *hist_range, int *fcnt)
{
  int i, p1, st, len;
  char range_str[50], val_str[25];

  st = 1;
  range_str[0] = '\0';
  *fcnt = *nsds = *nd_info = 0;
  hist_range[0] = hist_range[1] = -111;
  for (i=1; i<argc; i++)
  {
    if (strcmp(argv[i], "-layer") == 0) *nd_info = 1;
    else if (is_arg_id(argv[i], "-sds") == 0)
      get_arg_val_arr(argv[i], sds_names, nsds);
    else if (is_arg_id(argv[i], "-range") == 0)
      get_arg_val(argv[i], range_str);
    else if (argv[i][0] == '-')
      fprintf(stderr, "Ignoring unknown option %s\n", argv[i]);
    else ++*fcnt;
  }
  if (*fcnt == 0) {
    st = -1;
    fprintf(stderr, "Missing input file . . \n");
  }
  if (range_str[0] != '\0')
  {
    p1 = charpos(range_str, ',', 0);
    if (p1 != -1)
    {
      len = (int)strlen(range_str);
      strmid(range_str, 0, p1, val_str);
      hist_range[0] = (int)atoi(val_str);
      strmid(range_str, p1+1, len-p1-1, val_str);
      hist_range[1] = (int)atoi(val_str);
      if (hist_range[0] > hist_range[1]) {
        st = -1;
        fprintf(stderr, "Invalid range option %s\n", range_str);
      }
    }
    else {
      st = -1;
      fprintf(stderr, "Invalid range option %s\n", range_str);
    } 
  }
  else 
    fprintf(stderr, "No range option Input. Using valid range from SDS or default range of the data type. \n");
  if (st == 1)
    if (*nsds == 0) fprintf(stderr, "No SDS name input. Reading all SDSs \n");
  return st;
}

void compute_sds_hist(char *fname, char **sds_names, int nsds, int nd_info, int *hist_range)
{
  void *data_in;
  sds_t sds_info;
  int bsq, rank;
  int n, m, max_m;
  int st_c, offset;
  int irange, n_val;
  int layer_id, n_layer;
  int irow, irank, isds;
  int nrows, ncols, ndata;
  int *fill_cnt, **hist_cnt;
  int32 start[4], edge[4];
  int st;
   
  sds_info.sd_id = -1;
  for (isds=0; isds<nsds; isds++)
  {
    sds_info.sds_id = -1;
    strcpy(sds_info.name, sds_names[isds]);
    for (irank=0; irank<4; irank++)
      start[irank] = edge[irank] = 0;
    sds_info.range[0] = sds_info.range[1] = -111;
    if (get_sds_info(fname, &sds_info) != -1)
    { 
      if (sds_info.data_type == 5)
      {
	if ((hist_range[0] == -111) && (hist_range[1] == -111))
	  fprintf(stderr, "SDS %s is of float data type. Needs Histogram range input\n", sds_info.name);
      }
      else 
      {
	if ((hist_range[0] == -111) && (hist_range[1] == -111))
	{
	  if ((sds_info.range[0] == -111) && (sds_info.range[1] == -111))
	  {
	    switch(sds_info.data_type) {
	      case 20: hist_range[0] = int8_range[0]; hist_range[1] = int8_range[1]; break; 
	      case 21: hist_range[0] = uint8_range[0]; hist_range[1] = uint8_range[1]; break; 
	      case 22: hist_range[0] = int16_range[0]; hist_range[1] = int16_range[1]; break; 
	      case 23: hist_range[0] = uint16_range[0]; hist_range[1] = uint16_range[1]; break; 
	      case 24: hist_range[0] = int32_range[0]; hist_range[1] = int32_range[1]; break; 
	      case 25: hist_range[0] = uint32_range[0]; hist_range[1] = uint32_range[1]; break; 
	    }
	  }
	  else
	  { hist_range[0] = sds_info.range[0]; hist_range[1] = sds_info.range[1]; }
	}
      }
      if ((hist_range[0] != -111) && (hist_range[1] != -111))
      {
        rank = sds_info.rank;
        n_layer = 1;
        if (rank == 1)
        {
          nrows = 1;
	  edge[0] = ncols = ndata = sds_info.dim_size[0];
        }
        else
        {
          bsq = ((rank == 2) || (sds_info.dim_size[0] < sds_info.dim_size[rank - 1])) ? 1 : 0;
          if (bsq == 1)
          {
            nrows = sds_info.dim_size[rank-2];
            ncols = ndata = edge[rank-1] = sds_info.dim_size[rank-1];
            for (irank=0; irank<rank-2; irank++)
            {
	      n_layer *= sds_info.dim_size[irank];
              ndata *= sds_info.dim_size[irank];
              edge[irank] = sds_info.dim_size[irank];
            }
            edge[rank-2] = 1;
	    max_m = sds_info.dim_size[1];
          }
          else
          {
	    nrows = sds_info.dim_size[0];
	    ncols = ndata = sds_info.dim_size[1];
	    for (irank=2; irank<rank; irank++)
	    {
	      n_layer *= sds_info.dim_size[irank];
	      ndata *= sds_info.dim_size[irank];
	      edge[irank] = sds_info.dim_size[irank];
	    }
	    edge[0] = 1;
	    edge[1] = sds_info.dim_size[1];
	    max_m = sds_info.dim_size[3];
          }
        }
  
	if ((nd_info == 0) || (rank <= 2))
	{
	  st_c = 0;
	  offset = 1;
	  n_layer = 1;
	}
  
        if ((data_in = (void *)calloc(ndata, sds_info.data_size)) == NULL)
	  fprintf(stderr, "Cannot allocate memory for data_in in read_sds_range()\n");
        n_val = hist_range[1] - hist_range[0] + 1;
	if ((hist_cnt = (int **)Calloc2D(n_layer, n_val, sizeof(int))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for hist_cnt in read_sds_values\n");	
	if ((fill_cnt = (int *)calloc(n_layer, sizeof(int))) == NULL)
	  fprintf(stderr, "Cannot allocate memory for fill_cnt in read_sds_values\n");	
        if ((data_in != NULL) && (hist_cnt != NULL) && (fill_cnt != NULL)) 
	{
          for (layer_id=0; layer_id<n_layer; layer_id++)
	    fill_cnt[layer_id] = 0;
          for (layer_id=0; layer_id<n_layer; layer_id++)
	    for (irange=0; irange<n_val; irange++)
	      hist_cnt[layer_id][irange] = 0;
          for (irow=0; irow<nrows; irow++)
          {
            if (rank == 1) start[0] = 0;
	    else
	    {
	      if (bsq == 1) start[rank-2] = irow;
	      else start[0] = irow;
	    }
	    if(read_sel_sds_data(&sds_info, start, edge, data_in) == FAIL)
	      fprintf(stderr, "Failed to read data row for SDS %s\n", sds_info.name);
	    else
            {
	      for (layer_id=0; layer_id<n_layer; layer_id++)
	      {
	        if (nd_info == 1) 
	        {
		  m = (rank == 3) ? -1 : layer_id%max_m;
		  n = (rank == 3) ? layer_id : layer_id/max_m;
		  compute_sds_start_offset(&sds_info, n, m, &st_c, &offset);
	        }
	        else ncols = ndata;

	        add_to_hist_from_row(&sds_info, data_in, ncols, st_c, offset, hist_cnt[layer_id], &fill_cnt[layer_id], hist_range);
	      }
	    }
          } /* for (irow=0;  . .  ) */
          print_sds_hist(&sds_info, hist_cnt, fill_cnt, n_layer, hist_range);
	}
	if(hist_cnt) Free2D((void **)hist_cnt);
	if(fill_cnt) free(fill_cnt);
        if(data_in) free(data_in);
	close_sds(&sds_info, sds_info.H5IN);
      }
    }
  } /* for (isds=0;  . .  ) */
  close_sd(&sds_info, sds_info.H5IN);
}
	        
void add_to_hist_from_row(sds_t *sds_info, void *data_in, int ncol, int st_c, int offset, int *hist_cnt, int *fill_cnt, int *hist_range)
{
  float sds_fval;
  int i, id, icol, sds_val;

  switch(sds_info->data_type)
  {
    case 5 : for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_fval = ((float *)data_in)[icol]; 
      	       sds_val = (int)sds_fval; 
      	       if (sds_fval == sds_info->fill_fval) *fill_cnt += 1;
	       else if ((sds_val >= hist_range[0]) && (sds_val <= hist_range[1]))
	       {
		 id = sds_val - hist_range[0];
		 hist_cnt[id] += 1;
	       }	
             }
    	     break;
    case 20: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((int8 *)data_in)[icol]; 
      	       if (sds_val == sds_info->fill_val) *fill_cnt += 1;
	       else if ((sds_val >= hist_range[0]) && (sds_val <= hist_range[1]))
      	       {
		 id = sds_val - hist_range[0];
		 hist_cnt[id] += 1;
      	       }
             }
    	     break;
    case 21: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((uint8 *)data_in)[icol]; 
      	       if (sds_val == sds_info->fill_val) *fill_cnt += 1;
	       else if ((sds_val >= hist_range[0]) && (sds_val <= hist_range[1]))
      	       {
		 id = sds_val - hist_range[0];
		 hist_cnt[id] += 1;
      	       }
             }
    	     break;
    case 22: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((int16 *)data_in)[icol]; 
      	       if (sds_val == sds_info->fill_val) *fill_cnt += 1;
	       else if ((sds_val >= hist_range[0]) && (sds_val <= hist_range[1]))
      	       {
		 id = sds_val - hist_range[0];
		 hist_cnt[id] += 1;
      	       }
             }
    	     break;
    case 23: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((uint16 *)data_in)[icol]; 
      	       if (sds_val == sds_info->fill_val) *fill_cnt += 1;
	       else if ((sds_val >= hist_range[0]) && (sds_val <= hist_range[1]))
      	       {
		 id = sds_val - hist_range[0];
		 hist_cnt[id] += 1;
      	       }
             }
    	     break;
    case 24: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((int32 *)data_in)[icol]; 
      	       if (sds_val == sds_info->fill_val) *fill_cnt += 1;
	       else if ((sds_val >= hist_range[0]) && (sds_val <= hist_range[1]))
      	       {
		 id = sds_val - hist_range[0];
		 hist_cnt[id] += 1;
      	       }
             }
    	     break;
    case 25: for (i=0, icol=st_c; i<ncol; icol += offset, i++)
    	     {
      	       sds_val = ((uint32 *)data_in)[icol]; 
      	       if (sds_val == sds_info->fill_val) *fill_cnt += 1;
	       else if ((sds_val >= hist_range[0]) && (sds_val <= hist_range[1]))
      	       {
		 id = sds_val - hist_range[0];
		 hist_cnt[id] += 1;
      	       }
             }
    	     break;
  }
}

void print_sds_hist(sds_t *sds_info, int **hist_cnt, int *fill_cnt, int n_layer, int *hist_range)
{
  int i, j, id, sum;
  char fval_str[25], dim_str[80];
  int min, max;
  double tot, tot2, mean, std; 

  if (sds_info->rank == 2) 
    sprintf(dim_str, "Dimension = (%d x %d)", sds_info->dim_size[0], sds_info->dim_size[1]);
  else if (sds_info->rank == 3) 
    sprintf(dim_str, "Dimension = (%d x %d x %d)", sds_info->dim_size[0], sds_info->dim_size[1], sds_info->dim_size[2]);
  else if (sds_info->rank == 4) 
    sprintf(dim_str, "Dimension = (%d x %d x %d x %d)", sds_info->dim_size[0], sds_info->dim_size[1], sds_info->dim_size[2], sds_info->dim_size[3]);
  if (sds_info->data_type == 5) sprintf(fval_str, "Fill Value = %f", sds_info->fill_fval);
  else sprintf(fval_str, "Fill Value = %d", sds_info->fill_val);
  fprintf(stdout, "%s:\t%s\t%s\n", sds_info->name, dim_str, fval_str);

  for (i=hist_range[0]; i<=hist_range[1]; i++)
  {
    id = i - hist_range[0];
    for (j=0, sum=0; j<n_layer; j++)
      sum += hist_cnt[j][id];
    if (sum != 0) { 
      fprintf(stdout, "%d", i);
      for (j=0; j<n_layer; j++)
        fprintf(stdout, "\t%d", hist_cnt[j][id]);
      fprintf(stdout, "\n");
    }
  }
  if (sds_info->data_type == 5)
    fprintf(stdout, "%f", sds_info->fill_fval);
  else 
    fprintf(stdout, "%d", sds_info->fill_val);
  for (j=0; j<n_layer; j++)
    fprintf(stdout, "\t%d", fill_cnt[j]);
  fprintf(stdout, "\n"); 
}


