/****************************************************************************
!C

!File: sds_math.c

!Description:
  Contains routines for simple arithmatic on two sds 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original December 1999 Version 1.2 (developed as sds_diff supporting only - operation)
  Modified June 2001. Version 1.4 ( supports +, -, *, and /) 
  Modified March 2013  by Xiaoping Zhang, Sigma Space Corp.
  	(added the fill value attribute to SDS for output hdf file)
  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	
!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "qa_tool.h"
#include "str_op.h"
#include "sds_rw.h"
#include "meta.h"
#include "main_util.h"
#include "alloc_mem.h"

#define HELP \
"NAME \n" \
"	sds_math - Perform simple arithmatic on two input SDS. \n" \
" \n" \
"SYNOPSIS \n" \
"	sds_math -help \n" \
"	sds_math -math=SDSname,f1,<op>,SDSname,f2,dt,f_nop1,f_nop2,f_nop3,f_ovf -of=<filename> \n" \
" \n" \
"	echo -math=SDSname,f1,<op>,SDSname,f2,dt,f_nop1,f_nop2,f_nop3,f_ovf -of=<filename> | sds_math\n" \
" \n" \
"DESCRIPTION \n" \
"	Perform simple arithmatic on two input SDS of same or different resolution. \n" \
"If the two input SDSs are of different resolution then resolution of one SDS must be\n" \
"an integral multiple of the other and the output difference SDS will be of higher of\n" \
"the two input resolutions. Output is an HDF file containing one or more SDS of \n" \
"pixelwise arithmatic result. The -math option may be repeated to perform other arithmatic \n" \
"operations. \n" \
"	Tool supports 2D/3D/4D SDS in band sequential or bit sequential interleaved format. \n" \
"If no layers are specified all layers are processed and the original rank of the SDS is \n" \
"retained. To specify particular layer in 3D or 4D SDS include the layer number as dot \n" \
"extension in the input SDS name. \n" \
"\n" \
"OPTIONS \n" \
"\t-help                              Print this help message \n" \
"\t-math=SDSname,f1,<op>,SDSname,f2,dt,f_nop1,f_nop2,f_nop3,f_ovf \n" \
"\t                                   Arithmatic expression containing SDS name, filename, \n" \
"\t                                   an operator (+, -, *, /), output datatype and fill values. \n\n" \
"\t                                   values for output data type are \n" \
"\t                                     INT8, UINT8, INT16, UINT16, INT32, UINT32, FLOAT32 \n\n" \
"\t                                   f_nop1, f_nop2 and f_nop3 are no operation fill values in \n" \
"\t                                   the two input SDS and the output SDS. The arithmatic operation \n" \
"\t                                   is not performed if SDS values in the input SDS1 is f_nop1 or \n" \
"\t                                   SDS2 is f_nop2. The output fill value in such case is f_nop3. \n\n" \
"\t                                   f_ovf is the fill value in the output SDS if the arithmatic \n" \
"\t                                   operation results in overflow. The overflow fill option \n" \
"\t                                   is not implemented for float data type.\n\n" \
"\t                                   If SDS is a 3D or 4D specify the dimension value as dot \n" \
"\t                                   extension in the sds name. e.g. \n" \
"\t                                     -math=SDS.n1,file,+,SDS.n1.m1,file \n\n" \
"\t                                   default values are: \n" \
"\t                                     dt = data type of SDS1 \n" \
"\t                                     f_nop1, f_nop2 = corresponding input SDS fill values. \n" \
"\t                                     f_nop3 = fill value of SDS1  \n" \
"\t                                     f_ovf = f_nop3 \n" \
"\t                                   For one or more parameters to take default values use * \n" \
"\t                                   as value. \n" \
"\t-of=<filename>                     Output filename \n" \
" \n" \
"Examples: \n" \
"\tsds_math -math=LST_Day_1km,MOD11A1.A2001044.h09v05.002.2001154214604.hdf,-,LST_Night_1km,MOD11A1.A2001044.h09v05.002.2001154214604.hdf,INT16,0,0,-1,-2 -of=diff_temp.hdf\n" \
"\tsds_math -math=sur_refl_b02,MOD09A1.A1999049.h12v03.001.1999265195217.hdf,/,sur_refl_b03,MOD09A1.A1999053.h12v03.001.1999250150605.hdf,INT16,*,*,*,*, -of=MOD09A1.h12v03.b03.49-53.hdf \n" \
"\tsds_diff -math=\"Angles.1.1,MODAGAGG.A1996214.h12v04.001.hdf,-,Angles.1.2,MODAGAGG.A1996214.h12v04.002.hdf,*,*,*,*,*\" -of=MODAGG_verdiff.hdf \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:    sds_math -math=SDSname,f1<op>SDSname,f2,dt,f_nop1,f_nop2,f_nop3,f_ovf -of=<filename> \n" \
" \n" \
"OPTIONS \n" \
"	-help                              Print this help message \n" \
"\t-math=SDSname,f1,<op>,SDSname,f2,dt,f_nop1,f_nop2,f_nop3,f_ovf \n" \
"\t                                   Arithmatic expression containing SDS name, filename, \n" \
"\t                                   an operator (+, -, *, /), output datatype and fill values. \n\n" \
"\t                                   values for output data type are \n" \
"\t                                     INT8, UINT8, INT16, UINT16, INT32, UINT32, FLOAT32 \n\n" \
"\t                                   f_nop1, f_nop2 and f_nop3 are no operation fill values in \n" \
"\t                                   the two input SDS and the output SDS. The arithmatic operation \n" \
"\t                                   is not performed if SDS values in the input SDS1 is f_nop1 or \n" \
"\t                                   SDS2 is f_nop2. The output fill value in such case is f_nop3. \n\n" \
"\t                                   f_ovf is the fill value in the output SDS if the arithmatic \n" \
"\t                                   operation results in overflow. The overflow fill option \n" \
"\t                                   is not implemented for float data type.\n\n" \
"\t                                   If SDS is a 3D or 4D specify the dimension value as dot \n" \
"\t                                   extension in the sds name. e.g. \n" \
"\t                                     -math=SDS.n1,file,+,SDS.n1.m1,file \n" \
"\t                                   default values are: \n" \
"\t                                     dt = data type of SDS1 \n" \
"\t                                     f_nop1, f_nop2 = corresponding input SDS fill values. \n" \
"\t                                     f_nop3 = fill value of SDS1  \n" \
"\t                                     f_ovf = f_nop3 \n" \
"\t                                   For one or more parameters to take default values use * \n" \
"\t                                   as value. \n" \
"\t-of=<filename>                     Output filename \n" \
"\n"

int parse_cmd_sds_math(int argc, char **argv, char **expr, int *n_op, char *f3);
int read_param(char *expr, char *sds1, char *sds2, char *f1, char *f2, char *op_t, char *dt, char *f_nop1,
		char *f_nop2, char *f_nop3, char *f_ovf);
void compute_sds_math(sds_t *sds1_info, sds_t *sds2_info, sds_t *sds3_info, char op_t, char *dt, char *f_nop, char *f_ovf);
void get_sds_param(sds_t *sds_info, int *n, int *m, int *rank, int *dim_size);
int check_sds_param(int rank1, int rank2, int *dim_size1, int *dim_size2, int *sc_dim, int *bd);
void check_fsds_id(char *sds1, char *sds2, char *f1, char *f2, int *st_sds, int *st_f);

int int8_min = -128;
int int8_max = 127;
int uint8_min = 0;
int uint8_max = 255;
int int16_min = -32768;
int int16_max = 32767;
int uint16_min = 0;
int uint16_max = 65535;
int int32_min = -2147483648;
int int32_max = 2147483647;
int uint32_min = 0;
uint uint32_max = 4294967295;

int main(int argc, char **argv)
{
  int i, i_op, n_op;
  int st1, st2;
  int st_f, st_sds;
  int stdin_argc, status;
  char f_nop1[10], f_nop2[10];
  char dt[10],  f_nop3[10], f_ovf[10];
  char op_t, **expr, **stdin_argv;
  char sds1[MAX_SDS_NAME_LEN], sds2[MAX_SDS_NAME_LEN]; 
  char f1[MAX_PATH_LENGTH], f2[MAX_PATH_LENGTH], f3[MAX_PATH_LENGTH];
  sds_t sds1_info, sds2_info, sds3_info;

  if (argc == 1 || ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0))))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  if (argc == 1)
  {
    if ((stdin_argv = (char **)Calloc2D(MAX_NUM_PARAM, MAX_PARAM_LENGTH, sizeof(char))) == NULL)
    {
      fprintf(stderr, "Cannot allocate memory for stdin_argv in sds_math(main)\n");
      exit(1);
    }
    if (parse_stdin(&stdin_argc, stdin_argv) == -1)
    {
      Free2D((void *)stdin_argv);
      exit(0);
    }
  }
  if ((expr = (char **)Calloc2D(MAX_NUM_OP, MAX_STR_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for expr in main()\n"); 
  else 
  {
    if (argc == 1)
    {
      status = parse_cmd_sds_math(stdin_argc, stdin_argv, expr, &n_op, f3);
      Free2D((void **)stdin_argv);
    }
    else
      status = parse_cmd_sds_math(argc, argv, expr, &n_op, f3);
    if (status == -1)
    {
      fprintf(stderr, "%s\n", USAGE);
      exit(0);
    }
    else if (status != 0)
    {

      /* MSZ 1/31/2018 silent version mismatch warning standard print out */
      setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

      /* MSZ 1/31/2018 Disable error reporting */
      H5Eset_auto2(H5E_DEFAULT, NULL, NULL);

      sds3_info.H5IN = 0;   /* nothing to do with input file format */
      sds3_info.H5OUT = 0;  /* so far, output to HDF4 */
      sds3_info.sd_id = sds3_info.sds_id = -1;

      if(open_sd(f3, &sds3_info, 'W') == FAIL)
	fprintf(stderr, "Cannot open output HDF file %s\n", f3);
      else
      { 
        for (i_op=0; i_op<n_op; i_op++) 
        {
          if (read_param(expr[i_op], sds1, sds2, f1, f2, &op_t, dt, f_nop1, f_nop2, f_nop3, f_ovf) == -1)
            fprintf(stderr, "Cannot process operation %s\n", expr[i_op]);
          else 
	  {
	    sds1_info.sd_id = sds2_info.sd_id = -1;
	    strcpy(sds1_info.name, sds1);
	    strcpy(sds2_info.name, sds2);
	    check_fsds_id(sds1, sds2, f1, f2, &st_sds, &st_f);
	    sds1_info.sds_id = sds2_info.sds_id = -1;
	    st1 = get_sds_info(f1, &sds1_info);
	    if (st_f == 1)
	    {
	      sds2_info.sd_id = sds1_info.sd_id;
	      if (st_sds == 1)
	      {
	        sds2_info.sds_id = sds1_info.sds_id;
	        sds2_info.rank = sds1_info.rank;
	        for (i=0; i<sds1_info.rank; i++)
		  sds2_info.dim_size[i] = sds1_info.dim_size[i];
	        sds2_info.data_type = sds1_info.data_type;
	        sds2_info.data_size = sds1_info.data_size;
	      }
	      else
	        st2 = get_sds_info(f2, &sds2_info);
	    }
	    else
	      st2 = get_sds_info(f2, &sds2_info);
	    if ((st1 != -1) && (st2 != -1))
	    {
	      if (f_nop1[0] != '\0')
	        switch(sds1_info.data_type)
		{
		  case 5: sds1_info.fill_fval = (float32)atof(f_nop1);
		  case 20: sds1_info.fill_val = (int8)atoi(f_nop1);
		  case 21: sds1_info.fill_val = (uint8)atoi(f_nop1);
		  case 22: sds1_info.fill_val = (int16)atoi(f_nop1);
		  case 23: sds1_info.fill_val = (uint16)atoi(f_nop1);
		  case 24: sds1_info.fill_val = (int32)atoi(f_nop1);
		  case 25: sds1_info.fill_val = (uint32)atoi(f_nop1);
		}
	      if (f_nop2[0] != '\0')
	        switch(sds2_info.data_type)
		{
		  case 5: sds2_info.fill_fval = (float32)atof(f_nop2);
		  case 20: sds2_info.fill_val = (int8)atoi(f_nop2);
		  case 21: sds2_info.fill_val = (uint8)atoi(f_nop2);
		  case 22: sds2_info.fill_val = (int16)atoi(f_nop2);
		  case 23: sds2_info.fill_val = (uint16)atoi(f_nop2);
		  case 24: sds2_info.fill_val = (int32)atoi(f_nop2);
		  case 25: sds2_info.fill_val = (uint32)atoi(f_nop2);
		}
	      compute_sds_math(&sds1_info, &sds2_info, &sds3_info, op_t, dt, f_nop3, f_ovf);
	    }
	    if (sds1_info.sds_id != -1)
	      close_sds(&sds1_info, sds1_info.H5IN);
	    if (sds2_info.sds_id != -1)
	    {
	      if  (st_f != 1)
		close_sds(&sds2_info, sds2_info.H5IN);
	      else if (st_sds != 1)
		close_sds(&sds2_info, sds2_info.H5IN);
	    }
          }
	  close_sd(&sds1_info, sds1_info.H5IN);
	  if (st_f != 1)
	    close_sd(&sds2_info, sds2_info.H5IN);
	}
	close_sd(&sds3_info, sds3_info.H5OUT);
      }
    }
  }
  return 0;
}

int parse_cmd_sds_math(int argc, char **argv, char **expr, int *n_op, char *f3) 
{
  int i, i_op, st;

  st = 1;
  *n_op = 0;
  f3[0] = '\0';
  for (i=1, i_op=0; i<argc; i++)
  {
    if (is_arg_id(argv[i], "-math") == 0)
    {
      get_arg_val(argv[i], expr[i_op]);
      i_op++;
    }
    else if (is_arg_id(argv[i], "-of") == 0)
      get_arg_val(argv[i], f3);
    else fprintf(stderr, "Ignoring unknown option %s\n", argv[i]);
  }
  if (strlen(f3) <= 0) {
    st = -1; fprintf(stderr, "Missing output filename\n");
  }
  if (i_op == 0) {
    st = -1; fprintf(stderr, "No math operation specified\n");
  }
  if (st == 1) *n_op = i_op;
  return st;
}
	    
          
int read_param(char *expr, char *sds1, char *sds2, char *f1, char *f2, char *op_t, char *dt, char *f_nop1, 
	char *f_nop2, char *f_nop3, char *f_ovf)
{
  int st = 1;
  int p1, p2;
  int i, nc, len;
  char op[80];

  dt[0] = '\0';
  f1[0] = f2[0] = '\0';
  sds1[0] = sds2[0] = '\0';
  f_nop3[0] = f_ovf[0] = '\0';
  f_nop1[0] = f_nop2[0] = '\0';
  len = (int)strlen(expr);
  for (i=0, p1=0; i<9; i++)
  {
    p2 = charpos(expr, ',', p1);
    if (p2 != -1)
    {
      nc = p2 - p1;
      if (nc > 0) {
        switch (i) {
          case 0: strmid(expr, p1, nc, sds1); break;
          case 1: strmid(expr, p1, nc, f1); break;
          case 2: strmid(expr, p1, nc, op);
		  if (strlen(op) == 1) *op_t = op[0];
		  else st = -1;
		  break;
          case 3: strmid(expr, p1, nc, sds2); break;
          case 4: strmid(expr, p1, nc, f2); break;
          case 5: strmid(expr, p1, nc, dt); 
          case 6: strmid(expr, p1, nc, f_nop1); break;
          case 7: strmid(expr, p1, nc, f_nop2); break;
          case 8: strmid(expr, p1, nc, f_nop3); break;
	}
      }
      p1 = p2 + 1;
    }
    else { st = -1; break; }
  }
  if (st == 1) {
    nc = len - p1;
    strmid(expr, p1, nc, f_ovf);
    if (strcmp(f_ovf, "*") == 0) f_ovf[0] = '\0';
    if (strcmp(f_nop1, "*") == 0) f_nop1[0] = '\0';
    if (strcmp(f_nop2, "*") == 0) f_nop2[0] = '\0';
    if (strcmp(f_nop3, "*") == 0) f_nop3[0] = '\0';
    if (strcmp(dt, "*") == 0) dt[0] = '\0';
  }
  return st;
}

void compute_sds_math(sds_t *sds1_info, sds_t *sds2_info, sds_t *sds3_info, char op_t, char *dt, char *f_nop, char *f_ovf)
{
  int k1, k2;
  int val1, val2;
  int nrows, ncols;
  int st_c1, st_c2;
  int ic1, ic2, ic3;
  int i, ir, sc_dim;
  int offset1, offset2;
  int n1, n2, m1, m2, bd;
  int rank1, rank2, rank3;
  int dim_sz1[4], dim_sz2[4];
  int ndata1, ndata2, ndata3;
  void *data1, *data2, *data3;
  int32 edge1[4] = {0, 0, 0, 0};
  int32 edge2[4] = {0, 0, 0, 0};
  int32 edge3[4] = {0, 0, 0, 0};
  int32 start1[4] = {0, 0, 0, 0};
  int32 start2[4] = {0, 0, 0, 0};
  int32 start3[4] = {0, 0, 0, 0};
  int fval_st;
  float fval1, fval2;
  double of_ovf, of_nop;
  double dval1, dval2, dval3;

  fprintf(stdout, "Processing SDS: %s %c %s\n", sds1_info->name, op_t, sds2_info->name);
  get_sds_param(sds1_info, &n1, &m1, &rank1, dim_sz1); 
  get_sds_param(sds2_info, &n2, &m2, &rank2, dim_sz2); 
  if (check_sds_param(rank1, rank2, dim_sz1, dim_sz2, &sc_dim, &bd) == -1)
    return;
  if (sds1_info->data_type != sds2_info->data_type)
    fprintf(stderr, "Input SDSs are of different data type: Continues anyway . . \n");

  sds3_info->sds_id = -1;
  sds3_info->rank = rank3 = rank1;
  for (i=0; i<rank3; i++)
    sds3_info->dim_size[i] = (bd == 1) ? dim_sz1[i] : dim_sz2[i];
  if (dt[0] == '\0') sds3_info->data_type = sds1_info->data_type;
  else {
    if (strcmp(dt, "FLOAT32") == 0) sds3_info->data_type = 5;
    else if (strcmp(dt, "FLOAT64") == 0) sds3_info->data_type = 6;    
    else if (strcmp(dt, "INT8") == 0) sds3_info->data_type = 20;
    else if (strcmp(dt, "UINT8") == 0) sds3_info->data_type = 21;
    else if (strcmp(dt, "INT16") == 0) sds3_info->data_type = 22;
    else if (strcmp(dt, "UINT16") == 0) sds3_info->data_type = 23;
    else if (strcmp(dt, "INT32") == 0) sds3_info->data_type = 24;
    else if (strcmp(dt, "UINT32") == 0) sds3_info->data_type = 25;
    else {
      fprintf(stderr, "Output data type %s not recognized. Set to default\n", dt);
      sds3_info->data_type = sds1_info->data_type;
    }
  }
  sds3_info->data_size = DFKNTsize(sds3_info->data_type);
  sprintf(sds3_info->name, "%s%c%s", sds1_info->name, op_t, sds2_info->name);
  if (f_nop[0] == '\0') {
    if (sds1_info->data_type == 5) of_nop = (double)sds1_info->fill_fval; 
    else of_nop = (double)sds1_info->fill_val;
  }
  else of_nop = (double)atof(f_nop);
  if (f_ovf[0] == '\0') of_ovf = of_nop; 
  else of_ovf = (double)atof(f_ovf);

  if (open_sds((char *)NULL, sds3_info, 'W') != -1)
  {    
    if (sds3_info->data_type == 5 || sds3_info->data_type == 6)
      write_attr_fvalf(sds3_info->H5OUT, sds3_info->sds_id, sds3_info->data_type, 1, of_nop);
    else
      write_attr_fvalf(sds3_info->H5OUT, sds3_info->sds_id, sds3_info->data_type, 1, (int)of_nop);
    ndata1 = compute_sds_ndata(sds1_info);
    ndata2 = compute_sds_ndata(sds2_info);
    ndata3 = compute_sds_ndata(sds3_info);
    if (bd == 1) ndata1 *= sc_dim;
    else ndata2 *= sc_dim;
    ndata3 *= sc_dim;
    if ((data1 = (void *)calloc(ndata1, sds1_info->data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for data1 in sds_diff: compute_sds_diff()\n");
    if ((data2 = (void *)calloc(ndata2, sds2_info->data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for data2 in sds_diff: compute_sds_diff()\n");
    if ((data3 = (void *)calloc(ndata3, sds3_info->data_size)) == NULL)
      fprintf(stderr, "Cannot allocate memory for data3 in sds_diff: compute_sds_diff()\n");
    if ((data1 != NULL) && (data2 != NULL) && (data3 != NULL))
    {
      rank1 = sds1_info->rank;
      rank2 = sds2_info->rank;
      get_sds_edge(sds1_info, edge1);
      get_sds_edge(sds2_info, edge2);
      get_sds_edge(sds3_info, edge3);
      compute_sds_start_offset(sds1_info, n1, m1, &st_c1, &offset1);
      compute_sds_start_offset(sds2_info, n2, m2, &st_c2, &offset2);
      compute_sds_nrows_ncols(sds3_info, &nrows, &ncols);
      for (ir=0; ir<nrows; ir++)
      {
        if ((rank1 == 2) || (sds1_info->dim_size[0] > sds1_info->dim_size[rank1-1]))
          start1[0] = (bd == 1) ? ir : ir/sc_dim;
        else 
	  start1[rank1-2] = (bd == 1) ? ir : ir/sc_dim;
        if ((rank2 == 2) || (sds2_info->dim_size[0] > sds2_info->dim_size[rank2-1]))
          start2[0] = (bd == 2) ? ir : ir/sc_dim;
        else 
	  start2[rank2-2] = (bd == 2) ? ir : ir/sc_dim;
        if ((rank3 == 2) || (sds3_info->dim_size[0] > sds3_info->dim_size[rank3-1]))
          start3[0] = ir;
        else 
	  start3[rank3-2] = ir;
	if(read_sel_sds_data(sds1_info, start1, edge1, data1) == FAIL)
	{
	  fprintf(stderr, "Cannot read dataline from SDS %s in compute_sds_diff()\n", sds1_info->name);
	  break;
	}
	if(read_sel_sds_data(sds2_info, start2, edge2, data2) == FAIL)
	{
	  fprintf(stderr, "Cannot read dataline from SDS %s in compute_sds_diff()\n", sds2_info->name);
	  break;
	}
        ic1 = st_c1; ic2 = st_c2;
	k1 = k2 = 0;
        for (ic3=0; ic3<ncols; ic3++)
        {
          switch(sds1_info->data_type) 
	  {
            case 5 : fval1 = ((float32 *)data1)[ic1]; break;
            case 20: val1 = ((int8 *)data1)[ic1]; break;
            case 21: val1 = ((uint8 *)data1)[ic1]; break;
            case 22: val1 = ((int16 *)data1)[ic1]; break;
            case 23: val1 = ((uint16 *)data1)[ic1]; break;
            case 24: val1 = ((int32 *)data1)[ic1]; break;
            case 25: val1 = ((uint32 *)data1)[ic1]; break;
	  }
          switch(sds2_info->data_type) 
	  {
            case 5 : fval2 = ((float32 *)data2)[ic2]; break;
            case 20: val2 = ((int8 *)data2)[ic2]; break;
            case 21: val2 = ((uint8 *)data2)[ic2]; break;
            case 22: val2 = ((int16 *)data2)[ic2]; break;
            case 23: val2 = ((uint16 *)data2)[ic2]; break;
            case 24: val2 = ((int32 *)data2)[ic2]; break;
            case 25: val2 = ((uint32 *)data2)[ic2]; break;
	  }
	  fval_st = 0;
	  if (sds1_info->data_type == 5) {
	    if (fval1 == sds1_info->fill_fval) fval_st = 1;
	  }
	  else if (val1 == sds1_info->fill_val) fval_st = 1;
	  if (sds2_info->data_type == 5) {
	    if (fval2 == sds2_info->fill_fval) fval_st = 1;
	  }
	  else if (val2 == sds2_info->fill_val) fval_st = 1;

	  if (fval_st == 0)
	  {
	    if (sds1_info->data_type == 5) dval1 = (double)fval1;
	    else dval1 = (double)val1;
	    if (sds2_info->data_type == 5) dval2 = (double)fval2;
	    else dval2 = (double)val2;

	    if (op_t == '-') dval3 = dval1 - dval2;
	    else if (op_t == '+') dval3 = dval1 + dval2;
	    else if (op_t == '/') dval3 = (float)dval1 / (float)dval2 * 10000;
	    else if (op_t == '*') dval3 = dval1 * dval2;

            switch(sds3_info->data_type) 
	    {
              case 5 : ((float32 *)data3)[ic3] = (float32)dval3;
		       break;
              case 20: if ((dval3 < int8_min) || (dval3 > int8_max))
			 ((int8 *)data3)[ic3] = (int8)of_ovf;
		       else ((int8 *)data3)[ic3] = (int8)dval3;
		       break;
              case 21: if ((dval3 < uint8_min) || (dval3 > uint8_max))
			 ((uint8 *)data3)[ic3] = (uint8)of_ovf;
		       else ((uint8 *)data3)[ic3] = (uint8)dval3;
		       break;
              case 22: if ((dval3 < int16_min) || (dval3 > int16_max))
			 ((int16 *)data3)[ic3] = (int16)of_ovf;
		       else ((int16 *)data3)[ic3] = (int16)dval3;
		       break;
              case 23: if ((dval3 < uint16_min) || (dval3 > uint16_max))
			 ((uint16 *)data3)[ic3] = (uint16)of_ovf;
		       else ((uint16 *)data3)[ic3] = (uint16)dval3;
		       break;
              case 24: if ((dval3 < int32_min) || (dval3 > int32_max))
			 ((int32 *)data3)[ic3] = (int32)of_ovf;
		       else ((int32 *)data3)[ic3] = (int32)dval3;
		       break;
              case 25: if ((dval3 < uint32_min) || (dval3 > uint32_max))
			 ((uint32 *)data3)[ic3] = (uint32)of_ovf;
		       else ((uint32 *)data3)[ic3] = (uint32)dval3;
		       break;
	    }
	  }
	  else
	  {
            switch(sds3_info->data_type) 
	    {
              case 5 : ((float32 *)data3)[ic3] = (float32)of_nop; break;
              case 20: ((int8 *)data3)[ic3] = (int8)of_nop; break;
              case 21: ((uint8 *)data3)[ic3] = (uint8)of_nop; break;
              case 22: ((int16 *)data3)[ic3] = (int16)of_nop; break;
              case 23: ((uint16 *)data3)[ic3] = (uint16)of_nop; break;
              case 24: ((int32 *)data3)[ic3] = (int32)of_nop; break;
              case 25: ((uint32 *)data3)[ic3] = (uint32)of_nop; break;
	    }
	  }
          if (bd == 0) {
            ic1 += offset1; ic2 += offset2;
          }
          else if (bd == 1) 
	  {
	    ic1 += offset1;
	    k2++;
	    if (k2 == sc_dim) {
	      k2 = 0; ic2 += offset2;
	    }
	  }
          else 
	  {
	    ic2 += offset2;
	    k1++;
	    if (k1 == sc_dim) {
	      k1 = 0; ic1 += offset1;
	    }
	  }
        }
	if(set_sel_sds_data(sds3_info, start3, edge3, data3) == FAIL)
	{
	  fprintf(stderr, "Cannot write dataline for SDS %s in compute_sds_diff()\n", sds3_info->name);
	  break;
	}
      } /* for (ir=0; . . .) */
      close_sds(sds3_info, sds3_info->H5OUT);
    }
    if (data1 != NULL) free(data1);
    if (data2 != NULL) free(data2);
    if (data3 != NULL) free(data3);
  }
}

int check_sds_param(int rank1, int rank2, int *dim_size1, int *dim_size2, int *sc_dim, int *bd)
{
  int i, st;
  int sc_sz[2];
  int sz1[2], sz2[2];

  st = 1;
  if (rank1 != rank2)
  {
    fprintf(stderr, "Input SDS are of different rank: %d \t %d\n", rank1, rank2);
    st = -1;
  }
  else
  {
    if (rank1 == 2)
    {
      sz1[0] = dim_size1[0]; sz1[1] = dim_size1[1]; 
      sz2[0] = dim_size2[0]; sz2[1] = dim_size2[1]; 
    }
    else
    {
      if (dim_size1[0] < dim_size1[rank1-1])
      {
        sz1[0] = dim_size1[rank1-2]; sz1[1] = dim_size1[rank1-1]; 
        sz2[0] = dim_size2[rank1-2]; sz2[1] = dim_size2[rank1-1]; 
      } 
      else
      {
        sz1[0] = dim_size1[0]; sz1[1] = dim_size1[1]; 
        sz2[0] = dim_size2[0]; sz2[1] = dim_size2[1]; 
      } 
    }
    if (sz1[0] == sz2[0]) *bd = 0;
    else if (sz1[0] > sz2[0]) *bd = 1;
    else *bd = 2;
    for (i=0; i<2; i++)
    {
      sc_sz[i] = (*bd == 1) ? sz1[i]%sz2[i] : sz2[i]%sz1[i];
      if (sc_sz[i] != 0) 
      {
	fprintf(stderr, "Input SDSs dimensions are not integral multiples\n");
	st = -1;
	break;
      }
    }
    for (i=0; i<2; i++)
      sc_sz[i] = (*bd == 1) ? sz1[i]/sz2[i] : sz2[i]/sz1[i];
    if (sc_sz[0] != sc_sz[1])
    {
      fprintf(stderr, "All dimensions of input SDSs are not of same mutliples\n");
      st  = -1;
    }
  }
  *sc_dim = sc_sz[0];
  return st;
}
    
void check_fsds_id(char *sds1, char *sds2, char *f1, char *f2, int *st_sds, int *st_f)
{
  int p;
  char org_sds1[MAX_SDS_NAME_LEN];
  char org_sds2[MAX_SDS_NAME_LEN];

  *st_f = (strcmp(f1, f2) == 0) ? 1 : 0; 
  p = charpos(sds1, '.', 0);
  if (p == -1) strcpy(org_sds1, sds1);
  else strmid(sds1, 0, p, org_sds1); 
  p = charpos(sds2, '.', 0);
  if (p == -1) strcpy(org_sds2, sds2);
  else strmid(sds2, 0, p, org_sds2); 
  *st_sds = (strcmp(org_sds1, org_sds2) == 0) ? 1 : 0;
}


