/****************************************************************************
!C

!File: sds_ts_stat.c

!Description:
  Compute pixelwise statistics of N input tiles 

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original July 2001 QA Tool Version 1.4
  Updateded March 2013.  by Xiaoping Zhang Sigma Space Corp.
	fixed the problem caused by no valid_range attribute for NPP and MODIS products. 

  Modified in February, 2018, Version 3.5.1 by Maosheng Zhao, SSAI.
        1)modify it so that HDF5 can be read.
        2)update all HDF4 API SDxxx() in the code to wrapped functions 
          to deal with HDF4/HDF5 in sds_rw.c
	3)for comp_stat(), std calculation has been controled so that the input
	  to sqrt() should be positive. Negative due to float issue and 0 will
	  just have std of 0..
	4) add caution information for the output SDS sum because sum can be
	 overflew because it use the SDS dataset, which may not be able hold
	 sum exceeds the data type limits, especailly for one- or two-byte integers.

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:

    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "hdf5.h"
#include "mfhdf.h"
#include "qa_tool.h"
#include "str_op.h"
#include "sds_rw.h"
#include "meta.h"
#include "main_util.h"
#include "alloc_mem.h"

#define HELP \
"NAME \n" \
"	sds_ts_stat - Compute pixelwise statistics of SDS in set of tiles. \n" \
" \n" \
"SYNOPSIS \n" \
"	sds_ts_stat -help \n" \
"	sds_ts_stat -sds=<sds_name,sds_min,sds_max,f_nop_in,f_nop_out,dt> -of=<filename> \n" \
"                   -param=avg,std,min,max,npix,sum f1 f2 f3. . fn\n" \
" \n" \
"DESCRIPTION \n" \
"	Compute statistics of sds values at each pixel for a set of input tiles. User can specify \n" \
"range of the sds values to be considered for computing statistics and a fill value to be excluded \n" \
"from consideration. Multiple SDSs are specified by repeating the option. Output is an HDF file \n" \
"containing one or more statistics parameter SDS. Statistics parameter include mean, standard deviation. \n" \
"minimum, maximum, number of pixels and sum of sds values at each pixel for each input SDS. By default\n" \
"all parameters are output. An input file is ignored if the user specified SDS is missing in the file. \n" \
"Be cautious that for the Sum of SDS, the value may overflow and unrelibale for one- or two-byte data types.\n" \
"In this case, the true sum will be the Mean of SDS multiply number of input files\n" \
"\n" \
"OPTIONS \n" \
"\t-help              Print this help message \n" \
"\t-sds=SDSname,sds_min,sds_max,f_nop_in,f_nop_out,dt\n" \
"\t                   Input sds name, minmum and maximum of sds values to be considered for \n" \
"\t                   statistics, no operation fill value in input SDS and no operation fill value \n" \
"\t                   in output SDS and output data type. Output data type of SDS number of pixels \n" \
"\t                   is always INT16. Each of the parameters except the SDS name can be replaced \n" \
"\t                   by * to assign respective default values. Default values are: \n" \
"\t                     sds_min,sds_max: sds attribute valid range. \n" \
"\t                     f_nop_in: sds attribute fill value \n" \
"\t                     f_nop_out: f_nop_in \n" \
"\t                     dt: input sds data type \n" \
"\t                   Values for data type: INT8, UINT8, INT16, UINT16, INT32, UINT32, FLOAT32 \n" \
"\t-param=avg,std,min,max,npix,sum \n" \
"\t                   Output statistics parameter. One or more parameter is specified in random \n" \
"\t                   order. Default: all parameters are computed. \n" \
"\t-of=<filename>     Output filename \n" \
" \n" \
"Examples: \n" \
"     sds_ts_stat -sds=sur_refl_b01,0,10000,-28672,-28672,INT16 \n" \
"                 -param=avg,std,min,max,npix,sum \n" \
"                 -of=MOD09A1_ts_stat.hdf \n"  \
"                 MOD09A1.A2016209.h09v05.006.2016222123023.hdf MOD09A1.A2016217.h09v05.006.2016226075317.hdf \n" \
" \n" \
"     sds_ts_stat -sds=SurfReflect_M5,0,10000,-28672,-28672,INT16 \n" \
"                 -param=avg,std,min,max,npix,sum \n" \
"                 -of=VNP09A1_ts_stat.hdf       \n" \
"                 VNP09A1.A2016289.h11v08.001.2017189045223.h5 VNP09A1.A2016297.h11v08.001.2017189113645.h5 \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage:    sds_ts_stat -sds=<sds_name,sds_min,sds_max,f_nop_in,f_nop_out,dt> -of=<filename> \n" \
"                      -param=avg,std,min,max,npix,sum f1 f2 f3. . fn \n" \
" \n" \
"OPTIONS \n" \
"\t-help              Print this help message \n" \
"\t-sds=SDSname,sds_min,sds_max,f_nop_in,f_nop_out,dt\n" \
"\t                   Input sds name, minmum and maximum of sds values to be considered for \n" \
"\t                   statistics, no operation fill value in input SDS and no operation fill value \n" \
"\t                   in output SDS and output data type. Output data type of SDS number of pixels \n" \
"\t                   is always UINT16. Each of the parameters except the SDS name can be replaced \n" \
"\t                   by * to assign respective default values. Default values are: \n" \
"\t                     sds_min,sds_max: sds attribute valid range. \n" \
"\t                     f_nop_in: sds attribute fill value \n" \
"\t                     f_nop_out: f_nop_in \n" \
"\t                     dt: input sds data type \n" \
"\t                   Values for data type: INT8, UINT8, INT16, UINT16, INT32, UINT32, FLOAT32 \n" \
"\t-param=avg,std,min,max,npix,sum \n" \
"\t                   Output statistics parameter. One or more parameters is specified in random \n" \
"\t                   order. Default: all parameters are computed. \n" \
"\t-of=<filename>     Output filename \n" \
"\n"

#define MAX_NSDS 10

int parse_cmd_sds_ts_stat(int argc, char **argv, char **expr, int *nsds, int *fcnt, char *out_fname, int *param_st);
int read_param(char *expr, char *sds_name, char *range1, char *range2, char *f_nop_in, char *f_nop_out, char *dt);
void comp_stat(sds_t *in_sds_info, int fcnt, int out_sd_id, char *range1, char *range2, char *f_nop_in, 
	char *f_nop_out, char *dt, int *param_st, int H5OUT);
int getSdsnameDim(char *sdsname_str, sds_t* sds_info, int *n, int *m);

int main(int argc, char **argv)
{
  char **expr = NULL;
  char range1[10], range2[10];
  char sds_name[MAX_SDS_NAME_LEN];
  char out_fname[MAX_PATH_LENGTH];
  char f_nop_in[10], f_nop_out[10], dt[10];
  int param_st[6];
  int fcnt, id, fid, isds, nsds, iarg, status;
  sds_t out_sds_info;
  sds_t *in_sds_info = NULL;
  int H5OUT = 0; /* so far, output to HDF4 */

  if ((argc==2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }

  /* MSZ 4/6/2018 silent version mismatch warning standard print out */
  setenv("HDF5_DISABLE_VERSION_CHECK", "2", 1);

  /* MSZ 4/6/2018 Disable error reporting */
  H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  
  if ((expr = (char **)Calloc2D(MAX_NUM_SDS, MAX_STR_LEN, sizeof(char))) == NULL)
    fprintf(stderr, "Cannot allocate memory for expr in main()\n");
  else
  {
    if ((status = parse_cmd_sds_ts_stat(argc, argv, expr, &nsds, &fcnt, out_fname, param_st)) == -1)
      fprintf(stderr, "%s\n", USAGE);
    else if (status != 0)
    {
      if ((in_sds_info = (sds_t *)calloc(fcnt, sizeof(sds_t))) == NULL)
        fprintf(stderr, "Cannot allocate memory for in_sds_info in sds_ts_stat\n");
      else
      {
	out_sds_info.H5OUT = H5OUT;
	out_sds_info.sd_id = out_sds_info.sds_id = -1;

	if(open_sd(out_fname, &out_sds_info, 'W') == FAIL)
	  fprintf(stderr, "Cannot create output HDF file: %s\n", out_fname);
        else
        {
          for (isds=0; isds<nsds; isds++)
          {
            if (read_param(expr[isds], sds_name, range1, range2, f_nop_in, f_nop_out, dt) == -1)
              fprintf(stderr, "Invalid argument %s. Argument is not processed\n", expr[isds]);
            else
            {
	      fprintf(stdout, "Processing SDS %s\n", sds_name);
	      for (iarg=1, fid=0; iarg<argc; iarg++)
	      {
	        if (argv[iarg][0] != '-') 
	        {
	          in_sds_info[fid].sd_id = in_sds_info[fid].sds_id = -1;
	          strcpy(in_sds_info[fid].name, sds_name);
	          if (get_sds_info(argv[iarg], &in_sds_info[fid]) == -1)
	            fprintf(stderr, "\tIgnoring input file %s\n", argv[iarg]);
	          else
		    fid++;	
	        }
              } /* for (iarg=1, . . ) */
	      if (fid > 0)
	        comp_stat(in_sds_info, fid, out_sds_info.sd_id, range1, range2, f_nop_in, f_nop_out, dt, param_st, H5OUT); 
            }
	    for (id=0; id<fid; id++)
	    {
	      close_sds(&in_sds_info[id], in_sds_info[id].H5IN);
	      close_sd(&in_sds_info[id], in_sds_info[id].H5IN);
	    }
          } /* for (isds=0; . . .) */
	  close_sd(&out_sds_info, out_sds_info.H5OUT);
        }
        if(in_sds_info) free(in_sds_info);
      }
    }
    if(expr) Free2D((void **)expr);
    printf("Done!\n");
  }
  return 0;
}
  
int parse_cmd_sds_ts_stat(int argc, char **argv, char **expr, int *nsds, int *fcnt, char *out_fname, int *param_st)
{
  int i, p1, p2, len, isds, st;
  char tmp_str[10], param_str[80];

  *fcnt = *nsds = 0;
  out_fname[0] = '\0';
  param_str[0] = '\0';
  for (i=1, st=1, isds=0; i<argc; i++)
  {
    if (is_arg_id(argv[i], "-sds=") == 0)
    {
      get_arg_val(argv[i], expr[isds]);
      isds++;
    }
    else if (is_arg_id(argv[i], "-of=") == 0)
      get_arg_val(argv[i], out_fname);
    else if (is_arg_id(argv[i], "-param=") == 0)
      get_arg_val(argv[i], param_str);
    else if (argv[i][0] != '-') ++*fcnt;
    else fprintf(stderr, "Ignoring unknown option %s\n", argv[i]);
  }
  if (strlen(out_fname) <= 0) {
    st = -1; fprintf(stderr, "Missing output filename\n");
  }
  else 
  {
    if (isds == 0) {
      fprintf(stdout, "No SDS name input. \n");
      st = 0;
    }
    else 
    {
      if (param_str[0] == '\0')
      {
	fprintf(stdout, "No output parameters specified. All parameters output.\n");
	for (i=0; i<6; i++)
	  param_st[i] = 1;
      }
      else 
      {
	for (i=0; i<6; i++)
	  param_st[i] = 0;
	len = (int)strlen(param_str);
	for (i=0, p1=p2=0; i<6; i++)
	{
	  p2 = charpos(param_str, ',', p1);
	  if (p2 == -1) strmid(param_str, p1, len-p1, tmp_str);
	  else strmid(param_str, p1, p2-p1, tmp_str);
	  if (strcmp(tmp_str, "sum") == 0) param_st[0] = 1;
	  else if (strcmp(tmp_str, "avg") == 0) param_st[1] = 1;
	  else if (strcmp(tmp_str, "std") == 0) param_st[2] = 1;
	  else if (strcmp(tmp_str, "npix") == 0) param_st[3] = 1;
	  else if (strcmp(tmp_str, "min") == 0) param_st[4] = 1;
	  else if (strcmp(tmp_str, "max") == 0) param_st[5] = 1;
	  else fprintf(stderr, "Cannot compute parameter %s\n", tmp_str);
	  if (p1 == -1) break;
	  else p1 = p2 + 1;
	}
      }
    }
    *nsds = isds;
  }
  return st;                                          
}
          
int read_param(char *expr, char *sds_name, char *range1, char *range2, char *f_nop_in, char *f_nop_out, char *dt)
{
  int st = 1;
  int p1, p2;
  int i, nc, len;

  dt[0] = '\0';
  sds_name[0] = '\0';
  f_nop_in[0] = f_nop_out[0] = '\0';
  len = (int)strlen(expr);
  for (i=0, p1=0; i<5; i++)
  {
    p2 = charpos(expr, ',', p1);
    if (p2 != -1)
    {
      nc = p2 - p1;
      if (nc > 0) {
        switch (i) {
          case 0: strmid(expr, p1, nc, sds_name); break;
          case 1: strmid(expr, p1, nc, range1); break;
          case 2: strmid(expr, p1, nc, range2); break;
          case 3: strmid(expr, p1, nc, f_nop_in); break;
          case 4: strmid(expr, p1, nc, f_nop_out); break;
        }
      }
      p1 = p2 + 1;
    }
    else { st = -1; break; }
  }
  if (st == 1) {
    nc = len - p1;
    strmid(expr, p1, nc, dt);
  }
  return st;                                                              
}

void comp_stat(sds_t *in_sds_info, int fcnt, int out_sd_id, char *range1, char *range2, char *f_nop_in,
	char *f_nop_out, char *dt, int *param_st, int H5OUT)
{
  int32 out_dt;
  int16 *sds_npix;
  float32 *sds_std;
  int fid, isds, nsds;
  int ic, st_c, offset;
  int in_rank, out_rank, rank1;
  sds_t out_sds_info[6];
  int ndata_in, range[2];
  int i, n, m, dim_sz[4];
  int nop_in, nop_out, npix;
  int irow, icol, nrows, ncols;
  int32 in_start[4], in_edge[4];
  int32 out_start[4], out_edge[4];
  double sum, sum2, avg, avg2;
  double min, max, std, sds_val;
  void *sds_sum, *sds_mean;
  void **sds_data, *sds_min, *sds_max;
  char out_sds_name[MAX_SDS_NAME_LEN];
  char *sds_str[] = {"Sum", "Mean", "Std", "Npix", "Min", "Max"};
  double rangef[2];
  double nopf_in, nopf_out;

  if (strcmp(dt, "*") == 0)
    out_dt = in_sds_info[0].data_type;
  else {
    if (strcmp(dt, "FLOAT32") == 0) out_dt = 5;
    else if (strcmp(dt, "INT8") == 0) out_dt = 20;
    else if (strcmp(dt, "UINT8") == 0) out_dt = 21;
    else if (strcmp(dt, "INT16") == 0) out_dt = 22;
    else if (strcmp(dt, "UINT16") == 0) out_dt = 23;
    else if (strcmp(dt, "INT32") == 0) out_dt = 24;
    else if (strcmp(dt, "UINT32") == 0) out_dt = 25;
    else {
      fprintf(stderr, "Output data type %s not recognized. Set to default\n", dt);
      out_dt = in_sds_info[0].data_type;
    }                                                                     
  }

  if (strcmp(range1, "*") == 0)
  {
    range[0] = in_sds_info[0].range[0];
  }
  else 
  {
    range[0] = (int)atoi(range1);  
  }
  if (strcmp(range2, "*") == 0)
  {
    range[1] = in_sds_info[0].range[1];
  }
  else 
  {
    range[1] = (int)atoi(range2);  
  }  
  
  if (strcmp(f_nop_in, "*") == 0)
  {
    if (in_sds_info[0].data_type == 5 || in_sds_info[0].data_type == 6)     
      nopf_in = in_sds_info[0].fill_fval;
    else
      nop_in = in_sds_info[0].fill_val;    
  }
  else 
  {
    if (in_sds_info[0].data_type == 5 || in_sds_info[0].data_type == 6)     
      nopf_in = atof(f_nop_in);
    else
      nop_in = (int)atoi(f_nop_in);        
  }  
  
  if (strcmp(f_nop_out, "*") == 0)
  {
    if (in_sds_info[0].data_type == 5 || in_sds_info[0].data_type == 6)     
      nopf_out = in_sds_info[0].fill_fval;
    else
      nop_out = in_sds_info[0].fill_val;    
  }
  else 
  {
    if (in_sds_info[0].data_type == 5 || in_sds_info[0].data_type == 6)     
      nopf_out = atof(f_nop_out);
    else
      nop_out = (int)atoi(f_nop_out);        
  } 

/*  get_sds_param(&in_sds_info[0], &n, &m, &out_rank, dim_sz);*/
  getSdsnameDim(in_sds_info[0].name, &in_sds_info[0], &n, &m);
  if ((n == -1) && (m == -1))
  {
    out_rank = in_sds_info[0].rank;
    for (i=0; i < out_rank; i++)
      dim_sz[i] = in_sds_info[0].dim_size[i];
  }
  else
  {
    out_rank  = 2;
    rank1 = in_sds_info[0].rank;
    if (in_sds_info[0].dim_size[0] < in_sds_info[0].dim_size[rank1-1])
    {
      dim_sz[0] = in_sds_info[0].dim_size[rank1-2];
      dim_sz[1] = in_sds_info[0].dim_size[rank1-1];
    }
    else
    {
      dim_sz[0] = in_sds_info[0].dim_size[0];
      dim_sz[1] = in_sds_info[0].dim_size[1];
    }
  }  

  nsds = 6;
  for (isds=0; isds<nsds; isds++)
    if (param_st[isds] == 1)
    {
      out_sds_info[isds].H5OUT = H5OUT;
      out_sds_info[isds].sds_id = -1;
      out_sds_info[isds].sd_id = out_sd_id;
      if (isds == 2) out_sds_info[isds].data_type = DFNT_FLOAT32;
      else if (isds == 3) out_sds_info[isds].data_type = DFNT_INT16;
      else out_sds_info[isds].data_type = out_dt;
      out_sds_info[isds].data_size = DFKNTsize(out_dt);;
      out_sds_info[isds].rank=out_rank;
      for (i=0; i<out_rank; i++)
        out_sds_info[isds].dim_size[i] = dim_sz[i];    
      sprintf(out_sds_name, "%s of %s", sds_str[isds], in_sds_info[0].name);
      strcpy(out_sds_info[isds].name, out_sds_name);
      if (open_sds((char *)NULL, &out_sds_info[isds], 'W') != -1)
      {
        if (out_sds_info[isds].data_type == 5 || out_sds_info[isds].data_type == 6)
	  write_attr_fvalf(out_sds_info[isds].H5OUT, out_sds_info[isds].sds_id, out_sds_info[isds].data_type, 1, nopf_out);
	else
	  write_attr_fval(out_sds_info[isds].H5OUT, out_sds_info[isds].sds_id, out_sds_info[isds].data_type, 1, nop_out);	
      }
    }

  ndata_in = compute_sds_ndata(&in_sds_info[0]);
  for (isds=0; isds<nsds; isds++)
    if (param_st[isds] == 1) {
      compute_sds_nrows_ncols(&out_sds_info[isds], &nrows, &ncols);
      break;
    }
  if ((sds_data = (void **)Calloc2D(fcnt, ndata_in, in_sds_info[0].data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_data \n");
  if ((sds_sum = (void *)calloc(ncols, in_sds_info[0].data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_sum \n");
  if ((sds_mean = (void *)calloc(ncols, in_sds_info[0].data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_mean \n");
  if ((sds_std = (float32 *)calloc(ncols, sizeof(float32))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_std \n");
  if ((sds_npix = (int16 *)calloc(ncols, sizeof(int16))) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_npix \n");
  if ((sds_min = (void *)calloc(ncols, in_sds_info[0].data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_min \n");
  if ((sds_max = (void *)calloc(ncols, in_sds_info[0].data_size)) == NULL)
    fprintf(stderr, "Cannot allocate memory for sds_max \n");

  get_sds_edge(&in_sds_info[0], in_edge); 
  get_sds_edge(&out_sds_info[isds], out_edge); 
  compute_sds_start_offset(&in_sds_info[0], n, m, &st_c, &offset);

  in_rank = in_sds_info[0].rank;
  for (i=0; i<in_rank; i++) 
    in_start[i] = 0;
  for (i=0; i<out_rank; i++) 
    out_start[i] = 0;
  for (irow=0; irow<nrows; irow++)
  {
    if ((in_rank == 2) || (in_sds_info[0].dim_size[0] > in_sds_info[0].dim_size[in_rank-1]))
      in_start[0] = irow;
    else in_start[in_rank-2] = irow;
    if ((out_rank == 2) || (out_sds_info[isds].dim_size[0] > out_sds_info[isds].dim_size[out_rank-1]))
      out_start[0] = irow;                      
    else out_start[out_rank-2] = irow;
    for (fid=0; fid<fcnt; fid++)
    {
      if (read_sel_sds_data(&in_sds_info[fid], in_start, in_edge, sds_data[fid]) == FAIL)
	fprintf(stderr, "Error reading data line %d from SDS %s\n", irow, in_sds_info[fid].name);
    }

    ic = st_c;
    for (icol=0; icol<ncols; icol++)
    {
      sum = sum2 = 0.0;
      npix = 0;
      for (fid=0; fid<fcnt; fid++)
      {
        switch(in_sds_info[0].data_type)
        {
	  case 5 : sds_val = ((float32 **)sds_data)[fid][ic]; break;
	  case 20: sds_val = ((int8 **)sds_data)[fid][ic]; break;
	  case 21: sds_val = ((uint8 **)sds_data)[fid][ic]; break;
	  case 22: sds_val = ((int16 **)sds_data)[fid][ic]; break;
	  case 23: sds_val = ((uint16 **)sds_data)[fid][ic]; break;
	  case 24: sds_val = ((int32 **)sds_data)[fid][ic]; break;
	  case 25: sds_val = ((uint32 **)sds_data)[fid][ic]; break;
        }

        if (in_sds_info[0].data_type == 5 || in_sds_info[0].data_type == 6)
	{	
	  if ((sds_val != nopf_in) && (sds_val >= range[0]) && (sds_val <= range[1]))
          {
	    sum += sds_val;
            sum2 = sum2 + sds_val*sds_val;
	    if (npix == 0) 
	      min = max = sds_val;
	    else {
	      if (sds_val < min) min = sds_val;
	      if (sds_val > max) max = sds_val;
	    }
	    npix++;
          }			
	}
	else
	{
	  if ((sds_val != nop_in) && (sds_val >= range[0]) && (sds_val <= range[1]))
          {
	    sum += sds_val;
            sum2 = sum2 + sds_val*sds_val;
	    if (npix == 0) 
	      min = max = sds_val;
	    else {
	      if (sds_val < min) min = sds_val;
	      if (sds_val > max) max = sds_val;
	    }
	    npix++;
          }	  
	}		
      } /* for (fid=0; . . ) */
      if (npix == 0)
        sum = avg = std = min = max = nop_out;
      else
      {
	avg = sum/(double)npix;
        avg2 = sum2/(double)npix;

	/* MSZ 4/6/2018. add the condition so that sqrt just deals with postive values */
	if((avg2 - avg*avg) <= 0.) std = 0.;
	else std = sqrt(avg2 - avg*avg);
      }
      sds_npix[icol] = (int16)npix;
      sds_std[icol] = (float32)std;
      switch(in_sds_info[0].data_type)
      {
	case 5 : ((float32 *)sds_mean)[icol] = (float32)avg;
		 ((float32 *)sds_sum)[icol] = (float32)sum;
		 ((float32 *)sds_min)[icol] = (float32)min;
	         ((float32 *)sds_max)[icol] = (float32)max;
		 break;
	case 20: ((int8 *)sds_mean)[icol] = (int8)avg;
		 ((int8 *)sds_sum)[icol] = (int8)sum;
		 ((int8 *)sds_min)[icol] = (int8)min;
	         ((int8 *)sds_max)[icol] = (int8)max;
		 break;
	case 21: ((uint8 *)sds_mean)[icol] = (uint8)avg;
		 ((uint8 *)sds_sum)[icol] = (uint8)sum;
		 ((uint8 *)sds_min)[icol] = (uint8)min;
	         ((uint8 *)sds_max)[icol] = (uint8)max;
		 break;
	case 22: ((int16 *)sds_mean)[icol] = (int16)avg; 
		 ((int16 *)sds_sum)[icol] = (int16)sum;
	         ((int16 *)sds_min)[icol] = (int16)min;
	         ((int16 *)sds_max)[icol] = (int16)max;
		 break;
	case 23: ((uint16 *)sds_mean)[icol] = (uint16)avg; 
		 ((uint16 *)sds_sum)[icol] = (uint16)sum;
	         ((uint16 *)sds_min)[icol] = (uint16)min;
	         ((uint16 *)sds_max)[icol] = (uint16)max;
		 break;
	case 24: ((int32 *)sds_mean)[icol] = (int32)avg; 
	         ((int32 *)sds_sum)[icol] = (int32)sum;
	         ((int32 *)sds_min)[icol] = (int32)min;
	         ((int32 *)sds_max)[icol] = (int32)max;
		 break;
	case 25: ((uint32 *)sds_mean)[icol] = (uint32)avg; 
	         ((uint32 *)sds_sum)[icol] = (uint32)sum;
	         ((uint32 *)sds_min)[icol] = (uint32)min;
	         ((uint32 *)sds_max)[icol] = (uint32)max;
		 break;
      }    
      ic += offset;
    }
    if (param_st[0] == 1)
    {
      if(set_sel_sds_data(&out_sds_info[0], out_start, out_edge, sds_sum) == FAIL)
        fprintf(stderr, "Error writing data line %d to SDS %s\n", irow, out_sds_info[0].name); 
    }
    if (param_st[1] == 1)
    {
      if(set_sel_sds_data(&out_sds_info[1], out_start, out_edge, sds_mean) == FAIL)
        fprintf(stderr, "Error writing data line %d to SDS %s\n", irow, out_sds_info[1].name); 
    }
    if (param_st[2] == 1)
    {
      if(set_sel_sds_data(&out_sds_info[2], out_start, out_edge, sds_std) == FAIL)
        fprintf(stderr, "Error writing data line %d to SDS %s\n", irow, out_sds_info[2].name); 
    }
    if (param_st[3] == 1)
    {
      if(set_sel_sds_data(&out_sds_info[3], out_start, out_edge, sds_npix) == FAIL)
        fprintf(stderr, "Error writing data line %d to SDS %s\n", irow, out_sds_info[3].name); 
    }
    if (param_st[4] == 1)
    {
      if(set_sel_sds_data(&out_sds_info[4], out_start, out_edge, sds_min) == FAIL)
        fprintf(stderr, "Error writing data line %d to SDS %s\n", irow, out_sds_info[4].name); 
    }
    if (param_st[5] == 1)
    {
      if(set_sel_sds_data(&out_sds_info[5], out_start, out_edge, sds_max) == FAIL)
        fprintf(stderr, "Error writing data line %d to SDS %s\n", irow, out_sds_info[5].name); 
    }
  }

  for (isds=0; isds<nsds; isds++)
    if (param_st[isds] == 1)
      close_sds(&out_sds_info[isds], out_sds_info[isds].H5OUT);
  if(sds_data) Free2D((void **)sds_data);
  if(sds_mean) free(sds_mean);
  if(sds_std) free(sds_std);
  if(sds_npix) free(sds_npix);
  if(sds_min) free(sds_min);
  if(sds_max) free(sds_max);
}

int getSdsnameDim(char *sdsname_str, sds_t* sds_info, int *n, int *m) 
{  
  int st = 1;
  int p1, p2, p11, len;
  char nd_ext[20], md_ext[20];
  char sds_name[MAX_SDS_NAME_LEN];

  *n = *m = -1;  
  nd_ext[0] = md_ext[0] = '\0';
  if (charpos(sdsname_str, '(', 0) == -1)
  {      
    if ((p1 = charpos(sdsname_str, '.', 0)) != -1)
    {		  
      sds_info->sds_id = -1;		  
      strmid(sdsname_str, 0, p1, sds_name);
      strcpy(sds_info->name, sds_name);
      if (get_sds_info((char *)NULL, sds_info) != -1)
      {
	p1++;
	len = (int)strlen(sdsname_str);
	if ((p2 = charpos(sdsname_str, '.', p1)) != -1) 
	{
	  strmid(sdsname_str, p1, p2-p1, nd_ext); 
	  *n = (int)atoi(nd_ext);
	  p2++;
	  strmid(sdsname_str, p2, len-p2, md_ext); 
	  *m = (int)atoi(md_ext);	
	}
	else
	{
	  strmid(sdsname_str, p1, len-p1, nd_ext);
	  *n = (int)atoi(nd_ext);	
	}
      } 
      else
      {
	p1++;
	if ((p11 = charpos(sdsname_str, '.', p1)) != -1) 
	{		      		      		      
	  strmid(sdsname_str, 0, p11, sds_name);
	  strcpy(sds_info->name, sds_name);	    
	  len = (int)strlen(sdsname_str);
	  sds_info->sds_id = -1;
	  if (get_sds_info((char *)NULL, sds_info) != -1)
	  {
	    p11++;
	    if ((p2 = charpos(sdsname_str, '.', p11)) != -1) 
	    {
	      strmid(sdsname_str, p11, p2-p11, nd_ext); 
	      *n = (int)atoi(nd_ext);
	      p2++;
	      strmid(sdsname_str, p2, len-p2, md_ext); 
	      *m = (int)atoi(md_ext);	
	    }
	    else
	    {
	      strmid(sdsname_str, p11, len-p11, nd_ext);
	      *n = (int)atoi(nd_ext);	
	    }
	  }		      
	}	
	else
	{
	  *n = -1;
	  *m = -1;	  
/*	  strcpy(sds_info->name, sdsname_str);
	  sds_info->sds_id = -1;
      	  if (get_sds_info((char *)NULL, sds_info) != -1)
	  {
	    *n = -1;
	    *m = -1;
	  }
	  else
	  {
	    fprintf(stderr, "Cannot find the SDS %s \n", sds_info->name);
	    st = -1;
	  }  */
	}	    		    		    
      }		   
    }
    else 
    {
      *n = -1;
      *m = -1;
    }
  }
  else 
    st = -1;
  if ((*n == 0) || (*m == 0))
  {
    fprintf(stderr, "Invalid layer number in SDS name: %s\n", sdsname_str);
    st = -1;
  }
  else		    
  {
    if (*n != -1) --*n;
    if (*m != -1) --*m;
  }
  /*
  if (sds_info.sds_id != -1)
    close_sds(&sds_info, sds_info.H5IN);*/
  return st;
}


