/****************************************************************************
!C

!File: tilecenter.c

!Description:
  compute the centroid of the input tile

!Input Parameters: (none)

!Input/Output Parameters: (none)

!Output Parameters: (none)

!Revision History:
  Original April 1999. QA Tool Version 1.2

!Team-unique Header:
  This software was developed by:
    Land Data Operational Product Evaluation (LDOPE) Team for the
    National Aeronautics and Space Administration, Goddard Space Flight
    Center

!References and Credits:
    Sadashiva Devadiga
    LDOPE                             Science Systems and Applications Inc.
    devadiga@ltpmail.gsfc.nasa.gov    NASA/GSFC Code 922
    phone: 301-614-5449               Greenbelt, MD 20771

!Design Notes: (none)

!END
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "qa_tool.h"

#define TILEID_LEN 8
#define LAT_FILL -99.00
#define LON_FILL -999.00
#define TILEID_INFO_FNAME "tile_bound_10deg.txt"

#define HELP \
"NAME \n" \
"	tilecenter - compute the centroid of the input tile \n" \
"\n" \
"SYNOPSIS \n" \
"	tilecenter [-help] \n" \
"	tilecenter [tileid]|[h v] \n" \
" \n" \
"       echo [tileid] [h v] | tilecenter \n" \
" \n" \
"DESCRIPTION \n" \
"	This routine computes the longitude and latitude of the centroid \n" \
" of an input tile. Tile is specified by either the tileid or the vertical \n" \
" and horizontal tile number \n" \
" \n" \
"OPTIONS \n" \
"	-help                       Display this help message \n" \
"	tileid                      8 digit tileid \n" \
"	h v                         horizontal and vertical tile numbers \n" \
"	                            separated by space \n" \
" \n" \
"Examples: \n" \
"	tilecenter 11012004 \n" \
"	tilecenter 12 04 \n" \
" \n" \
"Last update: Version 3.6, 02/12/2020 \n" \
"Please report problems to Sadashiva Devadiga (sadashiva.devadiga-1@nasa.gov) \n"

#define USAGE \
"usage: tilecenter [tileid] [h v] \n" \
" \n" \
"OPTIONS \n" \
"	-help                       Display this help message \n" \
"	tileid                      8 digit tileid \n" \
"	h v                         horizontal and vertical tile numbers \n" \
"	                            separated by space \n" \
" \n"

int get_tileid(char *s, int *hn, int *vn);
int get_line(FILE *fp, char *s);

int main(int argc, char *argv[])
{
  FILE *fp;
  int st = 1;
  int iline = 1;
  int ih, iv, hn, vn;
  float lon_min, lon_max;
  float lat_min, lat_max;
  float lon_cent, lat_cent;
  char line[MAX_LINE_LENGTH];
  char fname[MAX_PATH_LENGTH];
  char *anc_path;

  if ((argc < 2) || (argc > 3))
  {
    fprintf(stderr, "Incorrect of number of arguments input\n");
    fprintf(stderr, "%s\n", USAGE);
    exit(1);
  }
  if ((argc == 2) && ((strcmp(argv[1], "-h")==0) || (strcmp(argv[1], "-help")==0)))
  {
    fprintf(stderr, "%s\n", HELP);
    exit(0);
  }
  if (argc == 2) 
    st = get_tileid(argv[1], &hn, &vn);
  else if (argc == 3)
  {
    hn = (int)atoi(argv[1]);
    vn = (int)atoi(argv[2]);
  }
  if ((vn < 0) || (vn > 17) || (hn < 0) || (hn > 35))
    fprintf(stderr, "Input tile number h=%d v=%d is undefined\n", hn, vn);
  else if (st != -1)
  {
    anc_path = getenv(ANC_ENV);
    sprintf(fname, "%s/%s", anc_path, TILEID_INFO_FNAME);
    if ((fp = fopen(fname, "r")) == NULL)
      fprintf(stderr, "Cannot open the tile information ancillary file: %s\n", fname);
    else
    {
      while (get_line(fp, line) != 0)
      {
	iline++;
	if (line[0] != '#')
	{
	  if (sscanf(line, "%d %d %f %f %f %f", &iv, &ih, &lon_min, &lon_max,
			&lat_min, &lat_max) != 6)
	    fprintf(stderr, "Error detected in file %s at line %d \n", fname, iline);
	  else
	  {
	    if ((iv == vn) && (ih == hn))
	    {
	      if ((lon_min == LON_FILL) && (lon_max == LON_FILL) &&
		  (lat_min == LAT_FILL) && (lat_max == LAT_FILL))
		fprintf(stderr, "Input Tile h=%d v=%d is undefined\n", hn, vn);
	      else
	      {
		lon_cent = (lon_max + lon_min)/2.0;
		lat_cent = (lat_max + lat_min)/2.0;
		fprintf(stdout, "Centroid for tile h=%d v=%d: %f %f\n", hn, vn, lon_cent,
				lat_cent);
	      }
	      break;
	    }
	  }
	}
      }
      fclose(fp);
    }
  }
/*
  free(anc_path);
*/
  return 0;
}

int get_tileid(char *s, int *hn, int *vn)
{
  int i;
  int st = -1;
  char tmp[5];

  if (strlen(s) != TILEID_LEN)
    fprintf(stderr, "Input tileid %s is invalid\n", s);
  else
  {
    for (i=2; i<5; i++)
      tmp[i-2] = s[i];
    tmp[i-2] = '\0';
    *hn = (int)atoi(tmp);
    for (i=5; i<8; i++)
      tmp[i-5] = s[i];
    tmp[i-5] = '\0';
    *vn = (int)atoi(tmp);
    st = 1;
  }
  return st;
}

int get_line(FILE *fp, char *s)
{
  int i, c;

  i = 0;
  while ((c = fgetc(fp)) != EOF)
  {
    s[i++] = c;
    if (c == '\n') break;
  }
  if (i != 0) s[i++] = '\0';
  return i;
}


